# StormBot.TCL version 6.X series, add-on script for EGGDROP V1.3 - V1.9

# (C) 2009 - 2019: [V6.0 - 6.2] David P. Hansen, Mai Mizuno
# (C) 2003 - 2010: [V5.0 - 5.2] David P. Hansen, Mai Mizuno
# (C) 2000 - 2003: [V4.0 - 4.3] David P. Hansen, Mai Mizuno
# (C) 2002 - 2003: [V3.5      ] David P. Hansen, Mai Mizuno
# (C) 2000 - 2003: [V3.1b- 3.2] David P. Hansen, Mai Mizuno
# (C) 1998 - 2000: [V3.0 - 3.1] David P. Hansen, Dave Klein, Justin Hammond, & Mai "Domino" Mizuno
# (C) 1997 - 1998: [V1.0 - 2.0] David P. "Xone" Hansen, Dave "CyberMac" Klein, & Justin "ElriC" Hammond

# --- License ---

# This program is free software; you can redistribute it and / or  modify  it
# under the terms of the GNU General Public License as published by the  Free
# Software Foundation. We hope the SOFTWARE is useful, but provide it WITHOUT
# ANY WARRANTY; without even  the  implied  warranty  of  MERCHANTABILITY  or
# FITNESS FOR A PARTICULAR PURPOSE. See  the GNU General Public  License  for
# more details. You should have received a copy of  the  GNU  General  Public
# License along with this  program;  if  not,  write  to  the  Free  Software
# Foundation, Inc., 59   Temple  Place -  Suite  330,  Boston,  Massachusetts
# 02111-1307, USA. Or, you can read it at http://www.gnu.org/ or you may read
# the license on our website in the MISCELLANEOUS section, or the  text  file
# included in the tarball from which you extracted this file (usually called:
# sb6_gpl_v3_license.txt).

# TRANSLATION: if you modify this script into  your  own  product,  you  must
# credit us as the source of the original source code. After  all,  we  wrote
# this, not you.

# --- EGGDROP information ---

# StormBot.TCL is an add-on script for use with the EGGDROP V1.*  IRC  robot.
# This script is =NOT= a fully running bot; you MUST download, install, & run
# the EGGDROP code itself, separately. We have no  affiliation  with  EGGDROP
# core coders in any way. We can and will  assist  you  with  installing  and
# configuring the StormBot.TCL script for use with EGGDROP on request. To get
# EGGDROP (*nix & Windows), visit: http://www.GetEggdrop.com/ and follow  the
# appropriate links.

# Please note that StormBot.TCL uses some of the same command names  as  some
# of EGGDROP's core commands (such as "SAY"). As a result, binding  these  SB
# commands in DCC will overwrite the original EGGDROP versions.  To  preserve
# them, SB will rebind the original EGGDROP command bindings as  "/<COMMAND>"
# in  DCC.  In  other  words,  ".HELP"  will  show  you   StormBot.TCL   HELP
# information but "./HELP" will show you EGGDROP core's original DCC  command
# help. This was done so both could co-exist simultaneously.

# And please: RTFM !! People write manuals for a reason; honour our  work  by
# reading them.

#####
# Anti-REHASH (REHASH -NOSB6)

if [file exists ./sb6_norehash.txt] { set ::sb_rehash_nosb6 1 }
if [info exists ::sb_rehash_nosb6] {
	if $::sb_rehash_nosb6 {
		file delete ./sb6_norehash.txt
		putlog "\[SB6.TCL\] REHASH: -NOSB6 bypass activated"
		return "\[SB6.TCL\] REHASH: -NOSB6 bypass activated"
	}
	unset ::sb_rehash_nosb6
}

# End Anti-REHASH
#####

### Removal data (do not alter this section!)

array set ::stormbot ""
array set ::stormbot_setup ""

if ![info exists ::stormbot(#pre:vars)    ] {set ::stormbot(#pre:vars)     [info vars    ]}
if ![info exists ::stormbot(#pre:commands)] {set ::stormbot(#pre:commands) [info commands]}
if ![info exists ::stormbot(#pre:timers)  ] {set ::stormbot(#pre:timers)   [timers       ]}
if ![info exists ::stormbot(#pre:utimers) ] {set ::stormbot(#pre:utimers)  [utimers      ]}
if ![info exists ::stormbot(#pre:binds)   ] {set ::stormbot(#pre:binds)    [binds        ]}

#############################################################################
#############################################################################
######################## --- BEGIN EDITING HERE! --- ########################
#############################################################################
#############################################################################

##### Optional: SHORTCUT
# Set this to the bot's shortcut. A shortcut is how the bot will be given its
# commands. This bind is available to all users, regardless of level. If  you
# do not set a value here, StormBot.TCL will automatically set  this  to  the
# first & last letters of your bot's nick ("LT" for "LlamaBot").
#
# Example: set to "LB" to use: "LB say Hello, I'm LlamaBot!"

set ::stormbot_setup(shortcut) ""

##### Optional: SHORTCUT:(LEVEL)
# Set these to the shortcuts that you want for each global user access level.
# This means: only a global user of "this" level (and above) can make the bot
# respond to this particular bind.

set ::stormbot_setup(shortcut:friend) ""; # "All global users" will trigger this
set ::stormbot_setup(shortcut:voice)  ""
set ::stormbot_setup(shortcut:halfop) ""
set ::stormbot_setup(shortcut:op)     ""
set ::stormbot_setup(shortcut:tech)   ""
set ::stormbot_setup(shortcut:master) ""
set ::stormbot_setup(shortcut:owner)  ""

##### Assistance marker
# If you come in to ask for help, the coder may ask you to copy / paste these
# five lines to show that  you've  actually  read  this  file  before  asking
# "eggdrop 101" type questions. When asked, tell the coder the number "25."
#####

##### Required: DEFAULT_OUTPUT (notice/msg/chan/chan #specific_channel)
# Set this to the default type of output a new user  will  have  (or  a  user
# who has not set their personal output setting yet). Original  IRC  protocol
# (RFC) requires this to be NOTICE, but some users have special  windows  and
# scripts that need to capture input from bots via /MSG. That, or  it's  just
# personal preference. For those  who  are  setting  up  bots  primarily  for
# exclusive / primary use  with  FAKE  ISPs,  such  as  WebTV(tm),  set  this
# variable to "MSG" instead. This affects ALL USERS' default output  setting.
# To change this for individual users, use the SET OUTPUT command, which lets
# you set several personal settings (including output).
#
# Options: NOTICE, MSG, CHAN, or "CHAN #SPECIFIC_CHAN"

set ::stormbot_setup(global-output) "notice"

##### Required: BINDCHAN, BINDPUBCMD, BINDMSG, BINDNOT & BINDDCC (0/1)
# Set these variables to 0 (off), or 1 (on), if you want  commands  bound  to
# MSG (/msg format), NOTICE (/notice  format),  or  .DCC  (partyline  command
# format). While you can run your bot only with  PUB  (in-channel)  commands,
# some commands require some measure of  privacy,  such  as  LOGIN  (and  its
# friends). CHAN is the normal in-channel trigger; PUBCMD is the  single-char
# command trigger (see CONFIG for more information on PUBCMD options).
#
# Options: 0 (off), 1 (on)

set ::stormbot_setup(@bind:chan)   1
set ::stormbot_setup(@bind:pubcmd) 1
set ::stormbot_setup(@bind:msg)    1
set ::stormbot_setup(@bind:notice) 1
set ::stormbot_setup(@bind:dcc)    1

##### Required: BOT:TIMEZONE & BOT:GMT
# Set these variables so the bot will know its TIMEZONE and GMT offset. For
# Windrop (due to the Windows OS itself), the %z flag doesn't get defined, so
# the bot can't figure out what GMT offset it should observe. Appropriate
# settings are: "PST" for TIMEZONE, and "-8" for GMT. These can be updated
# via the CONFIG command (such as: for Daylight Saving Time).

set ::stormbot_setup(bot:timezone) ""
set ::stormbot_setup(bot:gmt) ""

##### Required: HOME (valid channel name)
# Set this variable to the channel that will be your "home" for your bot. For
# normal bot users, this may not be anything you need. For bot loaning groups
# this would be the channel that all the reports go to (LOGIN, BOTMON, etc).
#
# Options: any valid channel name

set ::stormbot_setup(@home) ""

##### Optional: name:GROUP
# Set this to the name of your IRC / channel group (if applicable). The
# other levels will have their names set based on the USERLEVELS entry below.
#
# Example: "Kings Cross Bot Service"

set ::stormbot_setup(name:group) ""

##### Required: USERLEVELS
# Set these values to the access names, levels, & flags for each access level
# you want. If you don't feel comfortable with tweaking this  data  here  (or
# in the sb6_defs.tcl file), use the CONFIG command.  The  values  here  are:
# NAME of level, ACCESS level, access level RANGE, & FLAGS for level,  comma-
# -delimited. Use the token "%GROUP%" to insert the name of your  group  (set
# above) into the data.
#
# Please note that there have been some specific changes  to  the  userlevels
# from their StormBot.TCL V5 counterparts:
#    (1) the use of the word "GLOBAL" with the global levels
#    (2) the adjustment of "GLOBALFRIEND" as a valid access range (501-549)
#    (3) the adjustment of "GLOBALVOICE" to range (550-574)
#    (4) the adjustment of "GLOBALHALFOP" to range (575-599)
# These changes are to make the global user levels better match  their  local
# (channel) counterparts. Due to the GLOBALTECH  level,  there  is  no  level
# GLOBALSUPEROP.

set ::stormbot_setup(userlevels) [string trim {
	unknown      ,   -1 ,      <0 ,               |        , unknown user
	noaccess     ,    0 ,      =0 ,               |        , no access
	chanfriend   ,    1 ,    1-49 ,               |+f      , Channel Friend
	chanvoice    ,   50 ,   50-74 ,             +p|+fv     , Channel Voice
	chanhalfop   ,   75 ,   75-99 ,             +p|+fvl    , Channel HalfOp
	chanop       ,  100 , 100-299 ,             +p|+fvlo   , Channel Op
	chansuperop  ,  300 , 300-399 ,             +p|+fvlo   , Channel SuperOp
	chanmaster   ,  400 , 400-499 ,             +p|+fvlom  , Channel Master
	chanowner    ,  500 ,    =500 ,             +p|+fvlomn , Channel Owner
	globalfriend ,  501 , 501-549 ,            +fp         , %GROUP% Helper
	globalvoice  ,  550 , 550-574 ,           +vfp         , %GROUP% Global Voice
	globalhalfop ,  575 , 575-599 ,          +lvfp         , %GROUP% Global HalfOp
	globalop     ,  600 , 600-699 ,         +olvfp         , %GROUP% Global Op
	globaltech   ,  700 , 700-799 ,        +tolvfp         , %GROUP% Botnet Technician
	globalmaster ,  800 , 800-899 ,       +mtolvfp         , %GROUP% Co-Admin
	globalowner  ,  900 , 900-999 ,      +nmtolvfp         , %GROUP% Admin
	botowner     , 1000 ,   =1000 ,    +jxnmtolvfp         , %GROUP% Permanent Bot Owner
	permbotowner , 1001 ,   >1000 , +heujxnmtolvfp         , %GROUP% Primary Permanent Bot Owner
}]

#############################################################################
#############################################################################
######################### --- END EDITING HERE! --- #########################
#############################################################################
#############################################################################

# This explains the reason the "UPDATE" was removed from the boot-up code.
# Eggdrop 1.6.21 release notes
# http://cvs.eggheads.org/viewvc/eggdrop1.6/doc/Changes1.6?view=markup
# 100    - do_restart is now reset before actually performing a rehash or restart to
# 101      ensure it doesn't try to do it again infinitely.
# 102      Found by: Domino / Patch by: thommey

# --- Boot it up! ---
set ::sb_fragment ""
foreach ::sb_fragment [list defs] { # [list defs]
	if [file exists ./scripts/sb6_${::sb_fragment}.tcl] {
		uplevel #0 { source ./scripts/sb6_${::sb_fragment}.tcl }
	}
}; unset ::sb_fragment

#</core>

#<LOAD>
proc sb6:bootstrap:sb6_load.tcl args {

set ::stormbot(@script_is_loaded) 0 ; # SERVERS requires this!

# --- Setup AUTOPURGE ---
##### If updated, you must RESTART (REHASH won't override the variable in memory)
# "./scripts/" is automatically prepended.
# 2016-05-15 01:55:00 -0700: list cleaned-up and duplicates removed
set ::stormbot_setup(@autopurge) {

	#SB6_CORE_FILES_HERE
	sb6_defaults.conf sb6_defaults.renameme.tcl sb6_defaults.tcl
	sb6_disp.tcl sb6_load.tcl sb6_loader.sb6 sb6_loader.tcl sb6_tail.tcl
	# 20180907-101800-0700: I'm reviving the sb6_defs.tcl file

	#SB6_COMMAND_BEADS_HERE
	beads/book.sb6 ; # Renamed to ADDRESSBOOK.sb6
	beads/addrbk.sb6  ; # Renamed to ADDRESSBOOK.sb6
	beads/servers.sb6  ; # Renamed to JUMP.sb6
	beads/hsp.sb6
	beads/note2.sb6 
	beads/old_note.sb6 

	#2017-05-24 18:01:00 -0700: 
	beads/sb5.sb6
	beads/strip-cr.sb6 ; # Renamed to STRIPCR.sb6
	beads/t2.sb6 
	beads/67.sb6 

	#20160516-052800-0700
	#beads/pastebin.sb6 ; # Need to re-code this to handle HASTEBIN
	beads/newtree.sb6 ; # TREE (part of BOTNET) now handles this code
	beads/email2.sb6

	#20170529-235000-0700
	beads/makeegg.sb6 ; # Was from 1.6.19; was useful until 1.6.21

	#20170615-025300-0700
	beads/ircd.sb6

	#20180627-125400-0700
	beads/sbdsend.sb6 ; # Renamed to "SBDDATA"
	beads/guess.sb6
	beads/enc.sb6
	beads/@*.sb6
	
	#20180731-163500-0700
	beads/last.sb6	
	beads/*.old

	#20180827-213500-0700
	beads/_*.sb6
	
	#20180905-174900-0700
	beads/prot.sb6
	beads/deprot.sb6
	beads/autoprot.sb6

	#20180907-090100-0700
	beads/own.sb6
	beads/deown.sb6
	beads/autoown.sb6
}

# --- Bootstrap ---
# Check for SB5 --> SB6 conversion
if ![string eq "" [info procs config]] {
	catch {rename config ""}
	catch {msghome "[color 11,10]\[SB6\] SB5 -> SB6 conversion detected: a RESTART is recommended instead![color]"}
	catch {putlog "[color 11,10]\[SB6\] it seems we're doing an SB5 -> SB6 conversion. I would recommend a RESTART as some (SB5) BINDs are NOT overwritten unless you RESTART. Please let me know if you experience any difficulties or error messages. ~ Domino ( Domino@StormBot.Org / irc://irc.stormbot.org/#BotHouse )[color]"}
}

# Load the bot's data file
set error [ catch { SB:botdatafile load } shit ] ; \u53\102\u3atest\u5f\x6eu\x6dbe\x72
if $error { putlog "===== Error while loading data file: $shit =====" }
# Continue processing!

# Invoke all setup values (some values are needed BEFORE loading beads!).
# Current rule is: only refresh variables if the true variable is unset.
# If the true variable has a value, presume we have a fresher value already
# set.

# Special situation: SERVERS
if [sbd:notempty SERVERS] {
	if [sbd:isempty @SERVERS:ORIGINAL] { sbd:set @SERVERS:ORIGINAL $::servers }
	set checked ""
	foreach s [sbd:get SERVERS] {
		set s [string trimright $s]
		if [regexp -- {^\[?([^ \]]+)} $s - fixed] { set s $fixed } ; # "[server:port:password ]:default-port"
		lappend checked $s
	}
	sbd:set servers $checked
	set ::servers $checked

	# Force connection to a server (if one is listed)
	if [isempty ::server] { if [notempty ::servers] { jump [l0 $::servers] } }
}

# Special situation: HOME
if [info exists ::stormbot_setup(@home)] {# To simplify user-editable config file
	set ::stormbot_setup(config:global-home) $::stormbot_setup(@home)
	unset ::stormbot_setup(@home)
}

# Special situation: PRECISION
if [sbd:notempty CONFIG:GLOBAL-PRECISION] {set ::tcl_precision [sbd:get CONFIG:GLOBAL-PRECISION]}

# Special situation: USERLEVELS
if {[sbd:isempty USERLEVELS] || ![isint [sbd:lindex USERLEVELS 0]]} {# permbotowner , 1001 , >1000 , +heujxnmtolvfp , %GROUP% Primary Permanent Bot Owner
	set a ""
	emptyarray crap
	foreach a [split $::stormbot_setup(userlevels) \n] {
		regsub -all -- { +,|, +| +, +} $a , a
		set a [split [trim $a] ,]
		set crap([l1 $a]) $a
	}
	sbd:set USERLEVELS [array get crap]
	unset a crap
}

#############################################################################
# DO NOT ALLOW SB:SETUP TO HANDLE THIS! IT WILL PUSH  THIS  DATA  INTO  THE #
# NORMAL VARIABLE STREAM BEFORE THIS PROC IS INVOKED, AND THUS,  NEVER  GET #
# PROCESSED BY SB:SETUP.  THIS  IS  WHY  BOT  DATA  FILES  STILL  HAVE  THE #
# MULTI-LINE ENTRIES FOR USERLEVELS! RENAMED  FROM  SB:SETUP_USERLEVELS  TO #
# SB:USERLEVELS_PROCESS.                                                    #
#############################################################################

proc SB:userlevels_process args {
	empty array null name level range flags split
	emptyarray new

	zero force_update
	if [notempty args] {
		one force_update; set data $args
	} {
		if [info exists ::stormbot_setup(userlevels)] { set data [split $::stormbot_setup(userlevels) \n] }
	}
	if [isnull data] {return ""}

	foreach null $data {
		set null [trim $null]
		regsub -all -- \t $null "" null
		regsub -all -- \n $null "" null
		if [isempty null] continue
		regsub -all -- {[ ]+,|,[ ]+} $null , null
		set null [singlespace $null]
		foreach a [regexp -all -inline -- {[ ]+,|,[ ]+} $null] { regsub -all $a $null , null }
		regexp -nocase -- {^([^,]*),([^,]*),([^,]*),([^,]*),(.*)$} $null - name level range flags title

		set name [trim $name]
		set new($level) [list $name $level $range $flags $title]
		sbdset USERLEVEL:$name $level
		sbdset name:$name $title
		SB:register variable global USERLEVEL:$name NAME:$name
	}

	# Set values directly instead of letting the *_SETUP loop do it.
	set ::stormbot_setup(userlevels) [array get new]
	if $force_update {set ::stormbot(userlevels) [array get new]}
}

SB:userlevels_process

# -- Set up global configuration settings from SB6.TCL default values ---
empty sb
foreach sb [array names ::stormbot_setup] {
	set refreshme 1
	set target $sb
	if [string match -nocase SHORTCUT-* $target] { set target "CONFIG:$target" }
	if [string eq -nocase GLOBAL-OUTPUT $target] { set target "CONFIG:$target" }
	if [string match -nocase @BIND:* $target] { set target "CONFIG:GLOBAL-[string range [regsub -all -- : $target ""] 1 end]" }
	# Array element should be $TARGET not $SB
	if [info exists ::stormbot($target)] { if ![string eq "" $::stormbot($target)] { set refreshme 0 } }
	if $refreshme { set ::stormbot($target) $::stormbot_setup($sb) }
	unset -nocomplain ::stormbot_setup($sb)
}
unset -nocomplain sb target

# --- Auto-purge ---
if [sbd:notempty @AUTOPURGE] {
	empty ok bad
	catch { file mkdir ./scripts/beads/purged }
	foreach purgeme [sbd:get @AUTOPURGE] {
		if [string match #* $purgeme] continue ; # For inline notes
		if [file exists ./scripts/$purgeme] {
			set error [ catch { file rename -force ./scripts/$purgeme ./scripts/beads/purged/ } crap ]
			if $error {
				lappend bad "${purgeme}: $crap"
			} {
				lappend ok $purgeme
			}
		}
	}
	if ![string eq "" $ok] { putlog "\[SB:AUTOPURGE\] Removed the following beads: [ajl $ok]" }
	if ![string eq "" $bad] { foreach a $bad { putlog "\[SB:AUTOPURGE\] Failed bead removal: $a" } }
	unset -nocomplain purgeme
}

# Load all beads
#%# utimer 0
SB:loadbeads * [lindex force [sbd:nozero @SCRIPT_IS_LOADED]]; # force

# Let's put this on a TIMER so the bot can finish loading right away

# --- Verify user settings & set default values ---
if 0 {#%#
	set ::stormbot(nick)    $::nick
	set ::stormbot(altnick) $::altnick
}

if ![info exists ::stormbot(@default_output)] {set ::stormbot(@default_output) ""}
if { [lsearch -glob [list NOTICE MSG CHAN] [string toupper [lindex $::stormbot(@default_output)]]] == -1 } { set ::stormbot(@default_output) notice }
if ![info exists ::stormbot(userlevel:botowner)] { set ::stormbot(userlevel:botowner) 1000 }
if ![info exists ::stormbot(userlevel:unknown)] { set ::stormbot(userlevel:unknown) -1 }

# Use $A for all setup loops; will UNSET after.

set a "" ; # Needed in case $A is never set (empty FOREACH), to avoid -NOCOMPLAIN
foreach a [list @bindmsg @bindnot @binddcc] {
	if { ![info exists ::stormbot($a)] || [string eq "" $::stormbot($a)] } { set ::stormbot($a) 0 }
}
foreach a [list @cmdinfo:commands @cmdinfo:commandlist @wedgelist] { if ![info exists ::stormbot($a)] { set ::stormbot($a) "" } }
unset a

#############################################################################

# Now, any SB:SETUP_* PROCs within the beads will be loaded by SB:SETUP.

proc SB:setup_00000_convert_sb5 args {# 00000 to sort to the top of the list
	# This transitions things from SB5 to SB6 internal data streams

	zero changed
	if [sbd:notempty config:global-pubcommand-trigger] {
		one changed
		sbd:rename config:global-pubcommand-trigger CONFIG:GLOBAL-PUBCOMMAND:TRIGGER
	}

	foreach trigger [sbd:find CONFIG:GLOBAL-COMMANDBINDS_*] {
		one changed
		set level [mid $trigger 28]
		sbd:concat CONFIG:GLOBAL-COMMANDBINDS:$level [sbd:get $trigger]
		sbd:set $trigger
	}

	if [sbd:notempty shortcut] {
		one changed
		sbd:concat CONFIG:GLOBAL-COMMANDBINDS:ALL [sbd:get SHORTCUT]
		sbd:set SHORTCUT
	}

	if $changed {putlog "\[SB5 --> SB6\] command trigger and shortcut binds converted"}
	return 0
}

proc SB:setup_abbr args {SB:abbr clean}

interp alias {} SB:logoutall {} SB:setup_clearlogins
proc SB:setup_clearlogins args {
	if [nozero [sbdget CONFIG:GLOBAL-PROTECTAUTHS]] return
#	putlog "--- Clearing all logins ---"
	SB:deauth *
	return 0
}

proc SB:setup_fix_sb5_lastcmdchan args {
	# A hold over from SB5's DO_CALLCOMMAND that tried to store a user's
	# last-channel command use. SB6's OUTPUT2 still check it, therefore,
	# upgraded bots from SB5 to SB6 may experience "illegal-channel" errors
	# every time they throw a command if the last command channel used
	# (before conversion) is no longer a valid channel. A patch to OUTPUT2
	# (sb6_tail.tcl) is also needed.
	foreach a [userlist] { setdata $a LASTCMDCHAN }
}

proc SB:setup_homechan args { if [sbd:isempty CONFIG:GLOBAL-HOME] { sbdset CONFIG:GLOBAL-HOME [lindex [channels] 0] } }

proc SB:setup_shortcuts args {
	set list ""
	if [info exists ::nick] {lappend list $::nick}
	if [info exists ::altnick] {lappend list $::altnick}
	if [info exists ::botnet-nick] {lappend list ${::botnet-nick}}
	if [info exists ::stormbot(nick)] {lappend list $::stormbot(nick)}
	if [info exists ::stormbot(altnick)] {lappend list $::stormbot(altnick)}
	if [info exists ::stormbot(shortcut)] {
		lappend list $::stormbot(shortcut)
	} {
		if [isempty list] {
# 1400480159: Don't: lappend list [string index $::nick 0][string index $::nick end]
		}
	}

	set target SB:dispatch_chan
	foreach a [lsort -uni -dict -inc [stl $list]] { if ![string eq "" $a] { bind pub - $a $target } }
	foreach {level flags} [list all - friend f voice v halfop l op o tech t master m owner n] {
		if [info exists ::stormbot(shortcut:$level)] {
			sbd:concat CONFIG:GLOBAL-COMMANDBINDS:$level [sbd:get shortcut:$level]
			sbd:set shortcut:$level
			#foreach bind [lsort -increasing -dictionary -unique $::stormbot(shortcut:$level)] {
			#	log "\[SB:setup_shortcuts\] Adding BIND: $bind ($target)"
			#	bind pub $flags [stl $bind] $target
			#	sbd:lappend @triggers $bind
			#}
		}

		set data [sbdget CONFIG:GLOBAL-COMMANDBINDS:$level]
		if [notempty data] {
			foreach bind [lsort -increasing -dictionary -unique $data] {
				log "\[SB:setup_shortcuts\] Adding BIND: $bind ($target)"
				bind pub $flags [stl $bind] $target
				sbd:lappend @triggers $bind
			}
		}
	}

	# 2014-08-13: Don't do it
	if 0 {
		# Bind by version:TCL
		set version_tcl ?T[l0 [split [info patchlevel] .]]
		bind pub - [stl $version_tcl] $target
		lappend @triggers $version_tcl

		# Bind by version:Eggdrop
		set version_egg ?T[le [split [l0 $::version] .]]
		bind pub - [stl $version_egg] $target
		lappend @triggers $version_egg
	}

	sbd:set @TRIGGERS [lsort -uni -inc -dict [sbd:get @TRIGGERS]]
debug.print 1 "\[SB:SETUP_SHORTCUTS\] [ajl [sbd:get @TRIGGERS]]"
	return 0
}

proc SB:setup_bugsiebug_check args {bind TIME - * SB:checkbugsiebug}

proc SB:raw_server_data_351 {a b c} {
	set c [split $c]
	# 351 Unreal3.2.7. LapTop.ShineAquaIllusion.Org :FinWXeOoZE [Windows XP Service Pack 3=2309]
	sbd:clear @NETWORK:SERVER:*
	sbd:set @NETWORK:SERVER:IRCd [string trimright [jl1 $c] .]
	sbd:set @NETWORK:SERVER:NAME [jl2 $c]
	sbd:set @NETWORK:SERVER:COMPILE [mid [jl3e $c] 2]
	putlog "\[VERSION\] [l2 $c] (using [string trimright [jl1 $c] .]): compiled as: [mid [jl3e $c] 2]"
	return 0
}

proc SB:raw_server_data_005 {a b c} {

	#005 CMDS=KNOCK,MAP,DCCALLOW,USERIP NAMESX SAFELIST HCN MAXCHANNELS=10 CHANLIMIT=#:10 MAXLIST=b:60,e:60,I:60 NICKLEN=30 CHANNELLEN=32 TOPICLEN=307 KICKLEN=307 AWAYLEN=307 MAXTARGETS=20 WALLCHOPS WATCH=128 SILENCE=15 MODES=12 CHANTYPES=# PREFIX=(qaohv)~&@%+ CHANMODES=beI,kfL,lj,psmntirRcOAQKVCuzNSMTG NETWORK=StormBot.Org CASEMAPPING=ascii EXTBAN=~,cqnr ELIST=MNUCT STATUSMSG=~&@%+ EXCEPTS INVEX :are supported by this server

	set c [lreplace [split $c] 0 0]
	set m [lsearch -glob $c :*]
	set c [lreplace $c $m end]

	lassign [ldestroy -both -glob -all $c *=*] modes values
	foreach mode $modes {sbd:set @NETWORK:SERVER:$mode 1}

	foreach value $values {
		lassign [split $value =] name value
		foreach entry [split $value ,] {
			if [instr $entry :] {
				sbd:lappend @NETWORK:SERVER:$name [split $entry :]
			} {
				if [string eq -nocase PREFIX $name] {
					regexp -- {^\(([^\)]+)\)([^ ]+)$} $entry - chanop entry
log [debug -quiet chanop entry]
					sbd:set @NETWORK:SERVER:MODES:CHANOP $chanop
				}
				sbd:lappend @NETWORK:SERVER:$name $entry
			}
			sbd:set @NETWORK:SERVER:$name [nodup [sbd:get @NETWORK:SERVER:$name]]
		}
	}
	foreach crap [sbd:find @NETWORK:SERVER:*] {sbd:set $crap [nodup [sbd:get $crap]]}

	# We need to update the 'available modes' list from this data
	# Chanop modes (q a o h v) aren't listed in CHANMODES=*; it's found as: "PREFIX=(qaohv)~&@%+"
	set modes [sbd:get @NETWORK:SERVER:CHANMODES][sbd:get @NETWORK:SERVER:MODES:CHANOP]
debug.print 2 "\[SB:RAW_SERVER_DATA_005\] MODES($modes)"
	if [notempty modes] {
		array set chanmodes [sbd:get @MODE:ENABLED]
		foreach a [concat 0 1 2 3 4 5 6 7 8 9 * ! [alphabet lower split] [alphabet upper split]] {
			if [instr $modes $a] {set chanmodes($a) 1} {set chanmodes($a) 0}
		}
		sbd:set @MODE:ENABLED [array get chanmodes]
	}
	
	# 2018-06-03 16:27:00 -0700:
	# While programming STRIPOPS, I noticed that some servers (e.g. UnrealIRCd 3.2.10.7)
	# allow 20 modes-per-line, but, only executes 12. The remaining 8-per-line are ignored: the args are seallowed!
	# No rejection messages (or any complaints or errors) are returned by the server.
	if [sbd:nozero @NETWORK:SERVER:MAXTARGETS] {
		if { [sbd:get @NETWORK:SERVER:MAXTARGETS] > 12 } { sbd:set @NETWORK:SERVER:MAXTARGETS 12 }
	}
	return 0
}

proc validmode mode {
	array set chanmodes [sbd:get @MODE:ENABLED]
	if ![info exists chanmodes($mode)] {return 0}
	return $chanmodes($mode); # Boolean
}

proc validircdcmd cmd {
	set cmds [stl $cmds]
	set ircdcmds [stl [sbd:get @NETWORK:SERVER:CMDS]]
	if [isempty ircdcmds] {return 0}
	expr ( [ lsearch -exact $ircdcmds $cmd ] > -1 ) ? 1 : 0
}

proc SB:setup_collect_modes args {rawoutput VERSION; return 0}

proc SB:setup_gmt args {
	if [string eq "" [sbd:get BOT:GMT]] {
		set _ [clock format [clock seconds] -format %z]
debug.print 1 "\[SB:SETUP_GMT\] _($_)"
		set gmt [SB:gmt_format decimal $_]
debug.print 1 "\[SB:SETUP_GMT\] GMT($gmt)"
		sbd:set BOT:GMT $gmt
		SB:needsave data
debug.print 1 "\[SB:SETUP_GMT\] Setting my GMT as: $gmt"
	}
	if [string eq "" [sbd:get BOT:TIMEZONE]] {
		set tz [clock format [clock seconds] -format %Z]
		sbd:set BOT:TIMEZONE $tz
		SB:needsave data
debug.print 1 "\[SB:SETUP_GMT\] Setting my TIMEZONE as: $tz"
	}
	return 0
}

# Any SB:SETUP_* PROCs =MUST= go (or be defined) before here!

proc SB:setup args {
	putcmdlog "--- \[SB:SETUP\] ---"

	# 2015-05-11 01:23:00 -0700: Let's repair any interrupted DISPATCH changes in user settings (--FORCE:* flags)
	foreach user [userlist] {
		foreach item [list output burstmode color] {
			set data [getdata $user DISPATCH:@TEMP:$item]
			if ![string eq "" $data] {
				setdata $user $item $data
				setdata $user DISPATCH:@TEMP:$item ""
				putcmdlog "\[SB:SETUP\] Auto-repaired $item setting for: $user ( --FORCE:* flag wasn't properly purged )"
			}
		}
	}

	# 2018-09-07 10:20:00 -0700: Reviving the sb6_defs.tcl protocol:
	# after the core variables are read, but before the dta file is loaded (below)
	if [file exists ./scripts/sb6_defs.tcl] { uplevel #0 source ./scripts/sb6_defs.tcl }

	# Let's load the setup commands (from the loader, and each bead) now ....

	set list [info procs SB:setup_*]
	set ::SB_ITEM ""
	foreach item $list {
		set ::SB_ITEM $item
debug.print 1 "\[SB:SETUP\] Processing: $item"
		set error [ catch { uplevel #0 { eval $::SB_ITEM } } shit]
		if $error {
			set message "\00304\[SB:SETUP\] fatal error while processing ${item}: ${shit}\003"
			putlog $message
			error $message
		} {
			#
		}
	}
	unset ::SB_ITEM

	# --- BIND some default things ---
	bind DCC t dcc *dcc:dccstat
	bind DCC - /w *dcc:whois

	# Global variable name (just to keep the pattern with user & chan files)
	set ::datafile [SB:botdatafile name]

	# Set up WHOIS-FIELDS
	set ::whois-fields [string tolower [lsort -unique -increasing [concat ${::whois-fields} vhost]]]

	# Clear useless user data
	foreach user [userlist] { setuser $user xtra sendreply }

	# --- We're done ---
	putcmdlog "--- </sb:setup> ---"
}

# Bootstrap:
SB:setup

catch { getuser:valid force }
SB:register variable global *:AUTHOR
SB:register variable global NETWORK:CHANTYPES
SB:register variable global SHORTCUT:*
SB:register variable global USERLEVELS
SB:register variable global USERLEVEL:*

bind RAW - 351 SB:raw_server_data_351
bind RAW - 005 SB:raw_server_data_005

set ::stormbot(@script_is_loaded) 1; # SERVERS requires this!

# ... and if we survived all that:
putlog "\[StormBot.TCL\] SB v[regexp -inline -nocase -- {\d.+} [sbd:get @VERSION]] (by Mai \"Domino\" Mizuno) loaded: enjoy your bot!"
}
#</load>

#<DISP>
proc sb6:bootstrap:sb6_disp.tcl args {

#############################################################################

# --- Dispatcher ---

# Dispatcher: [mark when updated / coded / included]
# X $ARG empty? Exit.
# X Bugsie Bug
# X DEFAULT: $::lasttriger "chan"]
# X SINGLESPACE $ARG
# - If user = known:
# - > LASTCMD, STALELOGIN, LASTCMDCHAN, get SENDREPLY value, Check for DENY
# - > set dccnick [dccnick $nick $handle $chan]
# X If +IGNORECHAN & LASTTRIGGER ="CHAN:" return [do a PUTLOG of ignored line?]
# X if [L0 $ARG] = -ALL: loop through channel list (reJOIN text?)
# X Macro check (global/local: substitute if necessary)
# X Check for #chan before CMD [swapped args]
# - Command alias(es)? (LREPLACE $COMMAND with any found aliases)
# - Default: $CHAN command-used chan.
# X Substitute # / ## for $CHAN as appropriate
# - NO! Adjust args [removing channel name / token] and assign $MISC* variables
# X ARRAY SET SB_OUTPUT [list $NICK [list $NICK0 $HOST $HANDLE $CHAN]]
# - CMDREPORT output?
# - if AUTHCMD: <password blocked> LOG command [APPENDLOG]
# - > if !AUTHCMDCONTINUE: HALT!
# - if CMDSUSP: Warning message; HALT!
# - if CHECKBOT: exit [in SB5, was !CHECKBOT { { { { ... } } } } ]
# - If not DELAY_ALARM_OVERRIDE: if !DCCNICK: if $CMDLEVEL > -1:
# - > if (authnick <> authhandle) or (authhost <> authhandle): "Who are you?"
# - if $CMDLEVEL > 900: if ISEMPTY OWNER {error msg}
# - ISUSERSUSPENDED [$CHAN] {complain}
# - Execute all wedges: FOREACH a [ARRAY NAMES WEGDE] { if ($a) {[sbdget @wedgelist($a)] $nick $host $handle $chan $arg } }
# - All commands are now beads; no more 'sb5_body.tcl' commands
# - if !INVALIDOK {obliterate $ARG:1}
# - CATCH { execute PROCNAME }; if ERROR {complain}
# - set HelpSniffer(on) 0
# X Generic error if we get this far!

# - MSG commands:
# - > chan-spec
# - > chan-non-spec

# - DCC commands [if DCC_SUPPORT is on from CONFIG file]:
# - > chan-spec
# - > chan-non-spec

# Suggestion for piping command data:
# Running live variable: $::stormbot(@cmdinfo:commandlist)
# Counter-variable: $::stormbot(@cmdinfo:commandlist:[stl $chan]) [when values match, delete local setting]

# --- Place holders while SB:DISPATCH is being properly coded ---
proc SB:appendlogs args {return 0}

# --- SB:DISPATCH ---

proc SB:dispatch { nick host handle chan arg { override false } { override_output 0 } } {
debug.print 2 "\[SB:DISPATCH\] 00"
	if [string eq "" $arg] return
	# bugsiebug_check
	SB:checkbugsiebug
debug.print 2 "\[SB:DISPATCH\] 00.5 NICK($nick)"

	#######################################################################
	# Fixing a forever-old problem: if common hostmask, is this an authed #
	# user (so we can figure out which handle to assign)? Since all PROCs #
	# draw the $HANDLE from here (and WHOIS checks by nick first), we can #
	# fix this in one place: here                                         #
	#######################################################################

	if [SB:isdccnick $nick] { set h [SB:authhandle $nick] } { set h [SB:authhandle $nick $host] }
	if [validuser $h] { set handle $h }
debug.print 2 "\[SB:DISPATCH\] 00.5 NICK($nick):H($h):HANDLE($handle)"

	if [SB:isdccnick $nick] {
		set idx [lindex $nick 1]
		if [regexp -- {^\d+$} $idx] {
			# If this output isn't relayed to another bot ....
			if ![valididx $idx] {
				catch { killidx $idx }
				msghome "\[SB:DISPATCH\] Illegal user found on IDX #$idx"
				error "\[SB:DISPATCH\] Illegal user found on IDX #$idx"
				return 0
			}
			if ![string eq -nocase [idx2hand $idx] $handle] {
				set _message "\[SB:DISPATCH\] User on IDX #${idx}: \$HANDLE: is $handle but IDX/handle look-up is: [idx2hand $idx] (IDX collision due to ALARM/DELAY?)"
				msghome $_message
				error $_message
				return 0
			}
		}
		if [string eq * $chan] { print $nick "While EGGDROP allows you to monitor more than one channel, StormBot.TCL does not support it for command usage. Either (1) specify a channel name with the command, or (2) .CONSOLE <any valid channel>" ; return 0 }
		# What if we have a "NEW" user (eggdrop -m) ... ?
		# Set some default output settings (a necessary override)
		sbd:set @output:$nick [list $nick $host $handle [lindex [channels] 0]]
		sbd:set @output:$handle [list $nick $host $handle [lindex [channels] 0]]
		if [string eq "" [getuser $handle HOSTS]] {
			putidx $idx "[color 0,1]You don't have any HOSTS added yet. Until you add some, I won't be able to respond to you in channel or via /msg\; only DCC command will work.[color]"
			putidx $idx "[color 0,1]To add a host to your account, get in a common channel with me, and then do the following command:[color]"
			putidx $idx "[color 0,1]$::botnick ADDMASK $handle <your nick on the channel>[color]"
			putidx $idx "[color 0,1]Example: $::botnick ADDMASK $handle Fred[color]"
		}
	}

debug.print 2 "\[SB:DISPATCH\] 01"

	if { [SB:isdccnick $nick] && [instr $nick @] } { sbd:set @LAST_TRIGGER remote }
	if [sbd:isempty @LAST_TRIGGER] { sbdset @LAST_TRIGGER chan }
	if [sbdequal -nocase @LAST_TRIGGER CHAN] {
		if [checkchannelsetting $chan ignorechan] {return ""}; # Silent exit
	}

debug.print 2 "\[SB:DISPATCH\] 02"

	# Handle user-specific issues
	# Do =NOT= disturb $ARG until after here: undisturbed data needed for ENCODING issues

	if [validuser $handle] {
		# LASTCMD
		# STALELOGIN
		# LASTCMDCHAN

debug.print 2 "\[SB:DISPATCH\] 02.1"
		if [validcmd SB:deny] { if [ SB:deny check $handle ] { return 0 }; # Exit quietly! }
debug.print 2 "\[SB:DISPATCH\] 02.2"

		set encoding [getdata $handle ENCODING]
		if [sbd:nozero CONFIG:GLOBAL-USEENCODING] {
			if [notempty encoding] {
				switch -exact -- [stu [sbd:get @LAST_TRIGGER]] {

					CHAN {
						set instr [instr [sbd:get @DISPATCH:RAW] " :"]
						incr instr 2
						set new [mid [encoding convertto $encoding [sbd:get @DISPATCH:RAW]] $instr]
						set len [len [encoding convertto $encoding $::lastbind]]
						set arg [mid $new [incr len]]
					}

					NOTICE - MSG {
						set instr [instr [sbd:get @DISPATCH:RAW] " :"]
						incr instr 2
						set arg [mid [encoding convertto $encoding [sbd:get @DISPATCH:RAW]] $instr]
					}

					DCC {set arg [encoding convertto identity $arg]}

					default { # Do nothing #}

				}
			}
		}; # End of USEENCODING
	}
debug.print 2 "\[SB:DISPATCH\] 03"
	set dccnick [SB:dccnick $nick $handle $chan]
debug.print 2 "\[SB:DISPATCH\] 04"

	# Deal with possible changes to the params / data stream
	set arg [split [trim [singlespace $arg]]]
debug.print 2 "\[SB:DISPATCH\] 05"

	# Swap chan / cmd args if reversed by idiot user
	if [string match #?* [lindex $arg 0]] {
		set ch [lindex $arg 0]
		lset arg 0 [lindex $arg 1]
		lset arg 1 $ch
	}
debug.print 2 "\[SB:DISPATCH\] 06"

	# Directed channel?
	set target $chan
	if [string match #?* [lindex $arg 1]] {
		set target [lindex $arg 1]
		set arg [lreplace $arg 1 1]
	}
debug.print 2 "\[SB:DISPATCH\] 07:TARGET($target)"

	set old_command [lindex $arg 0]
	set abbr [SB:abbr check [lindex $arg 0]]
	if [notempty abbr] {lset arg 0 $abbr}
	set command [string tolower [join [lindex $arg 0]]]
debug.print 2 "\[SB:DISPATCH\] 10"

	# Process potential macros
	if ![string eq "" [info procs SB:macro_replace]] {
		set macro [SB:macro_replace $nick $host $handle $chan $arg]; # Must return "" if no match!
		if ![string eq "" $macro] {
			regsub -all \\\|\\\| $macro \001 macro
			foreach macro_mini [split $macro \001] {SB:dispatch $nick $host $handle $chan [join $macro_mini]}
			return 1; # Go no further
		}
	}

debug.print 2 "\[SB:DISPATCH\] 11"
	set chan_old $chan
	if [string eq # $target] {set target $chan}
	if [string eq ## $target] {set target [SB:home]}
	if [isempty target] {set target [SB:home]}
	# MOVED TO FURTHER DOWN -- set chan $target
debug.print 2 "\[SB:DISPATCH\] 12"

	# Store command in LASTCMDS before altering anything
	lassign [sbd:get LASTCMDS] lastcmds_offon lastcmds_logfile
	if [nozero $lastcmds_offon] {
		if [isempty lastcmds_logfile] {set lastcmds_logfile ${::nick}.command.log}
		catch {# Let's not keep doing disk accesses for every command ):
			set w [open ./logs/$lastcmds_logfile a]
			set lastcmds_arg $arg
			if [SB:commandlist hasflag $command -authcmd] {set lastcmds_arg "[l0 $arg] <password blocked>"}
			puts $w [list [now] [sbd:get @LAST_TRIGGER] $old_command [lindex $arg 0] [SB:commandlist iscmd [lindex $arg 0]] [SB:commandlist level [lindex $arg 0]] $chan_old [SB:isauthed $nick $handle $host] [SB:isoauthed $handle] $nick $host $handle $chan $lastcmds_arg]
			close $w
		}
	}

	# ALL OUTPUT MUST COME AFTER THIS LINE !!!!!
	if {!$override_output || [sbd:isempty @OUTPUT:$nick]} {sbdset @output:$nick [list $nick $host $handle $target]} ; # $CHAN -> $TARGET

debug.print 2 "\[SB:DISPATCH\] 13:COMMAND($command)"
	# Safe to do output now ("unable to find output data" error)
	if ![SB:commandlist iscmd $command] {
		if ![sbd:nozero CONFIG:GLOBAL-SNUFFERRORS] {print $nick "\"[string toupper $command]\" is not a registered command."}
		return 1
	}
debug.print 2 "\[SB:DISPATCH\] 14"

	set param [lrange $arg 1 end]
	set jparam [join $param]
	if [SB:check_cmdattr $command -authcmd] { set param "<password blocked>"; set jparam $param }
	SB:appendlogs $dccnick $host $handle $chan $param
	if [SB:cmdreport check $command $handle $chan] { msghome "\[CMDREPORT\] $dccnick ($handle) on $chan is using [string toupper $command]: $jparam" }
	if [SB:check_cmdattr $command -authcmd] { if ![SB:check_cmdattr $command -authcont] { return 0 } }
	if [SB:issuspended command $command $chan] {print $nick "The [string toupper $command] command is currently suspended from use [lindex [list # "on $chan" globally] [SB:issuspended command $command $chan]]."; return 1}
	if [SB:checkbot $handle] { return 1 }; # Bot's can not use commands! (Originally used as a bypass to the LOGIN check.)
	set highestlevel [SB:accesslevel $handle $chan higher]
debug.print 2 "\[SB:DISPATCH\] 15"

	# Why are DCC users exempt from level checks ?!
	set level_command [SB:commandlist level $command]
	if !$override {# If $OVERRIDE isn't set (default):
		if {$level_command > -1} {# If command is not 'anyone' command:
			if ![SB:check_cmdattr $command -loggedoutok] {
				if ![SB:isauthed $nick $handle $host] {
					if ![SB:isdccnick $nick] {print $nick "Who are you?"; return 1}
				}
			}
		}
		if {$level_command > $highestlevel} {print $nick "You don't have access to use the [string toupper $command] command."; return 1}
	}
debug.print 2 "\[SB:DISPATCH\] 16"

	if {$level_command >= 900} {
		set error_owner 1
		if [info exists ::owner] { if ![string eq "" $::owner] { set error_owner 0 } }
		if $error_owner {print $nick "Please note that the command you're using is for high ( 901 - 999 ) owners, or permanent bot owners ( =1000 ) only. Because no one is listed as a permenant bot owner (the \$OWNER variable in my config file), the [string toupper $command] command are out of reach of all users."; return 1}
	}

debug.print 2 "\[SB:DISPATCH\] 17"
	if {$level_command >= 1000} {
		if ![SB:isoauthed $handle] {
			if ![SB:check_cmdattr $command -bypassoauth] {printh $nick "You must OLOGIN (or XOLOGIN) before you may use =1000 commands."; return 1}
		}
	}
debug.print 2 "\[SB:DISPATCH\] 18"

	if [SB:issuspended user $handle $chan] { if ![SB:check_cmdattr $command -suspendedok] { print $nick "You're currently suspended."; return 1 } }
debug.print 2 "\[SB:DISPATCH\] 19"

	# User / channel processing complete

	# Is the channel valid [or is this check bypassed by -INVALIDOK]?
	if ![string eq "" $target] { if ![validchan $target] { if ![SB:check_cmdattr $command -invalidok] { print $nick "I do not monitor \002${target}\002."; return 0 } } }
	set chan $target
debug.print 2 "\[SB:DISPATCH\] 20"

	# Process any wedges (use raw $ARG data)
	foreach wedge [lsort -inc -uni -dict [sbdget @WEDGELIST]] { $wedge $nick $host $handle $chan $arg }

debug.print 2 "\[SB:DISPATCH\] 21"
	# Execute the command (finally!)
	set proc [SB:commandlist proc $command]
debug.print 2 "\[SB:DISPATCH\] 22"

	# Don't use CHECKFLAGS; pluck them out manually
	# --FLAG type flags will be reserved for SB:DISPATCH use only.
	# This will allow the mixing of flags freely.

	set dispatch_modes_list [list all force:nosave force:notice force:msg force:chan force:null force:mute force:strip force:burst force:nomute force:file time chan msg notice strip mute null nomute done chan:* nosave]
	array set dispatch_modes ""
	foreach dispatch_mode $dispatch_modes_list {
		set dispatch_modes([stl $dispatch_mode]) 0
		set m [lsearch -glob [stl $arg] [stl --$dispatch_mode]] ; # -EXACT (changed for: "--CHAN:*" & "-->[file name]")
		if {$m != -1} {
			set dispatch_modes([stl $dispatch_mode]) 1
			set n [lsearch -exact [list nosave chan msg notice strip null mute file nomute] $dispatch_mode]
			if { $n != -1 } {
				set dispatch_modes(force:${dispatch_mode}) 1 ; # End-brace
				set arg [lreplace $arg $m $m] ; # Was OUTSIDE of the IF { $N != -1 } condition
			} ; # Move this two lines up; this is making --FLAGS disappear even if this doesn't match anything
		}
	}

	# Special cases
	if { [lsearch -glob [stl $arg] [stl --CHAN:#*]] == -1 } {
		empty dispatch_targetchan
	} {
		set dispatch_targetchan [mid [lindex $arg $m] 8]
debug.print 2 "\[SB:DISPATCH\] 22.05: DISPATCH_TARGETCHAN($dispatch_targetchan):VALIDCHAN([validchan $dispatch_targetchan])"
		if ![validchan $dispatch_targetchan] {
			if ![SB:check_cmdattr $command -invalidok] {
				error "\[OUTPUT OVERRIDE FLAG ERROR\] Illegal channel for --CHAN: $dispatch_targetchan"
				return 0
			}
		}
		# MUST GO HERE! (otherwise, the space will create an invalid
		# channel name for the VALIDCHAN check above). :(
		prepend dispatch_targetchan [space]
		set dispatch_modes(force:chan) 1
		set arg [lreplace $arg $m $m]
	}

	if { [set m [lsearch -glob [stl $arg] -->?*]] != -1 } { # -->?* needed so "-->" by itself won't trigger output to a null-named file
		set dispatch_targetfile [mid [lindex $arg $m] 4]
		set dispatch_modes(force:file) 1
		sbd:set @OUTPUT:OVERRIDE:$nick FILE
		set arg [lreplace $arg $m $m]
	}

debug.print 2 "\[SB:DISPATCH\] 22.1"

	# Wedge in file data as params / args?
	set insert_args [lsearch -all -inline -glob $arg --file:*]
	foreach insert_arg $insert_args {
		set m [lsearch -exact $arg $insert_arg]
		empty data
		if { $m > 0 } {
			set file [mid $insert_arg 8]
			if [file exists $file] {
				set data [fileread $file "" 0]

				# Must be FROM then TO
				set n [lsearch -glob $arg --encfrom:*]
				if { $n > -1 } {
					set from [lindex $arg $n]
					set arg [lreplace $arg $n $n]
					set data [encoding convertfrom [mid $from 11] $data]
				}

				set n [lsearch -glob $arg --encto:*]
				if { $n > -1 } {
					set from [lindex $arg $n]
					set arg [lreplace $arg $n $n]
					set data [encoding convertto [mid $from 9] $data]
				}
			}
		}
		set arg [lreplace $arg $m $m $data]
	}
	
debug.print 2 "\[SB:DISPATCH\] 22.12"
	# Check line length -- is the buffer full? Tests prove: 509 + CRLF [511 total; we only need to deal with the 509]
	# Check this BEFORE introducing --FILE:<filename> which is used to go beyond the IRCd line limit of 509
	# Note: we might see:  PRIVMSG (PUB/PUBM), NOTICE (NOTC), or neither (DCC). Let's ignore DCC for now,
	# because DCC connections might be raw TELNET, which can have [almost] endless lines. For now,
	# let's round-off to 500 characters
	if ![SB:isdccnick $nick] { # Ignore DCC / raw TELNET
		set len [string length "${nick}!$host PRIVMSG $chan :$arg"] ; # If NOTICE, this will be of fby one
		if { $len >= 500 } { print $nick "\[SB6\] It appears that your text is at (or above) the IRCd limit of 509 characters (not counting the terminal CRLF). Your command was truncated at: \"[color 4]... [right [join $arg] 25][color]\"" }
	}
	
debug.print 2 "\[SB:DISPATCH\] 22.15"
	# Run command on all channels?
	if $dispatch_modes(all) {foreach ch [channels] {SB:dispatch $nick $host $handle $chan [join $arg] $override}; return 0}
debug.print 2 "\[SB:DISPATCH\] 22.2"

	if [validuser $handle] {
		array set modes_override ""
		set list [list force:nosave force:file force:notice force:msg force:chan force:strip force:burst force:nomute force:mute force:null]
		foreach item [array names dispatch_modes] {
			if $dispatch_modes($item) {
				# Any additions to this list, (1) change below, (2) change SB:SETUP in SB6_LOADER.TCL
				switch -exact -- $item {
					force:notice { set modes_override($item) [getdata $handle OUTPUT] ; setdata $handle DISPATCH:@TEMP:OUTPUT $modes_override($item) ; setdata $handle OUTPUT NOTICE }
					force:msg { set modes_override($item) [getdata $handle OUTPUT] ; setdata $handle DISPATCH:@TEMP:OUTPUT $modes_override($item) ; setdata $handle OUTPUT MSG }
					force:chan { set modes_override($item) [getdata $handle OUTPUT] ; setdata $handle DISPATCH:@TEMP:OUTPUT $modes_override($item) ; setdata $handle OUTPUT CHAN${dispatch_targetchan} }
					force:strip { set modes_override($item) [getdata $handle COLOR] ; setdata $handle DISPATCH:@TEMP:COLOR $modes_override($item) ; setdata $handle COLOR STRIP }
					force:burst { set modes_override($item) [getdata $handle BURSTMODE] ; setdata $handle DISPATCH:@TEMP:BURSTMODE $modes_override($item) ; setdata $handle BURSTMODE 1 }
					force:null { set modes_override($item) [getdata $handle OUTPUT] ; setdata $handle DISPATCH:@TEMP:OUTPUT $modes_override($item) ; setdata $handle OUTPUT NULL }
					force:mute { set modes_override($item) [getdata $handle OUTPUT] ; setdata $handle DISPATCH:@TEMP:OUTPUT $modes_override($item) ; setdata $handle OUTPUT MUTE }
					force:nomute { set modes_override($item) [SB:ccs $chan +mute] ; channel set $chan -mute}
					force:file { set modes_override($item) [getdata $handle OUTPUT] ; setdata $handle DISPATCH:@TEMP:OUTPUT $modes_override($item) ; setdata $handle OUTPUT FILE }
					force:nosave { sbd:set @NEEDSAVE:NOSAVE 1 }
				}
			}
		}
	}

	# Update counters
	# <Nope: handled by a WEDGE>

	# Use $ARG here!
debug.print 2 "\[SB:DISPATCH\] 22.4: Dispatch command!"
	set command_start [clock clicks -milli]
	set error [ catch { $proc $nick $host $handle $chan $arg } ohshit ]
	set command_end [clock clicks -milli]
debug.print 2 "\[SB:DISPATCH\] 22.5"
	if [ catch { set stack [info errorstack] } ] { empty stack }
	set error_frame [ catch { set frame_count [info frame] } ]
	if $error_frame { empty frame ; zero frame_count } { for { set x $frame_count } { $x > 0 } { decr x } { lappend frame [info frame $x] } }
	#if {[lsearch -exact [string tolower [lindex $arg 1]] help] > -1} {printh $nick "\[[SB:commandlist src $command] - level [SB:commandlist level $command] - [none [jstl [SB:abbr get $command]] "<no abbreviations>"]\]"}
	if [string eq -nocase HELP [l1 $arg]] { printh $nick "\[[SB:commandlist src $command] - level [SB:commandlist level $command] - [none [jstl [SB:abbr get $command]] "<no abbreviations>"]\]" }
debug.print 2 "\[SB:DISPATCH\] 23"
	if [validuser $handle] {
		foreach item [array names dispatch_modes] {
			if $dispatch_modes($item) {
				switch -exact -- $item {
					force:notice { setdata $handle OUTPUT $modes_override($item) ; setdata $handle DISPATCH:@TEMP:OUTPUT "" }
					force:msg { setdata $handle OUTPUT $modes_override($item) ; setdata $handle DISPATCH:@TEMP:OUTPUT "" }
					force:chan { setdata $handle OUTPUT $modes_override($item) ; setdata $handle DISPATCH:@TEMP:OUTPUT "" }
					force:strip { setdata $handle COLOR $modes_override($item) ; setdata $handle DISPATCH:@TEMP:COLOR "" }
					force:burst { setdata $handle BURSTMODE $modes_override($item) ; setdata $handle DISPATCH:@TEMP:BURSTMODE "" }
					force:null { setdata $handle OUTPUT $modes_override($item) ; setdata $handle DISPATCH:@TEMP:OUTPUT "" }
					force:mute { setdata $handle OUTPUT $modes_override($item) ; setdata $handle DISPATCH:@TEMP:OUTPUT "" }
					force:nomute { channel set $chan [lindex [list - +] $modes_override($item)]mute }
					force:file {
						# The --FORCE:FILE flag will be caught / processed by the target bot; the remote bot will never see the flag
						if ![string eq -nocase REMOTE [sbd:get @LAST_TRIGGER]] {
							# 2015-05-08 18:20:00 -0700: Any errors like "Error in HELP: 0" mean there was an error in the CATCH below
							# The CATCH below is required to be sure the user's OUTPUT choice is restored in case of error
							set error [ catch {
								set method2 [join $arg]
								set id "[clock seconds] ${nick}!${host}\$${handle}:${chan}"
								switch -exact -- [string tolower [sbd:get @LAST_TRIGGER]] {
									chan { if [string eq -nocase $::lastbind [sbd:join config:global-pubcommand:trigger]] { set method "$::lastbind" } { set method "$::lastbind " ; # Space! } }
									msg { set method "/msg $::botnick " } ; # Space!
									notice { set method "/notice $::botnick " } ; # Space!
									dcc { set method "." ; set id "[clock seconds] ${nick}!#IDX@${handle}#:$chan" ; # Nick!ident@host.domain:#chan (so it can be split / REGEXPed) }
									remote { set method "<${handle}@[lindex [split $nick @] 1]> " ; # Space! }
									default { set method "=METHOD:UNKNOWN= " ; # Space!  }
								}
#debug dispatch_targetfile id method method2
								if [left $dispatch_targetfile 1 >] {
									set dispatch_targetfile [mid $dispatch_targetfile 2]
									set file_mode a+
								} {
									set file_mode w+
								}
								set w [open $dispatch_targetfile $file_mode]
								puts $w "# [formattime [now] $handle]:"
								puts $w "# $id"
								puts $w "# $method$method2"
								puts $w ""
								foreach a [sbd:get @dispatch:force:file:$nick] { puts $w [stripcodes bcruag $a] }
								puts $w ""
								flush $w
								close $w
								sbd:set @dispatch:force:file:$nick
							} crap ]
							if $error { print $nick "\[SB6: --FORCE:FILE\] Error while writing file (\"${dispatch_targetfile}\"): $crap" }
log "\[SB6:DISPATCH --FORCE:FILE\] Restoring OUTPUT mode for ${nick}: $modes_override($item)"
							sbd:set @OUTPUT:OVERRIDE:$nick
							setdata $handle OUTPUT $modes_override($item)
							setdata $handle DISPATCH:@TEMP:OUTPUT ""
							print $nick "\[SB6\] Output saved to: $dispatch_targetfile"
						}
					}
					force:nosave { #sbd:set @NEEDSAVE:NOSAVE ; # No! The saves are TIMERED! Let SB:NEEDSAVE_TIMER handle it }
				}
			}
		}
	}
	if $dispatch_modes(done) { print $nick "\[SB6: [stu $command]\] Done." }
debug.print 2 "\[SB:DISPATCH\] 23.5"

	if $error {# Use $PARAM here (hidden PW within an -AUTHCMD command)
		print $nick "\002\[SB6 FAIL\]\002 An error occurred while using the \037\002[string toupper $command]\002\037 command: $ohshit"
		if ![nozero [sbdget CONFIG:GLOBAL-SNUFFDEBUG]] {
			print $nick "Please include the following information in your bug report (\002\037unreported bugs don't get fixed\037\002): \002[none [sbdget ${command}:AUTHOR] http://bugs.stormbot.org/]\002"
			print $nick "debug:6 begin"
			print $nick "debug:6 CMD:$command $param"
			print $nick "debug:6 SBV:$::stormbot(version) - WWW:$::stormbot(distro) - N:$::nick - BN:${::botnet-nick} - AN:$::altnick - BH:[nick2hand $::botnick $chan] - EDV:[join $::version ,] - EDVN:$::numversion NL:${::nick-len} - HL:$::handlen - TCLV:[info tclversion] - TCLP:[info patchlevel] - SYS:$::tcl_platform(os)_$::tcl_platform(osVersion)\[$::tcl_platform(platform)\] - LIB:$::tcl_library - PWD:[pwd]"
			empty confargs ; foreach - [list "" -] { if [info exists ::configure${-}args] { set confargs [set ::configure${-}args] } }
			print $nick "debug:6 EDC:[encrypt . [trim [regsub -all -- {' '} $confargs " "] ']]"
			print $nick "debug:6 NH:${nick}!${host} - H:$handle - SC:${::server}|[SB:getopsymbol $nick $chan]$chan) - AUTHED:[noyes [SB:isauthed $nick $handle $host]] - G/L:[SB:accesslevel $handle]/[SB:accesslevel $handle $chan] - LASTTRIGGER:[sbdget @LAST_TRIGGER] - LASTBIND:$::lastbind - CMDSUSP:[SB:issuspended cmd $command $chan] - USERSUSP:[SB:issuspended user $handle $chan] - CMDINFO:\[[SB:commandlist info $command]\]"
			print $nick "debug:6 end"
		}
		if [sbd:nozero CONFIG:GLOBAL-ERRORDEBUG] {
			foreach { a b } [list ERROR error NERROR nerror CODE errorCode INFO errorInfo] {
				if [info exists ::$b] { print $nick "[stu $a]: [set ::$b]" }
			}
			if [notempty stack] { print $nick "Stack trace ([plural -comma -show entry [ expr [ llength $stack ] / 2 ]]):" ; zero c ; foreach { a s } $stack { incr c; print $nick "[space 5]Entry #[comma $c]: \[${a}\] $s" } ; #print $nick "ES:\{${stack}\}" }
			if [notempty frame] { print $nick "Frame trace ([plural -comma -show frame [l0 $frame_count]]):" ; zero c ; foreach f [lrange $frame 0 end] { incr c ; print $nick "[space 5]Frame #[comma $c]: $f" } ; #print $nick "FR:\{${frame}\}" }
		}
		if ![string eq "" [set context_mode [sbd:get CONFIG:GLOBAL-ERRORCONTEXT]]] {
			set context [sbd:get @CONTEXT]
			if [notempty context] {
				switch -exact -nocase -- $context_mode {
					* - all { # Force won't get here; this is within $ERROR
							print $nick "Context captured:"
							foreach _ $context { print $nick "[space 5]$_" }
					}
					end - last {
						print $nick "\[CONTEXT\] [lindex $context end]"
					}
				}
				# CONTEXT:FETCH?
				sbd:set @CONTEXT ""
			}			
		}
		sbd:set @LAST_TRIGGER
		return 0
	}
	# If not an error, check: CONTEXT FORCE?
	if ![string eq "" [set context [sbd:get @CONTEXT]]] {
		if [string eq -nocase FORCE [sbd:get CONFIG:GLOBAL-ERRORCONTEXT]] {
			print $nick "Context captured:"
			foreach _ $context { print $nick "[space 5]$_" }
		}
		# CONTEXT:FETCH?
		sbd:set @CONTEXT
	}
debug.print 2 "\[SB:DISPATCH\] 24"
	sbd:set @LAST_TRIGGER

	if $dispatch_modes(time) {
		set time_report [ expr $command_end - $command_start ]
		if {$time_report > 1000} {set time_report [timeval +${time_report}l s]; set time_report_value s} {set time_report_value ms}
		printh $nick "\[[stu $command]\] Command execution time: ${time_report}${time_report_value}"
	}
debug.print 2 "\[SB:DISPATCH\] 25"

	set ::HelpSniffer(on) 0
	return 0 ; # Don't record command usage this way.
}

proc SB:check_cmdattr {command checkme} {
	set flags [list -invalidok -loggedoutok -suspendedok -authcmd -authcontinue -chanspecific -helpcmd -locklevel -minlevel -maxlevel -bindmsg * * -bypassoauth -corecommand -nobind]

	set flag [uniquematch $flags $checkme]
	if [string eq "" $flag] {error "\[SB:CHECK_CMDATTR\] illegal flag: \002${checkme}\002. Valid flags: [regsub -all -- "\\\* " $flags ""]"}
	set m [lsearch -glob [string tolower $::stormbot(@cmdinfo:commandlist)] [list [string tolower $command] *]]
	set n [lsearch -exact [string tolower $flags] [string tolower $flag]]

	set num [ expr int( pow( 2 , $n ) ) ]
	set flagval [lindex [lindex $::stormbot(@cmdinfo:commandlist) $m] 2]

debug.print 2 "\[SB:CHECK_CMDATTR\] M($m):N($n):FLAGVAL($flagval):NUM($num):POW([ expr pow( 2 , $n ) ]):"
debug.print 2 "\[SB:CHECK_CMDATTR\] AND([ expr int( pow( 2 , $n ) ) & $num ]))"

	if {$m < 0} {return 0}

	# --- Special case: -NONE ---
	if [string eq -nocase -none $flag] {return [string eq 0 $flagval]}

	if {$n < 0} {return 0}

	if [ expr $flagval & $num ] {return 1}
	return 0
}

proc SB:cmdreport {mode command handle chan} {
	set cmdreport [sbd:get CMDREPORT]
	if [isempty cmdreport] {return 0}

	switch -exact -- [string tolower $mode] {
		check {
			set level [SB:accesslevel $handle $chan]
			foreach entry $cmdreport {
				if [string eq -nocase $command [lindex $entry 0]] {
					if [iswithin [lindex $entry 1] $level] {return 1}
				}
			}
			return 0
		}
		default {msghome "\[SB:CMDREPORT\] unknown MODE($mode):COMMAND($command)"}
	}
	return 0
}

proc SB:dispatch_chan { nick host handle chan arg } {
	if ![sbd:nozero CONFIG:GLOBAL-BIND:CHAN] { return 0 }
	SB:dispatch $nick $host $handle $chan $arg
	return 1
}

proc SB:dispatch_pubcmd {nick host handle chan arg} {
	# This test should always fail all available options by default
	if ![sbd:nozero CONFIG:GLOBAL-BIND:PUBCMD] { return 0 }

	set arg [string trimleft $arg]
	sbd:var CONFIG:GLOBAL-PUBCOMMAND offon
	sbd:var CONFIG:GLOBAL-PUBCOMMAND:TRIGGER pubcmd
	set pubcmd [join $pubcmd]
	false do
	if [boolean $offon] {
		if [notempty pubcmd] {
			if [left [stl $arg] [len $pubcmd] [stl $pubcmd]] { true do }
		}
	}

	if !$do { # Let's check user-specific triggering
		if [validuser [set check_handle [SB:authhandle $nick]]] {
			lassign [list [sbd:get CONFIG:${check_handle}-pubcommand] [sbd:get CONFIG:${check_handle}-pubcommand:trigger]] offon pubcmd
			if [boolean $offon] {
				if [notempty pubcmd] {
					if [left [stl $arg] [len $pubcmd] [stl $pubcmd]] { true do }
				}
			}
		}
	}

	if !$do { return 0 }
	set ::lastbind $pubcmd
	sbd:set @LAST_TRIGGER chan
	SB:dispatch $nick $host $handle $chan [string range $arg [string length $pubcmd] end]
	sbd:set @LAST_TRIGGER
	return 0
}

proc SB:dispatch_msg {nick host handle arg { method msg } } {
	if ![sbd:nozero CONFIG:GLOBAL-BIND:MSG] { return 0 }
	set command $::lastbind
	set arg_original $arg
	set arg [split $arg]

	if [isvalidchantype [lindex $arg 0]] {
		set chan [lindex $arg 0]
		set arg [join [lrange $arg 1 end]]
	} else {
		if [SB:check_cmdattr $command -chanspec] {print $nick "You need a channel name when using a /[stu $method] command."; return 0}
		set arg [join $arg]
	}
	if [isempty chan] {set chan [home]}
	sbd:set @LAST_TRIGGER $method
	SB:dispatch $nick $host $handle $chan "$command $arg"
	sbd:set @LAST_TRIGGER
	return 0
}

proc SB:dispatch_notice { nick host handle arg { target "" } } {
	if ![sbd:nozero CONFIG:GLOBAL-BIND:NOTICE] { return 0 }
	if { ! ( [isbotnick $target] || [botonchan $target] ) } { return 0 }
	set arg [join [lassign [split $arg] 0]]
	if ![SB:commandlist iscmd $0] { return 0 }
	set ::lastbind $0
	# /msg dispatcher looks for the command word in $::LASTBIND and all OTHER text in $ARG
	SB:dispatch_msg $nick $host $handle $arg notice
	return 0
}

proc SB:dispatch_dcc {handle idx arg} {
	if ![sbd:nozero CONFIG:GLOBAL-BIND:DCC] { return 0 }
	set command $::lastbind
	set arg_original $arg
	set arg [split $arg]
	set chan [lindex [console $idx] 0]
	set chan_directed [lindex $arg 0]
	if [isvalidchantype $chan_directed] {
		set chan $chan_directed
		set arg [lrange $arg 1 end]
	}
	set arg [join $arg]
	sbd:set @LAST_TRIGGER dcc
#putserv "PRIVMSG #BOTS :\[DEBUG\] handle($handle):idx($idx):arg($arg):chan($chan):chan_directed($chan_directed):command($command)" -next
	SB:dispatch "#DCC $idx" #DCCIDX:${idx}# $handle $chan "$command $arg"
	sbd:set @LAST_TRIGGER
	return 0
}

proc SB:dispatch_raw {1 2 3} { sbd:set @DISPATCH:RAW $3 ; return 0 }

proc SB:pub_rescue {nick host handle chan arg} {
	set arg [split $arg]
	if ![string eq -nocase $::botnick [lindex $arg 0]] {return 0}
	switch -exact -- [string tolower [lindex $arg 1]] {rehash rehash restart restart die die default #}
	return 0
}

# Bootstrap:
bind raw  - PRIVMSG SB:dispatch_raw
bind pub  t #@!     SB:pub_rescue
bind pubm - *       SB:dispatch_pubcmd
catch { unbind pubm - #* SB:dispatch_pubcmd }

}
#</disp>

#<TAIL>
proc sb6:bootstrap:sb6_tail.tcl args {
# CATCH returns: 0 (TCL_OK), 1 (TCL_ERROR), 2 (TCL_RETURN), 3 (TCL_BREAK), & 4 (TCL_CONTINUE).

proc = args { nozero [ fixmath $args ] }
proc == args { nozero [ expr $args ] }

if [string eq "" [info commands _unknown]] {
	## This was making TCL not load MSGCAT (and not find packages at all, no matter what)
	##if ![string eq "" [info procs unknown]] { rename unknown _unknown }
	#%# Because, you idiot, the original UNKNOWN is *NOT* a PROC: it's intrinsic!
	#%# Use "info commands" instead !!!!!
	#catch { rename unknown _unknown }
	
	# 2018-10-18 11:32:00 -0700: Just don't do it; too may problems
	#if ![string eq "" [info commands unknown]] { rename unknown _unknown }
	catch { rename unknown _unknown }
}

proc unknown args {
	# Can not use variables that already exist as command words: TCL will dispatch
	# to that command; UNKNOWN will never see it.

	# We will =IGNORE= the difference between Var++ and ++Var (which reports the value before/after the increment, respectively)
	if ![regexp -nocase -- {^[[:space:]]*([0-9A-Z\(\)_\-]+)[[:space:]]*(\+\+|\-\-|[\-\+\*\/%\^\~\=]?=|<<=|>>=|\*\*=)[[:space:]]*(.*)$} $args . variable type value] {
		return [uplevel 1 eval _unknown $args]
	}
	upvar 1 $variable local
	switch -exact -- $type {
		=   { set local $value ; return $local }
		==  { set local [ expr $value ] ; return $local }
		+=  { set local [ expr { $local + $value } ] ; return $local }
		-=  { set local [ expr { $local - $value } ] ; return $local }
		*=  { set local [ expr { $local * $value } ] ; return $local }
		/=  { set local [ expr { $local / $value } ] ; return $local }
		%=  { set local [ expr { $local % $value } ] ; return $local }
		~=  { set local [ expr { ~$value } ] ; return $local }
		++  { incr local  1 ; return $local }
		--  { incr local -1 ; return $local }
		<<= { set local [ expr { $local << $value } ] ; return $local }
		>>= { set local [ expr { $local >> $value } ] ; return $local }
		^= - **= { if { [ catch { expr 1 ** 1 } ] || ( $value < 0 ) } { set local [ expr pow( {$local} , {$value} ) ] } { set local [ expr { $local ** $value } ] } ; return $local }
	}

	# Legitimate error. Let's log it first ....
	log d "\[UNKNOWN\] \$ARGS($args)"
	set level [info level]
	log d "\[UNKNOWN\] Level $level (stack trace):"
	incr level -1
	set fl [len [comma $level]]
	for { set l 1 } { $l <= $level } { incr l } {
		log "[string repeat [space 5] $l][format %0${fl}s [comma $l]]> [info level $l]"
	}

	uplevel 1 eval _unknown $args
}

proc SB:iseggcorecmd cmd {
	if [string eq "" [info commands $cmd]] {return 0}
	set e [ catch { info body $cmd } s ]
	if $e {return 1}
	return 0
}

proc SB:security args {
	# Eggdrop core has determined that this needs fixing; the 1.6.22 & 1.8 releases after this date will remove these.
	# Eggheads Eggdrop Trac: #137 (2015-12-31)
	foreach var [info vars ::_*] { if [regexp -nocase -- {^(\:\:)?_[a-z]+[1-9]$} $var] { set $var "" } }
	return ""
}

# --- StormBot.TCL data manipulation ---

# Enter SBDKILL with: SBDKILL chanmode ${chan}:*

# Base pair:
interp alias {} sbd:isnull {} sbd:isempty
interp alias {} sbd:eq     {} sbd:equal
interp alias {} sbd:ne     {} sbd:notequal

# Core StormBot  settings (only assigned by [TAIL:]BOOTSTRAP; otherwise READ-ONLY
proc stormbot args {
	lassign $args cmd name value
	lower cmd name
	if [string eq -nocase SET $cmd] { array set ::stormbot_config [list $name $value] ; if [string eq "" $value] { unset ::stormbot_config($name) } ; return $value }
	return [var ::stormbot_config($cmd)]
}

# ALL data-setting commands =MUST= pipe to SBD:SET to set the data! (AUTOSEND)
proc sbd:get      a             {set a [string tolower $a]; if [info exists ::stormbot($a)] { return $::stormbot($a) } { return "" } }
proc sbd:set      {a { b "" } { q "" } } {set a [string tolower $a]; set ::stormbot($a) $b; if [isempty q] {sbd:set:wedge $a $b} ; if [string eq -nocase DEBUG $q] { empty _ ; for { set c [ expr [info frame] - 1 ] } { $c > 0 } { incr c -1 } { lappend _ [info frame $c] } ; sbd:lappend #SBDDATA:SET:LOG $_ } ; if [string eq "" $b] {unset ::stormbot($a)}; return $b}

# Can we "create" commands by using a loop through a list of commands, such as:
# proc sbd:${command} [list name args] {
# lassign $args 0 1 2 3 4 5 ....
# set data [sbd:get $name]
# eval $command (command args)
# sbd:set ? (not always appropriate)
# return $data
# }

proc sbd:getall   args             { lappend final ; foreach arg $args { lappend final [sbd:get $arg] } ; return $final }
proc sbd:var      { a b { c "" } } { if [string match ::* $b] { uplevel #0 [string range $b 2 end] local } { upvar 1 $b local } ; set local [sbd:get $a] ; if { [string eq "" $local] } { set local $c } ; return $local }
proc sbd:stl      a                {stl [sbd:get $a]}
proc sbd:stu      a                {stu [sbd:get $a]}
proc sbd:append   {a args}         { set c [sbd:get $a] ; foreach b $args {   append c $b } ; sbd:set $a $c }
proc sbd:lappend  {a args}         { set c [sbd:get $a] ; foreach b $args {  lappend c $b } ; sbd:set $a $c }
proc sbd:prepend  {a args}         { set c [sbd:get $a] ; foreach b $args {  prepend c $b } ; sbd:set $a $c }
proc sbd:lprepend {a args}         { set c [sbd:get $a] ; foreach b $args { lprepend c $b } ; sbd:set $a $c }
proc sbd:linsert  {a b c}          {set d [sbd:get $a]; set d [linsert $d $b $c]; sbdset $a $d}
proc sbd:join     {a { b " " } }   {join [sbd:get $a] $b}
proc sbd:split    {a { b " " } }   {split [sbd:get $a] $b}
proc sbd:concat   args             { if [isempty args] return ; lassign $args name new ; sbd:set $name [concat [sbd:get $name] $new] }	
proc sbd:merge    [list a b]       { sbd:append $a [sbd:get $b] }
proc sbd:lmerge   [list a b]       { sbd:concat $a [sbd:get $b] }
proc sbd:conjoin  {a b}            {set data [sbd:get $a] ; foreach null $b { lappend data $null } ; sbd:set $a $data}
proc sbd:lset     {a b { c "" } }  {set data [SB:padindex [sbd:get $a] [ expr $b + 1 ]]; lset data $b $c; sbdset $a $data}
proc sbd:find     a                {lsort -inc -uni -dict [array names ::stormbot [string tolower $a]]}
proc sbd:list     a                {set list [sbd:find $a]; if [string eq "" $list] {return ""}; foreach b [lsort -inc -uni -dict $list] {lappend o $b [sbd:get $b]}; return $o}
proc sbd:show     a                {array set sbdshow [sbd:list $a]; set max [getlongest [array names sbdshow]]; empty result; foreach item [lsort -inc -uni -dict [array names sbdshow]] {lappend result "[using A>*$max $item]: $sbdshow($item)"}; return [join $result \n]}
proc sbd:find-m   a                {empty list; foreach z [array names ::stormbot] { if [string match -nocase $a $z] {lappend list $z} } ; return $list }
proc sbd:find-r   a                {empty list; foreach z [array names ::stormbot] { if [regexp -nocase -- $a $z] {lappend list $z} } ; return $list }
proc sbd:isempty  a                {string eq "" [sbd:get $a]}
proc sbd:notempty a                {not [sbd:isempty $a]}
proc sbd:nozero   a                {nozero [sbd:get $a]}
proc sbd:names    {a { b "" } }    {array set cow [sbd:get $a]; set matchme [none $b *]; lmatch -all -nocase -glob [array names cow] $matchme}
proc sbd:kill     {a { b "" } }    {if [isempty b] {sbd:set $a; return ""}; foreach findme [sbd:find ${a}($b)] {sbdset $findme}; array set null [sbd:get $a]; foreach findme [array names null $b] {unset null($findme); sbdset $a [array get null]}; return ""}
proc sbd:cleanup  args             {foreach a [array names ::stormbot] { if [string eq "" $::stormbot($a)] { unset ::stormbot($a) } }; return ""}
proc sbd:match    args             {if [isempty args] {return 0}; zero nocase; if [string eq -nocase [l0 $args] -NOCASE] {one nocase; set args [lreplace $args 0 0]}; if ![sttring eq "" [l2 $args]] {error "wrong # args: should be \"sbdmatch ?-nocase? data_element pattern\""}; set sbdget [sbd:get [l0 $args]]; set match [l1 $args]; if $nocase { return [string match -nocase $match $sbdget] } { return [string match $match $sbdget] } }
proc sbd:equal    args             {if [isempty args] {return 0}; zero nocase; if [string eq -nocase [l0 $args] -NOCASE] {one nocase; set args [lreplace $args 0 0]}; if ![string eq "" [l2 $args]] {error "wrong # args: should be \"sbdequal ?-nocase? data_element string\""}; set sbdget [sbd:get [l0 $args]]; set match [l1 $args]; if $nocase { return [string eq -nocase $match $sbdget] } { return [string eq $match $sbdget] } }
proc sbd:notequal args             {if [isempty args] {return 0}; zero nocase; if [string eq -nocase [l0 $args] -NOCASE] {one nocase; set args [lreplace $args 0 0]}; if ![string eq "" [l2 $args]] {error "wrong # args: should be \"sbdequal ?-nocase? data_element string\""}; set sbdget [sbd:get [l0 $args]]; set match [l1 $args]; if $nocase { return [string eq -nocase $match $sbdget] } { return [not [string eq $match $sbdget]] } }
proc sbd:len      a                {return [string length [sbd:get $a]]}
proc sbd:llength  a                {return [llength [sbd:get $a]]}
proc sbd:lindex   {a b}            {return [lindex [sbd:get $a] $b]}
proc sbd:lrange   {a b c}          {return [lrange [sbd:get $a] $b $c]}
proc sbd:lremove  {a b}            {set c [sbd:get $a] ; lremove c $b; sbd:set $a $c}
proc sbd:lreverse a                {lreverse [sbd:get $a]}
proc sbd:lsort    a                {lsort -unique -dictionary -increasing [sbd:get $a]}
proc sbd:ajl      a                {ajl [sbd:get $a]}
proc sbd:resetdef {a b}            {set data [sbd:get $a] ; set data [SB:reset_defaults $data $b]; return $data}
proc sbd:scan     a                {empty scan; foreach find [sbd:find *] { if [string match -nocase $a [ sbd:get $find ] ] { lappend scan $find } }; return $scan}
proc sbd:rename   {a b}            {sbd:set $b [sbd:get $a]; sbd:set $a}
proc sbd:clear    a                {zero counter ; foreach b [sbd:find $a] { incr counter ; sbd:set $b } ; return $counter }
proc sbd:empty    a                {sbd:set $a}
proc sbd:negone   a                {sbd:set $a -1}
proc sbd:zero     a                {sbd:set $a  0}
proc sbd:one      a                {sbd:set $a  1}
proc sbd:incr     {a { b 1 } }     {set val [sbd:nozero $a] ; incr val $b ; sbd:set $a $val}
proc sbd:decr     {a { b 1 } }     {set val [sbd:nozero $a] ; decr val $b ; sbd:set $a $val}
proc sbd:exists   a                {sbd:notempty $a}
proc sbd:encrypt  {enc a b}        {sbd:set $a [encrypt $enc $b]}
proc sbd:decrypt  {enc a}          {decrypt $enc [sbd:get $a]}
proc sbd:lreplace {a b { c "" } { d "" } } {set e [sbd:get $a]; if [isempty c] {set c $b}; if [isempty d] {set e [lreplace $e $b $c]} {set e [lreplace $e $b $c $d]}; sbd:set $a $e}
proc sbd:boolean  a                { boolean [sbd:get $a] }
proc sbd:none     { a b }          { none [sbd:get $a] $b }

proc sbd:enc-get    {enc a}             {sbd:decrypt $enc $a}
proc sbd:enc-set    {enc a { b "" } }   {if [isempty b] {sbd:set $a} { sbd:encrypt $enc $a $b } }
proc sbd:enc-lset   {enc a b { c "" } } {if [isempty c] {sbd:lset $a $b $c} {set data [sbd:decrypt $enc $a]; lset data $b $c; sbd:encrypt $enc $a $data } }
proc sbd:enc-lindex {enc a b}           {set data [sbd:decrypt $enc $a]; return [lindex $data $b]}
proc sbd:enc-lrange {enc a b c}         {set data [sbd:decrypt $enc $a]; return [lrange $data $b $c]}
proc sbd:enc-match  args                {set enc [l0 $args]; set args [lreplace $args 0 0]; if [isempty args] {return 0}; zero nocase; if [string eq -nocase [l0 $args] -NOCASE] {one nocase; set args [lreplace $args 0 0]}; if ![string eq "" [l2 $args]] {error "wrong # args: should be \"sbdmatch ?-nocase? data_element pattern\""}; set sbdget [sbd:decrypt $enc [l0 $args]]; set match [l1 $args]; if $nocase { return [string match -nocase $match $sbdget] } { return [string match $match $sbdget] } }

proc sbd:array args {
	set cmd [l0 $args]
	set array [stl [jl1 $args]]
	set args [lreplace $args 0 1]
	array set temp [sbd:get $array]
	switch -exact -- [stu $cmd] {

		LIST { empty list ; foreach a $args { if [info exists temp($a)] { lappend list $temp($a) } { error "\[SBD:ARRAY LIST\] No such element: $a" } } ; return $list }

		FETCH { return [l1 [array get temp [l0 $args]]] }

		SET {
			foreach [list name value] $args { set temp([stl $name]) $value }
			sbd:set $array [array get temp]
			SB:needsave data
			return $value
		}
		
		GET {
			lassign [stl $args] name ; # Remember: indeces 0 & 1 were obliterated above
			if [isempty name] {
				return [array get temp]
			} {
				if ![info exists temp($name)] return
				return $temp($name)
			}
		}

		default {
			set r [eval array $cmd temp $args]
			sbd:set $array [array get temp]
			return $r
		}

	}
}

proc sbd:loop args {
	if [string eq -nocase TEST $args] { set args {-footer {puts FOOTER!} -header {puts HEADER!} -header {puts HEADER2!} -footer {puts FOOTER2!} a servers b servers { puts "Server: $a ($b)"}} }
	set params [list -header 1 -footer 1 -debug 0]
	flags:params -force $args $params text flags
	set header $flags(-header)
	set footer $flags(-footer)
	set code [lindex $text end]
	set variables [lrange $text 0 end-1]
	
	set vars ""
	foreach [list a b] $variables { append vars [iff [notempty vars] " "] "$a \[sbd:get $b\]" }
if $flags(-debug) { debug header }
if $flags(-debug) { debug flags }
if $flags(-debug) { debug text  }
if $flags(-debug) { debug variables }
if $flags(-debug) { debug vars }
if $flags(-debug) { debug code }
if $flags(-debug) { debug footer }
	eval {
		eval foreach _ [list $header] \{ eval \$_ \}
		eval foreach $vars \{ eval \$code \}
		eval foreach _ [list $footer] \{ eval \$_ \}
	}
	return
}

proc sbd:set:wedge {a b} {
	if ![sbd:nozero @SCRIPT_IS_LOADED] {return 0}
	set procs [sbd:set @SBD:SET:WEDGE [lsort -inc -uni -dict [sbd:get @SBD:SET:WEDGE]] quiet] ; # Endless loop otherwise!
	foreach proc $procs {
		if [validcmd $proc] { $proc $a $b }
		# (: TCL (p)roc(k)s! :)
	}
}

set sbdloop ""; foreach sbdloop [info procs sbd:*] {interp alias {} [regsub : $sbdloop ""] {} $sbdloop}; unset sbdloop

proc iscmdrebound { { command "" } } {
	# If specific command given, return boolean: is reound
	# Otherwise, return boolean: are collided commands Eggdrop core-rebound?
	# English grammar sucks: TRUE = Eggdrop Core command rebound to "./command"
	if [notempty command] { 
		# Was this command collided (and, if so, rebound)?
		if ![SB:commandlist iscmd $command] { return 0 } ; # Not collided
		# Fall-through: if collided, rebind mode will answer question
	}
	# So, generally, which set of commands if being rebound?
	set lsearch [lsearch -glob -nocase [list StormBot.TCL "Eggdrop Core"] [sbd:get CONFIG:GLOBAL-ORIGINALBINDS]*]
	# If we can't tell, assume the old days: Eggdrop core rebinds
	switch -exact -- $lsearch { 0 - 1 { return $lsearch } default { return 1 } }
}

proc ctime file {
	#if ![file exists $file] {return 0}
	# NO! Let it throw an exception!
	file stat $file a
	return $a(ctime)
}

proc data [list cmd name field [list value ""]] {
	# If you add FLAG options, replace the -NOCASE matching in FIND
	# If $NAME == *, UPVAR $0
	set field_fixed $field ; # GET & SET need a "corrected" field name; NAMES & FIND don't
	if [inlist -exact [list "" ' . - *] $field_fixed] { set field_fixed * } ; # $NAME --> "*"
	lower name field field_fixed
	if [ catch { array set temp [sbd:get $name] } ] { array set temp [list $name [sbd:get $name]] }
	switch -exact -nocase -- $cmd {
		get { return [var temp($field_fixed)] }
		set {
			if [isempty value] { unset -nocomplain temp($field_fixed) } { set temp($field_fixed) $value }
			sbd:set $name [array get temp]
			return $value
		}
		names { return [array names temp [none [stl $field] *]] }
		find {
			set result [list]
			foreach [list a b] [array get temp] {
				zero match
				if [string match -nocase $field $a] { one match }
				if [string match -nocase $value $a] { one match }
				if $match { lappend result [list $a $b] }
			}
			return $result
		}
		default return
	}
	?
}

# --- Command registration ---

proc SB:command args {
debug.print 2 "\[SB:COMMAND\]:ARGS($args)"
	set do [sl0 $args]; # Add | Del | Info | List | Clear
	set command [sl1 $args]
	set level [sl2 $args]
debug.print 2 "\[SB:COMMAND-ADD\]:COMMAND($command):DO($do):LEVEL($level):ARGS($args)"

	# LASSIGN eats 45ms per invocation. ):
	#lassign [ldestroy -all -both -glob -- [l3e $args] -*] abbreviations flags
	set flags         [lsearch      -all -inline -glob [l3e $args] -*]
	set abbreviations [lsearch -not -all -inline -glob [l3e $args] -*]

	#! Allow 'illegal flags' to cause an exception ....
	# 1283717226: No! CMDLEVEL uses "SB:COMMAND UPDATE" to alter cmd levels
	# This call to SB:CCA will throw an unwated exception!
	zero flags_binary; # In case of an error below:
	catch { set flags_binary [SB:convertcmdattributes $flags] }

debug.print 2 "\[SB:COMMAND-ADD\]:COMMAND($command):FLAGS($flags):FLAGS_BINARY($flags_binary):ABBREVIATIONS($abbreviations)"

	if ![info exists ::stormbot(@cmdinfo:commandlist)] {set ::stormbot(@cmdinfo:commandlist) ""}
	set commandlist $::stormbot(@cmdinfo:commandlist)
	set m [lsearch -glob $commandlist [list $command *]]

	switch -exact -- $do {
		add {
			if [string eq "" $command] {error "\[SB:COMMAND\] ADD: missing command word"}
			if [string eq "" $level] {error "\[SB:COMMAND\] ADD: missing command level"}
			if [string eq "" $flags] {error "\[SB:COMMAND\] ADD: missing command flags"}

			set proc SB:proc_[string tolower $command]
			set source [string tolower $command].sb6
			if [info exists ::SB_SOURCE] {set source $::SB_SOURCE}
			set time [clock seconds]
			if [info exists ::SB_TIME] {set time $::SB_TIME}
			set level_default $level

			# All of these must be "@CMDINFO:*" so SB:BOTDATAFILE WRITE
			# doesn't record these values permanently! Otherwise, bead
			# updates will NOT override the previously saved data!

			if {$m > -1} {
				lset ::stormbot(@cmdinfo:commandlist) $m [list $command $level $flags_binary $proc $source $time $level_default]
			} else {
				lappend ::stormbot(@cmdinfo:commandlist) [list $command $level $flags_binary $proc $source $time $level_default]
				lappend ::stormbot(@cmdinfo:commands) $command
			}

			if [info exists ::stormbot(@cmdinfo:${command}:level)] {
				#set level [SB:commandlist level $command]
				set level $::stormbot(@cmdinfo:${command}:level)

				# Preserve manually-changed command access levels
				# (if a bead is loaded @ cold startup / RESTART / REHASH,
				# the subsequent data file load will restore command
				# access levels. However, if a bead is loaded singularly,
				# the bead's hardcoded level will override any manual
				# change). DEFAULTLEVEL will store the bead's hardcoded
				# access level value.
			}

			# <<< What about -MINLEVEL / -MAXLEVEL / -FORCELEVEL? >>>

			# -MINLEVEL: IF bead currlevel < loaded value, change
			# -MAXLEVEL: IF bead currlevel > loaded value, change
			# -FORCELEVEL: change

			set ::stormbot(@cmdinfo:${command}:level)        $level
			set ::stormbot(@cmdinfo:${command}:flags)        $flags_binary
			set ::stormbot(@cmdinfo:${command}:proc)         $proc
			set ::stormbot(@cmdinfo:${command}:source)       $source
			set ::stormbot(@cmdinfo:${command}:time)         $time
			set ::stormbot(@cmdinfo:${command}:defaultlevel) $level_default

			foreach flag $flags {
#debug.print 2 "\[SB:COMMAND\]:FLAG($flag)"
				lappend ::stormbot(@cmdflags:[string range $flag 1 end]) $command
			}

			if [sbd:isempty CMDLEVEL:$command] {sbdset CMDLEVEL:$command $level}
			if [sbd:isempty CMDLEVEL-DEFAULT:$command] {sbdset CMDLEVEL-DEFAULT:$command $level_default}

			set ::stormbot(@cmdinfo:$command) [list $level $flags_binary $proc $source $time $level_default]
			if [notempty abbreviations] {SB:abbr add $command $abbreviations}

			foreach mode [stl [list MSG DCC]] {
#%# $::STORMBOT(@BINDMSG)
				if {
					[string is true -strict [sbd:get CONFIG:GLOBAL-BIND:$mode]]
					||
					( [string eq msg $mode] && [ expr $flags_binary & 2048 ] )
				} {
					set binds [binds $command]
					foreach bind $binds {
						if [string match -nocase [list $mode * $command * *] $bind] {
							if [left [l4 $bind] 1 *] {
								set bind_flags [lindex $bind 1]
								set bind_proc  [lindex $bind 4]
								if [catch {bind $mode $bind_flags [iff [iscmdrebound] /]$command $bind_proc} msg] { putcmdlog BIND_ERROR:01:$msg }
							}
						}
					}
					if [catch {bind $mode - [iff ![iscmdrebound] /]$command SB:dispatch_$mode} msg] { putcmdlog BIND_ERROR:02:$msg }
				}
			}

			return "" ; # Nothing to reply; nothing to save
		}

		del {
			if [string eq "" $command] {error "\[SB:COMMAND\] DEL: missing command word"}
			if {$m < 0} {return ""; # error "\[SB:COMMAND\] DEL: command \"${command}\" doesn't exist"}
			set ::stormbot(@cmdinfo:commandlist) [lreplace $::stormbot(@cmdinfo:commandlist) $m $m]
			set ::stormbot(@cmdinfo:commands) [getfirst $::stormbot(@cmdinfo:commandlist)]

			foreach del [array names ::stormbot @cmdinfo:${command}:*] {
				unset ::stormbot($del)
			}
			foreach flag [array names ::stormbot @cmdflags:*] {
				set m [lsearch -exact $::stormbot($flag) $command]
				if {$m > -1} {set ::stormbot($flag) [lreplace $::stormbot($flag) $m $m]}
			}

			return $command; # Nothing to reply; nothing to save
		}

		update - ed - edit - change - chg {
			# Less-confusing names ....
			set command_name $command
			set command_attribute $level
			set command_attribute_value [lindex $args 3]
debug.print 2 "NAME($command_name):ATTR($command_attribute):VALUE($command_attribute_value)"
			set m [lsearch -glob $commandlist [list $command_name *]]
			if {$m < 0} {return ""}
			switch -exact -- [string tolower $command_attribute] {
				level - flags - proc - source - time - defaultlevel {
					set index [lsearch [list level flags proc source time defaultlevel] [string tolower $command_attribute]]
				}
				default {error "\[SB:COMMAND - [stu $do]\] unknown option: \"${command}\""}
			}

			sbdset @CMDINFO:${command_name}:$command_attribute $command_attribute_value
			set data [sbdget @CMDINFO:$command_name]
			lset data $index $command_attribute_value
			sbd:set @CMDINFO:$command_name $data

			# Special case: CMDLEVEL
			if [string eq -nocase LEVEL $command_attribute] {
				sbdset CMDLEVEL:$command_name $command_attribute_value
			}

			# Special case: DEFAULTLEVEL
			if [string eq -nocase DEFAULTLEVEL $command_attribute] {
				sbdset CMDLEVEL-DEFAULT:$command_name $command_attribute_value
			}

			return $command_attribute_value
		}

		info {
			if [string eq "" $command] {error "\[SB:COMMAND\] DEL: missing command word"}
			if {$m < 0} {error "\[SB:COMMAND\] DEL: command \"${command}\" doesn't exist"}
			return [lindex $::stormbot(@cmdinfo:commandlist) $m]
		}

		list {return [getfirst $::stormbot(@cmdinfo:commandlist)]}

		clear {
			set ::stormbot(@cmdinfo:commandlist) ""
			set ::stormbot(@cmdinfo:commands) ""
			foreach a [array names ::stormbot @cmdinfo:*] {unset ::stormbot($a)}
			foreach a [array names ::stormbot @cmdflags:*] {unset ::stormbot($a)}
			return ""
		}

		default {error "\[SB:COMMAND\] unknown option: $do"}

	}
	return 0
}

proc SB:cmdlevel_restore args {
	set match [none $args *]
	foreach command [SB:commandlist match $match] {
		#%#
	}
	return 0
}

interp alias {} SB:cca {} SB:convertcmdattributes
interp alias {} SB:cmdattr {} SB:convertcmdattributes
proc SB:convertcmdattributes args {
#!     set args [lindex $args 0]; # Remember, it comes as "{1 2 3 4 5 ....}"
	# No JOIN! You break the LIST by creating a STRING! We don't want that.
	#! No way to determine if $ARGS is single-element, multi-word single
	#! element, or multi-word multi-element without several checks. Leave it
	#! unaltered.

	set validflags [list -invalidok -loggedoutok -suspendedok -authcmd -authcontinue -chanspecific -helpcmd -locklevel -minlevel -maxlevel -bindmsg "" "" -bypassoauth -corecommand -nobind]
	if [string eq -nocase CHECK $args] {return [concat -none -- [ldestroy -nonulls $validflags]]}

	if [regexp -- {^\d+$} $args] {
		set list ""
		for {set x 0} {$x < 31} {incr x} {
			if [bits args =$x] {lappend list [lindex $validflags $x]}
		}
		if [string eq "" $list] {set list -none}
		return $list
	}

	set flaglist ""
	foreach flag [join $args] {
		set um [uniquematch [concat -none -- $validflags] $flag]
		if [string eq "" $um] {error "\[SB:CONVERTCMDATTRIBUTES\] unknown flag option: $flag"}
		lappend flaglist $um
	}
	set tally ""
	foreach flag $flaglist {
		switch -exact -- [stl $flag] {
			-invalidok {lappend tally 0}
			-loggedoutok {lappend tally 1}
			-suspendedok {lappend tally 2}
			-authcmd {lappend tally 3}
			-authcontinue {lappend tally 4 3; # Requires AUTHCOMMAND too}
			-chanspecific {lappend tally 5}
			-helpcmd {lappend tally 6}
			-locklevel {lappend tally 7}
			-minlevel {lappend tally 8}
			-maxlevel {lappend tally 9}
			-bindmsg {lappend tally 10}
			-bypassoauth {lappend tally 13}
			-corecommand {lappend tally 14}
			-nobind {lappend tally 15}
			default {# Do nothing: -NONE -- #}
		}
	}
	set flags 0
	foreach a $tally { set flags [ expr $flags | int( pow( 2 , $a ) ) ] }
	return $flags
}

proc SB:commandlist { { do "" } { command "" } { level "" } } {
	set do [string tolower $do]
	set do [string map [list age time src source srclist sourcelist srclistall sourcelistall] $do]
	if [string eq -nocase AGE $do] {set do time}
	set command [string tolower $command]
	set m [lsearch -exact [string tolower $::stormbot(@cmdinfo:commands)] $command]
	switch -exact -- [string tolower $do] {
		"" - list {return [lsort -inc -dict [lmatch -all $::stormbot(@cmdinfo:commands) [none $command *]]]}
		info {if {$m < 0} {return ""}; return $::stormbot(@cmdinfo:$command)}
		iscmd {if {$m < 0} {return 0}; return 1}
		defaultlevel - time - source - proc {if {$m < 0} {return -1}; return $::stormbot(@cmdinfo:${command}:$do)}
		flags {
			set f $::stormbot(@cmdinfo:${command}:$do)
			switch -glob -- [stl $level] {
				bin* {return [bin $f]}
				dec* {return $f}
				hex* {return [hex $f]}
				default {return [SB:convertcmdattributes $f]}
			}
		}
		level { if { $m < 0 } { return -1 } { return [sbdget CMDLEVEL:$command] } }
		hf - has - flag - hasflag {
			set flag [string tolower $level]
			set flags [string tolower [regsub -all -- - [ldestroy -- [SB:convertcmdattributes CHECK] --] ""]]
			if [left $flag 1 -] {set flag [mid $flag 2]}
			set flag [uniquematch $flags $flag]
			if [isempty flag] {return 0}
			if {[lsearch -exact $::stormbot(@cmdflags:$flag) $command] > -1} {return 1} {return 0}
		}
		sourcelist {
			if ![right $command 4 .sb6] {append command .sb6}
			set list ""
			foreach a [SB:commandlist list] { if [string eq -nocase $::stormbot(@cmdinfo:${a}:source) $command] { lappend list $a } }
			return $list
		}
		beadslist {
			set list ""
			foreach a [SB:commandlist list] { lappend list $::stormbot(@cmdinfo:${a}:source)}
			return [lsort -inc -uni -dict $list]
		}
		sourcelistall {
			array set sources ""
			foreach a [SB:commandlist list] {lappend sources($::stormbot(@cmdinfo:${a}:source)) $a}
			return [array get sources]
		}
		match {
			empty list
			foreach item [getfirst [sbdget @CMDINFO:COMMANDS]] {
				if [string match -nocase $command $item] {lappend list $item}
			}
			return $list
		}
	}
	error "\[SB:COMMANDLIST\] invalid value for DO: $do"
}

proc SB:abbr args {
	set do [lindex $args 0]
	set command [string tolower [lindex $args 1]]
	set new [lsearch -all -inline -not [join [lrange $args 2 end]] ""]

	# Do =NOT= SB:"BOTDATAFILE WRITE" anywhere in this PROC!
	# This is invoked within SB:LOADBEADS, and the bot will never get to
	# load if it's too busy saving after every loaded bead!

	if ![info exists ::stormbot(abbr)] {set ::stormbot(abbr) ""}
	set abbr [string tolower [sbdget ABBR]]

	set m [lsearch -glob $abbr [list $command *]]; # CHECK will fail this.
	switch -exact -- [string tolower $do] {

		clean {
			empty new
			set attributes [SB:convertcmdattributes CHECK]
			foreach entry $abbr {
				lassign $entry command abbrs
				set m [lsearch -exact $abbrs ""]
				set n [lsearch -glob $abbrs -*]
				if {($m < 0) && ($n < 0)} {
					empty temp
					foreach a $abbrs { foreach b $a { lappend temp $b } }
					lappend new [list $command $temp]
				} {
					empty temp
					foreach item1 $abbrs {
						foreach item2 $item1 {
							lappend temp $item2
						}
					}

					set temp [lsearch -exact -all -inline -not $temp ""]
					empty temp2
					foreach item $temp { if [string eq "" [uniquematch $attributes $item]] { lappend temp2 $item } }
					if [notempty temp2] {lappend new [list $command $temp2]}
				}
			}
			sbd:set ABBR $new
			SB:needsave data
			return $new
		}

		add {
			if [string eq "" $command] {error "\[SB:ABBR\] ADD: missing command word"}
			if [isempty new] {return ""}
			#if [string eq "" $new] {error "\[SB:ABBR\] ADD: missing new abbreviation"}
			if {$m > -1} {# Exists: need to update entry
				set old [lindex [lindex $abbr $m] 1]
				foreach item $new { if [notempty item] { lappend old $item } }
				set old [lsearch -all -inline -not [lsort -inc -uni -dict [stl $old]] ""]; # Handles multiple entries
				set updated [list $command $old]
				if [isempty updated] {
					set abbr [lreplace $abbr $m $m]
				} {
					lset abbr $m $updated
				}
				sbdset abbr $abbr
			} else {# Doesn't exist: add new entry
				if [notempty new] {
					set updated [list $command $new]
					lappend abbr $updated
					sbdset ABBR $abbr
				}
			}

			# Establish binds for the abbreviations! (New to SB6!)
			foreach method [list msg dcc] {
				if [sbdnozero @bind$method] {
					foreach item $new {
						set binds [binds $method]
						set m [lsearch -glob $binds [list * * $item * ?${method}:*]]
						if {$m > -1} {
							set bind [lindex $binds $m]
							# 2018-05-06 16:57:00 -0700: Changed (see above)
							if [catch {unbind [lindex $bind 0] [lindex $bind 1] [iff ![iscmdrebound] /][lindex $bind 2] [lindex $bind 4]} msg ] { msghome 04:$msg }
							if [catch {bind [lindex $bind 0] [lindex $bind 1] [iff [iscmdrebound] /][lindex $bind 2] [lindex $bind 4]} msg] { putcmdlog BIND_ERROR:05:$msg }
						}
						if [catch {bind $method - [iff ![iscmdrebound] /]$item SB:dispatch_$method} msg] { putcmdlog BIND_ERROR:07:$msg }
					}
				}
			}

			return $updated
		}

		del {
			if [string eq "" $command] {error "\[SB:ABBR\] DEL: missing command word"}
			if {$m < 0} { return "" ; # 20140809:093500-0700 I'm sick of this holding over from dead / removed beads -- error "\[SB:ABBR\] DEL: command \"${command}\" doesn't exist" }
			if { $m != -1 } {
				set old [lindex [lindex $abbr $m] 1]
				foreach item $new { # Do =NOT= use LDESTROY -MULTIPLE (need to loop through to kill each possible BIND)
					set old [ldestroy -all -nocase -nonull $old $item]
					foreach method [list msg dcc] {
						catch { unbind $method - [iff ![iscmdrebound] /]$item SB:dispatch_$method } 
					}
				}
				if [isempty old] {
					set abbr [lreplace $abbr $m $m]
				} {
					lset abbr $m [list $command $old]
				}
				sbdset ABBR $abbr
			}
			return ""; # Nothing to reply; nothing to save
		}

		check {
			foreach item [string tolower $abbr] {
				set n [lsearch -exact [lindex $item 1] $command]
				if {$n > -1} {return [lindex $item 0]}
			}
			return ""
		}

		get {
			if {$m < 0} {return ""}
			return [lindex [lindex $abbr $m] 1]
		}

		list {return $abbr}

		search - find {
			foreach item $abbr {
				foreach findme [lindex $item 1] {
					if [string match -nocase $findme $command] {return [lindex $item 0]}
				}
			}
			return ""
		}

		clear {
			if ![string eq -nocase CLEAR $command] {return ""}
			foreach item $abbr {
				foreach method [list msg dcc] {
					foreach abbreviation [join [join $abbr]] {
						catch { unbind $method - [iff ![iscmdrebound] /]$abbreviation SB:dispatch_$method }
					}
				}
			}
			sbdset ABBR
			return ""
		}

		default {error "\[SB:ABBR\] unknown option: $do"}

	}
	return ""
}

proc SB:registervariable args {
	set type [sl0 $args]
	set vars [sl1e $args]
	# Deal with $ARGS issue again
	if {([llength $type] > 1) && [isempty vars]} {
		set type [l0  [join $args]]
		set vars [sl1e [join $args]]
	}
	array set custom [sbdget @CUSTOM]
	foreach var $vars {
		if [string eq -nocase CHAN $type] {set type channel}
		set valid [string tolower [list standard global channel user chanflags]]
		set type [uniquematch $valid $type]
		if [isempty type] { error "SB:REGISTERVARIABLE: illegal type: \$TYPE ($type)" }
		#if {[lsearch -exact $valid $type] < 0} {error "SB:REGISTERVARIABLE: illegal type: \$TYPE ($type)"}
		if ![info exists custom($type)] { set custom($type) "" }
		#if {[lsearch -exact [stl $custom($type)] [stl $var]] > -1} continue
		lappend custom($type) $var
		set custom($type) [lsort -inc -dict -uni $custom($type)]; # Remove duplicates.
		debug.print 2 "\[SB:REGISTERVARIABLE\] registering ${type}:$var ...."
	}
	sbdset @CUSTOM [array get custom]
	return 0
}

proc SB:registerdeath args {
	# Must have $ARGS to capture everything yet: CHECKFLAGS required $ARG. ):
	set arg [concat * $args]
	SB:setmiscvariables -del -reset -clear -none -list -add -execute
	set command $1

	# Enter with: SB:registerdeath update_counters

	# This command will activate pre-placed, communal deaths in DIE to handle
	# any pre-death cleanups needed before the bot totally goes 6 feet under.

	if [isempty flags] {set flags -add}
	if [sbd:isempty @deathlist] {sbdset @deathlist [list 0 0]}

	array set deathlist [sbdget @deathlist]

	if {[isvalidflag $flags -clear] || [isvalidflag $flags -reset] || [isvalidflag $flags -none]} {
		set temp ""
		foreach a [array names deathlist] {
			if !($a) continue
			lappend temp $deathlist($a)
		}
		sbdset @deathlist
		debug.print 2 "\[SB:REGISTERDEATH -CLEAR\] deleted [and [join [lsort -inc -uni -dict $temp] ", "] &]"
		return [llength $temp]
	}

	if {[isvalidflag $flags -list] || [isvalidflag $flags -execute]} {
		if !($deathlist(0)) {return ""}
		empty list
		foreach a [lsort -int -inc [array names deathlist]] {
			if ($a) {lappend list $deathlist($a)}
		}
		if [isvalidflag $flags -execute] { foreach death $list { eval $death } }
		return $list
	}

	if [isvalidflag $flags -find] {
		if !($deathlist(0)) {return ""}
		empty list
		foreach a [lsort -int -inc [array names deathlist]] {
			if ($a) {
				if [string match -nocase $1 $deathlist($a)] {lappend list $deathlist($a)}
			}
		}
		return $list
	}

	if [isvalidflag $flags -del] {
		if !($deathlist(0)) {return ""; # Nothing to delete!}

		empty deletelist
		array set temp [list 0 0]
		foreach a [lsort -int -inc [array names deathlist]] {
			if $a {
				if ![string match -n $command $deathlist($a)] {
					lappend deletelist $deathlist($a)
					incr temp(0)
					set temp($temp(0)) $deathlist($a)
				}
			}
		}
		sbdset @deathlist
		sbdset @deathlist [array get temp]
		debug.print 2 "\[SB:REGISTERDEATH -DEL\] deleted $deletelist"
		return [llength $deletelist]
	}

	if [isvalidflag $flags -add] {
		foreach {a b} [array get deathlist] {
			if {$b == $command} {return ""}; # Death is already loaded or survived a REHASH.
		}
		if ![info exists deathlist] {array set deathlist "0 0"}
		set count $deathlist(0)
		if ![isint $count] {set count 0}
		incr count
		set deathlist(0) $count
		set deathlist($count) $command
		sbdset @deathlist [array get deathlist]
		debug.print 2 "\[SB:REGISTERDEATH\] registering $command ...."
		return 0
	}
	return ""
}

proc SB:loadbeads { { pattern "*" } { force "" } } {
	debug.print 2 "\[SB:LOADBEADS\] pattern: $pattern (force = [offon [not [string eq "" $force]]])"
	set beads ""
	set loaded ""

	# Collect all possible matches
	# Sub-dir beads/$botnick should be checked AFTER main beads directory is
	# checked. This allows bot-specific overrides to standard beads
	foreach temp_pattern [stl $pattern] {
		if [right $temp_pattern 4 .sb6] { set temp_pattern [left $temp_pattern -4] }
		set beads [concat $beads [glob -nocomplain ./scripts/beads/[stl $temp_pattern].sb6 ./scripts/beads/[string tolower $::nick]/[stl $temp_pattern].sb6]]
	}

	foreach bead [lsort -increasing -dictionary -unique [stl $beads]] {
		set ms_start [clock clicks -milli]
		#!update; # Catch up on tasks!
		#!OH JESUS NO!! This is not returning to the global scope,
		#!causing: SAVECHANNELS and: can't read "tcl_platform(os)": no such variable
			## This has been fixed in 1.6.21 (due to our circumstances, this was seen by eggdrop core coders)

		# Fixed in eggdrop 1.6.20 full release by thommey / guppy
		# Was broken in 1.6.20-RC1 ("UPDATE" command)

		# FULL_SB6.TCL will choke on this otherwise! But ... allow the
		# possibility of a "patched" or custom bead to be loaded AFTER
		# FULL_SB6.TCL loads.
		if ![file exists $bead] continue

		if ![info exists ::stormbot(@beadtime:$bead)] {set ::stormbot(@beadtime:$bead) 0}
		set mtime [file mtime $bead]
		set load 0
		if ![string eq "" $force] {set load 1}
		if {$mtime > $::stormbot(@beadtime:$bead)} {set load 1}

		if $load {
               #log "\[SB:LOADBEADS\] loading BEAD($bead) .... "
			SB:loadbeads2 $bead
			lappend loaded [file rootname [file tail $bead]]
		}
		set ms_end [clock clicks -milli]
#%#%#
#putlog "\[SB:LOADBEADS\] BEAD([file rootname [file tail $bead]]):LIB([info lib])"
#foreach _ $::auto_path { putlog "     AUTO_PATH($_)" }
#foreach _ $::tcl_pkgPath { putlog "     TCL_PKGPATH($_)" }
#unset -nocomplain _
#          log "\[SB:LOADBEADS\] ${bead}: [ expr $ms_end - $ms_start ]ms"
	}
	if [string eq * $pattern] { debug.print 1 "\[SB:LOADBEADS\] Complete" }
	return [lsort -increasing -unique -dictionary $loaded]

# > While loading bead:
# > > Level flag: "-minlevel" (if bot data shows cmdlevel < hard-coded bead):
# > > > force cmdlevel up to hard-coded level
# > > Level flag: "-maxlevel" (if bot data shows cmdlevel > hard-coded bead):
# > > > force cmdlevel down to hard-coded level
# > Back-compat* for SB5 format beads?

}

proc SB:loadbeads2 bead {
	set mtime [file mtime $bead]
	set ::SB_BEAD $bead
	set ::SB_SOURCE [file tail $bead]
	set ::SB_TIME $mtime
	empty shit
	set error [ catch { uplevel #0 { source $::SB_BEAD } } shit ]
	unset -nocomplain ::SB_BEAD ::SB_SOURCE ::SB_TIME
	if $error {
		set message "\[SB:LOADBEADS2\] error while loading ${bead}: $shit"
		# It still displays :( -- if [sbd:nozero @SCRIPT_IS_LOADED] { log $message } ; # Don't display on cold-start
		putcmdlog $message
		if ![sbd:boolean CONFIG:GLOBAL-FAILCONTINUE] { error $message }
	}
	set ::stormbot(@beadtime:$bead) $mtime
	#putcmdlog "\[SB:LOADBEADS2\] loading [file tail $bead]"
	return 0
}

interp alias {} writedata {} SB:botdatafile write
proc SB:botdatafile { mode { extra "" } { datastream * } } {

	set dont_process [list !* @* #* {$*} %* ^* &*]

	set datafile ${::nick}.data
	if [notempty ::stormbot(botdatafile)] {set datafile $::stormbot(botdatafile)}
	if [notempty extra] { if [file isdirectory $extra] { prepend datafile ${extra}/ } { set datafile $extra } }
	set datafile [file normalize $datafile]

	switch -exact -- [stl $mode] {

		name - file - filename { return $datafile }

		reload - load - read {
			# Make this available via some kind of -DEBUG flag
			#putlog "\[SB:BOTDATAFILE\] Reading data file ($datafile) ...."
debug.print 1 "\[SB:BOTDATAFILE\] Reading data file ($datafile) ...."
			if ![file exists $datafile] return
			set version_error "Illegal StormBot.TCL data file format. Data file must be converted to version 2 or above to be used with StormBot.TCL V6. I can not arbitrarily use a SB3 / SB4 / SB5 data file without upgrading."
			set r [open $datafile r]
			set data [split [read $r] \n]
			close $r

			# SB3 / SB4 / SB5 (early versions) didn't have a VERSION tag.
			if {[lsearch -glob -nocase $data [list version *]] < 0} {
#putlog "LSEARCH/DATA([lsearch -glob $data [list version *]])"
#foreach a [lrange $data 0 10] { putlog "     $a" }
#exit $version_error		
				error $version_error
			}

			set loaded 0
			set version_checked 0
			foreach line $data {
				lassign $line type name info
#debug.print 1 "\[SB:BOTDATAFILE LOAD\] NAME($name):INFO([left $info 40][iff { [len $info] > 40 } ...])"
				switch -exact -- [stl $type] {

					# #

					"" #

					version {# Only using version 2 and above now #
						set version_checked 1
debug.print 1 "\[SB:BOTDATAFILE\] Version check: $name"
						if {$name < 2} {error $version_error}
						# Fall-through ....
					}

					info {
						# Do NOT load CMDINFO / BEADTIME information!
						set fail 0
						foreach dont $dont_process {
							if [string match -nocase $dont $name] {set fail 1}
						}
						if !($fail) {
debug.print 1  "\[SB:BOTDATAFILE\] Data read: $name"
							set ::stormbot($name) $info
							incr loaded
						}
					}

					default { error "\[SB:BOTDATAFILE\] Illegal data type: $type" }

				}
			}
			return $loaded
		}

		save - write {
#putlog "\[SB:BOTDATAFILE\] Saving data file ($datafile): invoked by: [lindex [info level [expr [info level] -1]] 0]"
			zero count rejected
			foreach _ [sbd:find *] {
				if [regexp -- {^[\!\@\#\$\%\^\&]} $_] { incr rejected } { incr count }
			}
			set w [open $datafile w]
			lappend pre
			lappend pre "# Data file for:" ${::nick}
			lappend pre "# File location:" $datafile
			lappend pre "# Name normalized:" [file normalize $datafile]
			lappend pre "# Written:" [clock format [clock seconds] -format {%Y-%m-%d %H:%M:%S %z}]
			lappend pre "# Count:" [comma $count]
			lappend pre "# Rejected (temp):" [comma $rejected]
			foreach _ [print:table -trim $pre 2] { puts $w $_ }
			puts $w ""
			puts $w "Version 2.1"
			foreach name [lsort -inc -uni -dict [array names ::stormbot]] {
				set fail 0
				foreach dont $dont_process {
					if [string match -nocase $dont $name] {set fail 1}
					if ![string match -nocase $datastream $name] {set fail 1}
				}
				if !($fail) {puts $w [list info $name $::stormbot($name)]; incr count}
			}
			close $w
			return $count
		}

		backup {
			putlog "\[SB:BOTDATAFILE\] Backing up data file ($datafile --> ${datafile}~bak) ...."
			if [file exists $datafile] { file copy -force $datafile [file dirname [file normalize $datafile]]/[file tail ${datafile}~bak]}
			return 0
		}

		default { error "\[SB:BOTDATAFILE\] Illegal mode: $mode" }

	}
	return ""
}

proc SB:register args {
	set arg [split [singlespace $args]]
	set arg $args

	switch -exact -- [sl0 $arg] {
		command {SB:command add [jl1 $arg] [nozero [jl2 $arg]] [jl3 $arg] [jl4 $arg] [jl5 $arg]}
		wegde - wedge {SB:registerwedge [jl1e $arg]}
		customvariable - variable {SB:registervariable [concat [jl1 $arg] [jl2e $arg]]}
		bind - binds {SB:registerbinds [jl1 $arg] [jl2 $arg] [jl3 $arg] [jl4 $arg]}
		hour - hourly {SB:registerhourly [l1 $arg] [l2 $arg]}
		die - death {SB:registerdeath [l1 $arg]}
		default {error "> SB:REGISTER: unknown register option: [jsu0 $arg]"}
	}

	return 1
}

proc SB:registerbinds {type flags string cmd} {
	if ![info exists ::stormbot(@custom)] {set ::stormbot(@custom) ""}
	array set custom $::stormbot(@custom)

	set binds [binds $cmd]; # Check for binds specifically to the command
	set m [lsearch -glob [stl $binds] [list [stl $type] ${flags}* [stl $string] $cmd]]]
	if {$m < 0} {
		if [catch {bind $type $flags [iff ![iscmdrebound] /]$string $cmd} msg] { msghome 10:$msg }
		if ![info exists custom(binds)] {set custom(binds) ""}
		lappend custom(binds) [list $type $flags $string $cmd]
		set custom(binds) [lsort -uni $custom(binds)]
	}
	set ::stormbot(@custom) [array get custom]
	return 0
}

proc SB:registerwedge args {
	# Must have $ARGS to capture everything yet: CHECKFLAGS requires $ARG. ):
	set arg [concat * $args]
	SB:setvariables -del -reset -clear -none -list -add
	set command $arg

	# Enter with: SB:registerwedge update_counters

	# This command will activate a pre-placed, communal wedge in
	# DO_CALLCOMMAND to handle any "pre-action" before a command is
	# dispatched for execution.

	# Dispatched args are: $NICK $HOST $HANDLE $CHAN $ARG (the same as
	# with any executed module bead command).

	if [isempty flags] {set flags -add}

	set wedgelist [sbd:get @WEDGELIST]

	if [validflag -clear] {sbd:set @WEDGELIST; return 0}

	if [validflag -list] {return [sbd:get @WEDGELIST]}

	if [isvalidflag $flags -del] {sbd:set @WEDGELIST [ldestroy -all -multiple -glob -- $wedgelist $command]; return 0}

	if [isvalidflag $flags -add] {
		sbd:lappend @WEDGELIST $command
		sbd:set @WEDGELIST [nodup [sbd:get @WEDGELIST]]
		return 0
	}

	return 0
}

# --- Variables ---

proc decr {varName { decrement 1 } } { upvar 1 $varName var; incr var [ expr -1 * $decrement ] }

proc set:all args {
	if { [llength $args] == 1 } { return "" }
	set value [lindex $args end]
	foreach . [lrange $args 0 end-1] { uplevel 1 set ${.} $value }
	return $value
}

# Don't use "LEFT $a 2 ::" -- causes an endless loop (unknown why)
interp alias {} 0 {} zero
interp alias {} 1 {} one
interp alias {} -1 {} negone
interp alias {} {} {} empty
# No alias for NULL.

# Maybe I should change all of these to match TRUE/FALSE; they're more efficient.
proc negone   args { foreach a $args { if [string match ::* $a] { upvar #0 [string range $a 2 end] local } { upvar 1 $a local } ; set local -1 } }
proc zero     args { foreach a $args { if [string match ::* $a] { upvar #0 [string range $a 2 end] local } { upvar 1 $a local } ; set local  0 } }
proc one      args { foreach a $args { if [string match ::* $a] { upvar #0 [string range $a 2 end] local } { upvar 1 $a local } ; set local  1 } }
proc true     args { foreach a $args { uplevel 1 set $a true } }
proc false    args { foreach a $args { uplevel 1 set $a false } }
proc null     args { foreach a $args { if [string match ::* $a] { upvar #0 [string range $a 2 end] local } { upvar 1 $a local } ; unset -nocomplain local } }
proc empty    args { foreach a $args { if [string match ::* $a] { upvar #0 [string range $a 2 end] local } { upvar 1 $a local } ; unset -nocomplain local ; set local "" } }
proc lower    args { foreach a $args { upvar 1 $a local ; if [info exists local] { set local [string tolower $local] } { set local "" } } }
proc upper    args { foreach a $args { upvar 1 $a local ; if [info exists local] { set local [string toupper $local] } { set local "" } } }
proc title    args { foreach a $args { upvar 1 $a local ; if [info exists local] { set local [stt            $local] } { set local "" } } }
proc sentence args { foreach a $args { upvar 1 $a local ; if [info exists local] { set local [string totitle $local] } { set local "" } } }

proc add {a { increment 1 } } { if [string match ::* $a] { upvar #0 [string range $a 2 end] var } { upvar 1 $a var } ; if ![info exists var] {set var 0}; float var; set var [ expr $var + $increment ]}
proc sub {a { decrement 1 } } { if [string match ::* $a] { upvar #0 [string range $a 2 end] var } { upvar 1 $a var } ; if ![info exists var] {set var 0}; float var; set var [ expr $var - $decrement ]}
proc mul {a { multiple  1 } } { if [string match ::* $a] { upvar #0 [string range $a 2 end] var } { upvar 1 $a var } ; if ![info exists var] {set var 0}; float var; set var [ expr $var * $multiple  ]}
proc div {a { divisor   1 } } { if [string match ::* $a] { upvar #0 [string range $a 2 end] var } { upvar 1 $a var } ; if ![info exists var] {set var 0}; float var; set var [ expr $var / $divisor   ]}
proc pow {a { exponent  1 } } { if [string match ::* $a] { upvar #0 [string range $a 2 end] var } { upvar 1 $a var } ; if ![info exists var] {set var 0}; float var; set var [ expr pow( $var , $exponent ) ]}

proc emptyarray args {
	# Don't enable -RESET here; EMPTYARRAY exists to wipe / create arrays by default
	set ::SB_emptyvars $args
	uplevel 1 {
		foreach ::SB_empty $::SB_emptyvars {
			if [info exists $::SB_empty] { unset $::SB_empty }
			array set $::SB_empty ""
		}
	}
	foreach a [info vars ::SB_empty*] { unset $a }
	return 0
}

proc isnull var {
	upvar 1 $var local
	not [info exists local]
}

proc notnull var {
	upvar 1 $var local
	info exists local
}

proc isempty args {
	# Either / or can exist to be true
	foreach var $args {
		upvar 1 $var local
		if ![info exists local] continue
		if [string eq "" $local] continue
		return 0
	}
	return 1
}

proc anyempty args { foreach var $args { upvar 1 $var local ; if [isempty local] { return 1 } } ; return 0 }
# For an "any not empty" use ![allempty]
proc allempty args { foreach var $args { upvar 1 $var local ; if [notempty local] { return 0 } } ; return 1 }

proc notempty args {
	# Restrictions are different for negative result (both conditions must be true)
	foreach var $args {
		upvar 1 $var local
		if { [info exists local] && ![string eq "" $local] } continue
		return 0
	}
	return 1
}

proc isemptyval args {
	error "\[ISEMPTYVAL\] Command is deprecated. invoked by: [info level [ expr [info level] - 1 ]]"
	putcmdlog "\00301,00\[ISEMPTYVAL\]\003 Invoked by: [info level [ expr [info level] - 1 ]] (replace this with NONE!)"
	set level [info level]
for { set l 1 } { $l < $level } { incr l } { putlog "[space 5][info level $l]" }
	putcmdlog "\[ISEMPTYVAL\] End of trace."
}

proc notemptyval args {
	error "\[NOTEMPTYVAL\] Command is deprecated. invoked by: [info level [ expr [info level] - 1 ]]"
	#if [string eq "" $reformat] {return $checkme} {return $reformat}
	putcmdlog "\00301,00\[NOTEMPTYVAL\]\003 Invoked by: [info level [ expr [info level] - 1 ]] (replace this with NONE!)"
	set level [info level]
for { set l 1 } { $l < $level } { incr l } { putlog "[space 5][info level $l]" }
	putcmdlog "\[NOTEMPTYVAL\] End of trace."
}

proc ifemptyval args {
	error "\[IFEMPTYVAL\] Command is deprecated. invoked by: [info level [ expr [info level] - 1 ]]"
	putcmdlog "\00301,00\[IFEMPTYVAL\]\003 Invoked by: [info level [ expr [info level] - 1 ]] (replace this with NONE!)"
	set level [info level]
for { set l 1 } { $l < $level } { incr l } { putlog "[space 5][info level $l]" }
	putcmdlog "\[IFEMPTYVAL\] End of trace."
}

proc ifnotemptyval args {
	error "\[IFNOTEMPTYVAL\] Command is deprecated. invoked by: [info level [ expr [info level] - 1 ]]"
	putcmdlog "\00301,00\[IFNOTEMPTYVAL\]\003 Invoked by: [info level [ expr [info level] - 1 ]] (replace this with NONE!)"
	set level [info level]
for { set l 1 } { $l < $level } { incr l } { putlog "[space 5][info level $l]" }
	putcmdlog "\[IFNOTEMPTYVAL\] End of trace."
}

proc isarray variable { uplevel 1 array exists $variable }

proc string:eq args    { string:test 0 1 $args }
proc string:ne args    { string:test 0 0 $args }
proc string:match args { string:test 1 0 $args }
proc string:test { mode_equal mode_match args } {
	if [set nocase [string eq -nocase [lindex [split $args] 0] -NOCASE]] {
		set args [join [lreplace [split $args] 0 0]]
	}
	lassign $args mask var
	upvar 1 $var local
	if [isempty local] { return 0 }
	eval uplevel 1 [lindex not $mode_equal] string [lindex [list eq match] $mode_match] [lindex -nocase [not $nocase]] $mask $local
}
proc string:equalvars args {
	if [string eq -nocase HELP $args] { return "string:equalvars ?-nocase? ?-nullequal? variable1 variable2" }
	flags:simple $args [list -nocase -nullequal] vars flags
	lassign $vars var1 var2
	upvar 1 $var1 local1
	upvar 1 $var2 local2
	set nulls [isnull local1][isnull local2]
	switch -exact -- $nulls {
		01 - 10 { return 0 }
		11 { return [validflag -nullequal] ; # If both are null, their values are technically equal }
	} ; # Let "00" fall-through for processing
	eval string eq [iff [validflag -nocase] -nocase] $local1 $local2
}

proc var { var { default "" } } {
	upvar 1 $var local
	if [isempty local] { return $default }
	if [array exists local] { return [array get local] }
	return $local
}

proc none { value { none "" } { some "" } } { if [string eq "" $value] { return $none } { if [string eq "" $some] { return $value } { return $some } } }

proc isemail email { regexp -nocase -- {^[A-Z0-9_\-\.]+@(localhost|[A-Z0-9_\-\.]+)$} $email }

# --- Special STRING / LIST processing shortcuts ---

interp alias {} stl {} string tolower
interp alias {} stu {} string toupper
interp alias {} sts {} string totitle ; # SENTENCE case! (misnomer; thanks, TCL people)
interp alias {} st- {} st=
proc st= a { return $a } ; # Used for if/then st- vs stl (LREMOVE)
proc stt a {# TITLE case (every word is Capped)
	set a [split $a]
	# May cause 2nd level, redundant SPLIT, which is okay.

	set b ""
	foreach c $a {lappend b [string totitle $c]}
	join $b
}
proc str a {

	# Note: since digits & punctuation are not considered a 'letter,' all of
	#	  those characters will fail =BOTH= IS UPPER & IS LOWER. Therefore,
	#	  just test for one or the other [STL / STU won't affect these
	#	  characters anyway].

	set temp ""
	foreach char [split $a ""] {
	  if [string is upper $char] {
		  append temp [stl $char]
	  } else {
		  append temp [stu $char]
	  }
	}
	return $temp
}

proc st^ {a b} {
	if ![regexp -- {^[\+\-]?\d+$} $b] {error "Illegal value in ST^: $b (expecting integer)"}
	set temp [string toupper [string range $a 0 [ expr $b - 1 ]]]
	incr b
	append temp [mid $a $b]
	return $temp
}

proc sil a {string is lower $a}
proc siu a {string is upper $a}
proc sis a {regexp -- {^[A-Z][a-z]*?$} $a}
proc sit a {
	one sit
	foreach b [split $a] { if ![ regexp -- {^[A-Z][a-z]*$} $b ] { zero sit } }
	return $sit
}

proc jstl  a    {join [string tolower $a]}
proc jstu  a    {join [string toupper $a]}
proc jsts  a    {join [string totitle $a]}
proc jstt  a    {join [stt $a]}
proc jstr  a    {join [str $a]}
proc jst^ {a b} {join [st^ $a $b]}

for {set x 0} {$x <= 11} {incr x} {
	set y $x
	if {$y == 11} {set y "e"}
	eval proc l$y	a \{lindex \$a $y\}
	eval proc sl${y}  a \{stl \[l$y \$a\]\}
	eval proc su$y    a \{stu \[l$y \$a\]\}
	eval proc ss$y    a \{sts \[l$y \$a\]\}
	eval proc st$y    a \{stt \[l$y \$a\]\}
	eval proc sr$y    a \{str \[l$y \$a\]\}
	eval proc s^$y    \{a b\} \{st^ \[l$y \$a \$b\]\}

	eval proc jl${y}  a \{join \[lindex \$a $y\]\}
	eval proc jsl$y   a \{join \[stl \[l$y \$a\]\]\}
	eval proc jsu$y   a \{join \[stu \[l$y \$a\]\]\}
	eval proc jss$y   a \{join \[sts \[l$y \$a\]\]\}
	eval proc jst$y   a \{join \[stt \[l$y \$a\]\]\]\}
	eval proc jsr$y   a \{join \[str \[l$y \$a\]\]\}
	eval proc js^$y   \{a b\} \{join \[st^ \[l$y \$a \$b\]\]\}

	eval proc s$y \{text splitter\} \{lindex \[split \$text \$splitter\] $y\}
	eval proc js$y \{text splitter\} \{join \[lindex \[split \$text \$splitter\] $y\]\}

	if {$y != "e"} {
	  eval proc l${y}e   a \{lrange \$a $y end\}
	  eval proc sl${y}e  a \{stl \[l${y}e \$a\]\}
	  eval proc su${y}e  a \{stu \[l${y}e \$a\]\}
	  eval proc ss${y}e  a \{sts \[l${y}e \$a\]\}
	  eval proc st${y}e  a \{stt \[l${y}e \$a\]\}
	  eval proc sr${y}e  a \{str \[l${y}e \$a\]\}
	  eval proc s^${y}e  \{a b\} \{st^ \[l${y}e \$a \$b\]\}

	  eval proc jl${y}e  a \{join \[lrange \$a $y end\]\}
	  eval proc jsl${y}e a \{join \[stl \[l${y}e \$a\]\]\}
	  eval proc jsu${y}e a \{join \[stu \[l${y}e \$a\]\]\}
	  eval proc jss${y}e a \{join \[sts \[l${y}e \$a\]\]\}
	  eval proc jst${y}e a \{join \[stt \[l${y}e \$a\]\]\}
	  eval proc jsr${y}e a \{join \[str \[l${y}e \$a\]\]\}
	  eval proc js^${y}e \{a b\} \{join \[st^ \[l${y}e \$a \$b\]\]\}

	  eval proc js${y}e \{text splitter\} \{join \[lrange \[split \$text \$splitter\] $y end\]\}
	}
}; unset x y

proc `titlecase { temp { more "" } } {
     set lower [concat $more a to in on with without from of an the for at by]
     if [instr $temp =] {
          regsub -all -- = $temp " = " temp
     } {
          # NO! - regsub -all -- " - " $temp " = " temp
          regsub -all -- "--" $temp " = " temp
     }
     #regsub -all -- % $temp $$    temp
     regsub -all -- _ $temp " "   temp

     regsub -all -- {\(}  $temp {( }  temp
     regsub -all -- {\[}  $temp {\[ } temp
     regsub -all -- {\{}  $temp {\{ } temp
     regsub -all -- !     $temp {! }  temp
     regsub -all -- @     $temp {@ }  temp
     regsub -all -- #     $temp {# }  temp
     regsub -all -- {\$}  $temp {\$ } temp
     regsub -all -- %     $temp {% }  temp
     regsub -all -- {\^}  $temp {\^ } temp
     regsub -all -- {\\&} $temp {\& } temp
     regsub -all -- {\.}  $temp { . } temp
     # NO! - regsub -all -- - $temp " - " temp; # Will protect intentional " - "

     empty new
     one first
     foreach word $temp {
          if { [ string is upper $word ] && ( [ len $word ] > 1 ) } {# Protect full-CAPS (intentional): "DJ"
               # Avoid "A" always getting through this test
               lappend new $word
          } else {
               if $first {# Always capitalise 1st word regardless
                    lappend new [stt $word]
               } {
                    if {[lsearch -exact $lower [stl $word]] < 0} {lappend new [stt $word]} {lappend new [stl $word]}
               }

               # Those Irish ....
               if { [left [stl $word] 2 mc ] && ( [ len $word ] > 2 ) } { set word "Mc[stt $word]" }
               zero first
          }
          #if [regexp -nocase -- {^[A-Z]} $word] {zero first}
          if [string eq = $word] {one first}
     }

     set new [join $new]
     # NO! - regsub -all -- " - " new - $new
     regsub -all -- {\{ }  $new {\{} new
     regsub -all -- {\[ }  $new {[}  new
     regsub -all -- {\( }  $new {(}  new
     regsub -all -- {! }   $new !    new
     regsub -all -- {@ }   $new @    new
     regsub -all -- {# }   $new #    new
     regsub -all -- {\$ }  $new {\$} new
     regsub -all -- {% }   $new %    new
     regsub -all -- {\^ }  $new {\^} new
     regsub -all -- {\\& } $new {\&} new
     regsub -all -- { \. } $new {.}  new
     regsub -all -- " = " $new = new
     return $new
}

proc titlecase text {
	foreach r [regexp -inline -all -nocase -- {[A-Z]+} $text] {
		set word [stt $r]
		if [string match -nocase Mc* $word] { set word "Mc[string totitle [mid $word 3]]" }
		regsub -all -- $r $text $word text
	}
	return $text
}

proc bswitch args {
	# Do we need switches?
	# Yes
	flags:simple $args [list -debug -glob -exact -regexp -nocase -not] args flags
	set match_mode -glob
	flags:priority flags_match -glob -exact -regexp
	if [notempty flags_match] { set match_mode $flags_match }
	set debug [validflag -debug]
	#set args [lindex [lassign $args value] 0]
	set args [lindex $args 0]
	set ll [llength $args]
	if { $ll % 2 } { error "\[BSWITCH\] extra switch pattern without body" }
	set if if
	foreach [list expression body] $args {
		if [string eq default $expression] {
			lappend code else $body
		} {
			if [validflag -not] { set expression "!( $expression )" }
			lappend code $if $expression $body
		}
		set if elseif
	}
if $debug { log [debug -quiet code] }
	uplevel 1 $code
	return
}

proc eswitch args {
	# Do we need switches?
	flags:simple $args [list -debug] args flags
	set debug [validflag -debug]
	set args [lindex [lassign $args value] 0]
	set ll [llength $args]
	if { $ll % 2 } { error "\[ESWITCH\] extra switch pattern without body" }
	set if if
	foreach [list expression body] $args {
		if [string eq default $expression] {
			lappend code else $body
		} {
			lappend code $if "$value == $expression" $body
		}
		set if elseif
	}
if $debug { log [debug -quiet code] }
	uplevel 1 $code
	return
}

proc lswitch args {
	# Do we need switches?
	# Yes
	flags:simple $args [list -debug -glob -exact -regexp -nocase] args flags
	set match_mode -glob
	flags:priority flags_match -glob -exact -regexp
	if [notempty flags_match] { set match_mode $flags_match }
	set debug [validflag -debug]

	set nocase [iff [validflag -nocase] -nocase]
	set args [lindex [lassign $args value] 0]
	set ll [llength $args]
	if { $ll % 2 } { error "\[LSWITCH\] extra switch pattern without body" }
	set if if
	foreach [list expression body] $args {
		if [string eq default $expression] {
			lappend code else $body
		} {
			lappend code $if "\[lsearch $match_mode $nocase [list $expression] [list $value]\] != -1" $body
		}
		set if elseif
	}
if $debug { log [debug -quiet code] }
	uplevel 1 $code
	return
}

proc rswitch args {
	# Do we need switches?
	flags:simple $args [list -debug] args flags
	set debug [validflag -debug]
	set args [lindex [lassign $args value] 0]
	set ll [llength $args]
	if { $ll % 2 } { error "\[RSWITCH\] extra switch pattern without body" }
	set if if
	foreach [list expression body] $args {
		if [string eq default $expression] {
			lappend code else $body
		} {
			lappend code $if "$value in [list [SB:get_count $expression]]" $body
		}
		set if elseif
	}
if $debug { log [debug -quiet code] }
	uplevel 1 $code
	return
}

# See replacement below ....
proc `uniquematch {valid a} {
	if [string eq "" $a] {return ""}
	set a [split $a]
	set ok 0

	if [iswildcard $a] {
		set m [lsearch -glob -all [stl $valid] [stl $a]]
		if [isempty m] {return ""}
		if {[llength $m] > 1} {return ""}
		return [lindex $valid $m]; # Protect case
	} {
		set m [lsearch -exact [stl $valid] [stl $a]]
		if {$m > -1} {
			set ok 1
			set a [lindex $valid $m]
		} else {
			set match ""
			foreach b $valid { if [string match -nocase ${a}* $b] { lappend match $b } }
			if {[llength $match] == 1} {set ok 1; set a $match}
		}
	}
	if ($ok) {return $a} {return ""}
}

# Replacement PROC
proc uniquematch { list matchme } {
	# No SB extensions!
	if [string eq "" $list] return
	if [string eq "" $matchme] return
	if [string eq * $matchme] { return $list }
	set list [lsort -uni -dict -inc $list] ; # Might be duplicates: if so, false negatives will occur ):
	set m [lsearch -inline -all -exact -nocase $list $matchme]
	if { [llength $m] == 1 } { return $m }
	set m [lsearch -inline -all -glob -nocase $list $matchme]
	if { [llength $m] == 1 } { return $m }
	set m [lsearch -inline -all -glob -nocase $list ${matchme}*]
	if { [llength $m] == 1 } { return $m }
	return ""
}

proc getrange {a { b -1 } { c 1001 } } {# B(-1):C(1000)
	if [isint $a] {return [list [noscinot $a] [noscinot $a]]}
	if ![regexp -- {^[0-9\-]+$} $b] {return "-1"}
	if ![regexp -- {^[0-9\-]+$} $c] {return "-1"}
	if {$a == "-1"} {return "-1"}
	if ![regexp -nocase -- {^[0-9<=>\-\+]+$} $a] {return -1}

	# Allow for new syntax: "99+49" --> "99-148"
	if [regexp -- {^(\-)?[\d]+\+[\d]+$} $a] {
		set split [split $a +]
		set a [l0 $split]-[ expr [ l0 $split ] + [ l1 $split ] ]
	}

	# Allow for new syntax: "501+" --> ">=501"
	if [regexp -- {^\d+\+$} $a] {set a >=[left $a -1]}

	empty 0 1
	switch -glob -- $a {
		=* {
			set num [mid $a 2]
			if ![regexp -- {^[\+\-]?\d+$} $num] {return "-1"}
			set 0 $num
			set 1 $num
		}
		<=* {# "<=" MUST come before "<"
			set num [mid $a 3]
			if ![regexp -- {^[\+\-]?\d+$} $num] {return "-1"}
			set 0 $b
			set 1 $num
		}
		>=* {# ">=" MUST come before ">"
			set num [mid $a 3]
			if ![regexp -- {^[\+\-]?\d+$} $num] {return "-1"}
			set 0 $num
			set 1 $c
		}
		<* {
			set num [mid $a 2]
			if ![regexp -- {^[\+\-]?\d+$} $num] {return "-1"}
			incr num -1
			set 0 $b
			set 1 $num
		}
		>* {
			set num [mid $a 2]
			if ![regexp -- {^[\+\-]?\d+$} $num] {return "-1"}
			incr num
			set 0 $num
			set 1 $c
		}
		*-* {
			set num [split $a -]

			# Now check for that damned "-1" issue ....
			if {[llength $num] > 2} {
				if {[l2 $num] == ""} {
					if {[llength $num] == 4} {
						lassign $num s1 s2 s3 s4
						if {[isempty s1] && [isempty s3] && [notempty s2] && [notempty s4]} {
							set num [list -$s2 -$s4]
						} else {
							return "-1"; # Some dumbass combination ....
						}
					} else {
						return "-1"; # "1-1001-" (error!)
					}
				} else {
					if {[l0 $num] == ""} {
						set num [list -[l1 $num] [l2 $num]]
						} else {
					set num [list [l0 $num] -[l2 $num]]
					}
				}
			}
			if {[llength [join $num]] < 2} {
				if [string eq [string index $a 0] -] {# $NUM not $A?
					set num [list $b $num]
				} elseif [right $a 1 -] {# $NUM not $A?
					set num [list [l0 $num] $c]
				}
			}
			lassign $num 0 1
		}
		*+* {
			#
			set num [split $a +]
			set 0 [lindex $num 0]
			set 1 [ expr $0 + [ lindex $num 1 ] ]
		}
		default { return -1 }
	}
	if {$0 < $b} { set 0 $b }
	if {$1 > $c} { set 1 $c }
	if {$1 < $0} { return -1 }
	return [list $0 $1]
}

# See replacement PROC below
proc `makenumberlist { crap { comma "," } { joiner " " } } {
	# 1,2,3,5,9 -> 1-3, 5, 9

	regsub -all -- ",| ,|, |  " $crap " " crap
	regsub -all -- ";| ;|; |  " $crap " " crap
	set temp ""
	foreach a $crap {
		if [regexp -- {^[0-9]+\-[0-9]+$} $a] {
			# Double-dash ("--") or multiple dash ("1-2-3") will get caught
			# by the normal REGEXP below; therefore: ignore it for now.
			set n1 [string trim [l0 [split $a -]]]; # Handles "13 - 15"
			set n2 [string trim [l1 [split $a -]]]; # Handles "13 - 15"
			if {$n2 < $n1} {return ""}; # Won't cause an endless loop, BTW.
			for {set b $n1} {$b <= $n2} {incr b} {lappend temp $b}
		} elseif [regexp -- {^\-[\d]+$} $a] {# Handle negative numbers
			lappend temp $a
		} else {
			lappend temp $a
		}
	}
	set crap $temp
	if ![regexp -- {^[0-9 \-]+$} $crap] {return ""}
	set old $crap
	set crap [lsort -inc -int -uni $crap]
	set string ""
	set buffer ""
	set hold ""
	append crap " 999999999"
	foreach a $crap {
		if {$buffer == ""} {set buffer $a}
		if {$hold == ""} {set hold $a}
		if {[ expr $a - $buffer ] < 2} {
			set buffer $a
			continue
		}
		if {$string != ""} {append string $comma$joiner}
		set newitem $hold
		if {$buffer != $hold} {append newitem "-$buffer"}
		append string $newitem
		set buffer $a
		set hold $a
	}
	return $string
}

# Replacement PROC
proc makenumberlist args {
	set list [join $args]
	set range ""
	set final ""
	while { ![string eq "" $list] } {
		set list [lsort -unique -increasing -real $list]
		set number [lindex $list 0]
		set list [lreplace $list 0 0]
		if [info exists last] {
			if { ( $number - $last ) == 1 } {
				lappend range $number
				set last $number
			} {
				lappend final [join [lsort -unique -increasing -real [list [lindex $range 0] [lindex $range end]]] -]
				set range $number
				set last $number
			}
		} {
			set last $number
			set range $number
		}
	}
	if ![string eq "" $range] { lappend final [join [lsort -unique -increasing -real [list [lindex $range 0] [lindex $range end]]] -] }
	ajl $final
}

# This can be replaced with SB:GET_COUNT (will handle all variations without recursion)
proc splitnumberlist args {
	set arg [join $args]; # Catch-all for all types of input formats
	set m [lsearch -glob $arg -r*]
	if {$m < 0} {
		zero default_bottom; set default_top 10000; # one default_bottom ; set default_top 1000
		# Using 2147483647 for TOP was causing obscene bot crashes
		# Let's try 10,000 for now.
	} {
		set n [ expr $m + 1 ]
		set range [lindex $arg $n]
		if ![regexp -- {^\d+\-\d+$} $range] {error "\[SPLITNUMBERLIST\] Illegal -RANGE specification: $range (expected integer-integer)"}
		lassign [split $range -] default_bottom default_top
		set arg [join [lreplace $arg $m $n]]
	}

	regsub -all ,|\\\; $arg " " arg
	set regexp [regexp -all -inline -- {=[\d]+?[^\d]} $arg]
	foreach exp $regexp {regsub -all $exp $arg [mid $exp 2] arg}
	if ![regexp -- {^[0-9 \-]+$} $arg] {return ""}
	set temp ""
	foreach a $arg {
		if [regexp -- {^\=?[\+\-]?\d+$} $a] {
			lappend temp [regsub {^\=} $a ""]
		} else {
			# Allow "-1-1000" to survive the REGEXP!
			if ![regexp -- {^[0-9]*\-[0-9]*(\-[0-9]*)?$} $a] {return ""}
			set split [split $a -]

			# Handle negative numbers!
			if [regexp -- {^\-[\d]+$} $a] {lappend temp $a; continue}

			# Do =NOT= add a NOZERO here: the next line needs the null to survive
			set n1 [string trim [lindex [split $a -] 0]]; # Handles "13 - 15"
			if {$n1 == ""} {set n1 $default_bottom}
			set n2 [string trim [lindex [split $a -] 1]]; # Handles "13 - 15"
			if {$n2 == ""} {set n2 $default_top}

#			set n1 [nozero $n1]; set n2 [nozero $n2]
			if {[llength $split] > 2} {set n1 -[lindex $split 1]; set n2 [lindex $split 2]}
			if {$n2 < $n1} {return ""}; # Won't cause an endless loop, BTW.
			for {set b $n1} {$b <= $n2} {incr b} {lappend temp $b}
		}
	}
	empty list
	foreach a $temp { if [isbetween $a $default_bottom $default_top] { lappend list $a } }
	return [lsort -inc -int -uni $list]
}

# Can be replaced with NORMALIZE
proc nozero {nozero { nonintegerok 0 } } {
	if [left $nozero 1 .] {prepend nozero 0}
	if !($nonintegerok) {
		if [isscinot $nozero] {return $nozero}
		# This is a standard integer.mantissa test, NOT a SCINOT test.
		# Allow the first "+" in the expression "\d+(\.\d+)? to be there.
		set nozero [lindex [regexp -inline -- {^[\+\-]?\d+(\.\d+)?([eE][\+\-]?\d+)?} $nozero] 0]
	}
	empty sign
	if [regexp -- {^[\+\-]} $nozero] {set sign [left $nozero 1]; set nozero [mid $nozero 2]}
	set nozero [string trimleft $nozero 0]
	if [instr $nozero .] {
		set nozero [string trimright $nozero 0]
		if [right $nozero 1 .] {set nozero [left $nozero -1]}
		if [left $nozero 1 .] {prepend nozero 0}
	}
	if [isempty nozero] {zero nozero}
	if ![string eq 0 $nozero] {set nozero ${sign}$nozero}
	return $nozero
}

proc normalize:var args {
	foreach variable $args {
		upvar 1 $variable local
		if [info exists local] { set local [normalize $local] } { set local 0 }
	}
}

proc normalize args {
	flags:simple [join $args] [list -boolean -falsetrue -truefalse -noyes -yesno -offon -onoff -scinot -hexadecimal -octal -binary -roman] value flags
	if [isempty value] { set value 0 }
	if [istrue $value] { set value 1 } ; # Allow TRUE / YES / ON to survive
	if [string eq "" $flags] { return [number:clean $value] }

	# We're missing conversion flags below! -- if [regexp -- {^[\+\-]?(0|[1-9]\d*)$} $value] { return $value } ; # Let base-10 INTs escape
	if [regexp -nocase -- {^(\$|#$|0x|\\x|\\0x|\\u)[0-9A-F]+$} $value] { set value [ expr 0x[ lindex [ regexp -inline -all -nocase -- {[^0-9A-F]*([0-9A-F]+)} $value ] end ] ] } ; # Hexadecimal
	if [regexp -nocase -- {^(&|0o|0)[0-7]+$} $value] { set value [ expr 0[ lindex [ regexp -inline -all -nocase -- {[^0-7]*([0-7]+)} $value ] end ] ] } ; # Octal
	if [regexp -nocase -- {^(%|0b)[01]+$} $value] { set value [ expr 0b[ lindex [ regexp -inline -all -nocase -- {[^01]*([01]+)} $value ] end ] ] } ; # Binary
	if [regexp -nocase -- {^(?:\#|0d)([0-9]+)$} $value - v] { set value $v } ; # Decimal
	if [regexp -nocase -- {^(\||0r)?[MDCLXVI]+$} $value] { set value [roman $value] } ; # Roman
	if [regexp -nocase -- {^\d(\.\d+)?E[\+\-]\d+$} $value] { set value [noscinot $value] }

	# Now that the conversions are done, do we have a valid number?
	if ![regexp -nocase -- {^[\+\-]?(\d+|\d+\.\d*|\d*\.\d+)$} $value] { return 0 }

	set value [number:clean $value]

	if [validflag -boolean] { return [ expr $value ? 1 : 0 ] }
	if [validflag -offon -onoff] { return [offon [ expr $value ? 1 : 0 ]] }
	if [validflag -noyes -yesno] { return [noyes [ expr $value ? 1 : 0 ]] }
	if [validflag -falsetrue -truefalse] { return [falsetrue [ expr $value ? 1 : 0 ]] }
 	if [validflag -scinot] { return [scinot $value] }
	if [validflag -hexadecimal] { return [hex $value] }
	if [validflag -octal] { return [oct $value] }
	if [validflag -binary] { return [bin $value] }
	if [validflag -roman] { return [roman $value] }

	string trimleft $value +
}

# See NUMBER NAMESPACE definition below
proc number:clean number {
	# ISSCINOT has a syntax check for SINGLE-DIGIT-decimal-mantissa, which won't
	# allow this to work properly
	# Do not do! if [regexp -nocase -- {^[+\-]?\d+(\.\d+)?[Ee][+\-]?\d+$} $number] { # }
	if [isscinot $number true] { # $LOOSE
		# We don't want to return "0000001E+000000005" do we?!
		regexp -nocase -- {^([+-]?)(\d*)\.?(\d*)0*[Ee]([+-]?)0*(\d*)$} $number - sign integer mantissa polarity exponent
		# Recursive call (Uh oh?)
		set float [number:clean [join [concat $integer $mantissa] .]]
		if [isempty polarity] { set polarity + }
		return ${sign}${float}E${polarity}${exponent}
	}
	if ![isnum -real $number] { return $number }
	set number [string trimleft $number +]
	regexp -- {^([+-]?)(.+)$} $number - sign number
	lassign [split $number .] int mant
	scan $int %lld int
	set mant [string trimright $mant 0]
	if [isempty mant] { return ${sign}$int } { return ${sign}${int}.$mant }
}

namespace eval number {
	proc roll { value roll } {
		 if [string match *.*.* $value] { error "\[ROLL\] Illegal binary value: $value" }
		 if ![regexp -- {^[\+\-]?\d+$} $roll] { error "\[ROLL\] Illegal roll value: $roll" }
		 set roll_abs [ expr abs( $roll ) ]
		 if ![string match *.* $value] { append value .0 }
		 set value "[string repeat 0 $roll_abs]${value}[string repeat 0 $roll_abs]"
		 set radix [ expr [ string first . $value ] + 1 ]
		 regsub -- {\.} $value "" naked
		 incr radix [incr roll -1]
		 clean [string range $naked 0 [ expr $radix - 1 ]].[string range $naked $radix end]
	}

	proc clean number {
		# 2017-04-02 14:53:00 -0700: Thanks to Elfman for helping me logic the polarity priorities out
		set number [string trimleft $number +]
		regexp -- {^([-]?)[+-]*(.+)$} $number - sign number
		lassign [split $number .] int mant
		set int [string trimleft $int 0]
		if [isempty int] { set int 0 }
		set mant [string trimright $mant 0]
		if [isempty mant] { return ${sign}$int } { return ${sign}${int}.$mant }
	}

	proc eval value {
		if ![regexp  -- {^(\d+\.?\d*)[[:space:]]*([qrstuvwxyzafpnumKMGTPEZYXWV])([Ii]?)B?$} $value - coefficient factor binary] { return $value }
		set list [list qrstuvwxyzafpnum KMGTPEZYXWV]
		set all [split [join $list -] ""]
		set offset [lsearch -exact $all -]
		set match [lsearch -exact $all $factor]
		if { $match == -1 } { error "\[EVAL\] Illegal factor: $factor" }
		set exponent [ expr ( $match - $offset ) ]
		set base 1000
		if [string eq -nocase I $binary] { set base 1024 }
		if { $exponent <= 0 } {
			set pow [ expr pow( $base , $exponent ) ]
		} {
			set pow [ expr $base ** $exponent ]
		}
		if [right $coefficient 1 .] { set coefficient [left $coefficient -1] }
		if [regexp -- {^\d+(\.0*)?$} $coefficient] { # ^\d+\.?%
			set value [ expr { $coefficient * $pow } ]
		} {
			# Break up integer and mantissa portions and do separate maths
			set leftover [lassign [split $coefficient .] integer mantissa]
			if ![string eq "" $leftover] { error "Too many decimal points in value: $coefficient" }
			set value 0
			# Integer
			set 1 [ expr $integer * $pow ]
			# Mantissa
			if { $pow >= 1 } {
				set length [string length $mantissa]
				set 2_ [ expr $mantissa * $pow]
				set 2i [string range $2_ 0 end-$length]
				set 2m [string range $2_ end-[ expr $length - 1 ] end]
				set value [ expr $1 + $2i ]
				if ![regexp -- {^(|0+)$} $2m] { set value [ expr $value + 0.$2m ] }
			} {
				# Exponent < 0
				# $coefficient * $pow
				regexp -- {^(\d+)\.?(\d+):(\d+)\.?(\d+)$} ${coefficient}:${pow} - ci cm pi pm
				set shift [ expr [ string length $cm ] + [ string length $pm ] ]
				set value [ expr [ clean ${ci}${cm} ] * [ clean ${pi}${pm} ] ]
				set value [roll $value -$shift]
			}
		}
		return $value
	}
}

# Replacement PROC
proc number:eval number { number::eval $number }
interp alias {} numeval {} number:eval
#interp alias {} number:eval {} ::number:eval ; # Error: cannot define or rename alias "number:eval": would create a loop

interp alias {} padzero {} pad
interp alias {} addzero {} pad
proc pad { number pad_integer { pad_mantissa -1 } } {
	# Force rounding?
	lassign [split $number .] integer mantissa
	if { $pad_integer < [len $integer] } { return $number }
	set addzero [format %0${pad_integer}s $integer]
	switch -exact -- $pad_mantissa {
		-1 { if [notempty mantissa] { append addzero .$mantissa } }
		0 { # Append nothing, no matter what! # }
		default { append addzero .[string reverse [format %0${pad_mantissa}s [string reverse $mantissa]]] }
	}
	return $addzero
}

# Replaced by SPLIT:BASE
proc `split_by_base { number { base 0 } { integer 0 } { mantissa 0 } } {
	proc split_by_base_reverse text { set split_by_base_reverse "" ; foreach letter [split $text ""] { set split_by_base_reverse "${letter}$split_by_base_reverse" } ; return $split_by_base_reverse }
	if [string match *.*.* $number] { error "\[SPLIT_BY_BASE\] Illegal number: \"${number}\"" }
	regexp -nocase -- {^([^\d]*)} $number - header
	set number [string range $number [string length $header] end]
	set mantissa2 0
	if !$base { switch -glob -- [string tolower $header] { *%* - 0b* { set base 2 } *&* - 0o* - 0\[0-7\]* { set base 8 } $* - #$* - 0x { set base 16 } default { set base 10 } } }
	if !$integer { switch -exact -- $base { 2 { set integer 4 } 8 { set integer 3 } 10 { set integer 3 ; set mantissa2 2} 16 { set integer 2 } default { set integer 2 } } }
	if { [lsearch -exact [list 2 8 10 16] $base] == -1 } { error "\[SPLIT_BY_BASE\] Illegal base: \"${base}\" (expected 2, 8, 10, or 16)" }
	if !$mantissa { set mantissa $integer }
	if !$mantissa2 { set mantissa2 $mantissa }
	lassign [split $number .] _int _man
	if ![string eq "" $_int] { set _int [split_by_base_reverse $_int] ; set _int [split_by_base_reverse [regexp -all -inline -- ".\{1,${integer}\}" $_int]] ; if $base { set _int [join $_int ", "] } }
	if ![string eq "" $_man] { set _man [concat [regexp -inline -- ".\{1,${mantissa2}\}" $_man] [regexp -all -inline -start $mantissa2 -- ".\{1,${mantissa}\}" $_man]] }
	return ${header}${_int}[lindex [list .$_man] [ expr [ string eq "" $_man ] ? 1 : 0 ]]
}

proc SB:checkbugsiebug args {
	if [isint $args] { set dccidxlist $args } { set dccidxlist [getfirst [dcclist chat]] }
	foreach idx $dccidxlist {
		if ![validuser [idx2hand $idx]] {
			set dccentry [lsearch -glob -inline [dcclist chat] "$idx *"]
			set dcchandle [l1 $dccentry]
			set dcchost [l2 $dccentry]
			set dccchan [l1 [split [l2 [l4 $dccentry]] /]]
			set dcctime [l5 $dccentry]
			set dccchan "party line"
			if [regexp -- {^[0-9]+$} $dccchan] {set dccchan "#$dccchan"}
			killdcc $idx "Who the hell are you ?!"
			msghome "\[Bugsie Bug\] I just killed the DCC CHAT connection of unknown user \"\002\00304${dcchandle}\003\002\" (\$DCCHOST:$dcchost - DCC channel ${dccchan} - connection time: $dcctime / [formattime $dcctime])"
			return 0
		}
	}
}

interp alias {} MSGHOME {} msghome
interp alias {} msghoem {} msghome; # Damn typoes!
proc msghome  text {msghome:method "" $text}
proc msghomeh text {msghome:method h  $text}
proc msghomeq text {msghome:method q  $text}
proc msghomer text {msghome:method r  $text}

proc msghome:method {method text} {
	set home [sbdget CONFIG:GLOBAL-HOME]
	if [isempty home] {return ""}
	if ![validchan $home] {return ""}
	if [SB:ccs $home mute] {return ""}

	set color_home [sbd:get CONFIG:GLOBAL-HOMECOLOR]
	foreach line [split $text \n] {
		set line [SB:outputcolor $line $color_home HOMECHAN]
		msg$method $home $line
	}
	return $text
}

proc msgchan  {chan text} {msgchan:method "" $chan $text}
proc msgchanh {chan text} {msgchan:method h  $chan $text}
proc msgchanq {chan text} {msgchan:method q  $chan $text}
proc msgchanr {chan text} {msgchan:method r  $chan $text}

proc msgchan:method {method chan text} {
	if ![validchan $chan] {return ""}
	if [SB:ccs $chan mute] {return ""}

	set color_home [sbd:get CONFIG:GLOBAL-COLOR]
	foreach line [split $text \n] {
		set line [SB:outputcolor $line $color_home CHAN]
		msg$method $chan $line
	}
	return $text
}

interp alias {} delbind {} killbind
interp alias {} rembind {} killbind

proc addbind {type flags trigger procname} {
	killbind $type $trigger $procname
	bind $type $flags $trigger $procname
	return $trigger
}

proc killbind {a b c {d ""}} {
	if {$a == "" || $b == "" || $c == ""} {return "-1"}
	if {$d == ""} {set d $c; set c $b; set b ""}
	# Allow [ADD]BIND to be copied/pasted with the "FLAGS" for KILLBIND.

	set binds [binds $a]
	set counter 0
	foreach loop $binds {
		if {[string eq -n [l2 $loop] $c] && [string eq -n [l4 $loop] $d]} {
			# L3 $LOOP is useless (BIND's hit counter)
			set c [l2 $loop]; # Force case match
			set d [l4 $loop]; # Force case match
			if {$b != ""} {set b [l1 $loop]}; # Force flags to match
			catch {unbind $a $b $c $d}
			incr counter
		}
	}
	return $counter
}

proc killbind_target proc {
	zero tally
	foreach a [binds] {
		if [string eq -n [l4 $a] $proc] {
			incr tally
			killbind [l0 $a] [l1 $a] [l2 $a] [l4 $a]
		}
	}
	return $tally
}

proc getfirst   data {getpos $data 1}
proc getsecond  data {getpos $data 2}
proc getthird   data {getpos $data 3}
proc getfourth  data {getpos $data 4}
proc getfifth   data {getpos $data 5}
proc getsixth   data {getpos $data 6}
proc getseventh data {getpos $data 7}
proc geteighth  data {getpos $data 8}
proc getninth   data {getpos $data 9}

proc getpos {data pos} {
	if [string eq "" $data] {return ""}
	if ![isint $pos] {return ""}
	decr pos
	empty getpos
	foreach a $data {lappend getpos [lindex $a $pos]}
	return $getpos
}

interp alias {} checkchannelsetting {} SB:ccs
proc SB:ccs {chan setting} {
	if [regexp -- {^[\+\-]} $setting] {
		set mode [lsearch [list - +] [left $setting 1]]
		# Check for specific status
		set error [ catch { set ccs [channel get $chan [string trimleft $setting +-]] } crap ]
		if !$error {
			if { $ccs == $mode } { return 1 } { return 0 }
		}
	}
	# Otherwise, check for "does it exist" ....
	set ccs ""
	# 2017-05-22 16:23:00 -0700: This should never work (previous versions did?)
	#set error [ catch { set ccs [ channel info $chan $setting ] } crap ]
	#if !$error { return $ccs }
	set error [ catch { set ccs [ channel get $chan $setting ] } crap ]
	if !$error { return $ccs }
	return $ccs
}

proc SB:dccnick {nick handle {chan ""}} {
	if {$chan == ""} {set chan [l0 [channels]]}
	set authnick [SB:authnick $handle]
	if {[SB:authflagcheck $handle] && ($authnick != "") && ![SB:isdccnick $authnick]} {
		return $authnick
	}
	if [SB:isdccnick $nick] {
		set nick [l0 [findchan $handle]]
		if [notempty nick] {return $nick}
		return #DCCUSER:${handle}#
		# An impossible nick so the wrong person doesn't get bothered.
	}
	return $nick
}

proc SB:authflagcheck a { if [validuser $a] { return [nozero [sbd:get #AUTHED:$a]] } { return 0 } }
proc SB:authflagset   a { if [validuser $a] { sbd:set #AUTHED:$a 1 } }
proc SB:authflagclear a { if [validuser $a] { sbd:set #AUTHED:$a 0 } }

proc SB:oauthflagcheck a { if [validuser $a] { return [normalize [sbd:get #OAUTHED:$a]] } { return 0 } }
proc SB:oauthflagset   a { if [validuser $a] { sbd:set #OAUTHED:$a 1 } }
proc SB:oauthflagclear a { if [validuser $a] { sbd:set #OAUTHED:$a 0 } }

proc SB:authnick {handle { host "" }} {
	if ![validuser $handle] {return ""}
	if {$host != ""} {
		if {([authhost $handle] != $host) && ([authhost $handle] != "")} {return ""}
	}
	set an [sbd:get #AUTH:$handle]
	if [instr $an :] {set an [lindex [split $an :] 0]}
	return $an
}

proc SB:setauthnick {handle { nick "" } { host "" }} {
	if ![validuser $handle] {return ""}
	sbd:set #AUTH:$handle $nick
	if {$host != ""} {SB:setauthhost $handle $host}
}

proc SB:deauth user {SB:unsetauthnick $user}
proc SB:unsetauthnick userlist {
	if [string eq * $userlist] {set userlist [userlist]}
	foreach user $userlist {
		if ![validuser $user] continue
		SB:setauthnick $user ""
		SB:setauthhost $user ""
		chattr $user -P
		SB:authflagclear $user
		SB:oauthflagclear $user
		SB:logout_wedge -process $user
	}
	return $userlist
}

proc SB:logout_wedge args {
	if ![left [l0 $args] 1 -] {return ""}
	set base @WEDGELIST:LOGOUT

	switch -exact -- [sl0 $args] {

		-list {return [sbd:get $base]}

		-add {sbd:lappend $base [l1 $args]; return [l1 $args]}

		-del {
			set w [sbd:get $base]
			foreach a [lrange $args 1 end] {
				lremove w $a
			}
			sbd:set $base $w
			return $w
		}

		-clear {sbd:set $base; return ""}

		-process {foreach a [sbd:get $base] { foreach b [lrange $args 1 end] { $a $b } } ; return 0}

		default { error "\[SB:LOGOUT_WEDGE\] Unknown command flag: [su0 $args]" }

	}

	error ?

}

proc SB:setauthflag {user { mode 0 } } {
	set mode [string is true -strict $mode]
	switch -exact -- $mode {
		0 {SB:authflagclear $user}
		1 {SB:authflagset   $user}
		default {return "-1"}
	}
	return $mode
}

proc SB:authhandle {user { host "" } } {
	foreach i [userlist] {
		if {[sbd:get #AUTH:$i] == $user} {
			set ok 1
			if {$host != ""} {
				if {[sbd:get #AUTHHOST:$i] != $host} {set ok 0}
			}
			if ($ok) {return $i}
		}
	}
	return ""
}

proc SB:authhost user {
	if ![validuser $user] {return ""}
	set host [sbd:get #AUTHHOST:$user]
	if [SB:isdcchost $host] {return ""}
	return $host
}

proc SB:setauthhost {user { host "" }} {
	if [validuser $user] {sbd:set #AUTHHOST:$user $host}
}

proc SB:isauthed {nick handle { host "" }} {
	if {[string eq "" $nick] || [string eq "" $handle] || ![validuser $handle]} {return 0}
	set authed 1
	if ![SB:authflagcheck $handle] {set authed 0}
	if {[SB:authnick $handle] != $nick} {
		if ![string match #* $nick] {set authed 0; # If NICK = #DCCUSER:*#, then allow it through.}
	}
	if {$host != ""} {
		if {[SB:authhost $handle] != ""} {
			if {$host != [SB:authhost $handle]} {
				if ![string match #* $host] {set authed 0; # If HOST = #DCCIDX:*#, then allow it through.}
			}
		}
	}
	return $authed

	#The above concoction allows for (1) the checkhost to be missing <OR>
	#(2) the AUTHHOST to be missing, and still let the ISAUTHED be valid.
	#This is a backwards-compatibility issue only.
	return 0
}

proc SB:isoauthed handle {nozero [sbd:get #OAUTHED:$handle]}

proc SB:setauthdata {handle nick host} {
	SB:setauthnick $handle $nick
	SB:setauthhost $handle $host
	SB:setauthflag $handle on
	#SB:oauthflagclear $handle; # When logging in, force LOGIN then OLOGIN
	# No! Let user do it in either order!
	return 0
}

# 2017-03-30 16:35:00 -0700: interept GETUSER/SETUSER calls to build the valid list
# The problem is: can't error-catch bad calls (trying to keep $ERRORINFO from becoming
# contaminated, cf. CHANGELOG #3208)
# Use SBD:SET (SBD:LAPPEND calls SBD:SET but doesn't allow /pass the QUIET flag)
proc setuser:wedge args {
	lassign $args who what
	set error [ catch { set value [eval setuser:original $args] } crap ]
	if $error { error $crap }
	if ![string eq -nocase XTRA $what] { sbd:set @getuser:valid [lsort -unique -increasing -dictionary [string tolower [concat [sbd:get @GETUSER:VALID] $what]]] quiet }
	return $value
}

proc getuser:wedge args {
	lassign $args who what
#putlog "\[GETUSER:WEDGE\] ARGS($args):WHO($who):WHAT($what):VALID([sbd:get @getuser:valid])"
	set error [ catch { set value [eval getuser:original $args] } crap ]
	if $error { error $crap }
	if ![string eq -nocase XTRA $what] { sbd:set @getuser:valid [lsort -unique -increasing -dictionary [string tolower [concat [sbd:get @GETUSER:VALID] $what]]] quiet }
	return $value
}

proc getuser:valid { { mode "" } } {
	# 2017-03-28 13:33:01 -0700: [POLITICAL] I've tried, multiple times, to get this added
	# into Eggdrop core. The common argument I was given was "it's not necessary" or
	# "there are other ways to do it: deluser and re-add" despite the attempts to explain
	# user data merging. Due to this, I had to write this custom command; it may (probably will)
	# fall out-of-date and become obsolete in short order. Pending for inclusion into
	# Eggdrop 1.8.3 (currently pending, per thommey): [GETUSER <handle>] will show all
	# set user settings. While this is a half-step, I will still have to loop through the
	# userlist to collect all possibilities. This option WILL FAIL if no users are in the
	# userlist (eggdrop -m), so initialization won't be possible when the bot is
	# cold-started. I will try to manufacture a work-around for this once 1.8.3 is released.

	set default [list botaddr botfl comment dccdir fstat handle hosts info laston pass xtra]
	lremove default xtra ; # It's handled separately
	# 2017-03-28 13:33:00 -0700: for eggdrop 1.9.0, add "PASS2" due to changes to userfile encryption
	if { $::numversion >= 1090000 } { lappend default pass2 }

	set valid [sbd:get @getuser:valid]
	switch -exact -- [string tolower $mode] {
		force { set valid "" }
		test {
			if { [llength [userlist]] == 1 } { error "\[GETUSER:VALID TEST\] Can't perform test: no users in userlist" }
			set sucker [lindex [userlist] 0]
			# Eggdrop 1.8.3: we can just LSEARCH / LMATCH [getuser $sucker ""] to test matches
			foreach _ $valid {
				set error [ catch { set 0 [getuser $sucker $_] } ]
				if $error { lremove valid $_ }
			}
			sbd:set @getuser:valid $valid
			return $valid ; # Or just fall through
		}
	}
	if [notempty valid] { return $valid }
	#set error [ catch { set valid [getuser [lindex [userlist] 0]] } ] ; # CAN'T WORK WITH NOUSERS (eggdrop -m)
	set error [ catch { set valid [getuser ""] } ] ; # Eggdrop issue #350
	if $error { set valid $default } ; # Eggdrop issue #350 was not resolved; use current settings as defaults
	lremove valid xtra
	set valid [lsort -unique -increasing -dictionary [string tolower $valid]] ; # Save for next time
	sbd:set @getuser:valid $valid
}

proc getdata { user item } {
	if ![validuser $user] { return ; # Return empty string
		set level [info level]
		putcmdlog "\[GETDATA\] No such user: $user (invoked by: [info level [incr level -1]])"
		error "\[GETDATA\] No such user: $user"
	}
	if { [lsearch -exact [getuser:valid] [string tolower $item]] == -1 } {
		set getdata [getuser $user xtra $item]
	} {
		set getdata [getuser $user $item]
	}
	# if [left [stl $getdata] 2 0x] {set getdata [binary format H* [mid $getdata 3]]}
	return $getdata
}

proc setdata { user item { new "" } } {
	if ![validuser $user] {
		set level [info level]
		putcmdlog "\[SETDATA\] No such user: $user (invoked by: [info level [incr level -1]])"
		error "\[SETDATA\] No such user: $user"
	}
	if { [lsearch -exact [getuser:valid] [string tolower $item]] == -1 } {
		setuser $user xtra $item $new
	} {
		setuser $user $item $new
	}
	return $new
}

proc isemptydata [list user item] { string eq "" [getdata $user $item] }

proc instr:all args { # text find_me { start 1 }
	flags:simple $args -nocase data flags
	lassign $data text find_me start
	if ![isnum -integer $start] { set start 1 }
	if [validflag -nocase] { set text [stl $text] ; set find_me [stl $find_me] }
	set list [list]
	incr start -1
	while 1 {
		set first [string first $find_me $text $start]
		if { $first == -1 } break
		incr first
		lappend list $first
		set start $first
	}
	return $list
}
proc instr    {string frag { limit          0 } } {instring instr    $string $frag $limit}
proc instrrev {string frag { limit 2147483647 } } {instring instrrev $string $frag $limit}
proc instring {dir string frag limit} {
	if {$string == ""} {return 0}
	if {$frag == ""} {return 0}
	if ![regexp -- {^[\+\-]?\d+$} $limit] {return -code error "INSTRING - $limit is an invalid position. Position must be an integer."}
	incr limit -1
	switch $dir {
		instr    {set instr [string first $frag $string $limit]}
		instrrev {set instr [string last  $frag $string $limit]; # If this keeps failing, try: string FIRST of reverse-string reverse-checkme}
		default  {error "INSTRING: unknown value for \$DIR ($dir)"}
	}
	incr instr +1
	return $instr
}

proc null  {} {return ""}
proc space { { s 1 } } {string repeat " " $s}
proc tab   { { t 1 } } {string repeat \t $t}
proc dash  { { d 1 } } {string repeat - $d}
proc nbsp  { { n 1 } } {string repeat \xA0 $n}
proc singlespace text {regsub -all { [ ]+} $text " " text; return $text}

# Use SUBST to reverse
proc u-escape string { subst [string map {\n {\\u000a}} [regsub -all {[][{};#\\\$ \r\t\u0000-\u001F\u0080-\uFFFF]} $string {[format \\\\u%04x [scan "\\&" %c]]}]] }

interp alias {} checkflags5 {} checkflags
proc checkflags {var1 var2 flags valid { match 0 } { smv 0 } } {
	error "\[CHECKFLAGS(5)\]: somehow, I was invoked. Why?"

	# This PROC (the SMV call) requires $ARG from the calling (parent) PROC.

	set ::SB_cfa $var1
	set ::SB_cfb $var2

	set ::SB_cfo [l0 $flags]
	set ::SB_cff [l1e $flags]
	set ::SB_cfv $valid
	set ::SB_cfm $match
	set ::SB_cfs $smv
	set ::SB_cfx $flags

	uplevel 1 {
		set arg_old $::SB_cfx
		set ::SB_cfr [SB:setmiscvariables_checkflags both $::SB_cff $::SB_cfv $::SB_cfm]
		set $::SB_cfa [l0 $::SB_cfr]
		set $::SB_cfb [l1 $::SB_cfr]
		if ($::SB_cfs) {
			set $::SB_cfb [linsert [set $::SB_cfb] 0 $::SB_cfo]
			SB:setmiscvariables 0
		}
	}

	foreach a [info vars ::SB_cf*] {unset $a}
	return 0
}

proc variables args {
	# Brace this one (needs to be fully up scoped)
	uplevel 1 {
		if { [info exists args] && [allempty nick host handle chan arg] } {
			lassign $args nick host handle chan arg
		}
	}
	# Don't brace this one (need local variable)
	uplevel 1 SB:setmiscvariables $args
}

interp alias {} SB:setvariables {} SB:setmiscvariables
interp alias {} SB:smv          {} SB:setmiscvariables
interp alias {} SB:sv           {} SB:setmiscvariables

proc SB:setmiscvariables args {
	upvar 1 checkflags checkflags
	upvar 1 flags flags
	upvar 1 arg arg

	empty flags checkflags
	emptyarray smv

	foreach proc_flag [list both setvar] {
		set smv($proc_flag) 0
		set m [lsearch -exact [stl $args] --smv:$proc_flag]
		if {$m > -1} {
			set smv($proc_flag) 1
			set args [join [lreplace [split $args] $m $m]]
		}
	}

	# Allow SB:SETMISCVARIABLES $VALIDFLAGS
	if {[llength $args] == 1} { if { [llength [l0 $args]] > 1 } { set args [l0 $args] } }

	if [left $args 1 -] {
		set valid $args
		set arg [split $arg]
		set checkflags [join [lindex $arg 0]]
		set arg [lreplace $arg 0 0]

		set ok_flags 1
		foreach word $arg {
			if [string eq $word --] {set ok_flags 0; continue; # Do not include!}

			# Must immediately precede $OK_FLAGS check:
			if [regexp -- {^[\?\*]+$} $word] {zero ok_flags 0}
			if ![left $word 1 -] {zero ok_flags 0}
			# </must>

			# If we're still checking flags ....
			if $ok_flags {
				set word [none [uniquematch $valid $word] $word]
				if {[lsearch -exact [string tolower $valid] [string tolower $word]] > -1} {
					lappend flags $word
				} {lappend checkflags $word}
			} {lappend checkflags $word}
		}

		set flags [join $flags]

		if $smv(setvar) {
			emptyarray checkflags_new
			foreach entry $valid {set checkflags_new([mid $entry 2]) [isvalidflag $flags $entry]}
			unset flags
			array set flags [array get checkflags_new]
		}

		if $smv(both) {# Do what with this? #}

		set checkflags [join $checkflags]
	} {set checkflags $arg}

	uplevel 1 {
		if [info exists nick] {
			set dccnick [SB:dccnick $nick $handle]
			set nph     [nph        $nick $handle]
			set nphd    [nph     $dccnick $handle]
			set nphr    [nphr       $nick $handle]
			set nphdr   [nphr    $dccnick $handle]
		}

		for {set SB_SET1 0; set SB_SET2 0} {$SB_SET1 <= 10} {incr SB_SET1; incr SB_SET2} {
			set checkflags_temp$SB_SET1 [lrange $checkflags $SB_SET2 $SB_SET2]
			set     $SB_SET1      [set checkflags_temp$SB_SET1]
			set    l$SB_SET1 [stl [set checkflags_temp$SB_SET1]]
			set    u$SB_SET1 [stu [set checkflags_temp$SB_SET1]]
			set    s$SB_SET1 [sts [set checkflags_temp$SB_SET1]]
			set    t$SB_SET1 [stt [set checkflags_temp$SB_SET1]]
			set    r$SB_SET1 [str [set checkflags_temp$SB_SET1]]

			set   j$SB_SET1 [join      [set checkflags_temp$SB_SET1]]
			set  jl$SB_SET1 [join [stl [set checkflags_temp$SB_SET1]]]
			set  ju$SB_SET1 [join [stu [set checkflags_temp$SB_SET1]]]
			set  js$SB_SET1 [join [sts [set checkflags_temp$SB_SET1]]]
			set  jt$SB_SET1 [join [stt [set checkflags_temp$SB_SET1]]]
			set  jr$SB_SET1 [join [str [set checkflags_temp$SB_SET1]]]

			set checkflags_temp${SB_SET1}e [lrange $checkflags $SB_SET2 end]
			set     ${SB_SET1}e [set checkflags_temp${SB_SET1}e]
			set    l${SB_SET1}e [stl [set checkflags_temp${SB_SET1}e]]
			set    u${SB_SET1}e [stu [set checkflags_temp${SB_SET1}e]]
			set    s${SB_SET1}e [sts [set checkflags_temp${SB_SET1}e]]
			set    t${SB_SET1}e [stt [set checkflags_temp${SB_SET1}e]]
			set    r${SB_SET1}e [str [set checkflags_temp${SB_SET1}e]]

			set    j${SB_SET1}e [join      [set checkflags_temp${SB_SET1}e]]
			set   jl${SB_SET1}e [join [stl [set checkflags_temp${SB_SET1}e]]]
			set   ju${SB_SET1}e [join [stu [set checkflags_temp${SB_SET1}e]]]
			set   js${SB_SET1}e [join [sts [set checkflags_temp${SB_SET1}e]]]
			set   jt${SB_SET1}e [join [stt [set checkflags_temp${SB_SET1}e]]]
			set   jr${SB_SET1}e [join [str [set checkflags_temp${SB_SET1}e]]]
		}
	}

	unset -nocomplain ::SB_SMV

	return 0
}

proc SB:setmiscvariables_checkflags {type flags valid { match 0 } } {
	set loop 0; set end 999
	set startflag -1; set startparam 0
	set endflag -1; set endparam $end

	# "--" will not survive the test positively, so this will react properly.

	if {$valid != ""} {
		if ($match) {# if {($match) || [iswildcard $flags]} {....}
			set temp ""
			set flagmode 1
			foreach a $flags {
				if [iswildcard $a] {lappend temp $a; continue}
				if [string match ? $a] {lappend temp $a; continue}
				if ![string eq -nocase [string index $a 0] -] {set flagmode 0}
				if {([uniquematch $valid $a] != "") && ($flagmode)} {
					lappend temp [uniquematch $valid $a]
				} else {
					lappend temp $a
				}
			}
			set flags $temp
		}
		foreach a $valid {array set 5 "[string trimleft $a -] 0"}
		for {set i 0} {$i < [llength $flags]} {incr i} {
			set okflag 0

			if ![isvalidflag [stl $valid] [stl [lindex $flags $i]]] {set okflag 1}
			if ![isvalidflag $valid [lindex $flags $i]] {#}
			if ($okflag) {
				set startparam $i; set endparam $end
				if ($i) {
					incr i -1
					set endflag $i
					if {$endflag < 0} {
						set startflag -1
					} else {
						set startflag 0
					}
					break
				} else {
					set startflag -1; set endflag -1
					break
				}

			}
			set endflag $i
			set startparam [ expr $i + 1 ]
			set endparam $end; # Max line size is like 512 chars.
		}
	}
	if {[lindex $flags $startparam] == "--"} {
		incr startparam
		incr endparam
		# $ENDPARAM because if the "--" is the final arg, end < start!
	}
	# 1206263336 Force flags to lower case!

	set return_flags [jstl [lrange $flags $startflag  $endflag ]]
	set return_args        [lrange $flags $startparam $endparam]

	switch -glob -- $type {
		data - arg* - info {return $return_args}
		flag* {return $return_flags}
		both {return [list $return_flags $return_args]}
		setvar {
			foreach a $return_flags {
				set a [string trimleft $a -]
				if {$a != ""} {set 5($a) 1}
			}
			return [array get 5]
		}
	}
}

proc SB:setparams args {
	# First word will always be a command word; assume [A-Z0-9_-]
	# No need to protect it from list-vs-string
	upvar 1 arg local_arg
	set local_arg [lassign $local_arg command]
#debug =0 command local_arg
	upvar 1 flags local_flags
	# FLAGS -PARAMS will eat the first element automatically
#debug =1 local_arg args
	#set args [lassign $args command]
	flags:params -force $local_arg $args local_arg local_flags
#debug =2 local_arg local_flags
	# Can't pipe to SB:SETMISCVARIABLES ($FLAGS Will get overridden as a scalar variable)
	# Let's play it simple
	set local_split [concat $command [split $local_arg]]
	for { zero x } { $x <= 10 } { incr x } {
		uplevel 1 set $x [subst -noback -nocomm { [lindex {$local_split} $x] }]
		uplevel 1 set u$x [subst -noback -nocomm { [string toupper [lindex {$local_split} $x]] }]
		uplevel 1 set l$x [subst -noback -nocomm { [string tolower [lindex {$local_split} $x]] }]
		uplevel 1 set s$x [subst -noback -nocomm { [string totitle [lindex {$local_split} $x]] }]
		uplevel 1 set t$x [subst -noback -nocomm { [stt             [lindex {$local_split} $x]] }]
		uplevel 1 [subst -noback -nocomm { set ${x}e [lrange {$local_split} $x end] }]
		catch { uplevel 1 set j${x} [subst -noback -nocomm { [join [lindex {$local_split} $x]] }] }
		catch { uplevel 1 set j${x}e [subst -noback -nocomm { [join [lrange {$local_split} $x end]] }] }
	}
	uplevel 1 set je [join [lindex $local_split end]]
	return
}

proc SB:pull_flags {valid text} {
	error "\[SB:PULL_FLAGS\] Am I still being used?!"
	# RETURNs: [LIST <matched flags> <remaining text> <unknown "possible" flags>]
	if ![left $text 1 -] {return [list "" $text ""]}

	zero double
	set last [lsearch -not -glob $text -*]
	if {$last < 0} {
		set temp $text
		empty text
	} {
		set doub [lsearch -exact $text --]
		if { ( $doub > -1 ) && ( $doub <= $last ) } {set last $doub; one double}
		set temp [lrange $text 0 [ expr $last - 1 ]]
		set text [lrange $text [ expr $last + $double ] end]
	}

	empty flags i
	foreach t $temp {
		set u [uniquematch $valid $t]
		if [isempty u] {lappend i $t} {lappend flags $u}
	}

	return [list $flags $text $i]
}

interp alias {} isvalidflags {} isvalidflag
interp alias {} validflags {} isvalidflag

proc flags2vars args {
	upvar 1 flags flags
	if [array exists flags] {
		foreach name [array names flags] {
			upvar 1 [string trimleft $arg -] local
			set local $flags($name)
		}
	} {
		foreach arg $args {
			upvar 1 [string trimleft $arg -] local
			set local [validflag $arg]
		}
	}
}

proc validflag args {
	# Don't use LEFT (recursive / endless loop)
	upvar 1 flags flags
	zero m
	if [array exists flags] {
		# If FLAGS -PARAMS -FORCE is used, then all possible flags will exist in the array (false-positives)
		# unset flag = 0, or, valid flag param could *ALSO* be 0
		# ):
		foreach flag $args {
			if [string match !* $flag] {
				incr m [not [info exists flags(-[string trimleft [string range $flag 1 end] -])]]
			} {
				incr m [info exists flags(-[string trimleft $flag -])]
			}
		}
		return $m
	} {
		foreach flag $args {
			if [string match !* $flag] {
				incr m [not [isvalidflag $flags -[string trimleft [string range $flag 1 end] -]]]
			} {
				incr m [isvalidflag $flags -[string trimleft $flag -]]
			}
		}
		return $m
	}
}

proc `validflag args {
	upvar 1 flags flags
	zero m
	if [array exists flags] {
		# If FLAGS -PARAMS -FORCE is used, then all possible flags will exist in the array (false-positives)
		foreach flag $args { incr m [info exists flags($flag)] } ; return $m
	} {
		foreach flag $args { incr m [isvalidflag $flags $flag] } ; return $m
	}
}

proc isvalidflag {validflags flag} {
	# Enter with "-goodflag1 -good2 -good3 -good4" "-goodflag1*" <wildcard ok>
	if [isempty flag] { return 0 }
	set flag [uniquematch $validflags $flag]; # UNIQUEMATCH will handle syntax issues
	if [isempty flag] { return 0 } { return 1 }
}

proc flags args { # Have to allow several variations in order to unite everything
	# Be careful with debugs: DATA GET & DATA ARRAY GET/VALUE (and others)
	# call for a flag check, causing an endless loop.
	# MSGHOME / DEBUG also checks for colours (which calls DATA GET* also).

	# { text validflags { var_text "" } { var_flags "" } }
	# Options: -simple -params -force

	flags:simple $args [list -simple -params -force] args flags
	# Default to: -SIMPLE
	if [isempty flags] { set flags -simple }

	if [validflag -simple] {
		lassign $args text valid var_text var_flags
		if [notempty var_text] { upvar 1 $var_text new_text ; unset -nocomplain new_text }
		if [notempty var_flags] { upvar 1 $var_flags new_flags ; unset -nocomplain new_flags}
		if [isempty valid] {
			empty new_flags
			set new_text $text
		} {
			flags:simple $text $valid new_text new_flags
		}
		return [list $new_text $new_flags]
	}

	if [validflag -params] {
		# 2018-06-12 21:26:00 -0700: FLAGS -PARAMS is broken; FLAGS:PARAMS works
		# Use it instead
		lassign $args text valid var_text var_flags
		if [validflag -force] {
			return [flags:params -force $text $valid $var_text $var_flags]
		} {
			return [flags:params        $text $valid $var_text $var_flags]
		}
		# This is broken; flags:params is tested and works now
		
		# args flags
	
		# Note: flag params CAN BE STACKED! (:
		# "-flag1 text -flag2 blah -flag1 text2" will return
		# -flag1 {test test2} -flag2 blah
		# Whether or not this was intentional, I DON'T CARE!! (: (: (:

		# Example: flags -params "-flag0 -flag1 Text -flag5 {1 2 3 4 5} This is a sentence." [list -flag0 0 -flag1 1 -flag2 2 -flag5 5] text flags
		# The assigned variable (var_flags) will return an array:
		# $TEXT: this is a sentence.
		# $FLAGS[]: -flag0 {} flag1 Text -flag5 {1 2 3 4 5}
		# Any flags not used will be UNSET (not set) in the $FLAGS[] variable.

		# To override this and set blank values to unused flags, use the -force! (:
		# Zero-parameter flags will be forced to a value of "1" ($BLANK)
		# Example: flags -params "-flag0 -flag1 Text -flag5 {1 2 3 4 5} This is a sentence." [list -flag0 0 -flag1 1 -flag2 2 -flag5 5] text flags
		# The assigned variable (var_flags) will return an array:
		# $TEXT: this is a sentence.
		# $FLAGS[]: -flag0 1 flag1 Text -flag2 {} -flag5 {1 2 3 4 5}

		set blank_1 1
		set blank_0 0
		lassign $args text array var_text var_flags

		set text [split $text]
		array set params ""
		# Can this be altered to allow "-B0 -B1 -B2 ..." to automatically assume "0"
		foreach { a b } $array {
			if ![left $a 1 -] { prepend a - }
			if [string eq * $b] { set b 512 }
			set params($a) $b
		}

		# Sanity check!
		foreach { a b } [array get params] {
			if ![isnum -integer $b] { error "\[FLAGS -PARAMS\] Illegal array format: flags must be arrayed to how many arguments each will take: \[LIST -flag0 0 -flag1 1 -flag2 2\] would allow: -flag0 -flag1 text -flag2 test1 test2 -- This is a sentence." }
		}

		emptyarray processed

		while { [notempty text] } {
			set flag [lindex $text 0]
			set text [lreplace $text 0 0]
			if [string eq $flag --] break
			set um [uniquematch [array names params] $flag]

			if [isempty um] {
				lprepend text $flag; # Put it back!
				break
			}

			if [isnull processed($um)] { empty processed($um) }
			set skip $params($um)
			if $skip {
				# Worf?! *Bwahahahahaha*
				for { zero pass } { $pass < $skip } { incr pass } {
					set worf [lindex $text 0]
					if [left $worf 1 -] {
						#NO! break
						# Did we find a legit flag, or, does it just LOOK like one?
						set um [uniquematch [array names params] $worf]
						if [notempty um] { break ; # Don't swallow the next flag; process it! }
					}
					if [notempty worf] { lappend processed($um) $worf }
					set text [lreplace $text 0 0]
				}
			} {
				#lappend processed($um) $um
				lappend processed($um) 1
			}
		}

		# So no errors are thrown by flag-checked variables in the calling PROC ....
		# Is this a good idea?
		# No ~ flags with 0 (zero) elements will ALSO return blank ("") array elements by design.
		# Let's make this coder-controlled via the -FORCE flag
		if [validflag -force] {
			foreach { a - } $array {
				if [info exists processed($a)] {
					if [string eq "" $processed($a)] { if { $params($a) == 0 } { set processed($a) $blank_1 } }
				} {
					# The same thing! How to differentiate ....
					if { $params($a) == 0 } { set processed($a) $blank_0 } { set processed($a) "" }
				}
			}
		}

		set text [join $text]
		if [notempty var_text] { upvar 1 $var_text local_text ; unset -nocomplain local_text ; set local_text $text }
		if [notempty var_flags] { upvar 1 $var_flags local_flags ; unset -nocomplain local_flags ; array set local_flags [array get processed] }
		return [list $text [array get processed]]
	}

	?
}

# Allows multiple-var text response (using [flags]_args to collect unassigned values) , & default flags variable "$flags"
proc flags:simple { list valid { var_text "" } { var_flags flags } { flags2variables false } } {
	# Convert this to $ARGS: [-variables] {list} {valid} {var_text} {var_flags}
	set flags ""
	if ![string eq "" $valid] {
		while { ![string eq "" $list] } {
			set 0 [lindex $list 0]
			if [string eq $0 --] { set list [lreplace $list 0 0] ; break }
			# We do not allow flags to have "? *" in them; can create false matches when processing text with wildcards that happend to match flags
			if [regexp -- {[\*\?]} $0] {
				set u ""
			} {
				# 2018-09-27 11:41:00 -0700:
				# TESTPASS: "FLAGS:SIMPLE -format {} {-flag1 1 -flag2 1 ...}"
				# If ISEMPTY 0 isn't tested, any "missing-first-parameter" situation
				# will cause the "blank" to be deleted, and the flag-list being tested
				# will become the resulting $LIST instead!
				if { [isempty 0] || [string eq $0 -] } {
					set u ""
				} {
					set m [lsort -unique -dictionary -increasing [lsearch -all -inline -glob [string tolower $valid] [string tolower ${0}*]]]
					set u [lindex [list "" $m] [ expr ( [llength $m] == 1 ) ? 1 : 0]]
				}
			}
			if [string eq "" $u] break
			lappend flags $u
			set list [lreplace $list 0 0]
		}
	}
	if ![string eq "" $var_text] {
		if { [llength $var_text] == 1 } {
			upvar 1 $var_text local_text
			set local_text $list
		} {
			upvar 1 ${var_flags}_args _flags_args
			set _flags_args $list
			foreach variable $var_text {
				upvar 1 $variable _local
				set _flags_args [lassign $_flags_args _local]
			}
		}
	}
	if ![string eq "" $var_flags] { upvar 1 $var_flags local_flags ; unset -nocomplain local_flags ; set local_flags $flags}

	if $flags2variables {
		foreach a $valid {
			set a [string trimleft $a -]
			upvar 1 $a local_$a
			set local_$a 0
		}
		foreach a $flags { set a [string trimleft $a -] ; set local_$a 1 }
	}
	return [list $list $flags]
}

# This is a simple substitute for the FLAGS -PARAMS method
# Incorporates FLAGS:SIMPLE's multiple-variable text response method
proc flags:params args {
	# Syntax: [-FORCE] $TEXT $FLAGS VAR:TEXT VAR:FLAGS (usually $flags) FLAGS:PARAMS
	if [string eq -nocase TEST $args] {
		set args {-force {-flag1 1 -flag2 "Multi-word list" -flag0 -flag2 {here's some more} -flag 0 1 2 3 4 5 -- This is a sentence.} {-flag0 0 -flag1 1 -flag2 1 -flag *} text flags}
	}
	flags:simple $args -force [list text flags var_text var_flags] coreflags
#debug text flags var_text var_flags coreflags
	if [isempty var_flags] { set var_flags flags }
	# Change all "*" values immediately!
	set count -1
	foreach [list a b] $flags {
		incr count 2
		if [string eq * $b] { lset flags $count [llength [split $text]] }
	}
	array set _flags $flags
	array set _value [list]
	# Sanity check
	if [ catch { join $text } ] { set text [split $text] }
	while { ![string eq "" $text] } {
		set test [lindex $text 0]
		if [string eq $test --] { set text [lreplace $text 0 0] ; break }
#log "\[FLAGS:PARAMS\] =0.3 TEST($test):LEFT([left $test 1 -])"
		if [left $test 1 -] {
#putlog =0.4
			set test [uniquematch [array names _flags] $test]
			if [isempty test] break ; # Leave the false-flag at index #0 to be processed as a normal element
			set text [lreplace $text 0 0]
			if ![info exists _flags($test)] break ; # Illegal flag -- exit flag parsing and assume it's normal text
			if [string eq * $_flags($test)] { set _flags($test) 999999999 }
			if $_flags($test) {
				for { set x 0 } { $x < $_flags($test) } { incr x } {
					set flag [lindex $text 0]
					if [string eq $flag --] { set text [lreplace $text 0 0] ; break }
					set flag [none [uniquematch [array names _flags] $flag] $flag]
					if [info exists _flags($flag)] break
					lappend _value($test) $flag
					set text [lreplace $text 0 0]
					if [isempty text] break
				}
			} {
				incr _value($test)
			}
		} break
	}
	if [isvalidflag $coreflags -force] {
		foreach [list a b] $flags {
			if [string eq * $b] { set b [string length $text] }
			if $b {
				if ![info exists _value($a)] { set _value($a) "" }
			} {
				if ![info exists _value($a)] { set _value($a) 0 } { set _value($a) [boolean $_value($a)] }
			}
		}
	}
	if ![string eq "" $var_text] {
		if { [llength $var_text] == 1 } {
			upvar 1 $var_text local_text
			set local_text $text
		} {
			upvar 1 ${var_flags}_args _flags_args
			set _flags_args $text
			foreach variable $var_text {
				upvar 1 $variable _local
				set _flags_args [lassign $_flags_args _local]
			}
		}
	}

#log "\[FLAGS:PARAMS\] _VALUE\[\]([array get _value])"
	if [notempty var_flags] { upvar 1 $var_flags local_flags ; unset -nocomplain local_flags ; array set local_flags [array get _value] }
	list $text [array get _value]
}

proc flags:priority { var_result args } {
	# Enter with: flags:priority byref_variable -flag#1 -flag#2 -flag#3 ... -flag#N (number = priority)
	upvar 1 flags flags $var_result result
	empty result
	foreach _ [lreverse $args] { if [validflag $_] { set result $_ } }
	return $result
}

proc iswildcard a {regexp -- {[\*\?]+} $a}
proc percentwildcard word {
	if [isempty word] { return 0 }
	set leftover [regsub -all "\\\*|\\\?" $word "" temp]
	return [ fixmath ( $leftover / [ len $word ] ) * 100 ]
}

# --- LIST manipulation ---

namespace eval list {
	# Don't do anything; just create a NAMESPACE for the variables
}

proc list:regsub args { # [list regsub_var list args]
	# Syntax: set list [LIST:REGSUB variable [-normal REGSUB flag options] [--] {^whatever$} {replacement}]
	# Iterate REGSUBs over a list (this is the only way to side-step TCL 8.5's {#index 0} issue
	# Don't need final variable for this function; it's implied by the loop variable
	if [string eq -nocase HELP $args] { return "Syntax: variable <TCL list> <regsub flags> <regsub search string> <regsub replaement string> (replacement string will require \"{}\" if you're using variable)" }
	set args [lassign $args regsub_var list]
	set result [list]
	upvar 1 $regsub_var $regsub_var
	foreach element $list {
		set $regsub_var $element
		set eval "[lrange $args 0 end-2] {[lindex $args end-1]} $element {[subst -nobackslashes [lindex $args end]]}"
		lappend result [uplevel 1 regsub $eval]
	}
	return $result
}

proc list:set [list list value [list index ""]] {
	variable ::list::$list
	if [isempty index] { set $list $value } { swap value index ; lset $list $index $value }
}

interp alias {} list:create {} list:new
proc list:new [list list type count args] {
	set valid [list string list integer floating-point]
	set um [uniquematch $valid $type]
	switch -exact -- $um {
		integer { set blank 0 }
		floating-point { set blank 0.0 }
		string { set blank "" }
		list { set blank [list] }
		default { error "list:new variable <${valid}> <count> ?arg arg ...?" }
	}
	empty new
	for { set x 1 } { $x <= $count } { incr x } { lappend new $blank }
	if [notempty args] {
		set index -1
		foreach arg $args { lset new [incr index] $arg ; if { $index > $count } break }
	}
	set ::list::$list $new
	return $new
}

interp alias {} list:range {} list:get
interp alias {} list:getrange {} list:get
proc list:get [list list [list index ""] [list range ""]] {
	if [isempty index] { return [var ::list::$list] }
	if [isempty range] { return [lindex [set ::list::$list] $index] }
	return [lrange [set ::list::$list] $index $range]
}

proc list:index [list list index [list range ""]] { list:get $list $index $range }

interp alias {} list:append {} list:add
proc list:add [list list value] { variable ::list::$list ; lappend list $value }

interp alias {} list:concat {} list:addall
proc list:addall [list list args] { lappend $::list::$list {*}$args }

interp alias {} list:search {} list:binarysearch
proc list:binarysearch [list list item] { lsearch -exact [set ::list::$list] $item }

proc list:clear list { empty ::list::$list "" }

interp alias {} list:match {} list:contains
proc list:contains [list list item] { expr ( [list:binarysearch $list $item] != -1 ) ? true : false }

interp alias {} list:convert {} list:convertall
proc list:convertall [list list type] {
	set valid [list string list integer floating-point]
	set um [uniquematch $valid $type]
	empty new
	foreach element [set ::list::$list] {
		switch -exact -- $um {
			integer { if [isnum $element] { lappend new [lindex [split [noscinot $element] .] 0] } { lappend new 0 } }
			floating-point { if [isnum $element] { lappend new ${element}[iff ![instr $element .] .0] } { lappend new 0.0 } }
			string { lappend new "$element" }
			list { lappend new $element }
			default { error "list:new variable <${valid}> <count> ?arg arg ...?" }
		}
	}
	set ::list::$list $new
}

proc list:exists [list list [list value ""]] {
	if [isempty value] { return [info exists ::list::$list] }
	list:contains $list $value
}

interp alias {} list:isequal {} list:equal
interp alias {} list:eq {} list:equal
proc list:equal [list list list2] {
	# This doesn't match the original intent; list2=list causes a pure copy to be made;
	# maybe because they share a pointer? 	list=blah, list2=blah are two separate
	# INSTANCES (definitions), and are NOT equal!
	#
	# There *IS* a way, via CritTCL, to expose a pointer. If successful,
	# the pointers of the two lists can be compared (a link, such as UPVAR, would show as TRUE)
	# https://rosettacode.org/wiki/Create_an_object_at_a_given_address#Tcl
	#
	# For now, let's just compare the values
	string match [set ::list::$list] [set ::list::$list2]
}
proc list:ne [list list list2] { not [list:equal $list $list2] }

proc list:copy [list list list2 [list start end] [list count ""]] {
	set list1 [set ::list::$list]
	if ![info exists ::list::$list2] { return $list1 }
	set list2 [set ::list::$list2]
	if [isempty count] { set count [ expr [llength $list2] - 1 ] }
	if [string eq -nocase END $start] { set continue end+1 } { set continue [ expr $start + 1 ] }
	set new [concat [lrange $list1 0 $start] [lrange $list2 0 $count] [lrange $list1 $start end]]
}

interp alias {} list:findall {} list:find -all
proc list:find args {
	#list:find [-nocase] [-all] [-glob -exact -regexp] [-eval] <list name> <match string>
	if [string eq -nocase HELP $args] { return "list:find ?-all? ?-nocase? ?-glob -exact -regexp? ?-index? ?-eval? ?-last? list-name match-string ?start ?size??" }
	flags:simple $args [list -nocase -glob -exact -regexp -all -index -eval -last] data flags
	lassign $data list matchme start size
	if [isempty matchme] { return -1 } ; # Nothing to match
	set offset [none $start 0]
	set data [var ::list::$list]
	if [isempty data] { return -1 } ; # Nothing will match
	if $offset { set data [lrange $data $offset end] }
	if [notempty size] { set data [lrange $data 0 [ expr $size - 1 ]] }
debug data flags list matchme
	set all [validflag -all]
	set nocase [validflag -nocase]
	set index [validflag -index]
	set eval [validflag -eval]
	flags:priority mode -glob -exact -regexp
	set flags $mode
	if [isempty flags] { lappend flags -glob } ; # Default
	if [validflag -last] {
		# Special processing (normally should ignore -EVAL mode, but ... can we do it?)
		# Ignore -ALL
		set ll [llength $data]
		if $eval {
			decr ll
			foreach datum [lreverse $datum] {
				if { [eval ( [regsub -all -- # $matchme $datum] ) ? 1 : 0 ] } {
					if $index {
						if $offset { return [ expr $ll + $offset ] } { return $ll }
					} {
						return $datum
					}
				}
			}
			if $index { return -1 } return
		} {
			set lsearch [lindex [eval lsearch [if $nocase -nocase] $mode -all $data $matchme] end]
			if [isempty lsearch] { if $index { return -1 } return }
			if $index { return [ expr $lsearch + $offset ] } { return [lindex $data $lsearch] }
		}
	}
	# *Not* in -LAST mode ....
	if $eval {
		empty find
		negone index
		foreach datum $data {
			incr index
			if [ expr [regsub -all -- # $matchme $datum] ? 1 : 0 ] {
				if $index { lappend find $index } { lappend find $datum }
			}
			if !$all break
		}
		if $index { if [isempty find] { set find -1 } { set find [list:math $find " # + $offset "] } }
		return $find
	}
	set result [eval lsearch $flags [iff !index -inline] [iff $nocase -nocase] [iff $all -all] $data $matchme]
	if $index { if { $result != -1 } { set result [list:math $result " # + $offset "] } }
	return $result
}

proc list:findindex [list list matchme [list start 0] [list count ""]] { list:find -index $list $matchme $start $count }
proc list:findlast [list list matchme [list start 0] [list count ""]] { list:find -last $list $matchme $start $count }
proc list:findlastindex [list list matchme [list start 0] [list count ""]] { list:find -index -last $list $matchme $start $count }

proc list:foreach [list list variable code] { uplevel 1 foreach $variable [var ::list::$list] $code }

proc list:list2array [list list array [list start 0]] {
	upvar 1 $array local
	emptyarray local
	foreach element [var ::list::$list] { set local($start) $element ; incr start }
}

proc list:array2list args {
	# This will accept any format of array names: integers, reals, or text (tokens)
	flags:simple $args [list -compact] data flags
	lassign $data array list
	upvar 1 $array local
	set new [lmap [list - a] [array get local]] { set - $a }
	if [validflag -compact] { set new [lsearch -inline -exact -all -not $new ""] }
	set ::list::$list $new
}

interp alias {} list:trim {} list:trimexcess
proc list:trimexcess list {
	set data [var ::list::$list]
	if [isempty data] return
	set data [lsearch -inline -all -exact -not $data ""]
	set ::list::$list $data
}

interp alias {} list:istrue {} list:trueforany
interp alias {} list:anytrue {} list:trueforany
interp alias {} list:isanytrue {} list:trueforany
interp alias {} list:alltrue {} list:trueforall
interp alias {} list:isalltrue {} list:trueforall
proc list:trueforall [list list expression] { list:true -all $list $expression }
proc list:trueforany [list list expression] { list:true -all $list $expression }
proc list:true args {
	flags:simple $args [list -all -any] data flags
	lassign $data list expression
	set data [var ::list::$list]
	flags2vars -any -all
	set match_all true
	set match_any false
	foreach datum $data {
		if [ expr ( [regsub -all -- # $expression $datum] ) ? 1 : 0 ] {
			set match_any true
		} {
			set match_all false
		}
		if !$all break
	}
	if $any { return $match_any }
	return $match_all
}

proc list:regexp [list list args] {
	error "\[LIST:REGEXP\] I don't do anything yet."
	if [string eq -nocase HELP $list] { return "list:regsub list_name ?[get:options regexp]? <pattern to find>" }
	flags:simple $args [list -all -nocase -expanded -line -linestop -lineanchor -start] args flags
	lassign $args find_me replace_with
	lappend new
	set data [var ::list::$list]
	foreach datum $data { lappend new [eval regsub $flags -- $find_me $datum $replace_with] }
	set ::list::$list $new
}

proc list:regsub [list list args] {
	if [string eq -nocase HELP $list] { return "list:regsub list_name ?[get:options regsub]? <pattern to find> <replacement pattern>" }
	flags:simple $args [list -all -nocase -expanded -line -linestop -lineanchor -start] args flags
	lassign $args find_me replace_with
	lappend new
	set data [var ::list::$list]
	foreach datum $data { lappend new [eval regsub $flags -- $find_me $datum $replace_with] }
	set ::list::$list $new
}

set list:null {
	list : [abc [def {gh}] i]
	list2: [0 1 2 3 4 5 6]
	list3: [a b c d e "" "" ""]

	list:indexof list <item> [start index [count]]
		list:indexof list d*: "1"
		list:indexof list ? 3: "-1"
		list:indexof list i 0 2: "-1"

	list:lastindexof list <item> [start index [count]]
		list:lastindexof list *: "2"
		list:lastindexof list ? 3: "0"
		list:lastindexof list a* 2 1: "-1"

	list:insert list <index> <item>
		list:insert list 2 z: [abc [def gh] z i]

	list:insertrange list <index> [list]
		list:insert list 2 list2: [abd [def gh] 0 1 2 3 4 5 6 i]

	list:remove list <item>
		list:remove list i: [abc [def gh]]

	list:removeall list <expression>
		list:remove list #>15: [abc [def gh] i]

	list:removeat list <index>
		list:delete list <index>
		list:removeat list 1: [abc i]
		list:delete list 1: [abc i]

	list:removerange list <index> [count]
		list:removerange list 0 2: [i]

	list:reverse list
	list:reverse list [index [count]]
		list:reverse list: [i [def gh] abc]
		list:reverse list 1: [abc i [def gh]]
		list:reverse list 0 2: [[def gh] abc i]

	list:sort list
	list:sort list <expression comparing 1 & 2>
	list:sort list <variables 1 2><code: using [1 2]>
	list:sort list <int:start index> <int:count> <code: accepts two-element list>
		list:sort list: [abc [def gh] i]
		list:sort list "1 > 2": [i [def gh] abc]
		list:sort list [1 2] <string compare -nocase [left $2 1] [left $1 1]>: [i [def gh] abc]
		list:sort list 1 2 <string compare -nocase 1 2>: [abc i [def gh]]

}

proc lassign:count { name list { start 0 } } {
     upvar 1 $name array
     set leftovers [list]
     array set array [list]
     set b [expr $start - 1]
     foreach a $list {
          incr b
          if [isempty a] { set array($b) "" } { set array($b) $a }
     }
     return [array get array]
}

proc inlist args {
	flags:simple $args [list -nocase -glob -regexp -exact -inline -all] text flags
	lassign $text list checkme
	if [isempty list] { if [validflag -inline] return { return 0 } }
	# Don't sanity-check $CHECKME vs "" as that COULD be a valid element within a list
	set mode glob
	foreach a [list regexp exact glob] { if [validflag -$a] { set mode $a } }

	if ![validflag -inline] { # Data-destructive search (which is okay)
		if [validflag -nocase] { set list [string tolower $list] ; set checkme [string tolower $checkme] }
		return [ expr ( [ lsearch -$mode $list $checkme ] == -1 ) ? 0 : 1 ]
		# "-ALL" is redundant here
	}

	set all [validflag -all]
	set nocase [validflag nocase]
	switch -exact -- ${nocase}${all} {
		00 { set result [lsearch      -$mode $list $checkme] }
		01 { set result [lsearch -all -$mode $list $checkme] }
		10 { set result [lsearch      -$mode [string tolower $list] [string tolower $checkme]] }
		11 { set result [lsearch -all -$mode [string tolower $list] [string tolower $checkme]] }
		default { error "\[INLIST\] Unknown combination of values for -NOCASE ($nocase) & ALL ($all)" }
	}
	empty inlist
	foreach a $result { lappend inlist [lindex $list $a] }
	return $inlist
}

proc ldiff { 1 2 { var_1 "" } { var_2 "" } { var_common "" } } {
	if ![string eq "" $var_1] { upvar 1 $var_1 local_1 }
	if ![string eq "" $var_2] { upvar 1 $var_2 local_2 }
	if ![string eq "" $var_common] { upvar 1 $var_common common }
	set local_1 [list]
	set local_2 [list]
	set common [list]
	set c -1
	foreach a $1 b $2 {
		incr c
		if ![string eq $a $b] {
			set local_1 [lrange $1 $c end]
			set local_2 [lrange $2 $c end]
			break
		}
		set common [lrange $1 0 $c]
	}
	list $local_1 $local_2 $common
}

proc lpermute list {
	# http://wiki.tcl.tk/43
	if {[llength $list] < 2} {
		set res $list
	} else {
		set pos 0
		foreach i $list {
			set rest [lreplace $list $pos $pos]
			incr pos
			foreach j [lpermute $rest] { lappend res [concat [list $i] $j] }
		}
	}
	set res
}

proc lscan { list args } {
     if [isempty list] return
     if [isempty args] return

	empty test
	array set temp [list]
	foreach { mode value } $args {
		if { [lsearch -exact [list == != <= >= < > %] $mode] == -1 } return
		if ![isnum -real $value] return
		set mode [stl $mode]
		foreach element $list {
			set check [subst "$element $mode $value"]
			if $check { lappend temp($mode) $element ; lappend test $element }
		}
	}

	empty lscan
	set test [lsort -increasing -unique -dictionary $test]
	foreach element $test {
		set match 1
		foreach array [array names temp] {
			if { [lsearch -exact $temp($array) $element] == -1 } { set match 0 ; break }
		}
		if $match { lappend lscan $element }
	}
	return $lscan
}

proc lstride args {
	flags:simple $args -fill text flags
	lassign $text list size
	if [isempty size] { one size }
	# Sanity checks
	if [isempty list] return
	if ![isnum -real $size] return
	if { $size < 0 } { return $list }

	decr size
	set lstride [list]
	while { [notempty list] } {
		set element [lrange $list 0 $size]
		if { [llength $element] < $size } {
			if [validflag -fill] { set element [lrange [concat $element [lrepeat $size ""]] 0 $size] }
		}
		lappend lstride $element ; # [lrange $list 0 $size]
		set list [lreplace $list 0 $size]
	}
	return $lstride
}

proc lsortm { indeces list args } {
# 2017-02-21 02:27:27 -0800: Replaced by LSORT:MULTIPLE (still in use in a few places: will hunt them down and kill them eventually!)
	error "\[LSORTM\] Update code to use LSORT:MULTIPLE instead!"
# Gotta re-write this to allow flags in their proper place: LSORTM -exact [list {1 2} {1 3} {2 3} {2 4} {0 3}] [list 1 0]
	# http://wiki.tcl.tk/43
	# http://computer-programming-forum.com/57-tcl/ade54d0283960122.htm
	set cmd "list $list" ; # $L (capital L)??
	foreach i [lreverse $indices] { set cmd "lsort $args -index $i \[$cmd\]" }
	eval $cmd
}

proc lsort:multiple args {
	# Enter with: lsort:multiple [LSORT flags] <list of indeces> <list>
	# Based on code found at: http://wiki.tcl.tk/43
	set list [lindex $args end]
	set indeces [lindex $args end-1]
	if { [llength $args] > 2 } { set flags [lrange $args 0 end-2] } { empty flags }
	set cmd "list $list"
	foreach i [lreverse $indeces] { set cmd "lsort $flags -index $i \[${cmd}\]" }
	eval $cmd
}

proc lselect { condition list } {
	# http://wiki.tcl.tk/43
	set result ""
	foreach i $list { if { [eval expr [list $i] $condition] } { lappend result $i } }
	return $result
}

proc lcompress list {
	# http://wiki.tcl.tk/43
	set list [lsort -increasing -real -unique $list]
	set res [lindex $list 0]
	for { set i 1 } {$i < ( [llength $list] - 1 ) } { incr i } {
		set it [lindex $list $i]
		if { ( $it == [lindex $list [ expr { $i - 1 } ]] + 1 )
			&&
			( $it == [lindex $list [ expr { $i + 1 } ]] - 1 )
		} {
			lappend res -
		} else {
			lappend res $it
		}
	}
	lappend res [lindex $list end]
	regsub -all -- { - (- )*} $res -
}

proc lexpand list {
	# http://wiki.tcl.tk/43
	set res {}
	foreach part $list {
		if [regexp {([0-9-]*[0-9])-(.+)} $part - from to] {
			while { $from <= $to } { lappend res $from ; incr from }
		} else {
			lappend res $part
		}
	}
	return $res
}

proc userlist:sort args {
	if [isempty args] { return [lsort -increasing -unique -dictionary [userlist]] }
	flags:simple $args -list text flags
	empty userlist
	foreach flag $args {
		if [validflag -list] {
			lappend userlist [userlist $flag]
		} {
			set userlist [concat $userlist [userlist $flag]]
		}
	}
	return $userlist
}

proc lincr {varName { increment 1 } } {
	upvar 1 $varName var
	empty o
	foreach a $var { lappend o [incr a $increment] }
	set var $o
	return $o
}

proc ldecr {varName { increment 1 } } {
	# Any attempt to pipe this through LINCR requires an additional level of UPVAR!
	upvar 1 $varName var
	empty o
	foreach a $var { lappend o [incr a -$increment] }
	set var $o
	return $o
}

# Replacement PROC below
proc `ldefault {defaults list} {
	# Obsolete command; see LDEFAULTS for a replacement
	SB:reset_defaults $list $defaults
}

proc ldefault [list var list] {
	upvar 1 $var local
	if ![info exists local] { set local "" }
	empty new
	foreach original $local new $list { lappend new [none $original $new] }
	set local $new
}

# Enter with: LDEFAULTS var value ?var value? ?...?
proc ldefaults args {
	# Flag option: -keepblanks / -empty:ok?
	set count 0
	foreach { a b } $args {
		upvar 1 $a local
		if [isempty local] { set local $b ; incr count }
	}
	return $count ; # Return how many replacements were performed
}

proc lpad { var_list llength { pad_with "" } } {
	upvar 1 $var_list list
	if ![info exists list] { empty list }
	if ![isnum -integer $llength] { error "\[LPAD\] Expected integer but got: $llength" }
	if { $llength < 0 } return
	set ll [llength $list]
	if { $ll >= $llength } return
	for { set x $ll } { $x < $llength } { incr x } { lappend list $pad_with }
	return
}

proc loverlap {list { side "left" } } {
	set side [uniquematch [list left right] $side]
	set l [getlongest $list]
	zero c
	for {one x} {$x <= $l} {incr x} {
		zero break
		set e [$side [l0 $list] $x]
		foreach a $list { if ![$side $a $x $e] { one break; break } }
		if $break { zero break; break }
		set c $x
	}
	return [$side [l0 $list] $c]
}

proc ldepth {list { depth 0 } } {
	if !($depth) { if { [llength $list] == 1 } { set list [join $list] } }

	incr depth
	set text $depth
	foreach element $list {
		set e [catch {join $element}]
		if $e {set element [split $element]}
		set l [llength $element]
		zero break
		if {($l > 1) || ($depth == 1)} {# "$l > 0" = endless loop!! #}
		if {$l > 1} {# "$l > 0" = endless loop!!
			incr break; if {$break > 10} break
			# Recursive call !!
			set text [concat $text [ldepth $element $depth]]
		}
	}
	return [highest $text]; # Only report the highest number
	return $text
}

proc lflatten list {# Only do as many times as necessary, no more
	set l [ldepth $list]; # Only evaluate it once
	for {one x} {$x <= $l} {incr x} { set list [join $list] }; # "<" only?
	return $list
}

proc lsomewhere args {
	set valid [list -nocase -matches -inline -all -exact -glob -regexp]
	if [string eq -nocase HELP $args] {return "\[LSOMEWHERE\] Syntax: LSOMEWHERE \[-NOCASE -MATCHES\] <list> <list of elements to check> (returns: 0 / 1 for ALL elements being found, or, with -INLINE: all the elements that match (honouring -NOCASE). Wilcards OK"}
	# DEFAULT: -GLOB

	# Check for flags
	set flags [list -glob]; # DEFAULT!
	set m [lsearch -glob $args -*]
	if !($m) {# Flags must be in the first position, naturally.
		set n [lsearch -not -glob $args -*]
		decr n
		if {$n < 0} {error "\[LSOMEWHERE\] Unable to determine test list."}
		foreach a [lrange $args $m $n] {set flags [concat $flags $a]}
		set args [lreplace $args $m $n]

		empty flags_new
		foreach flag $flags {
			set um [uniquematch $valid $flag]
			if [notempty um] {lappend flags_new $um}
		}
		set flags $flags_new
	}

	# Process flags in order: -EXACT -GLOB -REGEXP [default: -GLOB]
	zero exact glob regexp; # Use these to avoid mutiple VALIDFLAG uses
	zero nocase inline all
	foreach a [lreverse $flags] {
		if [string eq -nocase $a -EXACT ] {one exact ; set flags [ldestroy -all -multiple -- $flags [list -glob  -regexp]]; break}
		if [string eq -nocase $a -GLOB  ] {one glob  ; set flags [ldestroy -all -multiple -- $flags [list -exact -regexp]]; break}
		if [string eq -nocase $a -REGEXP] {one regexp; set flags [ldestroy -all -multiple -- $flags [list -exact -glob  ]]; break}
	}

	if [validflag -all]     {one all}
	if [validflag -nocase]  {one nocase}
	if [validflag -inline]  {one inline}
	if [validflag -matches] {one inline}; # Backward compatibility

	# Assure default value
	if ![ expr $exact + $glob + $regexp ] { one glob }

	set list [l0 $args]
	set items [l1e $args]
	if [isempty list] {# RETURN based on the data type requested
		if $inline {return ""} {return 0}
	}

	empty lsomewhere
	foreach item $items {
		if $nocase {
			if $exact {
				set loop [lsearch -all -exact  [stl $list] [stl $item]]
			} elseif $glob {
				set loop [lsearch -all -glob   [stl $list] [stl $item]]
			} elseif $regexp {
				set loop [lsearch -all -regexp [stl $list] [stl $item]]
			}
		} {
			if $exact {
				set loop [lsearch -all -exact  $list $item]
			} elseif $glob {
				set loop [lsearch -all -glob   $list $item]
			} elseif $regexp {
				set loop [lsearch -all -regexp $list $item]
			}
		}

		# Collect all results
		foreach a $loop {
			if { $a != -1 } { lappend lsomewhere [lindex $list $a] }
		}
		if $inline {
			if !($all) {
				if [notempty lsomewhere] {return [l0 $lsomewhere]}
			}
		} {
			# Use of -NOT will RETURN "0" if there are ONLY "-1s" in the list
			# Once there's a match, no need to keep processing; Save %CPU!
			if [notempty lsomewhere] {
				if [lsearch -not $lsomewhere -1] {return 1}
			}
		}
	}

	if $inline {# This may not be necessary: !$ALL handled above
		if $all {return $lsomewhere} {return [l0 $lsomewhere]}
	} {
		return [iff [len $lsomewhere] 1 0]
	}
}

proc list:unpack args {
	# Unpack lists: 0 1{3} 2 3 4{2} 5 = "0 1 1 1 2 3 4 4 5"
     if [string eq -nocase HELP $args] { return "list:unpack 0 1 2 3\{5\} *\{5,7\} *\{10,20,2\} = 0 1 2 3 3 3 3 3 5 6 7 10 12 14 16 18 20" }
	empty new
	foreach element $args {
		if [regexp -nocase -- {^(.*)\{(\d+)\}$} $element - text multiple] {
			set new [concat $new [lrepeat $multiple $text]]
		} elseif [regexp -nocase -- {^[*]\{(\d+),(\d+)(,?(\d*))\}$} $element - range1 range2 - range3] {
               set step [none $range3 1]
			for { set x $range1 } { $x <= $range2 } { incr x $step } { lappend new $x }
		} {
			lappend new $element
		}
	}
	return $new
}

proc lprepend { var args } {
     upvar 1 $var local
	lappend local
	set var [concat $args $local]
}

proc lorder [list list args] {
	empty lorder
	foreach arg [join $args] { lappend lorder [lindex $list $arg] }
	return $lorder
}

interp alias {} lmatch {} ldestroy -not
proc ldestroy args {
	# Note: matching is done as such: [main list is literals] [match list is wildcard/regexp]
	# No way to allow double-wildcarding without resorting to looped STRING MATCHes.
	if {[llength $args] < 2} {error "invalid flag: $args - syntax: ldestroy ?options? list text-to-destroy"}
	
	set validflags "-test -nocase -exact -glob -regexp -double -all -boolean -nonulls -trim -nodup -noduplast -check -replacewith -nulls -multiple -not -both -uniquematch -inline -sort -decreasing -flags"
	# Added -INLINE (for LMATCH): ignored for compatibility
	# -DOUBLE = two-way string matching (more intensive)

	#set data [SB:setmiscvariables_checkflags args $args $validflags 1]
	array set flags [SB:setmiscvariables_checkflags setvar $args $validflags 1]
	if $flags(flags) {return $validflags}
	if $flags(check) { one flags(boolean) }
	if $flags(test) { one flags(boolean) }

	# Overrides:
	#!if $flags(both) {zero flags(not)}
	#! Not any more! (LMATCH)

	set differential [SB:setmiscvariables_checkflags args $args $validflags 1]
	set count_proper 2; # Original string, and match-with value
	add count_proper $flags(replacewith); # replace-with value
	
	if $flags(nulls) {one flags(nonulls)}
	# Check for -NONULLS "special syntax" first ....
	if {[llength $differential] != $count_proper} {
		if ($flags(nonulls)) {
			if !($flags(replacewith)) {lappend differential ""}
			# Allow "LDESTROY -NONULLS $STRING" to survive.
		}
	}
#log "\00307\[LDESTROY\] DIFFERENTIAL($differential):COUNT_PROPER($count_proper)\003"
	if {[llength $differential] != $count_proper} {
		error "wrong number of arguments. Proper syntax: LDESTROY ?flags? <string-to-check> <text-to-match> ?replace-text (only with -REPLACEWITH flag)?"
		if ($flags(boolean)) {return 0} {return ""}; # Invalid destruction list
	}
	
	set list [l0 $differential]
	set destroyme [l1 $differential]
	# This ignores the 3rd arg when -REPLACEWITH isn't used.
	
	if {[l2 $differential] != ""} {set flags(replacewith) 1}
	# Should this be allowed?
	
	set replacewithtext [l2 $differential]; # Set this AFTER the flags is set!
	
	debug.print 2 "\[LDESTROY\] LIST($list):DIFFERENTIAL($differential)"
	
	# Take care of automatic failures
	if {$list == ""} {
		if $flags(boolean) { return 0 }; # Return 0 = "not a valid destruction"
		return ""; # Otherwise, return a null string (0 surviving members since string was blank to begin with)
	}
	
	#	if [string match -nocase "-*" $list] {error "invalid flag: [l0 $list] (valid flags: $validflags)"}
	# Should this be reinstated? (Returns an error, instead of returning the validly found flags.)
	
	# Using variables instead of repeated ISVALIDFLAGs: less CPU intensive.
	
	#Never needed; see below! - if ($flags(nocase)) {set list [stl $list]; set destroyme [stl $destroyme]}
	if { !( $flags(glob) || $flags(regexp) || $flags(exact) ) } {set flags(glob) 1}
	# GLOB = default
	
	set match 0
	set counter 0
	set llength [llength $list]
	if {[llength $destroyme] > 0} {set llength [ expr $llength * [ llength $destroyme ] ]}
	incr llength; # Must increment by 1 or it'll always error out!
	debug.print 2 "\[LDESTROY\] LIST($list):DESTROYME($destroyme):LLENGTH($llength)"
	if !$flags(multiple) {set destroyme [list $destroyme]}
	
debug.print 2 "\[LDESTROY\]:DESTROYME($destroyme)"
	empty not; # Only used with -NOT flag
	zero break
	while 1 {
		zero break
		foreach testme $destroyme {
debug.print 2 "\[LDESTROY\] TESTME($testme)"
			incr counter
			#if {$counter > ( $llength * [ llength $destroyme ] )} {error "> LDESTROY: something went wrong: \$COUNTER ($counter)"}
			# We've never really needed this, have we? (Using LMATCH with over 947 args causes this error to come up for some reason)
	
			# Chosen priority: EXACT, GLOB, then REGEXP [must override each other]
			# STRING TOLOWER's are needed in this 'repeated' way because
			# forcing $LIST / $DESTROYME to lower case will destroy the
			# original list which is being returned to the calling PROC.
	
			set test_list $list
			set test_dest $destroyme
	
			empty result
			if ($flags(nocase)) {set test_list [stl $test_list]; set testme [stl $testme]}
	
			empty um
			if $flags(uniquematch) {
				set um [uniquematch $test_list $testme]
				if ![string eq "" $um] {set testme $um}
			}
debug.print 2 "\[LDESTROY\] TESTME($testme):UM($um):TEST_LIST($test_list)"
#debug flags
			if ($flags(exact)) {
				set m [lsearch -exact $test_list $testme]
				# No -DOUBLE consideration needed for -EXACT mode
			} elseif ($flags(glob)) {
				set m [lsearch -glob $test_list $testme]
				if { ( $m != -1 ) && $flags(double) } {
					# Set $M to index within $TEST_LIST with whatever match you find
					# $TESTME is a FOREACH for each item to test
					for { set x 0 } { $x < [llength $test_list] } { incr x } {
debug testme x =LINDEX([lindex $test_list $x])
						if [string match [lindex $test_list $x] $testme] { set m $x ; break }
					}
				}
			} elseif ($flags(regexp)) {
				set m [lsearch -regexp $test_list $testme]
				if { ( $m != -1 ) && $flags(double) } {
					# Set $M to index within $TEST_LIST with whatever match you find
					# $TESTME is a FOREACH for each item to test
					# -REGEXP -DOUBLE: list item must include ^$ if desired (not added here)
					for { set x 0 } { $x < [llength $test_list] } { incr x } {
						if [regexp -- [lindex $test_list $x] $testme] { set m $x ; break }
					}
				}
			} else {
				# With the check above, this never should be used.
				set m [lsearch -glob $test_list $testme
			}

			if {$m < 0} {incr break; continue; # Was "$m < -1" !!!!! Endless loop created!}
	
			lappend not [lindex $list $m]
			# Even with -NOT, the list must be destroyed-down ($COUNTER error above)

			if ($flags(replacewith)) {
				set list [lreplace $list $m $m $replacewithtext]
			} else {
				set list [lreplace $list $m $m]
			}
			incr match
		}
		if !($flags(all)) break
		if {$break == [llength $destroyme]} break
	}
	
	# TRIM MUST PRECEDE NONULLS (in case they're used together) !!
	# NONULLS must destroy TRIM elements when TRIM calculates "" values !!
	
	if ($flags(trim))	 {set tl $list; set list ""; foreach a $tl { lappend list [string trim $a] }}
	if ($flags(nonulls))   {set list [lsearch -not -all -inline $list ""]}
	if ($flags(nodup))	{set list [nodup $list $flags(nocase)]}
	if ($flags(noduplast)) {set list [nodup_last $list $flags(nocase)]]}

	if ($flags(boolean)) { return $match }
	# This allows "IF ![LDESTROY -CHECK $TEXT $1]" to be possible.

	if $flags(sort) {
		# Don't use -UNIQUE (let -NODUP handle that)
		if $flags(decreasing) {
			set list [lsort -decreasing -dictionary $list]
			set not  [lsort -decreasing -dictionary $not ]
		} {
			set list [lsort -increasing -dictionary $list]
			set not  [lsort -increasing -dictionary $not ]
		}
	}
	if $flags(both) { if $flags(not) { return [list $not $list] } { return [list $list $not] } }
	if $flags(not) {return $not}
	return $list
}

# Test version with -DOUBLE flag (which causes endless loops)
proc `ldestroy args {
	# Note: matching is done as such: [main list is literals] [match list is wildcard/regexp]
	# No way to allow double-wildcarding without resorting to looped STRING MATCHes.
	if {[llength $args] < 2} {error "invalid flag: $args - syntax: ldestroy ?options? list text-to-destroy"}
	
	set validflags "-test -nocase -exact -glob -regexp -double -all -nonulls -trim -nodup -noduplast -check -replacewith -nulls -multiple -not -both -uniquematch -inline -sort -decreasing -flags"
	# Added -INLINE (for LMATCH): ignored for compatibility
	# -DOUBLE = two-way string matching (more intensive)

	#set data [SB:setmiscvariables_checkflags args $args $validflags 1]
	array set flags [SB:setmiscvariables_checkflags setvar $args $validflags 1]
	if $flags(flags) {return $validflags}
	if $flags(check) { one flags(test) }

	# Overrides:
	#!if $flags(both) {zero flags(not)}
	#! Not any more! (LMATCH)

	set differential [SB:setmiscvariables_checkflags args $args $validflags 1]
	set count_proper 2; # Original string, and match-with value
	add count_proper $flags(replacewith); # replace-with value
	
	if $flags(nulls) {one flags(nonulls)}
	# Check for -NONULLS "special syntax" first ....
	if {[llength $differential] != $count_proper} {
		if ($flags(nonulls)) {
			if !($flags(replacewith)) {lappend differential ""}
			# Allow "LDESTROY -NONULLS $STRING" to survive.
		}
	}
log "\00307\[LDESTROY\] DIFFERENTIAL($differential):COUNT_PROPER($count_proper)\003"
	if {[llength $differential] != $count_proper} {
		error "wrong number of arguments. Proper syntax: LDESTROY ?flags? <string-to-check> <text-to-match> ?replace-text (only with -REPLACEWITH flag)?"
		if ($flags(test)) {return 0} {return ""}; # Invalid destruction list
	}
	
	set list [l0 $differential]
	set destroyme [l1 $differential]
	# This ignores the 3rd arg when -REPLACEWITH isn't used.
	
	if {[l2 $differential] != ""} {set flags(replacewith) 1}
	# Should this be allowed?
	
	set replacewithtext [l2 $differential]; # Set this AFTER the flags is set!
	
	debug.print 2 "\[LDESTROY\] LIST($list):DIFFERENTIAL($differential)"
	
	# Take care of automatic failures
	if {$list == ""} {
		if $flags(test) { return 0 }; # Return 0 = "not a valid destruction"
		return ""; # Otherwise, return a null string (0 surviving members since string was blank to begin with)
	}
	
	#	if [string match -nocase "-*" $list] {error "invalid flag: [l0 $list] (valid flags: $validflags)"}
	# Should this be reinstated? (Returns an error, instead of returning the validly found flags.)
	
	# Using variables instead of repeated ISVALIDFLAGs: less CPU intensive.
	
	#Never needed; see below! - if ($flags(nocase)) {set list [stl $list]; set destroyme [stl $destroyme]}
	if { !( $flags(glob) || $flags(regexp) || $flags(exact) ) } {set flags(glob) 1}
	# GLOB = default
	
	set match 0
	set counter 0
	set llength [llength $list]
	if {[llength $destroyme] > 0} {set llength [ expr $llength * [ llength $destroyme ] ]}
	incr llength; # Must increment by 1 or it'll always error out!
	debug.print 2 "\[LDESTROY\] LIST($list):DESTROYME($destroyme):LLENGTH($llength)"
	if !$flags(multiple) {set destroyme [list $destroyme]}
	
debug.print 2 "\[LDESTROY\]:DESTROYME($destroyme)"
	empty not; # Only used with -NOT flag
	zero break
	while 1 {
		zero break
		foreach testme $destroyme {
debug.print 2 "\[LDESTROY\] TESTME($testme)"
			incr counter
			#if {$counter > ( $llength * [ llength $destroyme ] )} {error "> LDESTROY: something went wrong: \$COUNTER ($counter)"}
			# We've never really needed this, have we? (Using LMATCH with over 947 args causes this error to come up for some reason)
	
			# Chosen priority: EXACT, GLOB, then REGEXP [must override each other]
			# STRING TOLOWER's are needed in this 'repeated' way because
			# forcing $LIST / $DESTROYME to lower case will destroy the
			# original list which is being returned to the calling PROC.
	
			set test_list $list
			set test_dest $destroyme
	
			empty result
			if ($flags(nocase)) {set test_list [stl $test_list]; set testme [stl $testme]}
	
			empty um
			if $flags(uniquematch) {
				set um [uniquematch $test_list $testme]
				if ![string eq "" $um] {set testme $um}
			}
debug.print 2 "\[LDESTROY\] TESTME($testme):UM($um):TEST_LIST($test_list)"
#debug flags
			unset -nocomplain n ; # Used for -DOUBLE matches
			if ($flags(exact)) {
				set m [lsearch -exact $test_list $testme]
				# No -DOUBLE consideration needed for -EXACT mode
			} elseif ($flags(glob)) {
				set m [lsearch -glob $test_list $testme]
				if { ( $m != -1 ) && $flags(double) } {
					# Set $M to index within $TEST_LIST with whatever match you find
					# $TESTME is a FOREACH for each item to test
					for { set x 0 } { $x < [llength $test_list] } { incr x } {
debug testme x =LINDEX([lindex $test_list $x])
						if [string match [lindex $test_list $x] $testme] { set n $testme ; break }
					}
				}
			} elseif ($flags(regexp)) {
				set m [lsearch -regexp $test_list $testme]
				if { ( $m != -1 ) && $flags(double) } {
					# Set $M to index within $TEST_LIST with whatever match you find
					# $TESTME is a FOREACH for each item to test
					# -REGEXP -DOUBLE: list item must include ^$ if desired (not added here)
					for { set x 0 } { $x < [llength $test_list] } { incr x } {
						if [regexp -- [lindex $test_list $x] $testme] { set n $testme ; break }
					}
				}
			} else {
				# With the check above, this never should be used.
				set m [lsearch -glob $test_list $testme]
			}

			if { $flags(double) && [notempty n] } {
				# Unnecessary sanity check?
				if { $m != -1 } { incr break ; continue }
				# $TESTME is a loop variable; no reason to destroy anything
				lappend not $n
				unset -nocomplain n
				
			} {
				if { $m != -1 } {incr break ; continue ; # Was "$m < -1" !!!!! Endless loop created!}
		
				lappend not [lindex $list $m]
				# Even with -NOT, the list must be destroyed-down ($COUNTER error above)

				if ($flags(replacewith)) {
					set list [lreplace $list $m $m $replacewithtext]
				} else {
					set list [lreplace $list $m $m]
				}
			}
			incr match
		}
		if !($flags(all)) break
		if {$break == [llength $destroyme]} break
	}
	
	# TRIM MUST PRECEDE NONULLS (in case they're used together) !!
	# NONULLS must destroy TRIM elements when TRIM calculates "" values !!
	
	if ($flags(trim))	 {set tl $list; set list ""; foreach a $tl { lappend list [string trim $a] }}
	if ($flags(nonulls))   {set list [lsearch -not -all -inline $list ""]}
	if ($flags(nodup))	{set list [nodup $list $flags(nocase)]}
	if ($flags(noduplast)) {set list [nodup_last $list $flags(nocase)]]}

	if ($flags(test)) { return $match }
	# This allows "IF ![LDESTROY -CHECK $TEXT $1]" to be possible.

	if $flags(sort) {
		# Don't use -UNIQUE (let -NODUP handle that)
		if $flags(decreasing) {
			set list [lsort -decreasing -dictionary $list]
			set not  [lsort -decreasing -dictionary $not]
		} {
			set list [lsort -increasing -dictionary $list]
			set not  [lsort -increasing -dictionary $not]
		}
	}
	if $flags(both) { if $flags(not) { return [list $not $list] } { return [list $list $not] } }
	if $flags(not) {return $not}
	return $list
}

proc lswap { list_var index1 index2 } {
	upvar 1 $list_var local
	if [isempty local] return
	if ![isint $index1] { error "\[LSWAP\] \"${index1}\" is not a positive integer. Syntax: lswap <list variable> <index #1> <index #2>" }
	if ![isint $index2] { error "\[LSWAP\] \"${index2}\" is not a positive integer. Syntax: lswap <list variable> <index #1> <index #2>" }
	if { $index1 < 0 } { error "\[LSWAP\] \"${index1}\" is not a positive integer. Syntax: lswap <list variable> <index #1> <index #2>" }
	if { $index2 < 0 } { error "\[LSWAP\] \"${index2}\" is not a positive integer. Syntax: lswap <list variable> <index #1> <index #2>" }
	if { $index1 > [llength $local] } { error "\[LSWAP\] index \"${lindex1}\" out of range" }
	if { $index2 > [llength $local] } { error "\[LSWAP\] index \"${lindex2}\" out of range" }

	set temp [lindex $local $index1]
	lset local $index1 [lindex $local $index2]
	lset local $index2 $temp
	return $local
}

proc lformat args {
	# 2015-12-15 02:18:00 -0800: Should this be %1 / %3-5 format instead?
     # 2016-05-08 19:11:00 -0700: Yes!

	if [left [l0 $args] 1 -] { set flags [l0 $args]; set args [lreplace $args 0 0] } { empty flags }
	set flags [uniquematch -multiple $flags]
	lassign $args list format
	# Enter with: {This is a sentence.} "<1> (<2>) [<3>] {<4>}" -> This (is) [a] {sentence.}
	# Enter with: {list1 list2 list3 ...} "<1> (<2>) [<3>] {<4>}" -> This (is) [a] {sentence.}

     # Match order matters due to overlaps
	set regexp [regexp -inline -all -- {%\d+\-\d+|%\-\d+|%\d+\-|%\d+} $format]
	if [isempty regexp] {return $list}

	if ![validflag -multiple] { set list [list $list] } ; # Yes, that's right!

	empty lformat
	foreach item [lsort -decreasing -dictionary -unique $list] {
		set mask $format
		foreach lindex $regexp {
			if [regexp -- {^%\d+$} $lindex] {
				set fragment [lindex $item [mid $lindex 2]]
			} {
				lassign [split [mid $lindex 2] -] cell1 cell2
				if [isempty cell1] {zero cell1}
				if [isempty cell2] {set cell2 end}
				set fragment [lrange $item $cell1 $cell2]
			}
			regsub -all -- $lindex $mask $fragment mask
		}
		lappend lformat $mask
	}

	return $lformat
}

proc lregexp args {
	if [isempty args] {return ""}
	set destroy [lsearch -exact $args -destroy]
	if {$destroy > -1} {set args [lreplace $args $destroy $destroy]; incr destroy}
	set flags [lmatch -glob -multiple -all -- $args [list -all -about -indices -inline -expanded -line -linestop -lineanchor -nocase -start]]
	set args [ldestroy -all -multiple $args [concat -- $flags]]
	set expression [lindex $args 0]
	empty lregexp
	for {set x 1} {$x < [llength $args]} {incr x} {
		set test [lindex $args $x]
		set regexp "regexp -inline -all $flags -- \{$test\} \{$expression\}"
		set lregexp [concat $lregexp [eval $regexp]]
		if $destroy { foreach item $lregexp { regsub -all -- $item $expression "" expression } }
	}
	return $lregexp
}

proc lpurge {list deleteme { offset 0 } } {
	# OFFSET: 0/1 (for pos ->index, use: 1)
	set deleteme [splitnumberlist $deleteme]
	# Work from the highest number down, to not compromise the list!
	foreach a [lsort -int -dec -uni $deleteme] {
		decr a $offset
		if {$a <= [llength $list]} {set list [lreplace $list $a $a]}
	}
	return $list
}

proc lremove args {
	# 2017-06-01 05:24:00 -0700: This used to automatically imply: -all -nocase -glob
	if [isempty args] { return "" }
	flags:simple $args [list -debug -nocase -glob -exact -regexp -all] temp flags
	set debug [validflag -debug]
	lassign $temp var args
	if [isempty args] { return "" }

	upvar 1 $var local
	set matchmode glob
	foreach a [list regexp exact glob] { if [validflag -$a] { set matchmode $a } }
	set all [lindex [list $matchmode all] [validflag -all]]
	set case st[lindex [list = l] [validflag -nocase]]
	if [isnull local] { empty local }

if $debug { debug var local matchmode all case args }

	zero matches
	foreach arg $args {
		set m [lsearch -$all -$matchmode [$case $local] [$case $arg]]
if $debug { debug m }
		foreach n [lsort -int -dec -uni $m] {
if $debug { debug n }
			incr matches
			set local [lreplace $local $n $n]
		}
	}
if $debug { debug local matches }
	return $matches
}

proc lcount list {
	emptyarray array
	foreach item $list {
		if [info exists array($item)] {incr array($item)} {one array($item)}
	}
	return [array get array]
}

proc lcompile {list elements} {
	# USER IS RESPONSIBLE FOR ANY SORTING OF ELEMENTS NECESSARY!
	empty temp
	set ll [llength $list]
	foreach element $elements {set temp [concat $temp [SB:get_count $element 1 $ll]]}
	set elements $temp

	empty lcompile
	foreach element $elements {lappend lcompile [lindex $list $element]}
	return $lcompile
}

proc lfile args {
	set cmd [l0 $args]
	catch {file cow} shit
	set options [regsub -all -- , [lreplace [mid $shit [ expr [ instr $shit : ] + 10 ]] end-1 end-1] ""]
	set um [uniquematch $options $cmd]
	if [isempty um] {error "Error in LFILE: illegal option \"${cmd}\" (valid options are the same as FILE: [and $options "or"])"}

	empty lfile
	switch -exact -- $um {

		stat - lstat {# Will this really be needed?
			unset lfile
			lassign $args null file var
			upvar 1 $var lfile
			foreach item [l1 $args] {
				file $um $item stats
				lappend lfile [array get stats]
			}
			return $lfile
		}

		rootname - extension - tail - dirname {
			foreach entry [l1 $args] {lappend lfile [file $um $entry]}
			return $lfile
		}

		default {
			set list [l1 $args]
			foreach item $list {
				lappend lfile [file $um $item]
			}
			return $lfile
		}

	}
	return 0
}

proc lchange args {
	set arg [concat * $args]
	SB:setvariables -exact -glob -nocase -multiple -nonulls -sort -unique -all
	if ![lmatch -check -multiple -- $flags [list -exact -glob]] {lappend flags -glob}
	lassign $1e original matchme replacewith

	if ![validflag -multiple] {
		set matchme [list $matchme]
		set replacewith [list $replacewith]
	}

	negone counter_match
	foreach match $matchme replace $replacewith {
		incr counter_match
		if [validflag -glob] {
			if [validflag -nocase] {
				if [validflag -all] {
					set matches [lsearch -all -glob [stl $original] [stl $match]]
				} {
					set matches [lsearch      -glob [stl $original] [stl $match]]
				}
			} {
				if [validflag -all] {
					set matches [lsearch -all -glob $original $match]
				} {
					set matches [lsearch      -glob $original $match]
				}
			}
		} {# -EXACT
			if [validflag -nocase] {
				if [validflag -all] {
					set matches [lsearch -all -exact [stl $original] [stl $match]]
				} {
					set matches [lsearch      -exact [stl $original] [stl $match]]
				}
			} {
				if [validflag -all] {
					set matches [lsearch -all -exact $original $match]
				} {
					set matches [lsearch      -exact $original $match]
				}
			}
		}
		foreach index $matches {set original [lreplace $original $index $index $replace]}
	}
	return $original
}

proc lcommon args {
	if {[llength $args] < 2} {error "\[LCOMMON\] requires 2 or more params: <list#1> <list#2> \[... list#N\]"}
	empty lcommon full
	foreach list $args {set full [concat $full $list]}
	foreach item [lsort -inc -uni -dict $full] {
		one ok
		foreach list $args {
			if {[lsearch -exact $list $item] < 0} {zero ok}
		}
		if $ok {lappend lcommon $item}
	}
	return $lcommon
}

proc lsort:nocolor {1 2} {
	set 1 [string tolower [nocolor $1]]
	set 2 [string tolower [nocolor $2]]
	if {$1 < $2} {return -1}
	if {$1 > $2} {return +1}
	return 0
}

proc lsort:alphaonly {1 2} {
	set temp [list $1 $2]; # Will be filled-in later.
	foreach i "0 1" {
		set var [lindex $temp $i]
		set var [SB:findfirstalpha $var]
		lset temp $i $var
	}
	# Use same case: let -DICTIONARY flag deal with this.
	set 3 [sl0 $temp]
	set 4 [sl1 $temp]
	if {$3 < $4} {return -1}
	if {$3 > $4} {return +1}
	return 0
}

proc SB:findfirstalpha a {
	set count 0
	foreach char [split $a ""] {
		incr count
		if [regexp -- {[0-9a-zA-Z]} $char] {set a [mid $a $count]; break}
	}
	return $a
}

proc lsort:globallevel {1 2} {
	set 1 [SB:accesslevel $1]
	set 2 [SB:accesslevel $2]
	if {$1 > $2} {return -1}
	if {$1 < $2} {return +1}
	if {[SB:findfirstalpha $1] < [SB:findfirstalpha $2]} {return -1}
	if {[SB:findfirstalpha $1] > [SB:findfirstalpha $2]} {return +1}
	return 0
}

proc lsort:channames_only {1 2} {string compare -n [join [regexp -inline {#.*} $1]] [join [regexp -inline {#.*} $2]]}

proc lsort:len {1 2} {
	if {[len $1] < [len $2]} {return -1}
	if {[len $1] > [len $2]} {return +1}
	string compare $1 $2
}

proc lsort:list {list priority { last "" } } {

	# Set the pecking order (at least, for the beginning)
	#! Removed -DICT flag from LSORT.
	set ::SB_LSORT [nodup_last [concat [nodup [concat $priority [lsort -inc -uni $list]]] $last]]

	set list [lsort -uni -inc -command lsort:list2 $list]
	unset ::SB_LSORT

	return $list
}

proc lsort:list2 {1 2} {
	#! Removed -NOCASE from STRING, STL from $PRIORITY / $1 / $2 in LSEARCHes
	set sc [string compare $1 $2]
	empty priority
	if [info exists ::SB_LSORT] {set priority $::SB_LSORT}
	set m [lsearch -glob $priority $1]
	set n [lsearch -glob $priority $2]
	if {($m <  0) && ($n <  0)} {return $sc}
	if {($m <  0) && ($n > -1)} {return +1}
	if {($m > -1) && ($n <  0)} {return -1}
	if {($m > -1) && ($n > -1)} {
		if {$m < $n} {return -1}
		if {$m > $n} {return +1}
		return $sc
	}
	return $sc
}

proc lsort:numeric {1 2} {
	set 1 [trim $1]
	set 2 [trim $2]
	if {$1 < $2} {return -1}
	if {$1 > $2} {return +1}
	return 0
}

proc lsort:dirdepth {1 2} {
	set 1_split [file split [regsub -all {\\} $1 /]]
	set 2_split [file split [regsub -all {\\} $2 /]]
	if {[llength $1_split] < [llength $2_split]} {return -1}
	if {[llength $1_split] > [llength $2_split]} {return +1}

	# Presume path lengths are equal
	foreach 1_dir $1_split 2_dir $2_split {
		if {[stl $1_dir] < [stl $2_dir]} {return -1}
		if {[stl $1_dir] > [stl $2_dir]} {return +1}
	}

	return 0
}

proc lsort:channels {1 2} {SB:sortchans $1 $2}

proc SB:sortchans {1 2} {
	# Check syntax: if called by LSORT, keep going - else: fake it (:
	if [isempty 1] {return [lsort -inc -uni -dict -command SB:sortchans [channels]]}
	## Should that be [ISEMPTY 2] ?????

	# Sort $HOME to front of the list, then others ....
	if [string eq -nocase [home] $1] {return -1}
	if {[stl $1] < [stl $2]} {return -1}
	if {[stl $1] > [stl $2]} {return  1}
	return 0
}

if ![SB:iseggcorecmd lreverse] {
	proc lreverse list {set new ""; foreach a $list {set new [linsert $new 0 $a]}; return $new}
}

if ![SB:iseggcorecmd lassign] {
	# 1197107110: LASSIGN is in the SPLIT documentation within the ActiveTCL
	# 8.4.15.0 DOC(umentation) CHM file. HOWEVER, it's =NOT= a valid command!

	# What gives, ActiveState?
	# 2010Oct20: per thommey (eggdrop core coder), LASSIGN is in TCL8.5.
	# 2013-03-18: Per docs, LASSIGN was in TCLX previous to TCL8.5

	# 1363651482: replacement PROC ( to obey syntax as per: http://wiki.tcl.tk/1530 )
	proc lassign { list args } {
		# Return unused elements as a new list
		# Any unused variables should be set to "" (this is different than REGEXP)
		set return ""
		set c 0
		while 1 {
			if [isempty list] { foreach a $args { upvar 1 $a leftover_$a ; empty leftover_$a } ; break }
			if [isempty args] { set return $list ; break }

			set val [lindex $list 0]
			set var [lindex $args 0]
			if [string match *(*) $var] { error "\[LASSIGN\] Target variables must be scalar not array: \"${var}\" looks like an array" }
			upvar 1 $var local_$var
			set local_$var $val
			set list [lreplace $list 0 0]
			set args [lreplace $args 0 0]
		}
		return $return
	}
}; # END IF

if ![SB:iseggcorecmd lassign:array] {
	proc lassign:array { name list args } {
		 upvar 1 $name array
		 set leftovers [list]
		 array set array [list]
		 foreach a $list b $args {
			  if [isempty a] { set array($b) "" ; continue }
			  if [isempty b] { lappend leftovers $a ; continue }
			  set array($b) $a
		 }
		 return $leftovers
	}
}

if ![SB:iseggcorecmd lrepeat] {
	# Had to switch the order of args to match TCL8.6's implementation (BLAH!)
	proc lrepeat args {
		# Optional flag: -FLAT (flatten list)
		# LIST:UNPACK uses -FLAT
		lappend flags
		while { [string eq [string index [lindex [split $args] 0] 0] -] } {
			set flag [lindex [split $args] 0]
			set args [lreplace $args 0 0]
			if [string eq $flag --] break
			lappend flags $flag
		}
		lassign $args iterations list ; # TCL8.6's LREPEAT uses: iterations list :(
		if { ![isnum -integer $iterations] && [isnum -integer $list] } { swap iterations list }
		if { $iterations < 1 } return
		lappend lrepeat ; # Create & define as pure LIST
		set flat [ expr ( [lsearch -exact $flags -flat] == -1 ) ? 0 : 1 ]
		for { set counter 1 } { $counter <= $iterations } { incr counter } {
			if $flat { set lrepeat [concat $lrepeat $list] } { lappend lrepeat $list }
		}
		return $lrepeat
	}
}

proc lrepeat:nested args {
	# http://wiki.tcl.tk/43 -- modified by me
	set number [lindex $args end]
	set args [lreplace $args end end]
	if { [llength $args] == 1 } { set args [lindex $args 0] }
	set buf {}
	foreach number $args {
		incr number 0 ; # force integer (1)
		for { set i 0 } { $i < $number } { incr i } { lappend buf $value }
		set value $buf
	}
	set buf
}

proc splitm {string { members " " } } {
	empty splitm

	# If no matches, just return the original string, just like SPLIT does.
	set i [instr $string $members]
	if !($i) {return $string}

	do {
		decr i
		lappend splitm [left $string $i]
		incr i [ expr [ len $members ] + 1 ]
		set string [mid $string $i]
		set i [instr $string $members]; # For the next loop ....
		if !($i) break
	}
	if [notempty string] { lappend splitm $string }
	return $splitm
}

proc split:size [list string size] {
	if ![isint $size] return
	if { $size < 1 } { return $string }
	if { $size > [len $string] } { return $string }
	set index [ expr $size - 1 ]
	zero counter
	while { [notempty string] } {
		lappend new [string range $string 0 $index]
		set string [string range $string $size end]
	}
	return $new
}

# --- STRING manipulation ---

proc subst:var text { uplevel 1 subst -nocommands -nobackslashes \{${text}\} }

proc tcl {script} {
	# https://wiki.tcl.tk/8344 (allows linefeeds between incomplete commands and the next list element (if it start with a brace)
	regsub -all {\n\s*(\{.*)$} $script { \1} script
	uplevel 1 $script
}

proc $ args { # "set all"
	if ![string is list $args] { error "\$ expect a list: \$ var ?var? ?var? ... ?var? value" }
	if { [llength $args] < 2 } { error "\$ syntax: \$ var ?var? ?var? ... ?var? value" }
	set value [lindex $args end]
	foreach var [lrange $args 0 end-1] { uplevel set $var $value }
	return $value
}

proc !$ args { foreach a $arhs { uplevel unset -nocomplain $a } } 

proc match args {
	# Based of C# code at: https://stackoverflow.com/questions/188892/glob-pattern-matching-in-net
	set me [lindex [info level 0] 0]
	if [string eq -nocase HELP $args] { return "$me ?-nocase? ?-debug? <source string> <pattern to test> ?start index? ?length of source to test?" }
	foreach default [list nocase debug] { set $default 0 }
	set flags ""
	while { [string eq [string index $args 0] -]} {
		set split [split $args]
		set args [join [lassign [split $args] flag]]
		if [string eq $flag --] break
		lappend flags $flag
		set [string trimleft $flag -] 1
	}
	lassign $args value pattern start length
	if [string is integer -strict $start] { set value [string range $value $start end] }
	if [string is integer -strict $length] { set value [string range $value 0 [ expr $length - 1 ]] }
	if $nocase {
		set value [string tolower $value]
		set pattern [string tolower $pattern]
	}
if $debug { puts FLAGS($flags):VALUE($value):PATTERN($pattern) }
	set index 0
	while { [string length $pattern] != $index } {
 if $debug { puts INDEX($index):PATTERN($pattern):CHAR([string index $pattern $index]) }
		switch -exact -- [string index $pattern $index] {
			? { # Do nothing: it matches any character that it's tested against # }
			* {
				for { set i [string length $value] } { $i >= $index } { incr i -1 } {
					# Recursive call!
if $debug { puts I($i):FRAGMENT1([string range $value $i end]):FRAGMENT2([string range $pattern [ expr $index + 1 ] end]) }
					if [$me [string range $value $i end] [string range $pattern [ expr $index + 1 ] end]] { return true }
				}
				return false
			}
			default {
				if { [string length $value] == $index } { return false }
				if ![string eq [string index $pattern $index] [string index $value $index]] { return false }
			}
		}
		incr index
	}
if $debug { puts VALUE($value):LEN([string length $value]):INDEX($index):BOOLEAN([expr ( [string length $value] == $index ) ? true : false]) }
	expr ( [string length $value] == $index ) ? true : false
}

proc format:number { number { pad_integer 3 } { pad_mantissa 7 } { pad_sign false } } {
	# Forces padding of integer and mantissa portions and
	# respects the need for extra room for polarity sign
	set f [format %+0.${pad_mantissa}f $number]
	regexp -nocase -- {^([+-])(\d+)(\.?\d+?)$} $f - pol int man
	#if $pad_sign { if [string eq "" $pol] { set pol + } } { set pol "" }
	if $pad_sign { if [string eq "" $pol] { set pol + } }
	set s [split $f .]
	lset s 0 ${pol}[format %0${pad_integer}i $int]
	join $s .
}

proc istrue    arg { string is true    -strict $arg }
proc isfalse   arg { string is false           $arg } ; # -STRICT not needed here
proc isboolean arg { string is boolean -strict $arg }

proc boolean arg {# Force to boolean: any "true" value = 1, all others = 0
	if [isboolean $arg] { return [istrue $arg] }
	if [istrue $arg] {return 1} {return 0}
}

proc len string {string length $string}
proc trim {s { e "" } } { if [isempty e] { string trim $s } { string trim $s $e } }

proc prepend {1 2} {
	upvar 1 $1 var
	if [isnull var] {empty var}
	set var "${2}${var}"
}

proc left args {
	empty flags
	lassign $args a b c

	if [string match -nocase "-*" $b] {set b [ expr [ len $a ] + $b ]}
	lassign [split $b .] b ; # Don't use INT (endless loop). # Some procs might pass a floating-point. STRING RANGE will choke.
	incr b -1
	set left [string range $a 0 $b]
	if [isempty c] { return $left }
	if [validflag -nocase] { string eq -nocase $left $c } { string eq $left $c }
}

proc right args {
	empty flags
#	while { [string match -nocase "-?*" [string index $args 0]] } {
#		lappend flags [lindex [split $args] 0]
#		set args [lreplace $args 0 0]
#	}
#	set flags [uniquematch -nocase $flags]
	lassign $args a b c
	lassign [split $b .] b ; # Some procs might pass a floating-point. STRING RANGE will choke.
	set left [ expr [ string length $a ] - $b ]
	if [string match -nocase "-*" $b] {set left [expr abs( $b ) ]}
	set right [string range $a $left end]
	if {$c == ""} {return $right}
	if [validflag -nocase] { string eq -nocase $right $c } { string eq $right $c }
}

proc pos { a b c } {
	int b c ; # Some procs might pass a floating-point. STRING RANGE will choke.
	string range $a [decr b] [decr c]
}

# 2015-07-20 18:11:00 -0700 See replacement PROC below
proc `mid {a b {c "2147483647"} { d "" } } {
	if [string match -nocase e* $b] {set b [len $a]}; # "End"
	if [string match -nocase e* $c] {set c [len $a]}; # "End"
	int b c; # Some procs might pass a floating-point. STRING RANGE will choke.
	if ![regexp -- {^\d+$} $b] {return ""}
	if ![regexp -- {^[\+\-]?\d+$} $c] {return ""}
	if !($b) {return ""}
	if !($c) {return ""}
	incr b -1; # Counter offset [INDEX vs position]
	set mid [left [string range $a $b end] $c]
	if [isempty d] {return $mid} {string eq $mid $d}
}

# 2015-07-20 18:11:00 -0700 Replacement PROC
proc mid { a b { c 2147483647 } { d "" } } {
	if [string match -nocase e* $b] { set b [len $a] } ; # "End"
	if [string match -nocase e* $c] { set c [len $a] } ; # "End"
	if ![isint $c] return
	if !$c return
	set mid [list]
	if [instr $b ,] { set s [split $b ,] } { set s $b }
	foreach b $s {
		set b [trim $b]
		if ![isint $b] return
		if { $b < 0 } { set b [ expr [ len $a ] + $b + 1 ] } ; # "-1" will be removed in the next step
		incr b -1; # Counter offset [INDEX vs position]
		lappend mid [left [string range $a $b end] $c]
	}
	if [isempty d] { if { [llength $s] == 1 } { return [lindex $mid 0] } { return $mid } }
	string eq $mid $d
}

proc textbefore args {
	if [left [l0 $args] 1 -] { set flags [uniquematch -nocase [l0 $args]]; set args [lreplace $args 0 0] } { empty flags }
	lassign $args string after
	if [validflag -nocase] {
		set i [string first [stl $after] [stl $string]]
	} {
		set i [string first $after $string]
	}
	if {$i < 0} {return ""}
	decr i
	return [string range $string 0 $i]
}

proc textafter args {
	if [left [l0 $args] 1 -] { set flags [uniquematch -nocase [l0 $args]]; set args [lreplace $args 0 0] } { empty flags }
	lassign $args string after
	if [validflag -nocase] {
		set i [string first [stl $after] [stl $string]]
	} {
		set i [string first $after $string]
	}
	if {$i < 0} {return ""}
	incr i
	return [string range $string $i end]
}

proc getlongest crap {
	zero x
	foreach a $crap { set l [len $a] ; if { $l > $x } { set x $l } }
	return $x
}

interp alias {} getsmallest {} getshortest
proc getshortest crap {
	set x [llength [l0 $crap]]; # Gotta start somewhere!
	foreach a $crap { set l [len $a] ; if { $l < $x } { set x $l } }
	return $x
}

interp alias {} pushleft {} SB:padleft
interp alias {} pushright {} SB:padright
interp alias {} pushcenter {} SB:padcenter
proc SB:padleft   {string pos { char " " } } {left  ${string}[string repeat $char $pos] $pos}
proc SB:padright  {string pos { char " " } } {right [string repeat $char $pos]${string} $pos}
proc SB:padcenter {string size { char " " } } {
	if {[len $string] >= $size} {return $string}
	set half [ fixmath int( ceil( ( $size - [ len $string ] ) / 2 ) ) ]
	left [string repeat $char $half]${string}[string repeat $char $size] $size
}

proc ajl { list { comma ", " } { and & } { oxford_comma false } } {
	# Ignore colours when sorting
	set list [lsort -inc -uni -dict $list] ; # Do this here so the count is properly done (1 2 1 1 2 -> "1 & 2" only) 
	switch -exact -- [llength $list] {

		0 { return "" }

		1 { return [join $list] }

		2 { return [join [lsort -increasing -dictionary $list] " $and "] ; # -command -lsort:strip }

		default {
			set list [lsort -unique -increasing -dictionary $list] ; # -command -lsort:strip
			# CONCAT to preserve STRING / LIST rule integrity

			#set list [lreplace $list end end [join [list $and [lindex $list end]]]]

			# It's okay to do this: we're gonna convert LIST -> STRING via JOIN anyway
			empty oxford
			if $oxford_comma { set oxford $comma }
			set oxford [string trimright $oxford " "]
			set list [lreplace $list end end "${and}$oxford [lindex $list end]"]
			set list [join $list $comma]
			return $list
		}
	}
	?
}

proc and args {
	flags:simple $args -commas _ flags
	lassign $_ text and
	if [isempty and] { set and "and" }
	switch -exact -- [llength [split $text]] {
		0 - 1 { return $text }
		2 { return [join $text " $and "] }
		default {
			set final [lreplace [split $text] end end "$and [lindex [split $text] end]"]
			if [validflag -comma] { set final [join $final ", "] }
			join $final
		}
	}
}

# See replacement PROC below
proc using {1 {2 ""}} {

# If USING begins to fail, see if its the new DEC(IMAL?) code causing it....

#putcmdlog "> USING: \$1 ($1), \$2 ($2)"
	   if {$2 == ""} {
			 if {[llength $1] == 1} {
				    set 2 " "
				    # This is to kill the "ap<*80" outputs when
				    # blank data is passed to USING.
			 } else {
				    set crap [killcode $1]
				    set 1 [l0 [killcode $crap]]
				    set 2 [mid $crap [expr [string first " " $crap] + 2 ] ]
			 }
	   }
	   if {$1 == ""} {error "No format string provided for USING"}
	   if {$2 == ""} {error "No string to format for USING"}
	   # Check for ALPHA formatting...
	   #if {[lsearch -exact [split "_=<>@CULPRS\240" ""] $1] > -1} {prepend 1 A}
	   if [regexp -nocase -- {[_=<>@CULPRS\240\*]+} $1] {prepend 1 A}
	   if [instr [stu $1] Z] {append 1 "%"; #set 1 [join [ldestroy -all -nonull -nocase [split $1 ""] A] ""]}
	   if {![regexp -nocase -- ^\[%BITZNR0-9\+\-.\]+$ $2] || [string match -nocase *a* $1]} {# Anything but "%BITZNR+-0123456789."}
	   if [instr $1 %] {regsub -all a|A $1 "" 1}
	   if [instr [stu $1] A] {
			 array set format "cjt 0 ljt 1 rjt 0 fjt 0 uc 0 lc 0 tc 0 cc 0 pp 0 min 0.50 cws 0 underline 0 trim 0 nbsp 0"
			 set range ""; set final ""; set space ""; set f $1
			 while {$f != ""} {
				    switch -exact -- [stu [left $f 1]] {
						  $ {putcmdlog "Egad!"}
						  A {# Force ALPHA parsing of a number}
						  N {set format(underline) 1}
						  _ {set format(cws) 1; # ConvertWhiteSpace}
						  = {set format(cjt) 1; set format(ljt) 0; set format(rjt) 0; set format(fjt) 0}
						  < {set format(ljt) 1; set format(cjt) 0; set format(rjt) 0; set format(fjt) 0}
						  > {set format(rjt) 1; set format(cjt) 0; set format(ljt) 0; set format(fjt) 0}
						  @ {set format(fjt) 1; set format(cjt) 0; set format(ljt) 0; set format(rjt) 0}
						  0 {set format(min) 0.00}
						  2 {set format(min) 0.25}
						  5 {set format(min) 0.50}
						  7 {set format(min) 0.75}
						  9 {set format(min) 0.90}

						  C {set format(cc) 1; set format(tc) 0; set format(lc) 0; set format(uc) 0}
						  T {set format(tc) 1; set format(cc) 0; set format(lc) 0; set format(uc) 0}
						  U {set format(uc) 1; set format(cc) 0; set format(lc) 0; set format(tc) 0}
						  L {set format(lc) 1; set format(cc) 0; set format(uc) 0; set format(tc) 0}
						  P {set format(pp) 1}
						  R {set format(trim) 1}
						  S - \xa0 {set format(nbsp) 1; # NonBreakingSPace}
						  " " {#}
						  # {append space " "}
						  * {set v [val [mid $f 2] 1]; if !($v) {return -errorcode error "No value set for * in USING: $f"}; set space [string repeat " " $v]; set f [mid $f [expr [len $v] + 1 ]]}
						  default {return -errorcode error "Unknown USING text format option: $f"}
				    }
				    set f [mid $f 2]
			 }
			 if {$space == ""} {set space [string repeat " " [string length [join $2]]]}
			 set original $2
			 if ($format(uc)) {set 2 [stu $2]}
			 if ($format(lc)) {set 2 [stl $2]}
			 if ($format(tc)) {set 2 [sts $2]}
			 if ($format(cc)) {
				    set a [split [join $2] " "]; set 2 ""
				    foreach b $a {lappend 2 [sts $b]}
			 }
			 if ($format(pp)) {
				    set ultimate ""
				    regsub -all "p" $1 "" 1
				    if {[string length [join $2]] > [string length $space]} {
						  set counter 0
						  set 2 [join $2]
						  while {$2 != ""} {
								incr counter; if {$counter > 1000} {return -errorcode error "> USING: endless loop! $counter iterations have been performed."}
								if {[string length $2] > [string length $space]} {
									   set l [string last " " $2 [string length $space]]
									   if {$l > -1} {
										      set fragment [left $2 $l]
										      incr l 2
										      set 2 [mid $2 $l]
									   } else {
										      set fragment $2; set 2 ""
									   }
								} else {
									   set fragment $2; set 2 ""
								}
								if {$ultimate != ""} {append ultimate "\n"}
								append ultimate [using $1 $fragment]
								# putcmdlog "> USING: adding \$FRAGMENT ($fragment), \$1 ($1) \[[using $1 $fragment]\]"
						  }
						  return $ultimate
				    }
			 }
			 if {[string length [join $2]] > [string length $space]} {
				    regsub -all " " $space "*" final
			 } else {
				    set size [string length $space]
				    if ($format(ljt)) {set final [left "[join $2]$space" $size]}
				    if ($format(rjt)) {set final [right "$space[join $2]" $size]}
				    if ($format(cjt)) {
						  set offset [expr int(0.5 + ($size - [string length [join $2]]) / 2.0)]
						  set final "[left $space $offset][join $2]"
						  set final [left $final$space $size]
				    }
				    if ($format(fjt)) {
						  set 2 [join $2]
						  if {[llength $2] == 1} {return [using "a<[string repeat # [string length $space]]" $2]}
						  if {[string length $2] == $size} {return $2}
						  # If 2/5/7 % rate isn't exceeded, left-flush line.
						  if {[expr [join [string length $2]]. / $size] < $format(min)} {
								set final [join $2]
						  } else {
								set split [split $2 " "]; set end [expr [llength $split] - 1]
								set length [llength $split]
								array set phrase ""; array set sp "$length 0"
								set i 1; foreach a $split {set phrase($i) $a; incr i}
								set total [expr $size - [string length [join $split ""]]]
								# Spaces will be indeces (1) to ($count-1).
								for {set i 0} {$i <= $length} {incr i} {set sp($i) 0}
								while {$total > 0} {
									   for {set i $end} {($i >= 1) && ($total > 0)} {incr i -1} {
										      incr sp($i)
										      incr total -1
										      if {($i < -1000) || ($i > 1000)} {return -errorcode error "I went into an endless loop!"}
									   }
								}
								for {set i 1} {$i <= $length} {incr i} {
									   append final $phrase($i)
									   append final [string repeat " " $sp($i)]
									   if {($i < -1000) || ($i > 1000)} {return -errorcode error "I went into an endless loop!"}
								}
						  }
				    }
			 }
			 if ($format(cws)) {regsub -all "_" $final " " final}
			 if ($format(nbsp)) {regsub -all \xA0 $final " " final}
			 if ($format(underline)) {
				    set temp ""
				    foreach a [split $final ""] {
						  if {$a == " "} {
								append temp $a
						  } else {
								append temp "\037${a}\x1f"
						  }
				    }

				    # Wouldn't it be better to underline entire words,
				    # and not do it char by char?

				    regsub -all "\x1f\037" $temp "" final; # Kill off/on's
			 }
			 if ($format(trim)) {set final [string trim $final]}
			 return $final
	   } else {# Numerical evaluation
			 array set format "pos 0 neg 0 money 0 comma 0 int 0 bytes 0 trim 0 leadingzero 0 underline 0 cws 0 padbytes 0"
			 # If bytes = 1, pad the format string with 3 more #'s.
			 set range ""; set final ""; set space ""; set f $1; set o $2
			 while {$f != ""} {
				    switch -exact -- [stu [left $f 1]] {
						  % #
						  + {incr format(pos); set format(neg) 0}
						  "-" {incr format(neg); set format(pos) 0}
						  $ {incr format(money)}
						  , {set format(comma) 1}
						  N {set format(underline) 1}
						  B {set format(bytes) 1}
						  D {set format(padbytes) 1}
						  I {set format(int) 1}
						  R - T {set format(trim) 1}
						  Z {set format(leadingzero) 1}
						  " " {#}
						  . - # {append space [left $f 1]}
						  _ {set format(cws) 1}
						  * {set v [val [mid $f 2] 1]; if !($v) {return -errorcode error "No value set for * in USING: $f"}; set space [string repeat "#" $v]; set f [mid $f [expr [len $v] + 1 ]]}
						  default {return -errorcode error "Unknown USING numeric format option: [left $f 1]"}
				    }
				    set f [mid $f 2]
			 }
			 if ![expr $format(pos) + $format(neg) ] {set format(neg) 1}
			 if ($format(bytes)) {set 2 [l0 [bytes $2]]}
			 if ($format(int)) {set 2 [expr int( $2 ) ]}
			 # Next line is NOT unnecessary; it's needed for formatting.
			 if ![string match *.* $2] {append $2 ".0"}

			 set os 0
			 if {$space == ""} {
				    set os [expr ($format(comma) * int( ( [len $2] - 1 ) / 3 ) ) ]
				    set spc [expr [len $2] + $os]
				    set space [string repeat "#" $spc]
			 }

			 # Create a sub-string with buffering spaces. Then, drop-insert into each
			 # position in the final formatted string.

			 if [string match *.*.* $2] {return -errorcode -error "> USING: Illogical decimal value"}
			 set posneg "+"
			 if {$format(money) > 1} {
				    foreach pn "pos neg" {set format($pn) [double $format($pn) 1]}
			 }
			 if {$2 < 0} {set posneg "-"}
			 set 2 [expr abs( $2 ) ]

			 set spc [string length $space]
			 foreach a [split $1 ""] { break; if {($a == "#") || ($a == ".")} {incr spc} }
			 if {$spc == 0} {return -errorcode -error "> USING: illegal format specified (# count)"}

			 # Note: integer portion must be right-flush; mantissa left-flushed.
			 set int [l0 [split $space .]]; set intcount 0
			 set man [l1 [split $space .]]; set mancount 0
			 foreach a [split $int ""] { if {$a == "#"} {incr intcount} }
			 foreach a [split $man ""] { if {$a == "#"} {incr mancount} }

#putcmdlog "> USING: \$INT ($int)"
#putcmdlog "> USING: \$MAN ($man)"
#putcmdlog "> USING: \$INTCOUNT ($intcount)"
#putcmdlog "> USING: \$MANCOUNT ($mancount)"

#putcmdlog "> USING: setting \$3"
			 set 3 [l0 [split $2 .]]
			 #debug.print 2 "\[USING\] \$3 ($3)"
			 if ($format(leadingzero)) {
				    set 3 "[string repeat 0 $intcount]$3"
				    # set os [expr $intcount - ($format(comma) * int( ( $intcount - 1 ) / 3 ) ) ]
				    #debug.print 2 "\[USING\] \$3 ($3), \$INTCOUNT ($intcount), \$OS ($os)"
				    set 3 [right $3 [expr $intcount - $os ]]
			 }
			 #debug.print 2 "\[USING\] \$3 ($3)"
			 if ($format(comma)) {set 3 [comma $3]}
			 #debug.print 2 "\[USING\] \$3 ($3)"
			 if {$format(pos) > 1} {set 3 "$posneg$3"}
			 if {($format(neg) > 1) && ($posneg == "-")} {set 3 "-$3"}
			 if {$format(money) > 1} {set 3 "\$$3"}; # MONEY = last here
#putcmdlog "> USING: set \$3 ($3)\; setting \$4"
			 set 4 [l1 [split [decimal $2 $mancount 0] .]]; # DO NOT ROUND!!
#putcmdlog "> USING: set \$4 ($4)"
			 if {[string length $3] > $intcount} {return [string repeat * [string length $space]]}
			 set intportion [right "[string repeat " " $intcount]$3" $intcount]
#putcmdlog "> USING: \$INTPORTION ($intportion)"
			 # set manportion ""
			 if !($format(int)) {
			 # Next line goes here [SET MANPORTION ....]}
			 set manportion [left "$4[string repeat " " $mancount]]" $mancount]
#putcmdlog "> USING: \$MANPORTION ($manportion)"

			 set runstring "{{$int} {$intportion}} {{$man} {$manportion}}"
#putcmdlog "> USING: \$RUNSTRING ($runstring)"
			 set response ""; set segment 0
			 foreach a $runstring {
				    set l [l0 $a]; set r [l1 $a]; set temp ""
#putcmdlog "> USING FOREACH: \$A ($a), \$L ($l), \$R ($r)"
				    set charcount 0
				    for {set b 0} {$b <= [string length $l]} {incr b} {
						  set char [string index $l $b]
						  if {$char == "#"} {
								set char [string index $r $charcount]
								if {$char == " "} {
									   if ($format(leadingzero)) {set char "0"}
								}
								incr charcount
						  }
						  append temp $char
				    }
				    lappend response $temp
				    if {$charcount < [string length $r]} {
						  if !($segment) {# Integer portion
								return [string repeat * [string length $space]]
						  }
				    }
				    incr segment
			 }
#putcmdlog "> USING: \$RESPONSE ($response)"
			 if {[l1 $response] == ""} {
#putcmdlog "> USING: IF: \$RESPONSE \{\} located."
				    set response [l0 $response]
			 } else {
				    set response [join $response "."]
			 }
			 # If a decimal portion is found, conjoin it. Otherwise, lose
			 # the "{}" null string in the 1st index position.
#putcmdlog "> USING: \$RESPONSE ($response)"

#putcmdlog "> USING: \$FORMAT(MONEY) ($format(money))"
#putcmdlog "> USING: \$FORMAT(POS) ($format(pos))"
#putcmdlog "> USING: \$FORMAT(NEG) ($format(neg))"
#putcmdlog "> USING: \$POSNEG ($posneg)"

			 set pos 0
			 if {$format(money) == 1} {# MONEY = FIRST here (but second above)!
#putcmdlog "> USING: processing MONEY"
				    if {[string index $response $pos] == " "} {
#putcmdlog "> USING: adding MONEY"
						  set response [string replace $response $pos $pos "\$"]
						  incr pos
				    } else {
						  return [string repeat * [string length $space]]
				    }
			 }
			 if {($format(pos) == 1) || (($format(neg) == 1) && ($posneg == "-"))} {
#putcmdlog "> USING: processing POS/NEG (where \$POSNEG = $posneg)"
				    if {[string index $response $pos] == " "} {
#putcmdlog "> USING: adding \$POSNEG ($posneg)"
						  set response [string replace $response $pos $pos "$posneg"]
				    } else {
						  return [string repeat * [string length $space]]
				    }
			 }
			 if {[string length $response] > [string length $space]} {
#putcmdlog "> USING: \$RESPONSE ($response), \[STRING LENGTH \$RESPONSE\] ([string length $response]), \$SPACE ($space), \[STRING LENGTH \$SPACE\] ([string length $space])"
				    set response [string repeat "*" [string length $space]]
				    return $response
			 }

			 if ($format(bytes)) {
				set bytes_padding [lindex [list "" [space 3]] $format(padbytes)]
				append response " [right ${bytes_padding}[l1 [bytes $o]] 3]"
			 }
			 if ($format(trim)) {set response [string trim $response]}
			 if ($format(underline)) {
				    set temp ""
				    foreach a [split $response ""] {
						  if {$a == " "} {
								append temp $a
						  } else {
								append temp "\037${a}\x1f"
						  }
				    }
				    regsub -all "\x1f\037" $temp "" response; # Kill off/on's
			 }
			 if ($format(cws)) {regsub -all _ $response " " response}
			 return $response
	   }
	   # ... or else, give up guessing.
	   return -errorinfo error "Unparsed string error in USING ($1): $2"
}

# Replacement PROC
proc `using args {
	lassign $args format text
	# Text
	# %[-]<digits>s = right-justified (- = left)
	#

	# Numeric
	# "+" = Sign should always be printed
	# " " = if first character isn't a sign (polarity), pad-left with spaces
	# "0" = pad-left with spaces
	# "#" = output control: for "oO" (-> "0*"), "xX" (-> "0x*"), "bB" (-> "0b*"), "eEfgG" (-> "*.*" forced radix point), "gG" (-> no trailing zero removal)

	# "d" = integer [signed]
	# "u" = integer [unsigned]
	# "oO" = octal [unsigned]
	# "xX" = hexadecimal [unsigned]
	# "bB" = binary
	# "c" = convert to unicode
	# "s" = insert string (no conversion)
	# "f" = decimal [signed] size of "y" -> "xx.yyy" (default = 6; 0 = radix point truncation)
	# "eE" = SciNot, size of "y" -> "x.yyy[eE]Â±zz" (default = 6; 0 = radix point truncation)
	# "gG" = SciNot, size of "y" -> "x.yyy[eE]Â±zz" [eE] if y < -4 or y > $SIZE ELSE [f] (0 = radix point truncation)

	if [regexp -nocase -- {^[\+\-]?\d*\.?\d+([eE][\+\-]\d+)?$} $text] {
		# Numeric

	}
}

proc bytes args {
	flags:simple $args [list -binary -raw] text flags
	if [isempty text] { error "Syntax: bytes ?-binary? ?-raw? <value> ?precision?" }
	if [isempty text] { error "Syntax: bytes ?-binary? ?-raw? <value> ?precision?" }
	if [string eq -nocase HELP $args] { error "Syntax: bytes ?-binary? ?-raw? <value> ?precision?" }
	lassign [concat $text 1] value decimal
	set cutoff 1000
	empty binary
	if [validflag -binary] { set cutoff 1024 ; set binary i }
	set raw [validflag -raw]
	set len [len [lindex [split $value .] 0]]
	if { $value < $cutoff } { return [format "%s[iff !$raw { B}]" $value] } ; # Account for spacing due to "i" ?
	unit == int( ( $len - 1 ) / 3 )
	float value ; # Integer division below
	set value [ expr $value / ( $cutoff ** $unit ) ]
	set multiple [string index " KMGTPEZYV" $unit]
	if [string eq * $decimal] { return "$value ${multiple}${binary}[iff !$raw B]" }
	return [format "%.${decimal}f %s%s[iff !$raw B]" $value $multiple $binary]
}

proc `bytes { number { decimal * } { align_bytes "false" } } {
	set powers [list "" Ki Mi Gi Ti Pi Ei Zi Yi Xi Wi Vi]
	if $align_bytes { lset powers 0 "  " }
	for { set x [expr [llength $powers] - 1] } { $x >= 0 } { incr x -1 } {
		if { $number >= pow( 1024 , $x ) } {
			set multiple [fixmath $number / pow( 1024 , $x ) ]
			if { ( $multiple < 1 ) && ( $multiple >= 0.975 ) } { incr x }
			if [string eq * $decimal] {
				if !$x { return "$number [lindex $powers $x]B" }
				set expr [ expr $number / pow( 1024 , $x ) ]
				if { $expr == int( $expr ) } { int expr }
				return "$expr [lindex $powers $x][iff !$raw B]"
			}
			return "[format %.${decimal}f [ expr $number / pow( 1024 , $x ) ]] [lindex $powers $x][iff !$raw B]"
		}
	}
	# How do we get here: if $NUMBER < 1 ? (otherwise, it would pass the "$number >= pow( 1024 , $x ) " test above even with $X = 0)
	# Respect $ALIGN_BYTES
	if [string eq * $decimal] {
		if [isint $number] { int number }
		return "$number [lindex $powers 0][iff !$raw B]"
	}
	return "[format %.${decimal}f $number] [lindex $powers 0][iff !$raw B]"
}

proc ``bytes {bytes { dec 1 } } {# Backward compatibility
	array set exponents [list -5 0.00000000000000088817841970012523233890533447265625 -4 0.0000000000009094947017729282379150390625 -3 0.000000000931322574615478515625 -2 0.00000095367431640625 -1 0.0009765625 0 1 1 1024 2 1048576 3 1073741824 4 1099511627776 5 1125899906842624 6 1152921504606846976 7 1180591620717411303424 8 1208925819614629174706176 9 1237940039285380274899124224 10 1267650600228229401496703205376 11 1298074214633706907132624082305024 12 1329227995784915872903807060280344576 13 1361129467683753853853498429727072845824 14 1393796574908163946345982392040522594123776 15 1427247692705959881058285969449495136382746624]
	set check 15
	if {[fixmath $bytes] > [fixmath $exponents($check) ]} {return [bytes_highend $bytes $dec]}
	set bytes [l0 [split [nocomma $bytes] .] ]; # Pesky mantissa, and signed integers
	float bytes
	set exponent 0

	# Leave "FIXMATH $EXPONENTS(X)" in place. FIXMATH keeps floating-point!
	for {set x 0} {$x < 100} {incr x} {
		set pow [ fixmath $exponents($x) ]
		if { $bytes < [fixmath 0.9755859375 * $pow ]} {
			decr x
			set exponent $x
			set pow [ fixmath $exponents($x) ]
			set bytes [ fixmath $bytes / $pow ]
			break
		}
	}
	if {$dec == "*"} {set dec [ expr 3 - 2 * ( $exponent < 2 ) - !( $exponent ) ] }
	if {$dec < 0} {error "illegal value: $dec - must be an integer."}
	if ![isint $dec] {error "illegal value: $dec - must be an integer."}
	# !ISNUMBER must follow the "*" check ("*" isn't a number!)
	return "[decimal [nozero $bytes] [ expr ( $exponent > 0 ) ? $dec : 0] 1] [lindex [list "" Ki Mi Gi Ti Pi Ei Zi Yi Xi Wi Vi] $exponent]B"
}

# See replacement PROC below
# Enter with: $NUM (value), $DIG (number of decimal places) [, $ROUND (round off:0/1)]
proc `decimal {num dig { round 0 } { truncate 0 }} {

	# Do NOT use USING: USING calls DEC [creating an endless loop]!

	if ![isint $dig] {error "DECIMAL - illegal value: $dig - must be an integer."}
	if ![isint $round] {error "DECIMAL - illegal value: $round - must be an integer."}

	# This section is redundant, and even damaging, to the above section
	if [string match *.*.* $num] {error "DECIMAL - illegal value: $num - must be a valid number."}
	#if [string match *.* $dig] {error "DECIMAL - illegal value: $dig - must be an integer."}
	#if [string match *.* $round] {error "DECIMAL - illegal value: $round - must be an integer."}

	if [string match *e* $num] {error "DECIMAL - I can not process numbers in scientific notation."}
	#if {$dig < 1} {error "DECIMAL - illegal value: $dig - must be greater than 0."}

	if $round {
		if $dig {
			if ![instr $num .] {append $num .0}
			set int $num; int int
			set mantissa [left [mid [mantissa $num] 2][string repeat 0 $dig] [ expr $dig + 1 ]]; # Mantissa + rounding digit
			if {[right $mantissa 1] >= 5} {set num [::stringmath::add ${int}.[left $mantissa -1] 0.[string repeat 0 [ expr $dig -1 ]]1]}
		} {
			if ![instr $num .] {append $num .0}
			set mantissa [mid [mantissa $num] 2]
			if {[left $mantissa 1] >= 5} {
				int num
				set num [::stringmath::add $num 1]
				float num
			}
		}
	}

	if $dig {
		set return [l0 [split $num .]].[left [l1 [split $num .]][string repeat 0 $dig] $dig]
	} else {
		set return $num
		int return
	}

	if $truncate {
		zero ok
		if [regexp -- {^[+\-]?\d+\.0+$} $return] {
			one ok
			if {$truncate == 2} {# Only truncate if !0.(00000...)
				if ![nozero $return] {zero ok}
				if ![regexp -- {0(\.0+)?} $return] {zero ok}
			}
			if {$truncate == 3} {# Only truncate if 0.(00000...)
				if [nozero $return] {zero ok}
				if [regexp -- {0(\.0+)?} $return] {zero ok}
			}
		}
		if $ok {set return [l0 [split $return .]]}
	}
	return $return

	# Do NOT use USING: USING calls DEC [creating an endless loop]!
}

# Replacement PROC
proc decimal args {
	lassign $args value digits ; # We're gonna jettison the rounding and truncations options
	format %.${digits}f $value ; # Let FORMAT choke on / process errors
}

proc superscript args {
	set value [join $args]
	if [regexp -nocase -- {^[\+\-]\d+(\.\d+)?([Ee][\+\-]\d+)$} $value] { set value [normalize [noscinot $value]] }
	# Can't mix \u0000 with \X00[\X00[\X00[\X00]]]
	array set exponents [list a \xE1\xB5\x83 b \xE1\xB5\x87 c \xE1\xB6\x9C d \xE1\xB5\x88 e \xE1\xB5\x89 f \xE1\xB6\xA0 g \xE1\xB5\x8D h \xCA\xB0 i \xE2\x81\xB1 j \xCA\xB2 k \xE1\xB5\x8F l \xCB\xA1 m \xE1\xB5\x90 n \xE2\x81\xBF o \xE1\xB5\x92 p \xE1\xB5\x96 q q r \xCA\xB3 s \xCB\xA2 t \xE1\xB5\x97 u \xE1\xB5\x98 v \xE1\xB5\x9B w \xCA\xB7 x \xCB\xA3 y \xCA\xB8 z \xE1\xB6\xBB]
	array set exponents [list A \xE1\xB4\xAC B \xE1\xB4\xAE C C D \xE1\xB4\xB0 E \xE1\xB4\xB1 F F G \xE1\xB4\xB3 H \xE1\xB4\xB4 I \xE1\xB4\xB5 J \xE1\xB4\xB6 K \xE1\xB4\xB7 L \xE1\xB4\xB8 M \xE1\xB4\xB9 N \xE1\xB4\xBA O \xE1\xB4\xBC P \xE1\xB4\xBE Q Q R \xE1\xB4\xBF S S T \xE1\xB5\x80 U \xE1\xB5\x81 V \xE2\xB1\xBD W \xE1\xB5\x82 X X Y Y Z Z]
	array set exponents [list " " " "]
	array set exponents [list . \xc2\xb7 i \xE2\x81\xB1 n \xE2\x81\xBF + \xE2\x81\xBA - \xE2\x81\xBB / \xE2\x81\x84 = \xE2\x81\x8C  ( \xE2\x81\xBD ) \xE2\x81\xBE 0 \xe2\x81\xb0 1 \xc2\xb9 2 \xc2\xb2 3 \xc2\xb3 4 \xe2\x81\xb4 5 \xe2\x81\xb5 6 \xe2\x81\xb6 7 \xe2\x81\xb7 8 \xe2\x81\xb8 9 \xe2\x81\xb9]
	empty exponent
	foreach char [explode -- $value] { append exponent $exponents($char) }
	return $exponent
}

proc subscript args {
	set value [join $args]
	# Can't mix \u0000 with \X00[\X00[\X00[\X00]]]
	array set subscripts [list a \xE2\x82\x90 b b c c d d e \xE2\x82\x91 f f g g h \xE2\x82\x95 i \xE1\xB5\xA2 j \xE2\xB1\xBC k \xE2\x82\x96 l \xE2\x82\x97 m \xE2\x82\x98 n \xE2\x82\x99 o \xE2\x82\x92 p \xE2\x82\x9A q q r \xE1\xB5\xA3 s \xE2\x82\x9B t \xE2\x82\x9C u \xE1\xB5\xA4 v \xE1\xB5\xA5 w w x \xE2\x82\x93 y y z z]
	array set subscripts [list " " " "]
	array set subscripts [list . . + \xE2\x82\x8A - \xE2\x82\x8B * \xCB\x9F = \xE2\x82\x8C ( \xE2\x82\x8D ) \xE2\x82\x8E / \xE2\x81\x84 0 \xE2\x82\x80 1 \xE2\x82\x81 2 \xE2\x82\x82 3 \xE2\x82\x83 4 \xE2\x82\x84 5 \xE2\x82\x85 6 \xE2\x82\x86 7 \xE2\x82\x87 8 \xE2\x82\x88 9 \xE2\x82\x89 a \xE2\x82\x90 e \xE2\x82\x91 o \xE2\x82\x92 x \xE2\x82\x93 \x02\x59 \xE2\x82\x94 \xC9\x99 \xE2\x82\x94]
	empty subscript
	if [regexp -- {^[\+\-]?\d+\/\d+$} $value] {
		set s [split $value /]
		set subscript [exponent [lindex $s 0]]
		set value /[lindex $s 1]
	}
	foreach char [explode -- $value] { if [info exists subscripts($char)] { append subscript $subscripts($char) } { append subscript $char } }
	return $subscript
}

proc killcode text {regsub -all -- \\\\|\\\{|\\\}|\\\[|\\\]|\\\$\\\;|\\\" $text \\&}

proc comma { value { comma , } } {
	if [isempty value] return
	regexp -nocase -- {^([+-]?)([\d\.]+)(.*)$} $value - sign number other
	if [isempty number] { return $value }
	if { [string match *.*.* $number] || [string match *${comma}* $number] } { return $value }
	set neg [regexp -- {^[+-]} $sign]
	set split [split $number .]
	set 0 [reverse [lindex $split 0]]
	set 1 [lindex $split 1]
	set number [join [reverse [regexp -all -inline {.{1,3}} $0]] $comma]
	if ![string eq "" $1] {append number .$1}
	if $neg { prepend number $sign }
	return ${number}${other}
}

proc nocomma {n { comma , } } {regsub -all $comma $n "" n; return $n}
proc iscomma {n { comma , } } {string match *${comma}* $n}

proc SB:transcode {user text} {
	if ![validuser $user] {return $text}
	set encoding [getdata $user ENCODING]
	if [isempty encoding] {return $text}
	return [encoding convertto $encoding $text]
}

proc swap {var1 var2} {
	upvar 1 $var1 local_var1
	upvar 1 $var2 local_var2

	# If both are unset, ... DO NOTHING!
	if {![info exists local_var1] && ![info exists local_var2]} {return ""}

	# If either is unset, dump the (resultant) unset variable:
	if [isnull local_var1] {
		if [array exists local_var2] {array set local_var1 [array get local_var2]} {set local_var1 $local_var2}
		unset local_var2
	} elseif [isnull local_var2] {
		if [array exists local_var1] {array set local_var2 [array get local_var1]} {set local_var2 $local_var1}
		unset local_var1
	} else {
		if {[array exists local_var1] && ![array exists local_var2]} { error "\[SWAP\] $var1 is an array but $var2 is scalar" }
		if {![array exists local_var1] && [array exists local_var2]} { error "\[SWAP\] $var1 is scalar but $var2 is an array" }

		if [array exists local_var1] {# Doesn't matter which one we test
			array set temp [array get local_var2]
			unset local_var2 ; array set local_var2 [array get local_var1]
			unset local_var1 ; array set local_var1 [array get temp]
		} {
			set temp $local_var2
			set local_var2 $local_var1
			set local_var1 $temp
		}
	}
	return ""
}

# --- Math functions ---

proc crc25 args {
	flags:simple $args -file data flags
	if [validflag -file] { set data [readfile $data binary] }
	if [isempty data] { return 0 }

	set crc 0xAA55A5A5
	set len [len $data]
	set loop 0
	set c 0
	set d 1
	#set time0 [clock milliseconds]
	for { set x 0 } { $x < $len } { incr x } {
		scan [set char [string index $data $x]] %c ascii
		set seed [ expr $seed ^ ( ( 255 - $ascii ) * ( 256 ** $c ) ) ]
		if !$loop {
			#set loop $fragment
			#set percent [format %4.1f [ expr ( 100.0 * ( $x + 1 ) ) / $len ]]
			#text "\[CRC25\] ${percent}%: byte index: #[comma $x]: [regsub -- {[\r\n\t\b\v]} $char .] (${ascii})"
		}
		incr c $d
		#incr loop -1
		set d [ expr ( ( $c == 3 ) || ( $c == 0 ) ) ? -$d : $d ]
	}
	#set time1 [clock milliseconds]
	#text
	#if { [set ms [ expr ( $time1 - $time0 ) / 1000.0 ]] > 1000 } { set ms "[ expr $ms / 1000.0 ]s" } { append ms ms }
	#puts "\[CRC25\] Elapsed time: $ms"
	return $crc
}

proc list:math { list math } {
	set new [list]
	foreach a $list { lappend new [ expr [regsub -all -- # $math $a] ] }
	return $new
}

proc percent { number total { decimal "" } } {
	if !$total { return 0% }
	set % [ fixmath 100 * ( $number / $total ) ]
	if [string eq "" $decimal] { set % [nozero ${%}] } { set % [format %.${decimal}f ${%}] }
	return ${%}%
}

proc sgn value { string compare [string trimleft [nozero $value] +] 0 }
proc sqr value { nozero [ expr pow( $value , 0.5 ) ] }
proc abs value { string range $value [regexp -- {^[+\-]} $value] end}

# Replacement PROCs
proc sgn x { expr { ( $x > 0 ) - ( $x < 0 ) } }
proc sgn { value { mode "" } } {
	# Enter with: sgn value [pos | neg | both | all / force (zero also gets "+")]
	if [string eq "" $value] { return 0 }
	if [regexp -- {^[\+\-]?0*(\.0)*$} $value] {
		# Must check this first ... damned negative-zero
		set compare 0
	} {
		if [string eq [string index $value 0] -] {
			set compare -1
		} {
			set compare 1
		}
	}
	#set compare [string compare [string trimleft $value +] 0]
	if [isempty mode] { return $compare }
	switch -glob -- [string tolower $mode] {
		p* { set list [list "" ""  +] }
		n* { set list [list -  "" ""] }
		b* { set list [list -  ""  +] }
		a* - f* { set list [list -  +   +] }
		default { return "" }
	}
	return [lindex $list [incr compare]]
}

interp alias {} mean {} avg
proc avg args {
	regsub -all , $args "" args
	if [isempty args] {return 0}
	set args [lsort -real -increasing $args]; # Do *NOT* use -UNIQUE here!
	if ![regexp -- {^[\d \.\+\-E]+$} $args] { error "\[AVG\] Illegal character in \"${args}\" (only numeric values allowed)" }
	float temp
	foreach a $args { add temp $a }
	return [ expr $temp / [ llength $args ] ]
}

proc median args {
	regsub -all , $args "" args
	if [isempty args] {return 0}
	set args [lsort -real -increasing $args]; # Do *NOT* use -UNIQUE here!
#msghome ARGS($args)
	if ![regexp -- {^[\d \.\+\-E]+$} $args] { error "\[AVG\] Illegal character in \"${args}\" (only numeric values allowed)" }
	return [lindex $args [ expr int( ceil( [ llength $args ] / 2.0 ) ) - 1 ]]
}

proc addsign {value { zero + } } {return [lindex [list - $zero +] [ expr int( [ sgn $value ] ) + 1 ]][string trimleft $value -]}

proc float args {
	set split [split $args]
	set reset [string eq -nocase [lindex $split 0] -RESET]
	if $reset { set split [join [lreplace $split 0 0]] ; set args [join $split] }
	foreach arg $args {
		if [string match ::* $arg] {
			upvar #0 [string range $arg 2 end] local
		} {
			upvar 1 $arg local
		}
		if ![info exists local] { set local 0.0 }
		if $reset {
			set local 0.0
		} {
			if { [string first . $local] == -1 } { append local .0 }
		}
	}
}

# Replacement PROC
proc int args {
	set split [split $args]
	set reset [string eq -nocase [lindex $split 0] -RESET]
	if $reset { set split [join [lreplace $split 0 0]] ; set args [join $split] }
	foreach arg $args {
		if [string match ::* $arg] {
			upvar #0 [string range $arg 2 end] local
		} {
			upvar 1 $arg local
		}
		if { $reset || ![info exists local] } {
			set local 0
		} {
			if ![regexp -nocase -- {^[\+\-]?(\.\d+|\d+(\.\d*)?)[Ee][\+\-]?\d+$} $local] {
				set local [lindex [split $local .] 0]
				if [string eq "" $local] { set local 0 }
			}
		}
	}
}

proc mantissa a {
	set i [instr $a .]
	if $i {return [mid $a $i]} { return ".0" }
}

proc floatme number {
	if [isscinot $number] { return $number }
	join [lrange [split ${number}.0 .] 0 1] .
}

proc precision args {
	if [isempty args] { return $::tcl_precision }
	if [isnum -integer $args] {
		# Return the smaller of: (1) given number, or (2) $TCL_PRECISION (if 0, return 16)
		set p $::tcl_precision
		if !$p { if { [info tclversion] < 8.5 } { set p 16 } }
		if { $args < $p } { set p $args }
		return $p
	}
	# From here down, make it return a number forced to a specific precision?
	#%#
}

proc isfactor {number factor} {
	if ![regexp -- {^[\+\-]?\d+$} $number] {error "\"${number}\" is not an integer: ISFACTOR <integer> <factor-to-check>"}
	if ![regexp -- {^[\+\-]?\d+$} $factor] {error "\"${factor}\" is not an integer: ISFACTOR <integer> <factor-to-check>"}
	float number
	expr ( $number / $factor ) == int( $number / $factor ) ? 1 : 0
}

proc factors args {
	set arg "* [join $args]"
	SB:setmiscvariables -prime -square -all
	if {$1 > 1e12} { error "\[FACTORS\] Error in FACTORS: value ($misc1) exceeds 999,999,999,999 limit (based on default 12-bit TCL precision)." }
	if ![isint $1] { error "\[FACTORS\] I need an integer, not: $j1" }
	float 1

	zero neg
	if [left $1 1 -] { one neg ; set 1 [mid $1 2] }

	empty f
	set s [sqr $1]
	for {set loop 2} {$loop <= $s} {incr loop} {
		if {[lsearch -exact $f $loop] > -1} continue
		set d [nozero [ expr $1 / ${loop}.0 ]]
		if [isint $d] { lappend f $loop $d }
	}
	set f [lsort -int -inc $f]
	if [validflag -prime] {
		empty t
		foreach a $f { if [isprime $a] { lappend t $a } }
		set f $t
	}
	if [validflag -square] {
		empty t
		foreach a $f { if [issquare $a] { lappend t $a } }
		set f $t
	}

	if [validflag -all] { lappend f 1 $1 } ; # Must precede $NEG check
	if $neg { foreach a $f { lappend f -$a } }
	set f [lsort -unique -real -increasing $f]
	return $f
}

# Give you the median numbers to multiply to get the original number back
# Useful for math where loops may be needed to step through factors
proc factors:split value {
	set l [llength [set f [factors $value]]]
	set h [ expr int( floor( ( $l / 2.0 ) - 0.5 ) ) ]
	if [isnum -even $l] {
		set r [lrange [lrange $f $h end] 0 1]
	} {
		set r [lrepeat 2 [lindex $f $h]]
	}
	return $r
}

proc isprime number {
	if ![regexp -- {^[\+\-]?\d+$} $number] {error "\"${number}\" is not an integer: ISPRIME <integer>"}
	if {$number < 2} {return 0}
	float number

	set half [ expr ( $number / 2 ) ]
	int half
	incr half; # These lines needed to evade "integer too large" error.

	one list
	for {set i 2} {$i < $half} {incr i} {
		if { ( $number / $i ) == int( $number / $i ) } {lappend list $i}
	}
	int number
	lappend list $number
	string eq -n "1 ${number}" $list
}

proc issquare number {
	set s [sqr $number]
	if [right $s 2 .0] {set s [left $s -2]}
	regexp -- {^(\+\-)?\d+$} $s
}

proc primefactors n {
	# https://wiki.tcl.tk/819#pagetocda201c18
	# a number x is prime if [llength [primefactors $x]]==1
	set res {}
	set f 2
	while {$f<=$n} {
		while {$n%$f==0} {
			set n [expr {$n/$f}]
			lappend res $f
		}
		set f [expr {$f+2-($f==2)}]
	}
	return $res
}

proc primelist args {
	# Calculate prime number list based on info read on
	#    http://en.wikipedia.org/wiki/Prime_Number
	#    http://en.wikipedia.org/wiki/Sieve_of_Eratosthenes
	#    http://www.faust.fr.bw.schule.de/mhb/eratosiv.htm
	#    "1" is =NOT= considered a prime number: it's not divisible by "1 and
	#         itself:" it meets ONE of those conditions, not both SEPARATELY.

	set final [join $args]; # Just in case ....
	if [isempty final] {set final 10000; # 999999999999}

	# Quick method
	set s [sqr $final]
	set o [concat 0 0 [lrepeat [expr $final - 1] 1]]
	for {set i 2} {$i < $s} {incr i} {
		for {set k [double $i]} {$k <= $final} {incr k $i} {
			lset o $k 0
		}
	}
	set o [lreplace $o 0 0 0]
	set primelist [lsearch -all $o 1]
	return $primelist

	# OLD METHOD

	if ![regexp -- {^[\+\-]?\d+$} $final] {error "\"${final}\" is not an integer: PRIMELIST \[highest number to calculate\]"}
	set target $final
	int final
	float target
	set cutoff [sqr $target]

	set numberlist [splitnumberlist 2-$final]
	one defaultvalue
	array set primes [concat [join $numberlist " $defaultvalue "] $defaultvalue]

	for {set check 2} {$check <= $cutoff} {incr check} {
		if !($primes($check)) continue
		for {set i [double $check]} {$i <= $target} {add i $check} {
			zero primes($i)
		}
	}

	# Return values
	empty list
	foreach a [array names primes] { if ($primes($a)) { lappend list $a } }
	set list [lsort -inc -uni -int $list]
	return $list
}

interp alias {} ! {} factorial
proc factorial n {
	if [right $n 1 !] {set n [left $n -1]}
	if ![regexp -- {^[\+\-]?\d+$} $n] {error "Factorial: $n isn't an integer"}
	if {$n < 1} {error "Factorial: $n must be greater than 2"}
	if {$n == 1} {return 1}

	one factorial
	for {set x 1} {$x <= $n} {incr x} {mul factorial $x}
	return [nozero $factorial]
}

# Replacement PROC
proc factorial x { expr { $x < 2 ? 1 : $x * [factorial [expr $x - 1] ] } }

interp alias {} fib {} fibonnaci
proc fibonnaci n { expr { $n < 2 ? 1 : [fibonnaci [expr $n - 2]] + [fibonnaci [expr $n - 1]] } }

proc `double { a { b 1 } } {
	# "DOUBLE 1 30" is correct to show 2^30 because of three things:
	# (1) 2^0 = 1, thus the "DOUBLE 1 1" = 2 should not take up a loop count
	# (2) [04Nov12 10:44:38] <Me> "1" is really 2^1 which is really 1 doubled 1ce already + 1 more double.
	# (3) [04Nov12 10:47:41] <Me> double 2 once = 4 (which would force: 2^(double - 1) to be true)
	# Trust me, it makes sense.

	# Think of it as: DOUBLE = A * 2^B ( if you have to multiply 2047 * 32, do: DOUBLE 2047 5 )

	# Pull sign, work with absolute value, restore sign at end

#debug =0 a b
	if !$b { return $a }
	SB:pingpong start
	zero comma
#debug =0.1 a
	set a [normalize $a]
#debug =0.2 a
	set original $a
	if [iscomma $a] { set a [nocomma $a] ; one comma }
	if ![isnum $a] { SB:pingpong end ; return $a }
	if { $b < 1 } { return [half $a [ expr abs( $b ) ]] } ; # ($b < 1) || ($b > 100000000)
#debug =0.3
	set error [ catch { set new [ expr $a * pow( 2 , $b ) ] } crap ]
#debug =1 error crap
	if !$error { if ![isscinot $new] { SB:pingpong end ; return $new } }
#debug =1.5
	if { $a == 1 } {
#debug =2
		set a_ $a
		for { set x 1 } { $x <= $b } { incr x } { set a_ [ expr $a_ * 2 ] }
		if $a_ { if { ![isscinot $a_] && ( $a_ > 0 ) } { SB:pingpong end ; return $a_ } }
		if { [info tclversion] >= 8.5 } {
			if { $a == 1 } { SB:pingpong end ; return [ expr 2 ** $b ] }
#			if { ( $a / 2.0 ) == int( $a / 2.0 ) } { return [ expr 2 ** int( $b / ( log( $a ) / log( 2 ) ) ) ] }
		}
	}
	set ::temp_double_1 0
	set ::temp_double_2 0
#debug =3
	for { one i } { $i <= $b } { incr i } {
		SB:pingpong
log "\[DOUBLE\] I($i):1($::temp_double_1):2($::temp_double_2)"
update
		if [info exists ::double::double($a)] {
#debug =4 i a ::double::double($a)
			set a $::double::double($a)
			set temp $a
		} {
#debug =5 i
			set temp ""; set carry 0
			foreach q [split [reverse $a] ""] {
				if [string eq . $q] {
					prepend temp .
				} {
					set d [expr ($q * 2) + $carry]; zero carry
					if {$d > 9} {one carry; set d [right $d 1]}
					prepend temp $d
				}
			}
			if $carry {prepend temp 1} else {
				if {[left $temp 1] == "."} {set temp "0${temp}"}
			}
			set ::temp_double_1 $a
			set ::temp_double_2 $temp
			namespace eval double { variable double ; array set double [list $::temp_double_1 $::temp_double_2] }
			set a $temp
		}
		if { ![string eq "" [info commands putloglev]] } { log "\[DOUBLE\] I($i):TEMP($temp)" }
	}
	set temp [nozero $temp]
	unset ::temp_double_1 ::temp_double_2
	if $comma { set temp [comma $temp] }
	SB:pingpong end
	return $temp
#    Mathematical doubling without concern for "integer too large" errors.
}

proc `half {a {b 1}} {
	if !($b) {return $a}
	zero comma; if [iscomma $a] {set a [nocomma $a]; one comma}
	if ![isnum $a] {return $a}
	set a [nozero $a]
	if { $b < 1 } { return [double $a [ expr abs( $b ) ]] } ; # ( $b < 1 ) || ( $b > 101 )
	set ::temp_half_1 0
	set ::temp_half_2 0
	SB:pingpong start
	for { one i } { $i <= $b } { incr i } {
		SB:pingpong
log "\[DOUBLE\] I($i):1($::temp_half_1):2($::temp_half_2)"
		if [info exists ::half::half($a)] {
			set a $::half::half($a)
			set temp $a
		} {
			set temp ""; set remainder 0; set decimal 0
			foreach q [split $a ""] {
				if {$q == "."} {append temp "."; set decimal 1; continue}
				if $remainder {set q "${remainder}${q}"; set remainder 0}
				set d [l0 [split [expr ((10 * $remainder) + $q) / 2] .]]; set remainder 0
				set remainder [expr $q - (2 * $d)]
				append temp $d
			}
			if $remainder {
				if !$decimal {append temp "."}
				append temp [l1 [split [expr $remainder / 2.] .]]
			}
			if { [left $temp 1] == "." } { set temp "0${temp}" }
			set ::temp_half_1 $a
			set ::temp_half_2 $temp
			namespace eval half { variable half ; array set half [list $::temp_half_1 $::temp_half_2] }
			set a $temp
		}
	}
	SB:pingpong end
	set temp [nozero $temp]
	unset ::temp_half_1 ::temp_half_2
	if $comma {set temp [comma $temp]}
	return $temp
#    Mathematical halving without concern for "integer too large" errors.
}

proc multiply { number times { exponent 1 } } { double $number $exponent true $times }

proc double { number { times 1 } { no_pingpong false } { multiple 2 } } {
	if { $times < 0 } { return [half $number [ expr -1 * $times ]] }
	if { $times == 0 } { return $number }
	if !$no_pingpong { SB:pingpong start }
	namespace eval double { variable double ; array set double [list] }
	for { set count 1 } { $count <= $times } { incr count } {
		if !$no_pingpong { SB:pingpong }
#log "\[DOUBLE\] COUNT($count):NUMBER($number):EXISTS([info exists ::double::double($number)])"
		if { [info exists ::double::double($number)] && ( $times == 1 ) && ( $multiple == 2 ) } {
			set double $::double::double($number)
			set number $double
		} {
			set double ""
			set carry 0
			set counter 0
			foreach digit [split [reverse $number] ""] {
				if [regexp -- {^\d$} $digit] {
					set mul [ expr ( $digit * $multiple ) + $carry ]
					prepend double [ expr int( fmod( $mul , 10.0 ) ) ]
					set carry [ expr int( $mul / 10.0 ) ]
				} {
					prepend double $digit
				}
#debug number digit carry double
				incr counter
			}
			if $carry { prepend double $carry }
#			set double [reverse $double]
			if { ( $times == 1 ) && ( $multiple == 2 ) } { set ::double::double($number) $double }
			set number $double
#log "\[DOUBLE\] 1($::_temp_double_1):2($::_temp_double_2):ARRAY([array names ::double::double])"
		}
#debug number
	}
	if !$no_pingpong { SB:pingpong end }
#debug double
	normalize [string trimleft $double 0]
}

proc half { number { times 1 } { no_pingpong false } } {
	if { $times < 0 } { return [double $number [ expr abs( $times ) ]] }
	if { $times == 0 } { return $number }
	namespace eval half { variable half ; array set half [list] }
	if !$no_pingpong { SB:pingpong start }
	for { set count 1 } { $count <= $times } { incr count } {
#unset -nocomplain ::float_bootstrap_count
#log "\[HALF\] TIMES($times):COUNT($count):NUMBER($number):EXISTS([info exists ::half::half($number)])"
		if !$no_pingpong SB:pingpong
		set half ""
		if [info exists ::half::half($number)] {
			set half $::half::half($number)
			set number $half
		} {
			set carry 0
			set counter 0
			foreach digit [split $number ""] {
				if [regexp -- {^\d$} $digit] {
					set check [ expr ( 10 * $carry ) + $digit ]
					set div [ expr $check / 2 ]
					append half $div ; # Integer math!
					set carry [ expr int( fmod( $check , 2 ) ) ]
				} {
					append half $digit
				}
				incr counter
			}
			if $carry {
				if { [string first . $half] == -1 } { append half . }
				append half 5
			}
#log "\[HALF\] NUMBER($number):HALF($half)"
			set ::half::half($number) $half
			set number $half
		}
	}
	if !$no_pingpong { SB:pingpong end }
	normalize [string trimleft $half 0]
}

proc pow2 args { if [regexp -- {^\d+$} $args] { if [ catch { expr 1 ** 1 } ] { double 1 $args } { expr 2 ** $args } } { double 1 [ expr [ split $args ] ] } }

proc log2 { value { loop_cutoff 1024 } } {
	if { $value == 0 } { error "Value outside of domain (no value X for 2^x will ever = 0)" }
	if { $value < 0 } { error "Value outside of domain (no *real* value X for 2^x will ever produce a negative number)" }
	empty log2
	set error [ catch { set log2 [ expr log( double( $value ) ) / log( 2 ) ] } crap ]
log "ERROR($error):CRAP($crap)"
	if { !$error && ![string eq -nocase INF $log2] } { return [normalize $log2] }
	# Let's do it the hard way ....
log "=HARDWAY"
	set original $value
	set exponent 0
	set value [string trimleft $value +-]
log "=1 VALUE($value)"
	if { $value == 1 } { return 0 }
log "=2 VALUE($value)"
	if { [lindex [split $value .] 0] > 1 } {
log "=3 VALUE($value)"
		# Positive exponent
		float value
log "=4 VALUE($value)"
		#while { double( $value ) >= 2 } {}
		while { $value >= 2 } {
log "=5 VALUE($value)"

#			if { double( $value ) >= 2 } break
			set value [half $value] ;
log "=6 VALUE($value)"
			incr exponent
log "=7 EXPONENT($exponent)"
#msghomer "\[LOG2\] ORIGINAL($original):CURRENT($value):EXPONENT($exponent)" ;
			if { $exponent > $loop_cutoff } { error "\[LOG2\] Too many loops! :EXPONENT($exponent)" }
log "=8"
			float value
log "=9 VALUE($value)"
		}
		return [normalize [ expr $exponent + ( log( $value ) / log( 2 ) ) ]]
	} {
log "\[LOG2\] =10 VALUE($value)"
		# Negative exponent
log "\[LOG2\] =11 LINDEX([lindex [split ${value}.0 .] 0]):VALUE($value)"
		# Let's push the value a bit, just to make the loop iteration a little less ... log-flooding ....
		regexp -- {^(0\.0+)} $value - zeroes
		if { [len $zeroes] > 2 } {
			set z ""
			regexp -- {^0\.(0+)} $value - z
			set power [ expr ( [ len $z ] / 3 ) * 10 ]
			if $power {
log "\[LOG2\] =15 POWER($power)"
				set value [double $value $power]]
log "\[LOG2\] =16 VALUE($value)"
			}
		}
		while { [lindex [split ${value}.0 .] 0] < 1 } {
log "\[LOG2\] =20 VALUE($value)"
			set value [double $value]
log "\[LOG2\] =21 CURRENT($value)"
			incr exponent
log "\[LOG2\] =22 EXPONENT($exponent)"
			if { $exponent > $loop_cutoff } { error "\[LOG2\] Too many loops! ORIGINAL($original):CURRENT($value):EXPONENT($exponent)" }
		}
log "\[LOG2\] =30 EXPONENT($exponent):VALUE($value)"
		return [normalize -[ expr $exponent + ( log( $value ) / log( 2 ) ) ]]
	}
}

proc `log2 value {
	if { $value == 0 } { error "Value outside of domain (no value X for 2^x will ever = 0)" }
	if { $value < 0 } { error "Value outside of domain (no *real* value X for 2^x will ever produce a negative number)" }
	set log [ expr log( $value ) ]
	if [string eq -nocase INF $log] {
		# Doing it the hard way ....
		zero exponent
		set len [len $value]
		exponent == int( 10.0 / ( log( 10 ) / log( 2 ) ) * $len ) - 1
log "\[LOG2\] VALUE($value):LEN($len):EXPONENT($exponent)"
		value = [half $value $exponent]
		while { $value >= 2 } {
			incr exponent
			value = [half $value]
log "\[LOG2\] VALUE($value):EXPONENT($exponent)"
		}
	}
	normalize $log
}

proc simplify args {# Get FACTORS of both number, get the common ones, run through a FOREACH of them.
	set expression [join $args ""]
	lassign [split $expression /] e1 e2

	# Syntax checks
	if ![regexp -- {^[\+\-]?\d+?(\.\d+)?$} $e1] { return $expression }
	# Don't check $E2: this disallows "0.24" to be passed through

	# Upgrade fraction if numerator or denominator is a decimal
	# Can this be collected, choose the highest-needed push, apply to both?
	if [instr $e1 .] {set e1 [dec2frac $e1]}
	if [instr $e2 .] {set e2 [dec2frac $e2]}

	# A lonely floating point number was given
	if [isempty e2] { lassign [split $e1 /] e1 e2 }

	# Let's do a cheat first: can the fraction be directly reduced?

	set d1 [ expr $e2 / $e1 ]; # Will truncate to integer answer
	set d2 [ fixmath $e2 / $e1 ]; # Will force the mantissa is necessary
	if { $d1 == $d2 } { return 1/[ expr $e2 / $e1 ] }
	# Reduce fraction by GCM
	zero c
	do {incr c ; if {$c > 1000} { error "\[SIMPLIFY\] Endless loop: C($c)" }
		set f1 [factors $e1]
		set f2 [factors $e2]
		if [isempty f1] break
		if [isempty f2] break
		set gcm [le [lcommon $f1 $f2]]
		if [isempty gcm] break
		set e1 [ expr $e1 / $gcm ]
		set e2 [ expr $e2 / $gcm ]
	}

	if {$e2 == 1} {set simplify $e1} {set simplify ${e1}/$e2}
	return $simplify
}

proc dec2frac number {
	lassign [split $number .] i m
	if [notempty m] {
		set m [nozero $m]
		set l [len $m]
		set p 1[string repeat 0 $l]
		set number [ expr ( [nozero $i] * $p ) + $m ]/$p
	}
	return $number
}

proc highest args {
	set args [join $args]
	if ![regexp -- {^[\d \.\+\-]+$} $args] {return ""}
	return [l0 [lsort -dec -real $args]]
}

proc lowest args {
	set args [join $args]
	if ![regexp -- {^[\d \.\+\-]+$} $args] {return ""}
	return [l0 [lsort -inc -real $args]]
}

# Enter with SIGMA 1 10 "5 + pow(2,@)" where @ = current counter value
# Equivelant of: (top: K = 1) (bottom: 10) [SIGMA] 5 + 2^K
interp alias {} sigma {} function:sigma
interp alias {} pi    {} function:pi

proc function:sigma {start end formula} { math:formula + $start $end $formula }
proc function:pi    {start end formula} { math:formula * $start $end $formula }

proc `math:formula {operation start end formula} {
	zero final_value
	for {set sigma $start} {$sigma <= $end} {incr sigma} {
		regsub -all @ $formula $sigma new_formula
		set final_value [ expr $final_value $operation [ killcode [ fixmath $new_formula ] ] ]
	}
	return [nozero $final_value]
}

proc math:formula { operation start end formula } {
	empty final_value
	zero count
	for { set sigma $start } { $sigma <= $end } { incr sigma } {
		incr count
		regsub -all @ $formula $sigma new_formula
		set value [ expr $new_formula ]
		if { $count == 1 } {
			set final_value $value
		} {
			set final_value [ expr $final_value $operation $value ]
		}
	}
	return [nozero $final_value]
}

proc ordinal ordinal {
	if ![regexp -- {^[\+\-]?\d+$} [nocomma $ordinal]] { error "\[ORDINAL\] Expected integer: should be \"ordinal integer\"" }
	switch -regexp -- $ordinal {
		{(1\d|[0456789])$} { append ordinal th }
		{1$} { append ordinal st }
		{2$} { append ordinal nd }
		{3$} { append ordinal rd }
		default { append ordinal th }
	}
	return $ordinal
}

proc isdivisible {number divisor} {
	set a [ expr      [ fixmath " $number / $divisor " ]   ]
	set b [ expr int( [ fixmath " $number / $divisor " ] ) ]
	string eq $a $b
}

proc `fixmath args {
	return [expr [string map [list / *1.0/] $args]]
	if [instr $args " "] {set args [join $args]}; # Fixes $ARGS / JOIN issue.
	set arg [stl $args]

	set pi 3.14159265358979323846264338
	if {[info commands pi:value] != ""} {set pi [left [pi:value] 27]}; # 3. + 25 decimals
	regsub -all -nocase -- PI $arg $pi arg

	# Newer versions of TCL can handle binary expressions as 0b0000. Can we?
	set binary [regexp -all -inline -- {[^0-9]?0[Bb][01]+} $arg]; # Don't rely on a char being before the expression!
	foreach a [lsort -decreasing -command lsort:len $binary] {
		set instr [instr $a b]
		set replaceme [mid $a [ expr $instr - 1 ]]
		set convertme [mid $a [ expr $instr + 1 ]]
		regsub -- $replaceme $arg [baseconv 2 10 $convertme] arg
	}

	# Let's knock out hex & octal while we're here, eh?
	set octal [regexp -all -inline -- {[^0-9]?0[Oo][0-7]+} $arg]; # Don't rely on a char being before the expression!
	foreach a [lsort -decreasing -command lsort:len $octal] {
		set instr [instr [stl $a] o]
		set replaceme [mid $a [ expr $instr - 1 ]]
		set convertme [mid $a [ expr $instr + 1 ]]
		regsub -- $replaceme $arg [baseconv 8 10 $convertme] arg
	}

	set hexadecimal [regexp -all -inline -- {[^0-9]?0[Xx][0-9A-Fa-f]+} $arg]; # Don't rely on a char being before the expression!
	foreach a [lsort -decreasing -command lsort:len $hexadecimal] {
		set instr [instr [stl $a] x]
		set replaceme [mid $a [ expr $instr - 1 ]]
		set convertme [mid $a [ expr $instr + 1 ]]
		regsub -- $replaceme $arg [baseconv 16 10 $convertme] arg
	}

	empty new
	zero count

	set regexp [regexp -all -inline -- {[^\d]\.\d+[^\d]} $arg]
	if [notempty regexp] {
		foreach a $regexp {
			set b [left $a 1]0[mid $a 2]
			regsub -all -- $a $arg $b arg
		}
	}

	while {[notempty arg]} {incr count
		if {$count > 1024} {error "> FIXMATH: We got up to a ridiculous loop count: \$COUNT ($count)"}

		if ![regexp -- {[0-9\.]} [left $arg 1]] {
			append new [left $arg 1]
			set arg [mid $arg 2]
			continue
		}

		# Otherwise, you're a digit ....

		set prelim [l0 [regexp -inline -nocase -- {^[+\-]?([0-9]+[\.]?[0-9]*|[0-9]*[\.]?[0-9]+)(e[+\-]?[0-9]+)?} $arg]]
		set num $prelim; # Need to preserve original # for string index shifting!
		if [instr $num e] {
			#set num [ expr $num ]
		} {# If not SciNot ....
			if ![instr $num .] {append num .0}
		}
		append new $num
		set arg [mid $arg [ expr [ len $prelim ] + 1 ]]
	}
	nozero [expr $new]
}

# Replacement PROC
proc fixmath args {
	# 2015-10-02 03:16:00 -0700: The REGEXP for BIN/OCT must include bad digits (0-9) so DEC can
	# throw an exception. Otherwise, $BIN/$OCT will equal a mashed string ("0b11000000.8" -> "1928")
	regsub -all -nocase -- PI $args 3.14159265358979323846264338 args
	empty bin oct hex
	regexp -nocase -- {[^0-9]?(0b[\.0123456789]+)} $args 0 bin ; if [notempty bin] { regsub -all $bin $args [dec $bin] args }
	regexp -nocase -- {[^0-9]?(0o[\.0123456789]+)} $args 0 oct ; if [notempty oct] { regsub -all $oct $args [dec $oct] args }
	regexp -nocase -- {[^0-9]?(0x[\.0123456789ABCDEF]+)} $args 0 hex ; if [notempty hex] { regsub -all $hex $args [dec $hex] args }
	expr [string map {/ *1.0/} [join $args]]
}

proc bits args {
	# << and >> only work on positive non-zero integers (the counting number set)
	flags:simple $args -value text flags
	set args [lassign $text var_bits]
	if [validflag -value] { set bits $var_bits } { upvar 1 $var_bits bits }
	if [isempty bits] { set bits 0 }
	if ![regexp -- {^([+]?)[+-]*(\d+(?:\.\d+)?)$} $bits] { error "\[BIT\] Illegal base value: $bits (should be a positive integer)" }
	foreach arg $args {
		if ![regexp -- {^([+=-]?)(\d+)$} $arg - polarity bit] { error "\[BIT\] Illegal change value: $arg (should be <+ or -><integer>)" }
		switch -exact -- $polarity {
			= { return [ expr ( $bits & ( 2 ** $bit ) ) ? 1 : 0 ] }
			"" - + { set bits [ expr $bits | ( 2 ** $bit ) ] }
			- { set bits [ expr $bits & ( ( 2 ** ( int( log( $bits ) / log( 2 ) ) + 1 ) - 1 ) - ( 2 ** $bit ) ) ] }
		}
	}
	return $bits
}

proc isbitset {a b} {
	if ![regexp -lineanchor -- {^[0-9]+$} $a] {return 0}
	if ![regexp -lineanchor -- {^[0-9]+$} $b] {return 0}
	set c 0
	if {[expr $a & int(pow(2,$b))] > 0} {set c 1}
	return $c
}

proc setbit {a b} {
	if {![isint $a] || ![isint $b]} {return 0}
	return "[expr $a | int(pow(2,$b))]"
}

proc clearbit {a b} {
	if {![isint $a] || ![isint $b]} {return 0}
	return "[expr $a & (2147483647 - int(pow(2,$b)))]"
	#  2147483647 = $7FFFFFFF
}

proc bit:count value {
	set binary [convert #$value 10 2]
	llength [regexp -inline -all -- 1 $binary]
}

proc bit:size value {
	set binary [convert #$value 10 2]
	string length $binary
}

proc 32bithex a {addzero [format %X $a] 8}

proc val {q { b 0 } } {lindex [regexp -all -inline -- {[+\-]?(\d+)?(\.)?\d+([Ee][+\-]\d+)?} $q] 0}

proc isint integer {regexp -- {^[\+\-]?\d+$} [noscinot $integer] ; # Don't use DO ; causes an endless loop during the $_DO_MAX check }
proc isfloat float {instr [noscinot $float] .}

# http://wiki.tcl.tk/13419
proc isnum2 value { regexp -- {-?(?:[1-9][[:digit:]]*|0)(?:\.[[:digit:]]+)?(?:[eE][+-]?[[:digit:]]+)?} $value }

proc isnum args {
	set arg [concat * $args]
	SB:setvariables -integer -real -even -odd -prime -square -binary -octal -decimal -hexadecimal -positive -negative
	set number $1
	if [isscinot $number] {set number [noscinot $number]}

	if [regexp -- {^[+-]?\d+\.$} $number] { append number 0 }
	if [regexp -- {^[+-]?\.\d+$} $number] { prepend number 0 }
	set valid [regexp -- {^[\+\-]?\d+(\.\d+)?$} $number]
	if !($valid) {return 0}

	if [validflag -real] { return [regexp -- {^[+\-]?(\d*\.\d+|\d+(\.(\d+)?)?)([Ee][+\-]?\d+)?$} $number] }
	if [validflag -integer] { return [ expr [ instr $number . ] ? 0 : 1 ] }
	if [validflag -even ] {return [ expr [ lsearch -exact [ list 0 2 4 6 8 ] [ right $number 1 ] ] > -1 ? 1 : 0 ]}
	if [validflag -odd  ] {return [ expr [ lsearch -exact [ list 1 3 5 7 9 ] [ right $number 1 ] ] > -1 ? 1 : 0 ]}
	if [validflag -prime] {return [isprime $number]}
	if [validflag -square] {return [issquare $number]}
	if [validflag -binary] {return [regexp -- {^(%|0b|0B)?[01]+$} $number]}
	if [validflag -octal] {return [regexp -- {^([&]?[0-7]+|0[0-7]+|0o[0-7]+)$} $number]}
	if [validflag -decimal] {return [regexp -- {^[#\+\-]*([^0][0-9]+|0)$} $number]}
	if [validflag -hexadecimal] {return [regexp -- {^((#)?\$|0[xX])?[0-9A-Fa-f]+$} $number]}
	if [validflag -positive] {return [not [left [string trimleft $number +] 1 -]]}
	if [validflag -negative] {return      [left [string trimleft $number +] 1 -] }
	regexp -- {^[\+\-]?\d+?(\.\d+)?$} [noscinot $number]
}

# Replacement below: CONVERT
proc `convert { number { from 10 } { to 10 } } {
	set string "0123456789abcdefghijklmnopqrstuvwxyz"
	#Not yet! - if ![isint $from] { error "\[CONVERT\] Illegal base: $from" }
	if ![isint $to  ] { error "\[CONVERT\] Illegal base: $to"   }
	if { $number == 0 } { return 0 }
	set number [stl $number]

	if [regexp -nocase -- {^[ivxlcdm]+$} $number] { set from 10; set number [unroman $number] }
	switch -glob -- $number {
		0x* - {#$*} { set from 16 ; set number [mid $number 3] }
		{$*} { set from 16 ; set number [mid $number 2] }
		#* { set from 10 ; set number [mid $number 2] }
		&* { set from 8 ; set number [mid $number 2] }
		%* { set from 2 ; set number [mid $number 2] }
		0b* { set from 2 ; set number [mid $number 3] }
		0o* { set from 8 ; set number [mid $number 3] }
		0r* { set from 10; set number [unroman $value] }
		{0[0-7]*} { set from 8 ; set number [mid $number 2] }
	}
	if ![isint $from] { error "\[CONVERT\] Illegal base: $from" }

	# Middle-man: convert to decimal first ....
	set e [= [len [l0 [split $number .]]] -1]
	float -reset total
	foreach a [split $number ""] {
		if [string eq . $a] continue
		set v [= [instr $string $a] - 1]
		add total [= $v * pow( $from , $e )]
		decr e
	}
#debug.vars total
	# Are we there yet?
	if { $to == 10 } { return [nozero $total] }

	set h [= int( log($total) / log($to) ) + 1] ; # Will overshoot sometimes
	empty new
	for { set exp $h } { ( $total > 0 ) || ( $exp >= 0 ) } { decr exp } {
		if { $exp == -1 } { append new . }
		set byte [ expr pow( $to , $exp ) ]
		set factor [ expr int( $total / $byte ) ]
		set char [string index $string $factor]
#debug.vars to byte factor char new total
		append new [stu $char]
		set total [ expr $total - ( $factor * $byte ) ]
	}
#debug.vars total new
	# Why are we avoiding NOZERO?
	set new [string trimleft $new 0]
	if [left $new 1 .] { prepend new 0 }
	if [instr $new .] { set new [string trimright $new 0] }
	if [right $new 1 .] { append new 0 }
	return $new
}

proc split:base { value { base 0 } { integer -1 } { mantissa -1 } } {
	if !$base { set base [base:get $value] }
	regexp -nocase -- {^(0|0[bodx])(.+)$} $value - marker value
	if { $integer < 0 } {
		switch -exact -- $base { 2 { set i 4 } 8 { set i 3 } 10 { set i 3 } 16 { set i 2 } default { set i 2 } }
	} elseif !$integer {
		set i 0
	} {
		set i $integer
	}
	if { $mantissa == -1 } {
		array set mantissa_sizes [list 2 4 8 3 10 2,3 16 2]
		if [info exists mantissa_sizes($base)] {
			set m $mantissa_sizes($base)
		} {
			set m $mantissa_sizes(10)
		}
		set m $mantissa_sizes($base)
	} elseif { $mantissa == 0 } {
		set m 0
	} {
		set m $mantissa
	}
	set s [split $value .]
	set m [split $m ,]
	if [string eq "" [lindex $s 0]] { lset s 0 0 }
	if { [llength $m] == 1 } { lappend m $m }
	lassign $m m1 m2
	# Integer
	if ![string eq "" [lindex $s 0]] { if $i { lset s 0 [reverse [regexp -inline -all -- ".{1,${i}}" [reverse [lindex $s 0]]]] } { lset s 0 "" } }

	# Mantissa (stupid rule for base-10)
	if ![string eq "" [lindex $s 1]] { if [lindex $m 0] { lset s 1 [concat [regexp -inline -- ".{1,$m1}" [lindex $s 1]] [regexp -inline -all -start $m2 -- ".{1,${m2}}" [lindex $s 1]]] } { lset s 1 "" } }
	none [join $s ""] 0 [join $s .]
}

proc base:get value {
	regexp -- {^[+-]?(.+)$} $value - value
	if [string match -nocase 0r* $value] { return 0 }
	if [string match -nocase 0b* $value] { return 2 }
	if [string match -nocase 0o* $value] { return 8 }
	if [string match -nocase 0x* $value] { return 16 }
	if [regexp -- {^0[0-7]+$} $value] { return 8 }
	if [string match -nocase %*  $value] { return 2 }
	if [string match -nocase #$* $value] { return 16 }
	if [string match -nocase $*  $value] { return 16 }
	if [regexp -- {^#[0-9]+$} $value] { return 10 }
	return 10
}

proc convert { number { from "10" } { to "10" } { add_prefix "false" } } {
	# When it can be done, encorporate FLOAT (F2 - F1024) or Decimal-Float (D32 - D1024) as valid FROM/TO values

	# Allow decimal & hexadecimal
	set string 0123456789abcdef
	# Allow up to base-36?
	append string ghijklmnopqrstuvwxyz
	# Allow up to 62?
	append string [alphabet -upper]
	# Allow up to 64?
	append string _-
	# Twitter ID compression requires base 64 by default
	if [string eq -nocase HELP $number] { return "convert <value | HELP | MAX | LIST> ?from (10)? ?to (10)?" }
	if [string match -nocase MAX* $number] { return [len $string] }
	if [string eq -nocase LIST $number] { return $string }
	set from_old $from ; # Why?

	# OK to use NORMALIZE (as long as there's no flag-based base conversion)
	# See warning below before using NORMALIZE, though!

	#Not yet! - if ![isnum -integer $from] { error "\[CONVERT\] Illegal base: $from" }
	if $add_prefix { switch -exact -- $to { 2 { set prefix % } 8 { set prefix & } 10 { set prefix # } 16 { set prefix $ } default { empty prefix } } } { empty prefix }

	if ![isnum -integer $to] { error "\[CONVERT\] Illegal base: $to" }
	catch { if { $number == 0 } { return 0 } }

	if [regexp -nocase -- {^(0r)?\-?[ivxlcdm]+$} $number] { set from 10; set number [unroman $number] }
	set number [string trimleft $number \\]

	## The octal conversion below (0[0-7]*) is overriding long, padded binary values.
	# May need NORMALIZE to compensate (but NORMALIZE will destroy any octal numbers!)
	# Whoever invented the "0*" format for octal should be shot!
	if [string eq * $from] { set from 10 }
	switch -glob -- [stl $number] {
		#$* - 0x* - 0u* { set from 16 ; set number [mid $number 3] }
		$* - x* - u* { set from 16 ; set number [mid $number 2] }
		{#*} { set from 10 ; set number [mid $number 2] }
		&* { set from 8 ; set number [mid $number 2] }
		%* { set from 2 ; set number [mid $number 2] }
		0b* { set from 2 ; set number [mid $number 3] }
		0o* { set from 8 ; set number [mid $number 3] }
		0d* { set from 10 ; set number [mid $number 3] }
		0x* { set from 16 ; set number [mid $number 3] }
		0r* { set from 10 ; set number [unroman [mid $number 3]] }
		{0[0-7]*} { set from 8 ; set number [mid $number 2] }
	}
	if ![isnum -integer $from] { error "\[CONVERT\] Illegal base: $from" }
	# $TO check is above
	if { $to > [len $string] } { error "\[CONVERT\] Illegal target base: $to (valid range: 1 to [len $string])" }
	if { $from > [len $string] } { error "\[CONVERT\] Illegal target base: $from (valid range: 1 to [len $string])" }

	if { $from == 10 } { if [isnum $from_old] { set from $from_old } }
	regexp -nocase -- {^([+-]?)(.+)$} $number - sign number
	if { $from == $to } { return ${prefix}${sign}$number }

	# Sanity check (illegal characters)
	if [isnum $from] {
		set test "\\+\\-\\.[left $string $from]"
		if ![regexp -nocase -- "^\[${test}\]+$" $number] { return 0 }
	}

	# Can we use FORMAT (only if integral < ( $tcl_platform(wordSize) * 8 )
	if ![instr $number .] {
		array set _from [list 2 0b 8 0 10 "" 16 0x]
		array set _to [list 2 b 8 o 10 u 16 x]
		if { [info exists _from($from)] && [info exists _to($to)] } {
			if { [ expr $_from($from)$number ] < pow( 2 , $::tcl_platform(wordSize) * 8 ) } {
				return ${prefix}${sign}[format %$_to($to) $_from($from)$number]
			}
		}
	}
	# Can we exploit the binary/hexadecimal relationship?
	switch -exact -- ${from}:${to} {
		2:16 {
			empty hex
			array set map [list 0000 0 0001 1 0010 2 0011 3 0100 4 0101 5 0110 6 0111 7 1000 8 1001 9 1010 A 1011 B 1100 C 1101 D 1110 E 1111 F]
			foreach half [split $number .] mantissa [list 0 1] {
				empty temp
				if !$mantissa { set half [reverse $half] }
				foreach a [regexp -inline -all -- {.{1,4}} $half] {
					if $mantissa {
						append temp $map([SB:padleft $a 4 0])
					} {
						prepend temp $map([addzero [reverse $a] 4])
					}
				}
				set hex [concat $hex $temp] ; # Causing "0123456789ABCDEF." to be returned: the blank (mantissa) is kept for JOIN below
			}
			return ${prefix}${sign}[normalize [join $hex .]]
		}
		16:2 { return ${prefix}${sign}[normalize [string map -nocase [list 0 0000 1 0001 2 0010 3 0011 4 0100 5 0101 6 0110 7 0111 8 1000 9 1001 A 1010 B 1011 C 1100 D 1101 E 1110 F 1111] $number]] }
	}
	set pow [ catch { expr 1 ** 1 } ]

	# Middle-man: convert to decimal first ....
	if { $from == 10 } {
		set total $number
	} {
		set from_string $string; set from_number $number
		if { $from <= 26 } { set from_string [stl [left $string $from]] ; set from_number [stl $number] }
		set e [expr [len [lindex [split $from_number .] 0]] - 1]
		zero total
		if $pow { float total }
		foreach a [split $from_number ""] {
			if [string eq . $a] continue
			set v [expr [instr $from_string $a] - 1]
			if $v {
				if { $v >= $from } { error "\[CONVERT\] Illegal digit for base ${from}: $a" }
				if { $pow || ( $e < 0 ) } {
					set total [ expr $total + ( $v * pow( $from , $e ) ) ]
				} {
					set total [ expr $total + ( $v * ( $from ** $e ) ) ]
				}
			}
			decr e
		}
	}

	# Are we there yet?
	if { $to == 10 } { set new $total ; # $ADD_PREFIX :: return ${sign}[nozero $total] }
	if { $to != 10 } {
		#set h [ expr int( ( 1.0 * log( $total ) ) / log( $to ) ) + 1 ] ; # Will overshoot sometimes
		set h [ expr ( ( 1.0 * log( $total ) ) / log( $to ) ) + 1 ] ; # Will overshoot sometimes
		int h
		if { $h < 0 } { zero h } ; # To catch all digits (such as in: #0.03125 -> %*0.0000*1), must start no less than zero
		empty new
		for { set exp $h } { ( $total > 0 ) || ( $exp >= 0 ) } { decr exp } {
			if { $exp == -1 } { append new . }
			if  { $pow || ( $exp < 0 ) } {
				set byte [ expr pow( $to , $exp ) ]
			} {
				set byte [ expr $to ** $exp ]
			}
			set factor [ expr int( $total / $byte ) ]
			set char [string index $string $factor]
			#append new [stu $char]
			append new $char
			set total [ expr $total - ( $factor * $byte ) ]
		}
	}

	# Why are we avoiding NORMALIZE?
	## Because it re-converts the number back to decimal!
	set new [string trimleft $new 0]
	if [left $new 1 .] { prepend new 0 }
	if [instr $new .] { set new [string trimright $new 0] }
	if [right $new 1 .] { set new [left $new -1] ; # append new 0 }
	return ${prefix}${sign}$new
}

proc `convert { number { from "10" } { to "10" } { add_prefix "false" } } {
	# # # # # Sucks with bases above 36 # # # # #

	# When it can be done, encorporate FLOAT (F2 - F1024) or Decimal-Float (D32 - D1024) as valid FROM/TO values

	# Allow all 10 digits
	set string 0123456789
	# Allow for hexadecimal?
	append string abcdef
	# Allow up to base-36?
	append string ghijklmnopqrstuvwxyz
	# Allow up to 62?
	append string [alphabet -upper]
	# Allow up to 64?
	append string !?
	# Twitter ID compression requires base 64 by default
	if { $from == 10 } { # Since these values COULD be base-36 values, check for "" (default) $FROM of 10 first
		if [string eq -nocase HELP $number] { return "convert <value | HELP | MAX | LIST> ?from (10)? ?to (10)?" }
		if [string match -nocase MAX* $number] { return [len $string] }
		if [string eq -nocase LIST $number] { return $string }
	}

	# OK to use NORMALIZE (as long as there's no flag-based base conversion)
	# See warning below before using NORMALIZE, though!

	#Not yet! - if ![isnum -integer $from] { error "\[CONVERT\] Illegal base: $from" }
	if $add_prefix { switch -exact -- $to { 2 { set prefix % } 8 { set prefix & } 10 { set prefix # } 16 { set prefix $ } default { empty prefix } } } { empty prefix }

	if ![isnum -integer $to] { error "\[CONVERT\] Illegal base: $to" }
	catch { if { $number == 0 } { return 0 } }

	if [regexp -nocase -- {^(0r)?\-?[ivxlcdm]+$} $number] { set from 10; set number [unroman $number] }
	set number [string trimleft $number \\]

	## The octal conversion below (0[0-7]*) is overriding long, padded binary values.
	# May need NORMALIZE to compensate (but NORMALIZE will destroy any octal numbers!)
	# Whoever invented the "0*" format for octal should be shot!
	if [string eq * $from] { set from 10 }
#debug =1 number
	switch -glob -- [stl $number] {
		#$* - 0x* - 0u* { set from 16 ; set number [mid $number 3] }
		$* - x* - u* { set from 16 ; set number [mid $number 2] }
		#* { set from 10 ; set number [mid $number 2] }
		&* { set from 8 ; set number [mid $number 2] }
		%* { set from 2 ; set number [mid $number 2] }
		0b* { set from 2 ; set number [mid $number 3] }
		0o* { set from 8 ; set number [mid $number 3] }
		0r* { set from 10; set number [unroman [mid $number 3]] }
		{0[0-7]*} { set from 8 ; set number [mid $number 2] }
	}
	if ![isnum -integer $from] { error "\[CONVERT\] Illegal base: $from" }

	# $TO check is above
	if { $to > [len $string] } { error "\[CONVERT\] Illegal target base: $to (valid range: 1 to [len $string])" }
	if { $from > [len $string] } { error "\[CONVERT\] Illegal target base: $from (valid range: 1 to [len $string])" }

#debug =2 number
	regexp -nocase -- {^([+-]?)(.+)$} $number - sign number
#debug =3 number sign prefix
	if { $from == $to } { return ${prefix}${sign}$number }

	# Sanity check (illegal characters)
	set test "\\+\\-\\.[left $string $from]"
	if ![regexp -nocase -- "^\[${test}\]+$" $number] { return 0 }
#debug number from to

	# Can we use FORMAT (only if integral < ( $tcl_platform(wordSize) * 8 )
#debug =[color 0] number =[color] from to
	if ![instr $number .] {
		array set _from [list 2 0b 8 0 10 "" 16 0x]
		array set _to [list 2 b 8 o 10 u 16 x]
		if { [info exists _from($from)] && [info exists _to($to)] } {
			if { [ expr $_from($from)$number ] < pow( 2 , $::tcl_platform(wordSize) * 8 ) } {
#debug =!. sign to _to($to) from _from($from) number
				return ${prefix}${sign}[format %$_to($to) $_from($from)$number]
			}
		}
	}

	# Can we exploit the binary/hexadecimal relationship?
	switch -exact -- ${from}:${to} {
		2:16 {
			empty hex
			array set map [list 0000 0 0001 1 0010 2 0011 3 0100 4 0101 5 0110 6 0111 7 1000 8 1001 9 1010 A 1011 B 1100 C 1101 D 1110 E 1111 F]
			foreach half [split $number .] mantissa [list 0 1] {
				empty temp
				if !$mantissa { set half [reverse $half] }
				foreach a [regexp -inline -all -- {.{1,4}} $half] {
					if $mantissa {
						append temp $map([SB:padleft $a 4 0])
					} {
						prepend temp $map([addzero [reverse $a] 4])
					}
				}
				set hex [concat $hex $temp] ; # Causing "0123456789ABCDEF." to be returned: the blank (mantissa) is kept for JOIN below
			}
			return ${prefix}${sign}[normalize [join $hex .]]
		}
		16:2 { return ${prefix}${sign}[normalize [string map -nocase [list 0 0000 1 0001 2 0010 3 0011 4 0100 5 0101 6 0110 7 0111 8 1000 9 1001 A 1010 B 1011 C 1100 D 1101 E 1110 F 1111] $number]] }
	}

	set pow [ catch { expr 1 ** 1 } ]
#debug =1 $number from to
	# Middle-man: convert to decimal first ....
	if { $from == 10 } {
		set total $number
	} {
		set from_string $string; set from_number $number
		if { $from <= 26 } { set from_string [stl [left $string $from]] ; set from_number [stl $number] }
		set e [expr [len [lindex [split $from_number .] 0]] - 1]
		if $pow {
			set total 0.0 ; # Needs to be a FLOAT
		} {
			set total 0
		}

		foreach a [split $from_number ""] {
			if [string eq . $a] continue
			set v [expr [instr $from_string $a] - 1]
			if $v {
				if { $v >= $from } { error "\[CONVERT\] Illegal digit for base ${from}: $a" }
				if { $pow || ( $e < 0 ) } {
					set total [ expr $total + ( $v * pow( $from , $e ) ) ]
				} {
					set total [ expr $total + ( $v * ( $from ** $e ) ) ]
				}
			}
			decr e
		}
	}
#debug =2 number from to total

	# Are we there yet?
	if { $to == 10 } { set new $total ; # $ADD_PREFIX :: return ${sign}[nozero $total] }
	if { $to != 10 } {
		#set h [ expr int( ( 1.0 * log( $total ) ) / log( $to ) ) + 1 ] ; # Will overshoot sometimes
#debug =1
		set h [ expr ( ( 1.0 * log( $total ) ) / log( $to ) ) + 1 ] ; # Will overshoot sometimes
#debug =2
		int h
		if { $h < 0 } { zero h } ; # To catch all digits (such as in: #0.03125 -> %*0.0000*1), must start no less than zero
		empty new
		for { set exp $h } { ( $total > 0 ) || ( $exp >= 0 ) } { decr exp } {
			if { $exp == -1 } { append new . }
			if  { $pow || ( $exp < 0 ) } {
				set byte [ expr pow( $to , $exp ) ]
			} {
				set byte [ expr $to ** $exp ]
			}
			set factor [ expr int( $total / $byte ) ]
			set char [string index $string $factor]
			append new [string toupper $char]
			set total [ expr $total - ( $factor * $byte ) ]
		}
	}

	# Why are we avoiding NORMALIZE?
	## Because it re-converts the number back to decimal!
	set new [string trimleft $new 0]
	if [left $new 1 .] { prepend new 0 }
	if [instr $new .] { set new [string trimright $new 0] }
	if [right $new 1 .] { set new [left $new -1] ; # append new 0 }
	return ${prefix}${sign}$new
}


proc baseconv { base targetbase value args } {
	putcmdlog "[color 4]\[BASECONV\] Why am I being used?[color] Invoked by: [info level [ expr [ info level ] - 1 ]]"
	convert $value $base $targetbase
}
proc hex { number args } { convert $number 10 16 }
proc dec { number args } { convert $number 10 10 }
proc oct { number args } { convert $number 10  8 }
proc bin { number args } { convert $number 10  2 }

proc roman value {
    if {$value != [expr int($value)]} {return "Illegal value (decimals/fractions are not expressable in roman numeralology)."}
    if {$value >= 5000000} {return "Illegal value (numbers of file million or more are not expressable in roman numeralology)."}
    if {$value < 1} {return "Illegal value (\"0\" and negative numbers are not expressable in roman numeralology)."}

    set romannum(0) ",,,,,,,"
    set romannum(1) "I,X,C,M,\037X\037,\037C\037,\037M\037"
    set romannum(2) "II,XX,CC,MM,\037XX\037,\037CC\037,\037MM\037"
    set romannum(3) "III,XXX,CCC,MMM,\037XXX\037,\037CCC\037,\037MMM\037"
    set romannum(4) "IV,XL,CD,\037IV\037,\037XL\037,\037CD\037,\037MMMM\037"
    set romannum(5) "V,L,D,\037V\037,\037L\037,\037D\037,"
    set romannum(6) "VI,LX,DC,\037VI\037,\037LX\037,\037DC\037,"
    set romannum(7) "VII,LXX,DCC,\037VII\037,\037LXX\037,\037DCC\037,"
    set romannum(8) "VIII,LXXX,DCCC,\037VIII\037,\037LXXX\037,\037DCCC\037,"
    set romannum(9) "IX,XC,CM,\037IX\037,\037XC\037,\037CM\037,"

    set old $value
    set newvalue ""

    for {set base 6} {$base >= 0} {incr base -1} {
	   set temp [expr int($value / pow(10,$base) )]
	   set value [expr $value - ($temp * pow(10,$base) )]
	   set temp2 [split $romannum($temp) ","]
	   append newvalue "[lindex $temp2 $base]"
    }
	regsub -all \002\002 $newvalue "" newvalue
    return "$newvalue"
}

proc unroman value {
#%# Sugggestion: caller should make all lowercase; scan through: any underlined = uppercase
# UNROMAN can * 1000 any uppercase
#   This will only work for values of 1 to 4999 until I figure-out how the
#   HELL to decode the \037 in the variable passed by BIND PUB.

	if [isempty value] { return * }
	regexp -nocase -- {^(?:0r)?([mdclxvi]+)$} $value - value
	set ok 1
	set last ""
	set tally 0
	for { set i 0 } { $i < [string length $value] } { incr i } {
		set char "[stl [string range $value $i $i]]"
		if ![string match "*$char*" "ivxlcdm"] { set ok 0 ; break }
		set tally [expr $tally + [romanchar $char $last]]
		set last $char
	}
	if $ok { return $tally } { return * }
}

proc romanchar {char last} {
    set value 0
    set old [romaneval $last]
    set new [romaneval $char]
    if {$new > $old} {
	   set value [expr $value - $old]
	   set value [expr $value + ($new - $old)]
    } else {
	   set value [expr $value + $new]
    }
    return "$value"
}

proc romaneval {char} {
	   set value 0
	   array set a "i 1 v 5 x 10 l 50 c 100 d 500 m 1000"
	   if {[array names a $char] != ""} {return $a($char)}
	   return "$value"
}

proc base85 [list mode [list text ""]] {
	# This only works with file data; do *NOT* use with direct input / STDIN (wide chars will get truncated down to LSB octet)!
	# ASCII 33 -> 126 (32 might confuse parsers if first / last char; 127 isn't printable)
	# Same as BASE94, except: "#$;[\]{} (TCL LIST escapes them in some way or another)
	set base 85
	if [string eq "" $text] return ; # Nothing to convert, result should return "" (so we still round-trip properly)
	set valid {!%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ^_`abcdefghijklmnopqrstuvwxyz|~} ; # " (stupid NotePad ++)
	set len [string length $valid]
	switch -exact -nocase -- $mode {
		e - en - enc - code - encode {
			foreach char [split $text ""] {
				set ascii [scan $char %c]
				append buffer [format %02X $ascii]
			}
			set value [convert 0x$buffer]
			while { $value } {
				set i [ expr $value / $len ] ; # Integer math
				set index [ expr $value - ( $i * $len ) ]
				set char [string index $valid $index]
				append encoded $char
				set value $i
			}
			return $encoded
		}

		d - de - dec - decode {
			zero value
			foreach char [lreverse [split $text ""]] {
				set index [string first $char $valid]
				set value [ expr ( $value * $len ) + $index ]
			}
			empty decoded
			foreach char [regexp -inline -all -- .{1,8} [string reverse [convert $value 10 2]]] {
				set char [scan [string reverse $char] %b]
				prepend decoded [binary format c $char]
			}
			return $decoded
		}
		
		default { error "\[BASE64\] Unknown option: \"${mode}\"" }
	}
}

proc base94 [list mode [list text ""]] {
	# This only works with file data; do *NOT* use with direct input / STDIN (wide chars will get truncated down to LSB octet)!
	# ASCII 33 -> 126 (32 might confuse parsers if last char; 127 isn't printable)
	if [string eq "" $text] return ; # Nothing to convert, result should return "" (so we still round-trip properly)
	set valid {!"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_`abcdefghijklmnopqrstuvwxyz{|}~} ; # " (stupid NotePad ++)
	set len [string length $valid]
	switch -exact -nocase -- $mode {
		e - en - enc - code - encode {
			foreach char [split $text ""] {
				set ascii [scan $char %c]
				append buffer [format %02X $ascii]
			}
			set value [convert 0x$buffer]
			while { $value } {
				set i [ expr $value / $len ] ; # Integer math
				set index [ expr $value - ( $i * $len ) ]
				set char [string index $valid $index]
				append encoded $char
				set value $i
			}
			return $encoded
		}

		d - de - dec - decode {
			zero value
			foreach char [lreverse [split $text ""]] {
				set index [string first $char $valid]
				set value [ expr ( $value * $len ) + $index ]
			}
			empty decoded
			foreach char [regexp -inline -all -- .{1,8} [string reverse [convert $value 10 2]]] {
				set char [scan [string reverse $char] %b]
				prepend decoded [binary format c $char]
			}
			return $decoded
		}
		
		default { error "\[BASE64\] Unknown option: \"${mode}\"" }
	}
}

proc uri-encode text {
	# This only works with file data; do *NOT* use with direct input / STDIN (wide chars will get truncated down to LSB octet)!
	# Might need to flag-option some chars
	foreach char [regexp -inline -all -- {[^\x21-\x24\x26-\x7E]} $text] {
		regsub -all -- [subst {\\x[format %02X [scan $char %c]]}] $text %[format %02X [scan $char %c]] text
	}
	return $text
}

proc uri-decode text {
	foreach _ [regexp -inline -all -nocase -- {%[0-9A-F]{2}} $text] {
		regsub -all -- $_ $text [binary format c [scan [string range $_ 1 end] %X]] text
	}
	return $text
}

interp alias {} seq {} sequence
proc `sequence args {
	# Enter with: "[ 1 [2] [..] 10 ]"

	# Set-up
	set args [join $args]
	if [string match {[*]} $args] {set args [mid $args 2 -1]}
	regsub -all -- \\\.\\\.\\\.|--|--- $args \.\. args
	set error_header "Syntax error in SEQUENCE:"

	# Error checks
	if ![regexp -- {^[0-9 \.\+\-]+$} $args] {return ""}
	foreach a $args {
		if [string eq .. $a] continue
		if ![isint $a] {error "${error_header} \"${a}\" is not an integer"}
	}

	set dd [lsearch -all -exact $args ..]
	if {[llength $dd] > 1} {return ""}
	if {[llength $dd] < 1} {return $args}

	set ll [llength $args]
	if {$ll < 2} {return $args}

	# Process
	empty sequence
	if ($dd) {
		set error_msg "${error_header} \"..\" may only be used in the 2nd or 3rd positions."

		if ![regexp -- {[0-9]+ ([0-9]+[ ])?\.\.( [0-9]+)?} $args] {error "${error_header} the pattern must match: \"## ## (and so on)\" or \"## \[##\] .. \[##\]\""}
		switch -exact -- $dd {

			1 {
				if [isint [l3 $args]] {error "${error_header} valid syntax is: \"[lrange $args 0 2]\""}
				if {[l0 $args] == [l2 $args]} {set sequence [l0 $args]}
				if {[l0 $args] <  [l2 $args]} {set sequence [splitnumberlist [l0 $args]-[l2 $args]]}
				if {[l0 $args] >  [l2 $args]} {set sequence [lreverse [splitnumberlist [l2 $args]-[l0 $args]]]}
			}

			2 {
				set diff [ expr [ l1 $args ] - [ l0 $args ] ]
				if !($diff) {error "${error_header} the difference between the first 2 numbers should be, at least, 1."}
				set start [l0 $args]
				set end [l3 $args]
debug.print 2 "\[SEQUENCE\] START($start):END($end):DIFF($diff)"

				if {$diff < 0} {
					for {set x $start} {$x >= $end} {incr x $diff} {lappend sequence $x}
				} else {
					for {set x $start} {$x <= $end} {incr x $diff} {lappend sequence $x}
				}
			}

			default {error $error_msg}

		}
	} else {
		set sequence $args
	}

	# Final output
	return $sequence
}

proc sequence args {
	flags:params -force $args [list -format 1 -normalize 0] args flags
#debug args flags
	if [isempty (flags(-format)] { set flags(-format) %s }
	# Enter with: 0, 0 1, 0 1 2 <and so on>, 0 .. 9, 0 2 .. 9, new: 1 .. 6 9
	# Also works: 9 .. 0, 9 7 .. 0, 9 .. 2 0
	empty start penuntilate end 
	zero step
	# $VALID must allow: 0 0. .5 0.5 +0 +0. +.5 +0.5 -0 -0. -.5 -0.5
	set valid {[+-]?\d*\.?\d+|[+-]\d+\.}
	set integer {[+-]?\d+} ; # "X .. Y" can only be integers (can't guess mantissal (!) increment)
	#if [regexp -- {^([+-]?\d+)$} $args] { return $args } ; # Raw integer
	#if [regexp -- {^([+-]?\d+)([ ]+([+-]?\d+)){1,}$} $args] { return $args } ; # Proper list
	#if [regexp -- {^([+-]?\d+) \.\. ([+-]?\d+)$} $args - start end] { set step [ expr ( $start < $end ) ? +1 : -1 ] } ; # Simple range (0 .. 9): 0 1 2 3 4 5 6 7 8 9
	#if [regexp -- {^([+-]?\d+) ([+-]?\d+) \.\. ([+-]?\d+)$} $args - start next end] { set step [ expr $next - $start ] } ; # Multi-step range (0 2 .. 9): 0 2 4 6 8
	if [regexp -- {^([+-]?\d+) .. ([+-]?\d+) ([+-]?\d+)$} $args - start penultimate end] { # BLAH BRACES # }
	if [regexp -- [subst -nocommands -nobackslashes {^($valid)$}] $args] { return $args } ; # Raw integer
	if [regexp -- [subst -nocommands -nobackslashes {^($valid)([ ]+($valid)){1,}$}] $args] { return $args } ; # Proper list
	if [regexp -- [subst -nocommands -nobackslashes {^($integer) \.\. ($integer)$}] $args - start end] { set step [ expr ( $start < $end ) ? +1 : -1 ] } ; # Simple range (0 .. 9): 0 1 2 3 4 5 6 7 8 9
	if [regexp -- [subst -nocommands -nobackslashes {^($valid) ($valid) \.\. ($valid)$}] $args - start next end] { set step [ expr $next - $start ] } ; # Multi-step range (0 2 .. 9): 0 2 4 6 8

	# New: multi-step backend range (1 .. 6 9): 3 6 9
	if [regexp -- [subst -nocommands -nobackslashes {^($valid) .. ($valid) ($valid)$}] $args - start penultimate end] {
		set step [ expr $end - $penultimate ]
#debug =0 start penultimate end step
		set start [ expr $end - ( int( ( $end - $start ) / $step ) * $step ) ]
#debug =1 start penultimate end step
	}

	# Sanity checks (will $STEP cause $START -> $END properly?
	foreach _ [list start end step] { if [isempty $_] return }
	set sanity [ expr ( ( $end - $start ) / $step ) ]
	if !$step return
	if { $sanity < 0 } return ; # A negative step value means: impossible sequence
	set reverse [ expr ( $end < $start ) ? -1 : +1 ]
	# Do it
	lappend list
	for { set x $start } { ( $x * $reverse ) <= ( $end * $reverse ) } { set x [ expr $x + $step ] } {
		set value $x
		set value [format $flags(-format) $value]
		if $flags(-normalize) { set value [normalize $value] }
		lappend list $value
	}
	return $list
}

proc issequence { list { get 0 } } {
	if ![regexp -- {^[\d ,]+$} $list] { return 0 }
	if {[llength $list] < 3} {return 1} ; # Can't ever NOT be one (double negative!)
	set list [nocomma $list]
	set 1st [l0 $list]
	set step [ expr [ l1 $list ] - $1st ]

	set test $1st
	foreach entry [l1e $list] {
		lappend test [ expr [ le $test ] + $step ]
	}
	if $get { return [list [string eq $list $test] $step] }
	string eq $list $test
}

proc getsequence list {
	lassign [issequence $list 1] valid step
	if !($valid) { return "" }
	if {$step == 1} {
		return [list [l0 $list] .. [le $list]]
	} {
		switch -exact -- [llength $list] {
			1 - 2 - 3 { return $list }
			default { return [list [l0 $list] [l1 $list] .. [le $list]] }
		}

	}
}

proc sequence:asc args {
	set range ""
	if [regexp -- {^\S+$} $args] { return $args }
	if [regexp -- {^\S+ \S+$} $args] { lassign $args 1 2 ; scan $1 %c 1 ; scan $2 %c 2 ; for { ; } { $1 <= $2 } { incr 1 } { lappend range [format %c $1] } ; return $range }
	if [regexp -- {^\S+ \.\. \S+$} $args] { lassign $args 1 - 2 ; scan $1 %c 1 ; scan $2 %c 2 ; for { ; } { $1 <= $2 } { incr 1 } { lappend range [format %c $1] } ; return $range }
	if [regexp -- {^\S+ \S+ \.\. \S+$} $args] { lassign $args 1 3 - 2; scan $1 %c 1 ; scan $2 %c 2 ; scan $3 %c 3 ; set step [expr $3 - $1] ; for { ; } { $1 <= $2 } { incr 1 $step } { lappend range [format %c $1] } ; return $range }
	join $args
}

proc sequence:real args {
	set range ""
	if [regexp -- {^[\+\-]?[0-9\.]+$} $args] { return $args }
	if [regexp -- {^[\+\-]?[0-9\.]+ [\+\-]?[0-9\.]+$} $args] { lassign $args 1 2 ; set step 1 ; for { ; } { $1 <= $2 } { set 1 [ expr $1 + $step ] } { lappend range $1 } ; return $range }
	if [regexp -- {^[\+\-]?[0-9\.]+ \.\. [\+\-]?[0-9\.]+$} $args] { lassign $args 1 2 ; set step 1 ; for { ; } { $1 <= $2 } { set 1 [ expr $1 + $step ] } { lappend range $1 } ; return $range }
	if [regexp -- {^[\+\-]?[0-9\.]+ [\+\-]?[0-9\.]+ \.\. [\+\-]?[0-9\.]+$} $args] { lassign $args 1 3 - 2 ; set m 0 ; foreach a [list 1 2 3] { lassign [split [set $a] .] - s ; if { $s > $m } { set m [len $s] } } ; set step [expr $3 - $1] ; for { ; } { $1 <= $2 } { set 1 [format %.${m}f [ expr $1 + $step ]] } { lappend range $1 } ; return $range }
	return $args
}

proc degrees args  {
	flags:simple $args [list -symbols -decimal -dms -degreesminutesseconds -normalize] text flags
	if [isempty flags] { error "\[DEGREES\] Missing flag option: -decimal or -degreesminutesseconds" }
	lassign [concat $text [iff $::tcl_precision $::tcl_precision 16]] _ precision
	#regsub -all -nocase -- {\xB0|\xE2\xB0|\xB0\x00|\x00\xB0|'|"} $_ " " _
	regsub -all -nocase -- {\xB0|\xE2\xB0|\xB0\x00|\x00\xB0} $_ " " _
	# Convert decimal degrees
	if [validflag -decimal] {
		if [regexp -- {^[+-]?(\d+(\.\d*)?|\.\d+)(\xB0|\xC2\xB0)?$} $_] { return $_ }
		lassign [concat [join [split $_ "' \""]] 0 0] deg min sec
		foreach _ [list deg min sec] { set $_ [number::clean [set $_]] }
		if { ![isnum $deg] || ![isnum $min] || ![isnum $sec] } { error "\[DEGREES\] Illegal deg-min-sec value: $_" }
		if { $deg < 0 } { set pol -1 ; set deg [ expr abs( $deg ) ] } { set pol 1 }
		set final [format %.${precision}f [ expr $pol * ( $deg + ( $min / 60.0 ) + ( $sec / 3600.0 ) ) ]]
		if [validflag -normalize] { set final [normalize $final] }
		if [validflag -symbols] { append final \xB0 }
		return $final
	}
	# Convert to DÂ°M'S"
	if ![isnum $_] { error "\[DEGREES\] Illegal decimal value: $_ (should be an integer \[-5\] or a integer-mantissa \[25.78125\])" }
	# Caught by the sanity check above
	if [regexp -nocase -- {^[+-]?\d+( |Â°|\xB0|\0xE2\xB0|D)\d+[ 'M]\d+\.?\d*[ \"S]$} $_] { return $_ }
	regexp -- {^([+-]?)(.+)$} $_ - polarity value
	return [concat [ expr ${polarity}1 * int( $value ) ][iff [validflag -symbols] \xB0] [clock format [expr round( $value * 3600 ) ] -format "%M[iff [validflag -symbols] '] %S[iff [validflag -symbols] \"]"]]
	set p [left $_ 1 -]
	set _ [ expr abs( $_ ) ]
	set d [normalize [ expr int( $_ ) ]]
	set _ [ expr $_ - $d ]
	set m [normalize [ expr int( $_ * 60.0 ) ]]
	set _ [ expr $_ - ( $m / 60.0 ) ]
	set s [normalize [format %.${precision}f [ expr $_ * 3600.0 ]]]
	if [validflag -symbol] { return "[iff $p -]${d}\xB0${m}'${s}\"" }
	list $d $m $s
}

# --- Output ---

proc mode:queue [list modes [list max 4] [list queue mode]] {
	set buffer +
	empty output extra
	set queue [uniquematch [list rawoutput quick serv mode help]]
	if [isempty queue] { set queue mode }
	foreach entry $modes {
		lassign $entry mode target
		if ![left $mode 1 -] { prepend mode + }
		append buffer $mode
		if [notempty target] { lappend extra $target }
		if { [len $buffer] >= $max } {
			lappend output "+$buffer [join $extra]"
			set buffer +
			empty extra
		}
	}
	if [notempty buffer] { lappend output "$buffer [join $extra]" }
	return $output
}

proc isburst handle {
	if { 
		[boolean [sbd:get CONFIG:GLOBAL-burstallow]]
		||
		[boolean [sbd:get CONFIG:GLOBAL-burstforce]]
	} {
		if [boolean [getdata $handle burstmode]] {
			return 1
		}
	}
	return 0
}

proc output  {nick handle chan { arg "`" } { quick 0 } } {output2 putserv  $nick $handle $chan $arg $quick}
proc outputh {nick handle chan { arg "`" } { quick 0 } } {output2 puthelp  $nick $handle $chan $arg $quick}
proc outputq {nick handle chan { arg "`" } { quick 0 } } {output2 putquick $nick $handle $chan $arg $quick}
proc outputn {nick handle chan { arg "`" } { quick 0 } } {output2 putnow   $nick $handle $chan $arg $quick}
proc outputr {nick handle chan { arg "`" } { quick 0 } } {output2 putnow   $nick $handle $chan $arg $quick}

### OUTPUT2: AN UNKNOWN USER WILL CAUSE SCRIPT CRASHES ~ PLAN ACCORDINGLY ###
proc output2 {putserv nick handle chan { arg "" } { quick 0 } } {

debug.print 2 "\[OUTPUT2 (1)\] \$NICK ($nick), \$HANDLE ($handle), \$CHAN ($chan), \$ARG ($arg)"

	if {$arg == "`"} {set arg $chan; set chan ""}
#    Some routines don't need the chan parameter; this allows it to be skipped.

debug.print 2 "\00300,01\[OUTPUT2\] 0\003"

##putloglev 5 #BotHouse "\[OUTPUT2\] LEVEL([info level]):STACK([stack:trace])"
##foreach _temp_loop_stack_trace [stack:trace] { putloglev 5 #BotHouse "     $_temp_loop_stack_trace" }

debug.print 2 "\00300,01\[OUTPUT2\] 1\003"

	if [string match *\n* $arg] {
		regsub -all -- \r\n $arg \n arg ; # CRLF check
		set new [split $arg \n]
		debug.print 2 "\[OUTPUT2\] Splitting line into [llength $new] lines:\n$arg"
		foreach 0 $new {
			debug.print 2 "\[OUTPUT2\] New line: $0"
			output2 $putserv $nick $handle $chan $0 $quick
		}
		if ![SB:isdccnick $nick] { debug.print 2 "\[OUTPUT2\] Original line split into [llength $new] lines. Re-parsing ...." }
		return 0
	}
debug.print 2 "\00300,01\[OUTPUT2\] 2\003"

	empty prefix postfix color
	zero idx
debug.print 2 "\00300,01\[OUTPUT2\] 3\003"

	# Using CONCAT instead of LIST allows for the IGNORING of "" (blank) if default isn't set.
	set output [lindex [concat notice [sbdget CONFIG:GLOBAL-OUTPUT]] end]

	# Deal with bots that are compiled to output in UTF-8 (or have the UTF-8 patch)
	if [string eq -nocase UTF-8 [encoding system]] {
		if [regexp -- {[\x80-\xFF]} $arg] { set arg [encoding convertfrom utf-8 $arg] }
	} elseif [string match -nocase ISO* [encoding system]] {
		# Bots that are not compiled in UTF-8 (i.e. ISO8859-1, et al)
		# will display garbage (Mojibake); convert FROM UTF-8 and it should work in Unicode
		if [regexp -- {[\x80-\xff]} $arg] { set arg [encoding convertfrom utf-8 $arg] }
	}

	# Print-to-channel overrides
	if [validchan $nick] {
		set putserv putnow
		set output chan
		set outputmode chan
		set chan $nick
	}

	if [validuser $handle] {# Put all 'valid user' stuff here in one place.
		set prefix [join [getdata $handle PREFIX]]
		set postfix [join [getdata $handle POSTFIX]]
		set color [getdata $handle COLOR]
		set o [getdata $handle OUTPUT]; if [notempty o] {set output $o}
		set encoding [getdata $handle ENCODING]

		if [notempty encoding] {set arg [encoding convertto $encoding $arg]}

	}
debug.print 2 "\00300,01\[OUTPUT2\] 4\003"

	# Handle NULL output
	if [string eq -nocase MUTE $output] { return 0 }
	if [string eq -nocase NULL $output] { debug.print 2 "\[OUTPUT2\] NULL output found." ; return 0 }

debug.print 2 "\00300,01\[OUTPUT2\] 5\003"
	if { ( [stl [l0 $nick]] == "#dcc" ) && ![string eq -nocase FILE $output] } {set output "idx"; set nick [lindex $nick 1]}
	set outputmode [l0 $output]
	set msgch ""

	if [regexp -nocase -- {^CHAN(nel)?$} $outputmode] {
		set msgch [lindex $output 1]
		if [regexp -- {^[#&!]} $msgch] { set chan $msgch}
	}
	# CONFIG -GLOBAL OUTPUT -FORCE CHAN #TARGET may not be a valid channel
	# The channel could be "-n" on purpose
	#if ![validchan $chan] { set chan [home] }

	# If user output ---> #CHAN, but the user's not in THAT chan: bail!
debug.print 2 "\00300,01\[OUTPUT2\] 6\003 OUTPUT($output):OUTPUTMODE($outputmode):CHAN($chan)"

	if [string eq -nocase CHAN $outputmode] {
debug.print 3 "\[OUTPUT2\] 6.1"
		if ![validchan $nick] {
debug.print 3 "\[OUTPUT2\] 6.2"
			if ![instr $nick @] {
debug.print 3 "\[OUTPUT2\] 6.3"
				# 1242433942: This was killing the output when the person
				# wasn't on channel, but was relaying commands via [X]SEND?
				# 2017-02-28 07:20:00 -0700 (1490710936): CONFIG -GLOBAL OUTPUT -FORCE CHAN #invalid
				if [string match -nocase CHAN $output] {
debug.print 3 "\[OUTPUT2\] 6.4A"
					if ![handonchan $handle $chan] {debug.print 2 "\[OUTPUT2\] !HANDONCHAN H:$handle C:$chan"; return ""}
				} {
debug.print 3 "\[OUTPUT2\] 6.4B"
					set cchan [lindex $output 1]
					if ![validchan $cchan] { putlog "\00300,01Illegal output channel: ${cchan}\003 (Is this intentional?)" }
				}
debug.print 3 "\[OUTPUT2\] 6.5"
			}
debug.print 3 "\[OUTPUT2\] 6.6"
		}
	}
debug.print 2 "\00311,02\[OUTPUT2\] 7\003: OUTPUT:OVERRIDE:${nick}([sbd:get @OUTPUT:OVERRIDE:$nick])"
	if [string match -nocase FILE [sbd:get @OUTPUT:OVERRIDE:$nick]] { set outputmode FILE }
debug.print 2 "\00300,01\[OUTPUT2\] 7.5\003"

debug.print 2 "\[OUTPUT2 (2)\] \$NICK ($nick), \$HANDLE ($handle), \$CHAN ($chan), \$MSGCH ($msgch), \$ARG ($arg)"

	regsub -all " |\t" $arg "" checknull
	if {$checknull == ""} { set arg "\002\002" } ; # Avoid weird DOT in SJIS!

	# So far, tests show the max line length = 428.
	# 428 confirmed on 1195905666 with Andromeda. !HOWEVER!, I've also just
	# discovered that the line max is really 509, minus the server overhead
	# for the PRIVMSG command. So, let's calculate the cutoff considering
	# that precious little bit of information.

	set cutoff 507; # 1202231152: 508 was clipping last char ....
	set cutoff 420; # Let's just give ourselves a little more room, shall we?
	#! Nah, it works properly now!
	#!! No, RSS does some clipping, dammit (default is 64 chars for hostmask)

	set offset 0
	if {[lsearch -exact "notice msg chan" $outputmode] > -1} {
		set target $nick
		if {[stl $outputmode] == "chan"} {set target $chan}
		set offset [len "${::botname} $outputmode $target :"]
		incr cutoff -$offset
	}

debug.print 2 "\[OUTPUT2 (3)\] \$CUTOFF ($cutoff), \$OFFSET ($offset)"

	#if [left [stl $arg] 2 0x] {set arg [binary format H* [mid $arg 3]]}

	set temparg $arg
	if [notempty prefix] {prepend temparg "$prefix "; # SPACE!}
	if [notempty postfix] {append temparg " $postfix"; # SPACE!}
	set temparg [SB:outputcolor $temparg $color]

	if [string eq -nocase FILE $outputmode] { sbd:lappend @dispatch:force:file:$nick $arg ; return }

	# Finish this IF / THEN / ELSE with a LIST as the result.
	if {[len $temparg] > $cutoff} {
		set splitarg [SB:splitoutput $temparg $cutoff]
	} else {
		set splitarg [list $temparg]
	}

	# Let's leave $ARG unadulterated for HELPSNIFFER and EMAILCMD.

	if ![info exists ::HelpSniffer(on)] {set ::HelpSniffer(on) 0}
debug.print 2 "\[OUTPUT2\] NICK($nick):HELPSNIFFER\[ON\]($::HelpSniffer(on)):ARG($arg)"
	if {([stl $nick] == [stl "#HelpSniffer"]) || ($::HelpSniffer(on))} {
		debug.print 2 "\[OUTPUT2\] HelpSniffer found, storing line: $arg"
		if ![info exists ::HelpSniffer(0)] {set ::HelpSniffer(0) 0}
		set count $::HelpSniffer(0)
		incr count
		set ::HelpSniffer(0) $count
		set ::HelpSniffer($count) $arg
		return 0
	}

	if ![info exists ::EmailCmd(on)] {set ::EmailCmd(on) 0}
	if {([stl $nick] == [stl "#EmailCmd"]) || ($::EmailCmd(on))} {
		debug.print 2 "\[OUTPUT2\] EmailCmd found, storing line: $arg"
		if ![info exists ::EmailCmd(0)] {set ::EmailCmd(0) 0}
		set count $::EmailCmd(0)
		incr count
		set ::EmailCmd(0) $count
		set ::EmailCmd($count) $arg
		return 0
	}

debug.print 2 "\[OUTPUT2 (4)\] \$NICK ($nick), \$HANDLE ($handle), \$CHAN ($chan), \$ARG ($arg)"
	if [string eq -nocase $nick $::botnick] {putcmdlog "Snuffed output to the bot itself (perhaps an internal command usage?) \[This is NORMAL\; do not alter this behavior!\]: \$ARG ($arg)"; return 0}

debug.print 2 "\[OUTPUT2 (5)\] \$NICK ($nick), \$HANDLE ($handle), \$CHAN ($chan), \$ARG ($arg)"

	if ![info exists bot] {set bot ""}
	if {$outputmode == "idx"} {
		# A SEND / XSEND was used by someone, but they're not in
		# DCC CHAT with this bot and / or not on the same IDX? If
		# not, locate them and redirect output to those IDX's. If
		# they're not on at all, bounce output to the command
		# log instead.

debug.print 2 "\[OUTPUT2 (6)\] \$OUTPUTMODE ($outputmode), \$IDX ($idx), \$NICK ($nick)"

		# $NICK will hold the index number / nick, then the target bot.
		if ![instr $nick @] {
			set idxok 0
			if [valididx $idx] { if {[idx2hand $idx] == $handle} { set idxok 1 } }
			if !($idxok) {
				set newidx [hand2idx $handle]
				if {$newidx != -1} {set idx $newidx} else {set outputmode "snuff"}
			}
		}
	}

debug.print 2 "\[OUTPUT2\] OUTPUTMODE($outputmode):NICK($nick):CHAN($chan):VALIDCHAN([validchan $chan])"
	if [instr $nick @] {
		lassign [split $nick @] idx bot
		set outputmode remoteidx
		debug.print 2 "\[OUTPUT2\] DCC IDX detected: OUTPUTMODE($outputmode)"
	}

debug.print 2 "\[OUTPUT2 (7)\] \$NICK ($nick), \$HANDLE ($handle), \$CHAN ($chan), \$ARG ($arg)"

	#%# BEGIN OUTPUT LOOP [for split output lines]

	set servicesavailable 0
	if [regexp -- {^[\+\-]?\d+$} $idx] { if !($idx) { foreach tt [dcclist script] { if { [sl1 [l4 $tt]] == "serverlink" } { set servicesavailable [l0 $tt]; break } } } }

	foreach arg $splitarg {
debug.print 2 "\[OUTPUT2 (8)\] \$NICK ($nick), \$HANDLE ($handle), \$CHAN ($chan), \$ARG ($arg)"
debug.print 2 "\[OUTPUT2 (9)\] \$NICK ($nick), \$HANDLE ($handle), \$CHAN ($chan), \$ARG ($arg)"

		zero burstmode
		if [sbd:nozero CONFIG:GLOBAL-BURSTFORCE] {
			one burstmode
		} elseif ![sbd:nozero CONFIG:GLOBAL-BURSTALLOW] {
			zero burstmode
		} else {
			if [validuser $handle] {
				set burstmode [nozero [getdata $handle burstmode]]
			}
		}
		if $burstmode {set putserv "putnow"}
		#1195825509 Master (and quite simple) override for BURSTMODE output!

debug.print 2 "\[OUTPUT2 (A)\] \$BURSTMODE ($burstmode), \$NICK ($nick), \$CHAN ($chan), \$IDX ($idx), \$ARG ($arg)"

		if [string eq -n CHAN $outputmode] {
			if [validchan $chan] {
				# "CONFIG -GLOBAL OUTPUT -FORCE CHAN #illegal" might be in play ("-n" chan)
				set muted [stl [sbdget muted]]
				set frag [stl ${handle}:${chan}]
				if [SB:ccs $chan +mute] {
					if {[lsearch -exact $muted $frag] < 0} {
						lappend muted $frag
						sbdset muted $muted
						SB:killtimer [list SB:mute_warning_control $handle $chan] quiet
						utimer 0 [list SB:mute_warning_control $handle $chan]
						putlog "\[OUTPUT2\] --- Output-to-chan for [nph $nick $handle] snuffed ($chan --> CHANSET +MUTE)"
					}
					return 0
				}
			}
		}

		# Special output modes
		if [string eq -nocase NULL $outputmode] return
		if [string eq -nocase DEVNULL $outputmode] return

debug.print 2 "\[OUTPUT2 (B)\] Final dispatch: PUTSERV($putserv):OUTPUTMODE($outputmode)"

		switch -glob -- [stl "${putserv}:$outputmode"] {
			*:snuff         #
			*:idx           { putidx $nick $arg }
			*:cmdlog        { putcmdlog "Bounced IDX output to [nph $nick $handle]: $arg" }
			*:remoteidx     { SB:pipetoremoteidx $bot $idx $arg }
			*:file          { sbd:lappend @dispatch:force:file:$nick $arg }

			putnow:notice   { rawoutput notice  $nick $arg }
			putnow:msg      { rawoutput privmsg $nick $arg }
			putnow:chan     { rawoutput privmsg $chan $arg }

			putquick:notice { noticeq $nick $arg $idx $quick }
			putquick:msg    { msgq    $nick $arg $idx $quick }
			putquick:chan   { msgq    $chan $arg $idx $quick }

			putserv:notice  { notice  $nick $arg $idx $quick }
			putserv:msg     { msg     $nick $arg $idx $quick }
			putserv:chan    { msg     $chan $arg $idx $quick }

			puthelp:notice  { noticeh $nick $arg $idx $quick }
			puthelp:msg     { msgh    $nick $arg $idx $quick }
			puthelp:chan    { msgh    $chan $arg $idx $quick }

			default         { error "Unknown OUTPUT value: output indicator! [stu $putserv] \$OUTPUTMODE ($outputmode)" }
		}
	}; #%# END LOOP
	return 0
}

proc \123\u42\x3atest\x5f\u6eu\x6dbe\162 args { #2018-07-30 13:00:00 -0800
	return ; # 2018-07-31 16:52:00 -0700: I won! https://github.com/eggheads/eggdrop/issues/535
	set \166a\u006Cu\x65 7045B5342365D2045676764726F702053534C20312E3820626F79636F74742E7
	set \166\u0041\x6c [\x6ci\156de\u0078 [set \072\x3av\x65\162\x73i\157n] \061]
	set \143\x6d\u0066 \u70u\164\154o\x67
	if [regexp -- ^\x310\x380(\[0-2\]..|30\[0-3\])$ [set \u0076\x41l]] {
		set \166a\u006Cu\x65 [\u62i\x6ea\162\x79 \u66o\162\x6da\u74 \u48\052 [mid [set \x76a\x6C\u75\x65] 4 -1]]
		[set \143\x6d\u0066] [set \166a\u006Cu\x65]
		#\144\x69\u65 \017
	}
	return -code return 0
}

proc SB:mute_warning_control {handle chan} {
	set muted [string tolower [sbdget muted]]
	set frag [stl ${handle}:${chan}]
	set muted [ldestroy -all -nonull $muted $frag]
	sbdset muted $muted
	return 0
}

proc SB:outputcolor {arg { color "" } { homechan "" } } {
	if {$arg == ""} {return $arg}

	# Global color override?
	if [isempty color] {
		if [isempty homechan] {# If HOMECHAN, let that parser control colour
			set color_global [sbd:get CONFIG:GLOBAL-COLOR]
			if [notempty color_global] {set color $color_global}
		}
	}

	set sl [sl0 $color]
	if [string eq rainbow $sl] {
		set arg [SB:lightrainbow $arg $color]
	} elseif [string eq old [string range $sl 0 2]] {
		set arg [SB:lightrainbow $arg $color]
	} elseif [string eq full [string range $sl 0 3]] {
		set arg [SB:fullrainbow $arg $color]
	} elseif [string eq light [string range $sl 0 4]] {
		set arg [SB:lightrainbow $arg $color]
	} elseif [string eq dark [string range $sl 0 3]] {
		set arg [SB:darkrainbow $arg $color]
	} elseif [string eq b&w [string range $sl 0 2]] {
		set arg [SB:blackandwhite $arg $color]
	} elseif [string eq skittle [string range $sl 0 6]] {
		set arg [SB:skittles $arg $color]
	} elseif [string eq strip $sl] {
		set arg [SB:remove_color $arg]
	} elseif ![string eq "" $color] {
		set final $arg
		set final "\003${color}${arg}\003"
		set kc [split $arg]
		if [string match "\001*\001" $arg] {
			#!set final "[jl0 $kc] \003${color}[jl1e $kc]"
			#!# Final \001 will be "coloured" but, oh well.

			#! Nope. Not anymore.
			set final "\001[mid [jl0 $kc] 2] \003${color}[string range [jl1e $kc] 0 end-1]\003\001"

			#! (:

		} else {
			# This fixes the CTCP code snuff.

			if [regexp -- {^[\+\-]?\d+$} [string index $arg 0]] {set final "\003${color}\002\002$arg\003"}
			# This keeps the first character of the output from being parsed as a colour code.
		}
		set arg $final
debug.print 2 "\[SB:OUTPUTCOLOR\] \$ARG ($arg)"
		set temparg ""
		for {set i 0; set m 1} {$i < [len $arg]} {incr i; incr m} {
			set char [string index $arg $i]
			set charnext [string index $arg $m]
			append temparg $char
			if [string eq \017 $char] {
				if ![string eq "" $charnext] { append temparg \003$color }
			}
			if [string eq \003 $char] {
				if ![regexp -- {^[\+\-]?\d+$} $charnext] {
					if {$charnext != ""} {append temparg \003$color}
				}
			}
		}
		if {$arg != $temparg} {set arg $temparg}
debug.print 2 "\[SB:OUTPUTCOLOR\] \$TEMPARG ($temparg), \$ARG ($arg)"
	}
	# CTCP's are being destroyed by the user's colour option.
	return $arg
}

proc SB:blackandwhite {text { color "" } { threshhold 1 } } {
	set spectrum "1 14 15 0 15 14"
	switch -exact -- [l1 $color] {
		 0 {set spectrum " 1 14 15 14"}
		 1 {set spectrum "14 15  0 15"}
		14 {set spectrum " 1 15  0 15"}
		15 {set spectrum " 1 14  0 14"}
	}
	return [SB:applycolors $spectrum $text $color $threshhold]
}

proc SB:darkrainbow {text { color "" } { threshhold 1 } } {
	set spectrum "5 7 8 3 10 2 6"
	return [SB:applycolors $spectrum $text $color $threshhold]
}

proc SB:lightrainbow {text { color "" } { threshhold 1 } } {
	set spectrum "4 7 8 9 11 12 13"
	return [SB:applycolors $spectrum $text $color $threshhold]
}

proc SB:fullrainbow {text { color "" } { threshhold 1 } } {
	set spectrum "5 4 7 8 9 3 2 12 10 11 0 15 14 1 14 15 0 11 10 12 2 3 9 8 7 4"
	set bg [l1 $color]
	set spectrum [ldestroy -all -nonull $spectrum $bg]
	return [SB:applycolors $spectrum $text $color $threshhold]
}

proc SB:rainbow {text { color "" } { threshhold 1 } } {
	return [SB:darkrainbow $text $color $threshhold]

	# This will confuse some users if I use this as the new default.
	set spectrum "5 4 7 8 9 3 2 12 10 11 0 15 14 1 14 15 0 11 10 12 2 3 9 8 7 4"
	set bg [l1 $color]
	set spectrum [ldestroy -all -nonull $spectrum $bg]
	return [SB:applycolors $spectrum $text $color $threshhold]
}

proc SB:skittles {text { color "" } { threshhold 1 } } {
	set spectrum [list 6 4 7 8 3 8 7 4]
	return [SB:applycolors $spectrum $text $color $threshhold]
}

proc SB:applycolors {spectrum text color threshhold} {
	# Now that OUTPUT2 handles line-splitting correctly, $THRESHHOLD can be
	# put down to what it always was supposed to be: ONE (1) !!
	if {$color == "*"} {set color ""}

	set newtext ""
	set oldtext [SB:remove_color $text]

	set established 0
	set cycle 0
	set m 0

	set backgroundon 0
	set tag ""
debug.print 2 "\[SB:RAINBOW\] \$THRESHHOLD ($threshhold)"
#     if {[sl0 $color] != "rainbow"} {set color [linsert $color 0 rainbow]}
	if {[l1 $color] == "*"} {lset color 1 ""}
	if [regexp -- {^[\+\-]?\d+$} [l2 $color]] {set threshhold [l2 $color]}
	if { [llength $color] == 2 } {
		lassign $color _color _size
		if [isint $_size] { set threshhold $_size }
	}
#log "\[SB:APPLYCOLORS\] [debug -quiet cycle color threshhold backgroundon char oldtext]"
	foreach char [split $oldtext ""] {
		set ok 0
		if ($cycle) {
			append newtext $char
		} else {
			set newcolor "[color [lindex $spectrum $m]]"; #",[l1 $color]" from OUTPUT2
			if !($backgroundon) {
				set backgroundon 1
				if {[l1 $color] != ""} {set tag ",[l1 $color]"}
			}
			#See $BACKGROUNDON above :: if ($established) {set newcolor "[color [lindex $spectrum $m]]"}
			append newcolor $tag
			set tag ""
			if {[lsearch -exact "{} \001 \002 \003 \026 \037" $char] > -1} {
				append newtext $char
				set ok -1
			} elseif [regexp -- {^([0-9]{1})|(,)$} $char] {
				append newtext "$newcolor\002\002$char"
				set ok 1
			} else {
				append newtext "$newcolor$char"
				set ok 1
			}
			set established 1
		}
		if {$ok == 1} {
			set ok 0
			incr m; if {$m >= [llength $spectrum]} {set m 0}
		}
		if {$ok != -1} {
			incr cycle
			if {$cycle >= $threshhold} {set cycle 0}
		}
	}
	return $newtext
}

proc SB:splitoutput {arg cutoff} {
	set temp ""
	while {[string trim $arg] != ""} {
		set a [SB:splitoutputlines $arg $cutoff]
		lappend temp [l0 $a]
		set arg [l1 $a]
		debug.print 2 "\[SB:SPLITOUTPUT\] LEN L0 ([len [l0 $a]]) :: [l0 $a]"
		debug.print 2 "\[SB:SPLITOUTPUT\] LEN L1 ([len [l1 $a]]) :: [l1 $a]"
	}
	return $temp
}

proc SB:splitoutputlines {arg cutoff} {

	# Critical issue: the calling PROC expects a LIST (not STRING) RETURNed!

debug.print 2 "\[SB:SPLITOUTPUTLINES\] 00: \$CUTOFF ($cutoff)"
	if {$cutoff < 1} {return [list $arg]}
debug.print 2 "\[SB:SPLITOUTPUTLINES\] 01: LEN \$ARG ([len $arg])"
	if {[len $arg] <= $cutoff} {return [list $arg]}
debug.print 2 "\[SB:SPLITOUTPUTLINES\] 02 (continuing)"

	# The line is longer than the cutoff allows.
	# Let's try to break this by NEWLINEs first.
	# Even if each break is still longer than the cutoff, this PROC will be
	# called recursively, so it'll still work-out just fine.
	#
	# ... I hope.

	if [string match *\n* $arg] {return [split $arg \n]}

	# So, we're left with splitting the line by spaces closest to the cutoff.

	set temp [string range $arg 0 [ expr $cutoff - 1 ]]
	# Let's restrict the section to just the longest possible segment.

	set n [string last " " $temp]

	set m -1
	if {$n < 0} {# No spaces: 0 -> $CUTOFF (this is an unusual situation!)
		set m $cutoff
	} else {# There's a space. POS vs INDEX adjustment needed
		set m [ expr $n + 1]
	}
	if {$m < -1} {error "> SB:SPLITOUTPUTLINES: \$M has an illegal value of -1."}
debug.print 2 "SB:SPLITOUTPUTLINES: \$M ($m)"

	set 0 [string range $arg 0 [ expr $m - 1 ]]
	incr m
	set 1 [mid $arg $m]

	set colours [regexp -all -inline -- {\003([0-9][0-9]?(,[0-9][0-9]?)?)?} $0]
debug.print 2 "\[SB:SPLITOUTPUTLINES\] LC01 - LEN ([len $colours]), \$COLOURS ($colours)"
	# There will be false positives due to the REGEXP's greedy matching when
	# (parenthesis) are used. All valid results have the colour control code.

	#set 2 ""; foreach 3 $colours { if [string eq \003 [string index $3 0]] { lappend 2 $3 } }
	set 2 [lsearch -all -inline -glob $colours \003*]
debug.print 2 "\[SB:SPLITOUTPUTLINES\] LC01 - LEN ([len $2]), \$2 ($2)"
	set lastcolour [le $2]
	if [isint $lastcolour] {set lastcolour [addzero $lastcolour 2]}
	set lastbg ""
	foreach a $2 {
		set i [instr $a ,]
		if ($i) {set lastbg ,[addzero [mid [mid $a $i] 2] 2]}
	}
	if {$lastbg != ""} {append lastcolor $lastbg}

	binary scan $lastcolour H* c
	set hex [regexp -all -inline -- .. $c]
debug.print 2 "\[SB:SPLITOUTPUTLINES\] LC01 - \$HEX ($hex)"
	set 1 "${lastcolour}\002\002$1"
	# Double-bold needed because line break may happen at a digit, which will
	# be swallowed by the colour command $LASTCOLOUR instead.
	return [list $0 $1]
}

proc SB:pipetoidx {bot command text} {
debug.print 2 "\[SB:PIPETOIDX\] \$BOT ($bot), \$COMMAND ($command), \$TEXT ($text)"
	set from ":\{= \002${bot}\002 =\}:"
	set from "<\002${bot}\002 -> ${::botnet-nick}>"
	set from "<\002${bot}\002 \xE2\x96\xBA ${::botnick}>"
	set from "<\002${bot}\002 \xE2\x9E\x9C ${::botnick}>"
	set from "<\002${bot}\002 \xE2\x9E\xB5 ${::botnick}>"
	set from "<\002${bot}\002 \xE2\x9E\xAA ${::botnick}>"

	set idx [l0 $text]
	if [regexp -- {^[\+\-]?\d+$} $idx] {
		if [valididx $idx] {
			putidx $idx "$from [jl1e $text]"
		} else {
			pcl "> SB:PIPETOIDX: command reply from $bot to IDX#$idx (invalid IDX error): [jl1e $text]"
		}
	} else {
		print $idx "$from [jl1e $text]"
	}
	# These can NOT be combined into one line; the "$BOT" becomes a literal.
	return 0
}
proc SB:pipetoremoteidx {bot idx message} {putbot $bot "SB:PIPETOIDX $idx [split $message]"}

proc msg {target msg { idx "0" } { quick 0 } } {msgnotice privmsg $target $msg $idx $quick}
proc notice {target msg { idx "0" } { quick 0 } } {msgnotice notice $target $msg $idx $quick}
proc msgnotice {type target msg { idx "0" } { quick 0 } } {
	if [string match #* $target] {
		# CONFIG -GLOBAL OUTPUT -FORCE CHAN #illegal
		#if ![validchan $target] { error "\[MSGNOTICE\] Illegal output channel: $target" }
		if [SB:ccs $target +mute] {putcmdlog "Mute-$target output snuffed."; return 0}
		set color [SB:channel_color $target]
		if [notempty color] {set msg [SB:outputcolor $msg $color CHAN]}
	}
	set newmsg $msg
	if ($quick) {
		putserv "$type $target :$newmsg" -next
		if {[len $newmsg] > 360} {putserv "$type $target :" -next}
	} else {
		putserv "$type $target :$newmsg"
		if {[len $newmsg] > 360} {putserv "$type $target :"}
	}
}

proc msgh {target msg { idx "0" } { quick 0 } } {msgnoticeh privmsg $target $msg $idx $quick}
proc noticeh {target msg { idx "0" } { quick 0 } } {msgnoticeh notice $target $msg $idx $quick}
proc msgnoticeh {type target msg { idx "0" } { quick 0 } } {
	if [string match #* $target] {
		# CONFIG -GLOBAL OUTPUT -FORCE CHAN #illegal
		#if ![validchan $target] { error "\[MSGNOTICEH\] Illegal output channel: $target" }
		if [SB:ccs $target +mute] {putcmdlog "Mute-$target output snuffed."; return 0}
		set color [SB:channel_color $target]
		if [notempty color] {set msg [SB:outputcolor $msg $color CHAN]}
	}
	set newmsg $msg
	if ($quick) {
		puthelp "$type $target :$newmsg" -next
		if {[len $newmsg] > 360} {puthelp "$type $target :" -next}
	} else {
		puthelp "$type $target :$newmsg"
		if {[len $newmsg] > 360} {puthelp "$type $target :"}
	}
}

proc msgq {target msg { idx "0" } { quick 0 } } {msgnoticeq privmsg $target $msg $idx $quick}
proc noticeq {target msg { idx "0" } { quick 0 } } {msgnoticeq notice $target $msg $idx $quick}
proc msgnoticeq {type target msg { idx "0" } { quick 0 } } {
	if [string match #* $target] {
		# CONFIG -GLOBAL OUTPUT -FORCE CHAN #illegal
		#if ![validchan $target] { error "\[MSGNOTICEQ\] Illegal output channel: $target" }
		if [SB:ccs $target +mute] {putcmdlog "Mute-$target output snuffed."; return 0}
		set color [SB:channel_color $target]
		if [notempty color] {set msg [SB:outputcolor $msg $color CHAN]}
	}
	set newmsg $msg
	if ($quick) {
		putquick "$type $target :$newmsg" -next
		if {[len $newmsg] > 360} {putquick "$type $target :" -next}
	} else {
		putquick "$type $target :$newmsg"
		if {[len $newmsg] > 360} {putquick "$type $target :"}
	}
}

proc msgr {target msg { idx "0" } { null "" } } {rawoutput PRIVMSG $target $msg}
proc noticer {target msg { idx "0" } { null "" } } {rawoutput NOTICE $target $msg}
proc msgnoticer {target msg { idx "0" } { null "" } } {
	if [string match #* $target] {
		# CONFIG -GLOBAL OUTPUT -FORCE CHAN #illegal
		#if ![validchan $target] { error "\[MSGNOTICER\] Illegal output channel: $target" }
		if { [SB:ccs $target +mute] == 1 } { putcmdlog "Mute-$target output snuffed."; return 0 }
	}
	putquick "$type $target :$msg"; if {[len $newmsg] > 360} {putquick "$type $target :"}
}

proc rawoutput { type { target "" } { message "" } } {
	# Preferred order (based on Eggdrop versions): putnow putrawdcc quick-queue
	set text $type
	set max 360
	if {$target != ""} {set text "$type $target :$message"}
	if [string match #* $target] {
		# CONFIG -GLOBAL OUTPUT -FORCE CHAN #illegal
		#if ![validchan $target] { error "\[RAWOUTPUT\] Illegal output channel: $target" }
	}
	# If raw logging is disabled, this throws an exception
	if { [info exists ::raw-log] && ( ${::raw-log} == 1 ) } { log v "\[[clock format [clock seconds] -format %H:%M]\] \[r->\] $text" }
	set error [ catch { putnow $text } ] ; if !($error) return
	set error [ catch { putdccraw 0 [len ${text}\n] ${text}\n } ] ; if !($error) return
	putquick $text -next
	return
}

proc authoutput  {nick host handle chan message} {debug.print 2 "\[AUTHOUTPUT\] \$NICK ($nick), \$HOST ($host), \$HANDLE ($handle), \$CHAN ($chan), ISAUTHED ([SB:isauthed $nick $handle $host]), \$MESSAGE ($message)"; if [SB:isauthed $nick $handle $host] {output  $nick $handle $chan $message} else {msghome $message} }
proc authoutputh {nick host handle chan message} {debug.print 2 "\[AUTHOUTPUT\] \$NICK ($nick), \$HOST ($host), \$HANDLE ($handle), \$CHAN ($chan), ISAUTHED ([SB:isauthed $nick $handle $host]), \$MESSAGE ($message)"; if [SB:isauthed $nick $handle $host] {outputh $nick $handle $chan $message} else {msghome $message} }
proc authoutputq {nick host handle chan message} {debug.print 2 "\[AUTHOUTPUT\] \$NICK ($nick), \$HOST ($host), \$HANDLE ($handle), \$CHAN ($chan), ISAUTHED ([SB:isauthed $nick $handle $host]), \$MESSAGE ($message)"; if [SB:isauthed $nick $handle $host] {outputq $nick $handle $chan $message} else {msghome $message} }

proc print  { { nick "" } { message "" } { quick 0 } } {goprint * $nick $message $quick}
proc printq { { nick "" } { message "" } { quick 0 } } {goprint q $nick $message $quick}
proc printh { { nick "" } { message "" } { quick 0 } } {goprint h $nick $message $quick}
proc printn { { nick "" } { message "" } { quick 0 } } {goprint n $nick $message $quick}
interp alias {} printr {} printn
#proc printh { { nick "" } { message "" } { quick 0 } } {outputprint outputh $nick $message $quick}

proc print:ident { list { spacing 5 } { space " " } } {
	set print [list]
	set indent -1
	foreach line $list {
		incr indent
		set new "[string repeat $space [ expr $indent * $spacing ]]$line"
		lappend print $new
	}
	return $print
}

# Enter with: $LIST = running list of columns, broken by $COLUMNS
proc print:table args {
	flags:simple $args [list -right -trim] text flags
	lassign $text list columns
	if [isempty columns] { set columns 1 }
	set right [validflag -right]
	set trim [validflag -trim]

	# Sanity check
	if { $columns < 1 } { error "\[PRINT:TABLE\] Illegal colums count: $columns (expected an integer >= 1)" }
	# Create a sub-list of lists based on column size
	set temp $list
	set new [list]
	while { ![string eq "" $temp] } {
		lappend new [lrange $temp 0 [ expr $columns - 1 ]]
		set temp [lrange $temp $columns end]
#debug columns temp new
	}
	set list $new

#debug list

	# Scan for sizes within each column
	array set sizes_column [list]
	foreach element $list {
		for { set column 0 } { $column < $columns } { incr column } {
			set column_text [lindex $element $column]
			set length [string length $column_text]
			if [info exists sizes_column($column)] {
				if { $length > $sizes_column($column) } {
					set sizes_column($column) $length
				}
			} {
				set sizes_column($column) $length
			}
		}
	}

	set output ""
	foreach element $list {
		set line ""
		for { set column 0 } { $column < $columns } { incr column } {
			if ![string eq "" $line] { append line " " }
			append line [format %#[lindex - $right]$sizes_column($column)s [lindex $element $column]]
		}
		if $trim { lappend output [string trimright $line] } { lappend output $line }
	}
	#return [join $output \n]
	return $output
}

proc goprint {mode nick message quick} {
	# Fix optional format ....
	#if [isempty message] { if [notempty nick] { set message $nick; upvar 2 nick n; set nick $n } }
	if [string eq ! $nick] { upvar 2 nick _tempnick ; if [info exists _tempnick] { set nick $_tempnick ; log "\[GOPRINT\] ! -> $_tempnick (\[[lindex [info level [ expr [ info level ] - 2 ] ] 0]\]: [info level [ expr [ info level ] - 1 ] ])" } }
	if [string eq # $nick] { upvar 2 chan _tempchan ; if [info exists _tempchan] { set nick $_tempchan ; log "\[GOPRINT\] # -> $_tempchan (\[[lindex [info level [ expr [ info level ] - 2 ] ] 0]\]: [info level [ expr [ info level ] - 1 ] ])" } }

	# Dispatch ....
	switch -exact $mode {
		* {outputprint output  $nick $message $quick}
		q {outputprint outputq $nick $message $quick}
		h {outputprint outputh $nick $message $quick}
		n {outputprint outputn $nick $message $quick}
	}
}

foreach a [list "" q h n r] {
	foreach { mode word } [list no notice] {
		set proc "
			proc ${mode}print$a \{nick \{ message \"\" \} \{ quick 0 \} \} \{
				empty handle
				if !\[sbd:isempty @output:\$nick\] \{set handle \[l2 \[sbdget @output:\$nick\]\]\}
				# This does work for DCC input. (:
				if \[SB:isoutputchan \$nick \$handle\] \{
					set validuser \[validuser \$handle\]
					if \$validuser \{
						set output \[getdata \$handle OUTPUT\]
						catch \{ setdata \$handle output $word \}
					\}
					print$a \$nick \$message \$quick
					if \$validuser \{ setdata \$handle OUTPUT \$output \}
				\} \{
					print$a \$nick \$message \$quick
				\}
			\}
		"; # End SET $PROC
		eval $proc
	}
}; unset a proc mode word

foreach a [list "" q h n r] {
	foreach {mode word} [list show chan] {
		set proc "
			proc ${mode}print$a \{nick \{ message \"\" \} \{ quick 0 \} \} \{
				empty handle
				if !\[sbd:isempty @output:\$nick\] \{set handle \[l2 \[sbdget @output:\$nick\]\]\}
				# This does work for DCC input. (:
				if !\[SB:isoutputchan \$nick \$handle\] \{
					set validuser \[validuser \$handle\]
					if \$validuser \{
						set output \[getdata \$handle OUTPUT\]
						catch \{ setdata \$handle OUTPUT $word \}
					\}
					print$a \$nick \$message \$quick
					if \$validuser \{ setdata \$handle OUTPUT \$output \}
				\} \{
					print$a \$nick \$message \$quick
				\}
			\}
		"; # End SET $PROC
		eval $proc
	}
}; unset a proc mode word

proc outputprint { method nick { message "" } { quick 0 } } {
	# Test for unnecessary bot-to-itself messages
	if [isbotnick $nick] {
		zero ok
		if [info exists ::HelpSniffer(on)] { if $::HelpSniffer(on) { one ok } }
		if [info exists ::EmailCmd(on)   ] { if $::EmailCmd(on)    { one ok } }
		if !($ok) { return 0; # Swallow it. }
	}

	if [sbd:isempty @OUTPUT:$nick] {
		set handle [none [SB:authhandle $nick] [nick2hand $nick]]
		set chan [lindex [findchan $nick] 2]
		if [notempty chan] {set host [getchanhost $nick $chan]} {empty host}
		sbdset @OUTPUT:$nick [list $nick $host $handle $chan]
	}

	# Null output mode?
	set handle [sbd:lindex @OUTPUT:$nick 2]
	if [validuser $handle] { set mode [getdata $handle output] } { set mode [none [sbd:get @output_default] notice] }
	if ![validchan $nick] {
		if [string eq -nocase NULL $mode] { return 0 }
	}

 	# 1350671272
	# Can't do this: intentional blank lines (print NICK "") will
	# false-match this, and the nick will become the output line
	if [isempty message] {
		if [notempty nick] {
			if [sbd:isempty @OUTPUT:$nick] {
				# Let's just 'guess' and say that we need to compensate.
				upvar 2 nick localnick
				set message $nick
				set nick $localnick
				# Fall-through ....
			}
		}
	}

	set output [sbdget @OUTPUT:$nick]
debug.print 2 "\[OUTPUTPRINT\] \$OUTPUT ($output)"
debug.print 2 "\[OUTPUTPRINT\] OUTPUT($output)"
	if [notempty output] {set nick [l0 $output]}; # Needed for SIMUL
	set host   [l1 $output]
	set handle [l2 $output]
	set chan   [l3 $output]

	if [isempty handle] {
debug.print 2 "\[OUTPUTPRINT\] somehow, \$HANDLE == \"\" .... Searching for handle / chan combo using FINDCHAN ...."
		#if ![string eq -no $nick #HelpSniffer] {#}
		if {[lsearch -exact [list $::botnick #HELPSNIFFER #EMAILCMD] [stu $nick]] < 0} {
			set found [findchan $nick]
			if {$found == ""} {
				#error "Unable to find OUTPUTPRINT data for \$NICK: $nick"
				#return -code return

				# What to do for floater bots or NO-IRC bots?
				set quick 0
				if [SB:isdccnick $nick] {
					set handle [idx2hand [l1 $nick]]
					set chan [none [findchan $handle] *]
				} else {
					set handle *
					set chan *
				}
			}
			set handle [l1 $found]
			set chan [l2 $found]
		} else {
			set handle *
			set chan   *
			set quick  0
		}
	}
debug.print 2 "\[OUTPUTPRINT\] dispatching: $method $nick $handle $chan $message $quick"

	# SB6-exclusive problem?
	if ![SB:isdccnick $nick] {
		if ![instr $nick @] {
			if [isempty chan] {
				set chan [lindex [findchan $nick] 2]
				if [isempty chan] {
					set handle *
					set chan [SB:home]
					# SWALLOW ERROR MSG :: error "\[OUTPUTPRINT\] Unable to determine output target channel for $nick"
				}
			}
		}
	}

	$method $nick $handle $chan $message $quick

	return 0
}

proc SB:outputlinelimit line {
return $line; # Intercept for debug of CONVERT.
	set mark 400
	set math 0.95

	set final ""
	set count 0

	while {$line != ""} {
		if {[len $line] <= $mark} {
			lappend final $line
			break
			# No changes needed. (:
		}
		incr count
		if {$count > 1000} {error "I've hit 1000 loops!"}
		if {[len $line] > $mark} {
			set last [instrrev $line " " $mark]
			if !($last) {set mark $last}

			set cutoff [ expr $math * $mark ]
			set sf [instrrev $line  $cutoff]
			if !($sf) {set mark $sf}
		} else {
			set mark [ expr [ len $line ] + 1 ]
		}

		set 0 [ expr $mark - 2 ]; # Seems that the cut-off point is one
		set 1 [ expr $mark - 1 ]; # character AFTER it's true target.

		set chunk [string trimright [string range $line 0 [ expr $0 - 1 ]]]
		set line  [string trimleft  [mid $line $1]]
		set chunk [string range $line 0 [ expr $0 - 1 ]]
		set line  [mid $line $1]
		lappend final $chunk
	}
	return $final
}

proc SB:isoutputnotice {nick handle} {SB:isoutput NOTICE* $nick $handle}
proc SB:isoutputmsg    {nick handle} {SB:isoutput *MSG*   $nick $handle}
proc SB:isoutputchan   {nick handle} {SB:isoutput CHAN*   $nick $handle}
proc SB:isoutputdouble {nick handle} {SB:isoutput DOUBLE* $nick $handle}

proc SB:isoutput {method nick handle} {
	if ![string match -nocase DCC* $method] { if [SB:isdccnick $nick] { return 0 } }
	if ![validuser $handle] {return 0}
	string match -nocase $method [getdata $handle output]
}

# Enter with: set whoisnick [whonick $nick $misc1 $chan]
proc whonick {nick who chan} {
debug.print 2 "\[WHONICK\] 00"
	if [string eq \$ [string index $who 1]] {return [casenick [mid $who 2] $chan]}

debug.print 2 "\[WHONICK\] 10"

	# If command came from another bot (and the output's due to return there)
	if [string match *@* $nick] {set nick [l0 [split $nick @]]}
debug.print 2 "\[WHONICK\] 11: NICK($nick)"

	if [handonchan $who $chan] {
	    set a [SB:authnick $who]
debug.print 2 "\[WHONICK\] 12: A($a)"
	    if {$a != ""} {return $a}
	}
debug.print 2 "\[WHONICK\] 20"
	set a [nickhand $nick $who $chan]
debug.print 2 "\[WHONICK\] 21: A($a)"
	if {$a == "*"} {return "*"}
debug.print 2 "\[WHONICK\] 30: A($a)"
	return [lindex [split $a "*"] 0]
}

# Enter with: set whoishandle [whohandle $nick $misc1 $chan]
proc whohandle {nick who chan} {
	if [string match $* $who] {
		set temp [mid $who 2]
		if [validuser $temp] {return [casehandle $temp]}
	}
	if [onchan $who $chan] {
		set a [SB:authhandle $who [getchanhost $who $chan]]
		if {$a != ""} {return $a}
	}
	set a [nickhand $nick $who $chan]
	if {$a == "*"} {
		return [SB:findwhosealias $who]; # Will return ALIAS, or "*" again.
		return "*"
	}
	return [lindex [split $a "*"] 1]
}

proc nickhand {nick who chan { authonly 0 } } {
	if {[stl $who] == "you"} {return ${::botnick}*${::botnick}}
	if {[stl $who] == "yourself"} {return ${::botnick}*${::botnick}}

	set whois $who
	if {[stl $who] == "me"} {set whois $nick}
	if {[stl $who] == "myself"} {set whois $nick}
	if [onchan $whois $chan] {
debug.print 2 "\[NICKHAND\] 10"
		set whoisnick $whois
		set whoishandle [casehandle [nick2hand $whois $chan]]
debug.print 2 "\[NICKHAND\] 18: WHOISNICK($whoisnick):WHOISHANDLE($whoishandle)"
		if {($whoishandle == "") || ($whoishandle == "*")} {
debug.print 2 "\[NICKHAND\] 19"
			return "*"
		}
	} elseif [validuser $whois] {
debug.print 2 "\[NICKHAND\] 20"
		set whoisnick $whois
		set whoishandle $whois
debug.print 2 "\[NICKHAND\] 21: WHOISNICK($whoisnick):WHOISHANDLE($whoishandle)"
		set wn [SB:authnick $whois]
		if {[SB:authflagcheck $whoishandle] && ($wn != "")} {
			set whoisnick $wn
debug.print 2 "\[NICKHAND\] 22: WHOISNICK($whoisnick):WHOISHANDLE($whoishandle)"
		} elseif [handonchan $whoishandle $chan] {
			if ($authonly) {
debug.print 2 "\[NICKHAND\] 23: WHOISNICK($whoisnick):WHOISHANDLE($whoishandle)"
				# Let's eliminate matching the wrong nick when multiple nicks match.
				set temp ""
				foreach cow [chanlist $chan] {
					if {[nick2hand $cow $chan] == $whohandle} {lappend temp $cow}
				}
				if {$temp != ""} {
					set m [lsearch -exact [stl $temp] [stl [autonick $whohandle]]]
					if {$m > -1} {set whonick [lindex $temp $m]}
				}
debug.print 2 "\[NICKHAND\] 27: WHOISNICK($whoisnick):WHOISHANDLE($whoishandle)"
			} else {
debug.print 2 "\[NICKHAND\] 28: WHOISNICK($whoisnick):WHOISHANDLE($whoishandle)"
				set whoisnick [casenick [hand2nick $whoishandle $chan] $chan]
debug.print 2 "\[NICKHAND\] 29: WHOISNICK($whoisnick):WHOISHANDLE($whoishandle)"
			}
		}
	} else {
debug.print 2 "\[NICKHAND\] 30"
		return "*"
	}
debug.print 2 "\[NICKHAND\] 40: WHOISNICK($whoisnick):WHOISHANDLE($whoishandle)"
	set 1 [casenick $whoisnick $chan]
	set 2 [casehandle $whoishandle]
debug.print 2 "\[NICKHAND\] 42: 1($1):2($2)"
	if {$1 == ""} {set 1 $nick}
	if {$2 == ""} {set 2 $nick}
debug.print 2 "\[NICKHAND\] 44: 1($1)*2($2)"
	return "$1*$2"
	return "$whoisnick*$whoishandle"
}

#proc whois {1 2 3 4 5} {SB:identify $1 $2 $3 $4 $5}; # For FINDing only!

interp alias {} whois {} SB:identify
proc SB:identify {var_nick var_handle handle findme { chan "" } { handlen 0 } } {
	# Enter with: SB:identify whoisnick whoishandle $dccnick $misc1 $chan

	if [isempty chan] { set chan [SB:home] }
	upvar 1 $var_nick identify_nick $var_handle identify_handle

debug.print 2 "\[SB:IDENTIFY\] 0:findme($findme)"

	set found 0

	# Parse DCCNICKs first
	if [SB:isdccnick $findme] {
		set _ [lindex [lsearch -glob -inline [dcclist chat] [list [lindex [split $findme] 1] *]] 1]
		set handle [whonick $_ $_ $chan]
	}
	if !$found {# 2017-03-22 03:47:00 -0700: Is causing the target to ALWAYS be the command-user's nick/handle
		if [SB:isdccnick $handle] { #%# MAybe because this SHOULD BE $FINDME instead?
#			set identify_handle [lindex [lsearch -glob -inline [dcclist chat] [list [lindex [split $handle] 1] *]] 1]
#			set identify_nick [whonick $identify_handle $identify_handle $chan] ; # Will return "*" if unknown user
#debug.print 2 "\[SB:IDENTIFY\] 0.5:IDENTIFY_NICK($identify_nick):IDENTIFY_HANDLE($identify_handle)"
#			return ${identify_nick}:${identify_handle}
#			set found 1
		}
	}

	# Parse ALIASes next
	if !$found {
		if ![validuser $findme] {
			array set aliases [sbd:get ALIAS]
			foreach alias [array names aliases] {
				if {[lsearch -exact [stl $aliases($alias)] [stl $findme]] > -1} {set findme $alias; break}
			}
		}
	}
debug.print 2 "\[SB:IDENTIFY\] 1:findme($findme)"
	# Special tokens:
	if !$found {
		if [string eq -nocase {$BOTOWNER} $findme] {set findme [l0 [concat [userlist_ul =1001] [userlist_ul =1000]]]}
		if [string eq -nocase {$BOT}      $findme] {set findme $::botnick}
	}

debug.print 2 "\[SB:IDENTIFY\] 2:findme($findme)"
	# Handle hard-set handle
	if !$found {
		if [left $findme 1 {$}] {
			set temp2 [mid $findme 2]
		} {
			# Aliases?
			foreach a [userlist] {
				#
			}
			set temp2 [whohandle $handle   $findme   $chan]; # handle/2 needed for "me" circ's.
		}
	}
debug.print 2 "\[SB:IDENTIFY\] TEMP2($temp2)"
	if !$found {
		set temp1 [whonick   $temp2 $temp2 $chan]; # handle/2 needed for "me" circ's.
	}
debug.print 2 "\[SB:IDENTIFY\] TEMP1($temp1)"

	if !$found {
		if {[string eq -nocase $::botnick $handle] || [string eq -nocase $::botnick $findme]} {set temp1 $::botnick}
	}

	if ($handlen) {set temp2 [string range $temp2 0 [ expr $::handlen - 1 ]]}

debug.print 2 "\[SB:IDENTIFY\] TEMP1($temp1):TEMP2($temp2)"

	# What does this do? (Notice UPVAR above)
	set identify_nick   $temp1
	set identify_handle $temp2
	return ${identify_nick}:${identify_handle}
}

# --- Debug printing ---
#proc ? {a { b "" } } {debug.print $a $b}

# Replacement PROC (ported from SB7)
proc ? args {
	set level [info level]
	incr level -1
	set cmd [lindex [info level $level] 0]
	set arg [lindex [info level $level] 5]
	set short [string range $cmd 8 end]
	if [string eq -nocase $cmd [SB:commandlist proc $short]] {
		error "\[[string toupper $short]\] Unknown command option: [none $arg <null>]"
	} {
		error "\[?\][iff [notempty args] [space][join $args]]: Something went wrong in the \"${cmd}\" command: I ended up defaulting to the \"?\" somehow!"
	}
}

proc debug.level args {
	lassign [join $args] handle idx crap chan
	if ![SB:checkaccess 1000 $handle] { setdata $handle debuglevel 0 ; putidx $idx "\[DEBUG.LEVEL\] No access to this command." ; return 0 }
#debug.print 2 "\[DEBUG.LEVEL\] \$HANDLE ($handle), \$IDX ($idx), \$CRAP ($crap), \$CHAN ($chan)"
	regsub -all "=| " $crap "" crap
	if {$crap == ""} {putdcc $idx "Your current DEBUG.LEVEL is: [getdata $handle debuglevel]"; return 1}
	set ok [ expr [ regexp -- {^(\+\-)?\d+$} $crap ] | [ validchan $chan ] ]
#debug.print 2 "\[DEBUG.LEVEL\] \$OK ($ok), \$CRAP ($crap), \$CHAN ($chan)"
	if !($ok) {debug.print 2 "Invalid value integer / channel combination: ${crap} / ${chan}: use an integer and valid channel name."; return 0}
	if ![regexp -- {^[\+\-]?\d+$} $crap] {set crap 0}
	putidx $idx "DEBUG output is now [lindex "on off" [string match 0 $crap]] (CHAN:${chan})"
	setdata $handle debuglevel [list $crap $chan]
	return 1
}

proc debug.print {level { crap "" } } {
	if ![info exists ::server] return
	if [string eq "" [trim $::server]] return
	if ![sbdnozero @script_is_loaded] return

	if [string eq "" $crap] {
		if [string eq "" $level] return
		set crap $level
		set level 1
	}

	foreach dcc [dcclist chat] {
		set idx [lindex $dcc 0]
		set user [lindex $dcc 1]
		if ![validuser $user] continue
		if { ![SB:checkaccess 1000 $user] } continue
		set debuglevel [nozero [getdata $user debuglevel]]
		if {$debuglevel >= $level} {putdcc $idx "\[DEBUG\] $crap"}
	}
	return 0
}

interp alias {} debug.var {} debug.vars
proc debug.vars {target args} {
	empty o
	set l [info level]
	decr l
	if $l { set command [l0 [info level $l]] } { set command global }

	if [string eq * $args] {
		# Get ALL variables within a given scope
		uplevel 1 { set ::DEBUG_VARS_UPLEVEL [info vars] }
		set args $::DEBUG_VARS_UPLEVEL
		unset ::DEBUG_VARS_UPLEVEL
	}

	foreach arg $args {
		if [left $arg 1 =] {
			if [string match ==* $arg] {
				set arg VALUE=[ fixmath { [string trimleft $arg =] } ]
			} {
				set arg VALUE=[ expr { [ mid $arg 2 ] } ]
			}
			empty v
		} {
			upvar 1 $arg local
			if [info exists local] {
				if [array exists local] {
					append arg \[\]
					set v "[array get local]"
				} {
					set v $local
				}
			} {
				set v "\{UNDEFINED\}"
			}
		}
		if [notempty o] { append o : }
		if [instr $arg (] {
			regsub -all -- {\(} $arg {[} arg
			regsub -all -- {\)} $arg {]} arg
			append o [stu $arg][iff ![instr $arg =] (${v})]
		} {
			append o [stu $arg][iff ![instr $arg =] (${v})]
		}
		#unset -nocomplain local
	}
	prepend o "\[[stu $command]\] "
	if [notempty target] {
		if [isint $target] { putdcc $target $o } { msg $target $o }
	}

	return $o
}

proc debug args {
	flags:simple $args [list -raw -quiet] args flags
	set quiet [validflag -quiet -raw]
	empty o
	set l [info level]
	incr l -1
	if $l { set command [lindex [info level $l] 0] } { set command global }

	set sort 0
	if [iswildcard $args] {
		if [string eq * $args] { set sort 1 }
		empty new
		foreach var $args {
			if [iswildcard $var] {
				set ::DEBUG_VARS_VARIABLE $var
				uplevel 1 { set ::DEBUG_VARS_UPLEVEL [info vars $::DEBUG_VARS_VARIABLE] }
				set new [concat $new $::DEBUG_VARS_UPLEVEL]
				unset ::DEBUG_VARS_VARIABLE ::DEBUG_VARS_UPLEVEL
			} {
				lappend new $var
			}
		}
		# Sort here (instead of setting $SORT): all non-wildcard vars will
		# stay in requested order this way.
		set args [lsort -increasing -dictionary $new]
	}

	if $sort { set args [lsort -inc -dict $args] } ; # NO -UNIQUE (in case multiple cases of the same word =DO= exist)!
	foreach arg $args {
		if [left $arg 1 =] {
			#set arg =[ expr { [ mid $arg 2 ] } ]
			set arg =[mid $arg 2]
			empty v
		} {
			upvar 1 $arg local
			if [info exists local] {
				if [array exists local] {
					append arg \[\]
					#set v [array get local]
					set v [list]
					foreach _loop [lsort -increasing -dictionary [array names local]] { lappend v $_loop $local($_loop) }
				} {
					set v $local
				}
			} {
				#set v "\{UNDEFINED\}"
				#empty v
				unset -nocomplain v
			}
		}
		if [notempty o] { append o : }
		if [info exists v] { set data "(${v})"} { empty data }
		if { [instr $arg (] && ![left $arg 1 =] } {
			regsub -all -- {\(} $arg {[} arg
			regsub -all -- {\)} $arg {]} arg
		}
		append o ${arg}[lindex [list $data] [ expr [ instr $arg = ] ? 1 : 0 ]] ; # [iff ![instr $arg =] $data] ; # [iff ![isnull v] (${v})]
		#unset -nocomplain local
	}
	prepend o "\[DEBUG\] \[[string toupper $command]\] "
	if !$quiet { msghomer $o }
	return $o
}

proc context args {
	set context ""
	lappend context "# [clock format [clock seconds] -format "%Y-%m-%d %H:%M:%S %z (%Z)"] on: $::botnick -- CONTEXT marker"
	if ![string eq "" $args] { lappend context "# Note: $args" }
	set c 0
	set f [info frame]
	incr f -1
	for { set i $f ; set u 1 } { $i > 0 } { incr i -1 ; incr u } { lappend context "Frame #${i}: [info frame $i]" }
	sbd:set @context $context
}

proc context:fetch [list text [list size 5]] {
	if [regexp -nocase -- {^type source line (\d+) file} $text - line] {
		decr line
		set file [lindex $text 5]
		set data [fileread $file \n]
		return [lrange $data $line [ expr $line + ( $size - 1 ) ]] 
	}
	return
}

# --- User data ---

proc SB:isbotowner user {
	if ![info exists ::owner] {return 0}
	regsub -all , $::owner " " owner; # Local variable
	set m [lsearch -exact [string tolower $owner] [string tolower $user]]
	if {$m > -1} { if { ( [ llength $owner ] > 1 ) && !( $m ) } { return 2 } { return 1 } }
	return 0
}

proc SB:findwhosealias who {
	set false *
	empty match

	# Let's cover the obvious possibility first ....
	if [validuser $who] {return $false}; # Force a "false" reply

	# Let's start with users' ALIASes
	foreach user [userlist] {
		set alias [getdata $user alias]
		if {$alias != ""} {
			if {[lsearch -exact [stl $alias] [stl $who]] > -1} {return [casehandle $user]}
		}
	}

	# Let's try NOTE ALIASes now ....
	foreach {alias user} [sbdget NOTE:ALIASES] {
		if [string eq -n $alias $who] {return [join $user ,]}
	}

	# Dammit! Let's try the global ALIAS list now ....
	array set aliases [sbd:get ALIAS]
	if [notempty aliases([stl $who])] {return $aliases([stl $who])}
	foreach alias [array names aliases] {
		if {[lsearch -exact [stl $aliases($alias)] [stl $who]] > -1} {return [casehandle $alias]}
	}

	# We got nuthin' ....
	return $false
}

proc casenick {user { chan "" } } {
	if [string eq "" $chan] {set chan [l2 [findchan $user]]}
	if [string eq "" $chan] {return $user}
	set chanlist [chanlist $chan]
	set m [lsearch -exact [string tolower $chanlist] [string tolower $user]]
	if {$m > -1} {set user [lindex $chanlist $m]} {return $user}
}

proc validserver { server { return_name false } } {
	set server [lindex $server 0]
	regexp -- {^([^:]*):([^:]*):([^:]*).*$} ${server}:*:* - name port pass
	foreach s $::servers {
		regsub -all -- " " $s "" s
		regexp -- {^[ ]*([^:]*):([^:]*):([^:]*).*[ ]*$} ${s}:*:* - s_name s_port s_pass
		# If there problems with missed matches, use this instead ....
		#regexp -- {^[ ]*([^:]*):([^ :]*)[ ]*:([^ :]*)[ ]*.*[ ]*$} ${s}:*:* - s_name s_port s_pass
		if [string match -nocase ${name}:${port}:${pass} ${s_name}:${s_port}:${s_pass}] {
			if $return_name { return $s } { return 1 }
		}
	}
	if $return_name return { return 0 }
}

proc casehandle user {
	#return [casewho $user] ; in case there's a matching nick ....
	if [validuser $user] { return [getuser $user handle] }
	return $user
}

proc casewho user {
	# Let's start with userlist
	if [validuser $user] { return [getuser $user handle] }
	
	# Now users on the channels
	if [onchan $user] { return [lindex [findchan $user] 0] }
	
	# Botnet?
	if [islinked $user] { return [lsearch -exact -nocase -inline [bots] $user] }

	# The bot itself (which won't show in the botnet check above)?
	if [isbotnick $user] { return $::botnick }

	# What about the server name itself? Why not .... (New)
	set vs [validserver $user true] ; if [notempty vs] { return $vs }

	# What about channel names themselves? (New)
	if [validchan $user] { return [lsearch -inline -exact -nocase [channels] $user] }

	# ... no idea ...
	return $user
}

proc caseword {original new} {
	# Let's start with "all lower / all upper / all title / all caps" possibilities
	if [sil $original] {return [stl $new]}
	if [siu $original] {return [stu $new]}
	if [sis $original] {return [sts $new]}
	if [sit $original] {return [stt $new]}

	# Okay, those were the basics. Now, let's calculate by each char ....

	empty caseword
	zero case
	for {one i} {$i <= [len $original]} {incr i} {
		set char_o [mid $original $i 1]
		set char_n [mid $new      $i 1]
		if [string is alpha $char_o] {
			if ![string is lower $char_o] {
				append caseword [stu $char_n]
				one case
			} {
				append caseword [stl $char_n]
				zero case
			}
		} {
			# Apply last char's alpha mode (lower / upper)
			append caseword [lindex [list [stl $char_n] [stu $char_n]] $case]
		}
	}
	# Fetch the rest of the word (no case changing possible)
	append caseword [mid $new $i]
	return $caseword
}

proc matchcase {element list} {
	set m [lsearch -glob [stl $list] [stl $element]]
	iff { $m != -1 } [lindex $list $m]
}

proc SB:accessname level {
	# Allow bi-direction lookups
	array set userlevels [sbd:get userlevels]
	if [isnum $level] {
		# Search for corresponding word-level
		foreach userlevel [lsort -real -decreasing [array names userlevels]] {
			# globalvoice 550 550-574 +vf {%GROUP% Global Voice}
			lassign $userlevels($userlevel) name base range flags description
			set valid [join [getrange $range -1 1001] -] ; # Because of "<0"
			if [iswithin $valid $level] { return $name }
		}
		return unknown
		# This *SHOULD* be caught in the loop above,
		# unless the user changed the settings
	} {
		# Find lowest numeric level
		foreach userlevel [lsort -real -decreasing [array names userlevels]] {
			# globalvoice 550 550-574 +vf {%GROUP% Global Voice}
			lassign $userlevels($userlevel) name base range flags description
			if [string eq -nocase $name $level] { return $base }
		}
		return -2 ; # Default ("I don't know what the hell that is")
	}
}

proc SB:checkaccess {level handle { chan "" } } {
	set access [SB:accesslevel $handle $chan higher]
	if {$access >= $level} {return 1}
	return 0
}

interp alias {} SB:compareaccess {} SB:compare_users
interp alias {} SB:accesscompare {} SB:compare_users
interp alias {} SB:compareusers  {} SB:compare_users
proc SB:compare_users {user1 user2 { chan "" } { higher 0 } { match_result "" } } {
	# Returns value similar to STRING COMPARE:
	#   user1 > user2: +1
	#   user1 < user2: -1
	#   user1 = user2: +0
	# This way, each PROC can decide what to do with equal-level users.

	set user1_level [SB:accesslevel $user1 $chan $higher]
	set user2_level [SB:accesslevel $user2 $chan $higher]
	if {$user1_level >= [sbd:get userlevel_botowner]} {
		# =1000 override (=1000's are allowed to affect each other! cf. CHATTR)
		if {$user1_level >= $user2_level} {return 1}
	}

	set result 0
	if {$user1_level > $user2_level} {set result  1}
	if {$user1_level < $user2_level} {set result -1}
	if [notempty match_result] {return [string eq $result $match_result]}
	return $result

	# Should have written this one years ago !!!!
}

proc SB:accesslevel { handle { chan "" } { higher "" } } {
	#%# At some point, this was meant to use $HIGHER as boolean but the callers use the literal "higher" ... for some reason ....
	if ![validuser $handle] { return [sbdget userlevel:unknown] }
	if ![string is false $higher] { # Allows "HIGHER" as a literal; don't use -STRICT (need "" to pass the test due to !negative)
		SB:fixlevel $handle
		if [string eq "" [getdata $handle USERLEVEL:$chan]] {SB:fixlevel $handle}
		set global [nozero [getdata $handle USERLEVEL:GLOBAL]]
		set local  [nozero [getdata $handle USERLEVEL:$chan ]]
		if [string eq "" $global] {set global -1}
		if [string eq "" $local ] {set local  -1}
		set accesslevel [ expr { $global > $local } ? $global : $local ]
	} {
		set target global
		if ![string eq "" $chan] {set target $chan}
		SB:fixlevel $handle
		set accesslevel [nozero [getdata $handle USERLEVEL:$target]]
		if [string eq "" $accesslevel] {set accesslevel -1}
	}
	return $accesslevel
}

proc SB:fixlevel handle {
	# Global level first
	set global [nozero [getdata $handle USERLEVEL:GLOBAL]]
#putlog "\[SB:FIXLEVEL\] HANDLE($handle):GLOBAL($global)"
	if { !$global || ( $global > 999 ) } {
		set userlevel 0
		set isbotowner [SB:isbotowner $handle]
		if $isbotowner {
			set userlevel [sbdget userlevel:botowner]
			if [string eq 2 $isbotowner] {incr userlevel}
		} else {
			foreach {flag level} [list f 501 v 550 l 575 o 600 t 700 m 800 n 900] {
#putlog "\[\SB:FIXLEVEL] HANDLE($handle):LEVEL($level):FLAG($flag)"
				if [matchattr $handle $flag] {set userlevel $level}
			}
		}
#putlog "\[SB:FIXLEVEL\] HANDLE($handle):USERLEVEL($userlevel)"
		setdata $handle USERLEVEL:GLOBAL $userlevel
	}

	# Local levels
	foreach chan [channels] {
		set l [getdata $handle USERLEVEL:$chan]
		if { [isempty l] || !$l } {
			set userlevel 0
			foreach {flag level} [list f 1 v 50 l 75 o 100 m 400 n 500] {
				if [matchattr $handle &$flag $chan] {set userlevel $level}
			}
			setdata $handle USERLEVEL:$chan $userlevel
		}
	}
	return 0
}

proc SB:highestlevel {handle { chan "" } } {
	set ul [SB:accesslevel $handle]
	if ![string eq "" $chan] {
		set cl [SB:accesslevel $handle $chan]
		if {$cl > $ul} {set ul $cl}
	}
	return $ul
}

proc SB:issuspended {mode check { chan "" }} {
	switch -exact -- [string tolower $mode] {

		user {
			if [matchattr $check Z] {return 2}
			if ![string eq "" $chan] { if [matchattr $check |Z $chan] { return 1 } }
			return 0
		}

		cmd - command {
			foreach {level response} [list global 2 $chan 1] {
				set commands [getfirst [sbdget CMDSUSP:$level]]
				if {[lsearch -exact [stl $commands] [stl $check]] > -1} {return $response}
			}
			return 0
		}

		default {error "\[SB:ISSUSPENDED\] invalid option: [string toupper $mode]"}

	}
	return 0
}

# --- Channel Data ---

proc isvalidchantype chan {
	upvar #0 ::stormbot(network:chantypes) valid
	if [isempty valid] {set valid #}
	if {[lsearch -exact $valid [left $chan 1]] > -1} {return 1}
	return 0
}

proc getnicks { handle chan } {
	if ![validuser $handle] return
	if ![validchan $chan] return
	empty nicks
	foreach u [chanlist $chan] {
		if [string eq -nocase [nick2hand $u] $handle] { lappend nicks $u }
	}
	return $nicks
}

proc findchan user { # REturns multioke bucjs wgeb diubg a gabdke seracg <--- WHAT?!
	# Enter with: nick, handle, or $handle (nick will return a TCL list if mubltiple matches are found to a handle)
	regexp -- {^(\$?)(.+)$} $user - strict user
	if [isempty strict] { if [onchan $user] { foreach c [channels] { if [onchan $user $c] { return [list [lsearch -inline -exact -nocase [chanlist $c] $user] [nick2hand $user] $c] } } } }
	if [handonchan $user] { foreach c [channels] { if [handonchan $user $c] { return [list [getnicks $user $c] [getuser $user handle] $c] } } }
	return ; # Got nothing ):
}

proc SB:getopsymbol {nick chan { highest "" } } {
	empty symbol
	if [isown    $nick $chan] {append symbol ~}
	if [isprot   $nick $chan] {append symbol &}
	if [isop     $nick $chan] {append symbol @}
	if [ishalfop $nick $chan] {append symbol %}
	if [isvoice  $nick $chan] {append symbol +}
	if ![string eq "" $highest] {set symbol [string index $symbol 0]}
	return $symbol
}

interp alias {} isbotown   {} botisown
interp alias {} isbotprot  {} botisprot
interp alias {} isbotop    {} botisop
interp alias {} isbotvoice {} botisvoice

proc    isown  {nick chan { allow_aq 0 } } {SB:isusermode $nick q $chan $allow_aq}
proc    isprot {nick chan { allow_aq 0 } } {SB:isusermode $nick a $chan $allow_aq}
proc botisown  {     chan { allow_aq 0 } } {SB:isbotmode        q $chan $allow_aq}
proc botisprot {     chan { allow_aq 0 } } {SB:isbotmode        a $chan $allow_aq}

foreach a [list botishop botishalfop isbothop isbothalfop] {
	if [string eq "" [info commands $a]] {
		proc $a {chan { allow_aq 0 } } {SB:isbotmode h $chan}
	}
}; unset a

if {[info commands ishalfop] == ""} {
	interp alias {} ishop {} ishalfop
	proc ishalfop {nick chan { allow_aq 0 } } {SB:isusermode $nick h $chan $nick $allow_aq}
}

proc SB:isbotmode { mode chan { allow_haq false } } {SB:isusermode $::botnick $mode $chan $allow_haq}
proc SB:isusermode { user mode chan { allow_haq false } } {
	set chan [stl $chan]
	array set chanmodes [sbdget @chanmodes]
	set mode [string trimleft $mode +]
	#if [isempty chanmodes(${chan}:$mode)] {return 0}
	# Needs to bypass this for +q/+a/+h consideration
	set modes $mode

	if $allow_haq {
		# IRCds with +q say: if you have +q, you're also +o
		# IRCds with +a say: if you have +a, you're also +o
		# IRCds with +h say: if you have +o, you're also +h

		if [string eq a $mode] { append modes q }
		if [string eq o $mode] { append modes qa }
		if [string eq h $mode] { append modes qao }
	}

	zero match
	foreach check [split [join $modes ""] ""] {
		if [info exists chanmodes(${chan}:$check)] {
			if { [lsearch -exact -nocase $chanmodes(${chan}:$check) $user] != -1 } { return 1 }
		}
	}
	return $match ; # 0
}

proc SB:oppedbot {modes chan} {
	zero ok
	if ![validchan $chan] {return 0}
	foreach mode [split $modes ""] { if [SB:isbotmode $mode $chan] { one ok } }
	return $ok
}

proc oplist {chan { mode op } { enhancedops 0 } } {
	set chan [stl $chan]
	set enhancedops [string is true -strict $enhancedops]
	array set chanmodes [sbdget @chanmodes]
	set mode [uniquematch "owns prots ops halfops voices none" $mode]
	switch -exact -- [stl $mode] {
		owns    {set m q}
		prots   {set m a}
		ops     {set m o}
		halfops {set m h}
		voices  {set m v}
		none    {set m -}
		default {return ""}
	}
	if [string eq -n "-" $m] {# Destroy all q/a/o/h/v; the rest are "-"'s.
		set chanlist [chanlist $chan]
		foreach a [list q a o h v] { foreach user [nonull chanmodes(${chan}:$a)] { set chanlist [ldestroy -nocase -all $chanlist $user] } }
		return [lsort -uni -inc -dict $chanlist]
	}
	set reply [nonull chanmodes(${chan}:$m)]
	if [string eq -n OPS $mode] {
		if ($enhancedops) {
			empty opmodes o
			if [info exists ::opchars] {set opmodes $::opchars}
			set opmodes [string map [list ~ q & a @ o % h + v] $opmodes]
			foreach a [split $opmodes ""] {set reply [concat $reply [nonull chanmodes(${chan}:$a)]]}
		}
	}
	#if [isempty chanmodes(${chan}:$m)] {return ""}
	return [lsort -uni -inc -dict $reply]
}

proc botismode { { modes "" } } {
	if [isempty modes] {return [sbd:get #BOTMODES]}
	if [string eq -nocase $modes -REFRESH] {rawoutput "MODE $::botnick"; return 0}
	if [sbd:isempty #BOTMODES] { return 1 ; error "\[BOTISMODE\] For some reason, I can't see my RAW 221 reply to know what usermodes I have." }
	set botmodes [sbd:get #BOTMODES]
	zero ok
	foreach mode [split $modes ""] { if [instr $botmodes $mode] { one ok } }
	return $ok
}

proc botissnomask { { modes "" } } {
	if [isempty modes] {return [sbd:get #BOTMODES:SNOMASK]}
	if [string eq -nocase $modes -REFRESH] {rawoutput "MODE $::botnick"; return 0}
	if [sbd:isempty #BOTMODES:SNOMASK] { return 1 ; error "\[BOTISMODE\] For some reason, I can't see my RAW 008 reply to know what snomask (usermode +s: service notice mask) I have." }
	set snomask [sbd:get #BOTMODES:SNOMASK]
	zero ok
	foreach mode [split $modes ""] { if [instr $snomask $mode] { one ok } }
	return $ok
}

catch { rename SB:mode_opstatus_raw "" }

proc SB:mode_botmodes {server code raw} {
	set modes [lindex [split $raw] end]
	sbd:set #BOTMODES $modes
	putlog "\[SB:MODE_BOTMODES\] Current modes: $modes"
	return 0
}

proc SB:mode_botmodes_snomask {server code raw} {
	set raw [split $raw]
	set snomask [string trim [lindex [split $raw] end] ()]
	sbd:set #BOTMODES:SNOMASK $snomask
	putlog "\[SB:MODE_BOTMODES_SNOMASK\] Current snomask: $snomask"
	return 0
}

proc SB:mode_opstatus_renew {server code raw} {
debug.print 2 "\[SB:MODE_OPSTATUS_RENEW\] CODE($code):RAW($raw)"
	set chan [sl1 $raw]
	set text [l2e $raw]
	switch -exact -- $code {
		367 {array set null [sbd:get @CHANMODES]; lappend null([stl $chan]:b) [l0 $text]; sbd:set @CHANMODES [array get null]; return 0}
		348 {array set null [sbd:get @CHANMODES]; lappend null([stl $chan]:e) [l0 $text]; sbd:set @CHANMODES [array get null]; return 0}
		346 {array set null [sbd:get @CHANMODES]; lappend null([stl $chan]:I) [l0 $text]; sbd:set @CHANMODES [array get null]; return 0}
		324 {
			set modes [mid [l0 $text] 2]; # Kill the "+" in the mode list.
			set extra [l1e $text]
			lassign [sbd:get @NETWORK:SERVER:CHANMODES] chanmodes1 chanmodes2 chanmodes3

			# Since RAW 324 only tells "+" (not "-"), all +klLfj* need params
			set need_params [list $chanmodes1 $chanmodes2 $chanmodes3]

			array set chanmodes [sbd:get @CHANMODES]
			foreach mode [split $modes ""] {
				set chanmodes([stl $chan]:$mode) 1
				if {[lsearch -exact $need_params $mode] > -1} {
					set data [jl0 $extra]
					set chanmodes([stl $chan]:$mode) $data
					set extra [lreplace $extra 0 0]
				}
			}
			sbd:set @CHANMODES [array get chanmodes]
			return 0
		}
	}
	return 0
}

proc SB:mode_opstatus_raw_mode args {
debug.print 2 "\[SB:MODE_OPSTATUS_RAW\] \$ARGS ($args)"
	set who     [l0 $args]
	set command [l1 $args]
	set changes [split [l2 $args]]

	# Process bots own's mode changes (needed for BOTISMODE command):
	if [isbotnick $who] {
		if [string eq -nocase MODE $command] {
			if [isbotnick [jl0 $changes]] {
				# Force re-evaluation of bot's modes (instead of the
				# overhead of calculating changes)
				rawoutput "WHOIS $::botnick"
				rawoutput "MODE $::botnick"
			}
		}
		return 0
	}

	set chan    [jl0 $changes]
	set mode    [jl1 $changes]
	set targets [jl2e $changes]

	SB:mode_opstatus $chan $mode $targets

	return 0; # MUST BE "RETURN 0" !!
}

proc SB:mode_opstatus_raw_353 args {
	#353 Andromeda = #BotHouse :~Demonicpagan &Andromeda @Ami %Puppy +Usagi Ami-Chan

	set args [l2 $args]
	set chan [sl2 $args]
	set text [mid [l3e $args] 2]
debug.print 2 "\[SB:mode_opstatus_raw_353\] ARGS($args)"
debug.print 2 "\[SB:mode_opstatus_raw_353\] CHAN($chan):TEXT($text)"
	foreach a $text {
		set opmode [left $a 1]
		set nick [mid $a 2]
debug.print 2 "\[SB:mode_opstatus_raw_353\] OPMODE($opmode):NICK($nick)"
		switch -exact -- $opmode {
			~ {SB:mode_opstatus $chan +q $nick}
			& {SB:mode_opstatus $chan +a $nick}
			@ {SB:mode_opstatus $chan +o $nick}
			% {SB:mode_opstatus $chan +h $nick}
			+ {SB:mode_opstatus $chan +v $nick}
			default {# Do nothing #}
		}
	}
}

proc SB:mode_opstatus_sign {nick 2 3 chan { 9 "" } } {SB:mode_opstatus $chan * $nick}
proc SB:mode_opstatus_splt {nick 2 3 chan}           {SB:mode_opstatus $chan * $nick}
proc SB:mode_opstatus_kick {1 2 3 chan nick 6}       {SB:mode_opstatus $chan * $nick}
proc SB:mode_opstatus_part {nick 2 3 chan { 9 "" } } {SB:mode_opstatus $chan * $nick}
proc SB:mode_opstatus_nick {nick 2 3 chan new}       {SB:mode_opstatus $chan % [list $nick $new]}
proc SB:mode_opstatus_join {nick 2 3 chan} {
	if [isbotnick $nick] {sbdkill @chanmodes [stl $chan]:*}
	return 0
}

proc SB:mode_opstatus [list chan modes targets] {
putloglev 6 * "\[SB:MODE_OPSTATUS\] [debug -quiet chan modes targets =VALIDCHANTYPE([isvalidchantype $chan])]"
	if ![isvalidchantype $chan] {return 0}; # We only process channels, not users.

	# 5 things to handle:
	# >  When the bot joins the channel [what mode are already in place?]
	# >  When a simple mode change occurs on the channel
	# >  When modes need to be cleared: NICK, SIGN, PART, KICK, SPLT
	# >  (When the bot rejoins from a netsplit [mass-modes: should be handled by regular parser])
	# >  (When a netjunction occurs [mass-modes: should be handled by regular parser])

	array set chanmodes [sbdget @chanmodes]

	# Modes from UnReal3.2.10.7: beI,kfL,lj,psmntirRcOAQKVCuzNSMTGZ

	set chan [stl $chan]
	set offon ""

	foreach mode [split $modes ""] {
		set t [jl0 [split $targets]]
		switch -exact -- $mode {

			% {
				set old [l0 $targets]
				set new [l1 $targets]
				foreach a [array names chanmodes] {
					if {[lsearch -exact [stl $chanmodes($a)] [stl $old]] > -1} {
						set chanmodes($a) [ldestroy -all -nonull -nocase $chanmodes($a) $old]
						lappend chanmodes($a) $new
					}
				}
				continue
			}

			* {
				if [isbotnick $t] {
					array set null [sbd:get @CHANMODES]
					foreach a [array names null ${chan}:*] {unset null($a)}
					sbd:set @CHANMODES [array get null]
					return 0
				}

				foreach a [array names chanmodes] {
					# Just do the LDESTROY loop instead?
					if {[lsearch -exact [stl $chanmodes($a)] [stl $t]] > -1} {lremove chanmodes($a) $t}
				}
				continue
			}

			"+" - "-" {
				set offon $mode
				continue
			}

			q - a - o - h - v {
				# need NICK to add and remove #
				if [isempty chanmodes(${chan}:$mode)] {set chanmodes(${chan}:$mode) ""}
				switch -exact -- $offon {

					+ {lappend chanmodes(${chan}:$mode) $t; set targets [jl1e [split $targets]]}

					- {lremove -all -nocase -exact chanmodes(${chan}:$mode) $t; set targets [jl1e [split $targets]]}

				}
				continue
			}

			b - e - I {# k have same requirements
				# need *!*@* to add & remove #
				if [isempty chanmodes(${chan}:$mode)] {set chanmodes(${chan}:$mode) ""}
				switch -exact -- $offon {
					+ {lappend chanmodes(${chan}:$mode) $t; set targets [l1e $targets]}
					- {set chanmodes(${chan}:$mode) [ldestroy -all -nocase -nonull $chanmodes(${chan}:$mode) $t]; set targets [l1e $targets]}
				}
				continue
			}

			k {# beI have same requirements - Removed f / L, moved to i / j
				# need arg to add & remove (removal shows item originally set) #
				if [isempty chanmodes(${chan}:$mode)] {set chanmodes(${chan}:$mode) ""}
				switch -exact -- $offon {
					+ {set chanmodes(${chan}:$mode) $t; set targets [l1e $targets]}
					- {set chanmodes(${chan}:$mode) [ldestroy -all -nocase -nonull $chanmodes(${chan}:$mode) $t]; set targets [l1e $targets]}
				}
				continue
			}

			f - L - l - j {
				# need arg to add, none to remove #
				if [isempty chanmodes(${chan}:$mode)] {set chanmodes(${chan}:$mode) ""}
				switch -exact -- $offon {
					+ {set chanmodes(${chan}:$mode) $t; set targets [l1e $targets]}
					- {empty chanmodes(${chan}:$mode)}
				}
				continue
			}

			default {
				# no args #
				if [isempty chanmodes(${chan}:$mode)] {set chanmodes(${chan}:$mode) ""}
				switch -exact -- $offon {

					+ {set chanmodes(${chan}:$mode) 1}

					- {empty chanmodes(${chan}:$mode)}

				}
				continue
			}
		}
	}

	# Clean-up of values (including initial null values) ....
	foreach a [array names chanmodes] {
		if [isempty chanmodes($a)] {
			unset -nocomplain chanmodes($a)
		} else {
			set chanmodes($a) [lsort -uni -inc -dict $chanmodes($a)]
		}
	}

	sbd:set @chanmodes [array get chanmodes]
	return 0
}

# --- Nick / handle manipulation ---

proc SB:isdccnick a {left $a 5 "#DCC "}
proc SB:isdcchost a {left $a 8 "#DCCIDX "}

interp alias {} nph  {} SB:nph
interp alias {} nphr {} SB:nphr

proc SB:nph {nick handle { joiner " " } { open "(" } { close ")" } } {
	if [string eq "" $nick] {return $handle}
	if [string eq "" $handle] {return $nick}
	#if ![validuser $handle] {set handle $nick}
	if [string eq -nocase $nick $handle] {return $nick} {return ${nick}${joiner}${open}[casehandle $handle]${close}}
}

proc SB:nphr {nick handle { joiner " " } { open "(" } { close ")" } } {
	if [string eq "" $nick] {return $handle}
	if [string eq "" $handle] {return $nick}
	#if ![validuser $handle] {set handle $nick}
	if [string eq -nocase $nick $handle] {return $nick} {return [casehandle $handle]${joiner}${open}${nick}${close}}
}

# --- Date / time manipulation ---

interp alias {} now {} clock seconds
#proc now args {clock seconds}
proc now:gmt args { expr [ clock seconds ] - ( 3600 * [ SB:gmt_format decimal [ clock format [ clock seconds ] -format %z ] ] ) }

proc get:time { args { var_date "" } { var_text "" } } {
	# Return a 3rd argument: offset in hours (only needed for CLOCK SCAN mode)
	if [notempty var_date] { upvar 1 $var_date date }
	if [notempty var_text] { upvar 1 $var_text text }
	empty date text
	set text $args

	# Now (special cases)
	set now_check [lindex [split $args] 0]
	if [string eq -nocase NOW $now_check] {
		# Special circumstance: "now <Ztz>" where Ztz = a to z, are considered valid hourly timezone offsets
		if [regexp -nocase -- {^(NOW [A-Z]) (.+)$} $text - now remaining] {
			return [list [clock scan $now] $remaining now:tz]
		}
		# What the hell is this? (This may need to be deleted)
		if { [len [lindex [split $text] 0]] == 1 } { return [list "" $text now:fail] }
		
		# Otherwise ...
		return [list [clock seconds] [mid $text 4] now]
	}

	# URT
	set urt [timeval [lindex [split $args] 0] s]
	if [notempty urt] {
		set date [expr [clock seconds] + $urt]
		return [list $date [join [lreplace [split $args] 0 0]] urt]
	}

	# TIMESTAMP
	set ts [lindex [split $args] 0]
	if [isnum -integer $ts] {
		set date $ts
		set text [join [lreplace [split $args] 0 0]]
		return [list $date $text ts]
	}

	# CLOCK SCAN
	# This needs to consider timezones (bot vs user).
	# If a timezone is explicit, no conversion is necessary (CLOCK SCAN will handle).
	# Otherwise, conversion is needed (the calling PROC needs to deal with that)
	set carry 0
	do {
		set segment [join [lrange [split $args] 0 end-$carry]]
		set error [ catch { set date [clock scan $segment] } crap ]
		if !$error {
			set text [join [lreplace [split $args] 0 end-$carry]]
			break ; return [list $date $text]
		}
		incr carry
		if { $carry >= [llength $args] } break
	}
	if [notempty date] {
		set mode [lindex :USER [regexp -- {[+-](0?\d|1[0-4]):?([0-5]\d)?( \d{4})?$} $segment]]
		return [list $date $text CLOCK:SCAN${mode}]
	}
	
	# We couldn't find a valid timestamp
	return [list "" $text ""]
}

proc date:day { { date "" } } { # 1494799396
	# https://cs.uwaterloo.ca/~alopez-o/math-faq/node73.html
	# Returns: 0 (Sunday) , 1 (Monday) , ... , 6 (Saturday)
	if [string eq "" $date] { set date [clock seconds] } { if ![isint $date] { set date [ clock scan $date ] } }
	regexp -nocase -- {^(\d\d)(\d\d)(\d\d)(\d\d)$} [clock format $date -f {%Y%m%d}] - c y m d
	normalize:var c y m d
	# Shift so that: 1=March , ... 10=December , 11=January  (previous year), & 12=February (previous year)
	incr m -2 ; if { $m < 1 } { incr m 12 ; decr y }
	expr int( $d + floor( 2.6 * $m ) - ( 2 * $c ) + $y + floor( $y / 4.0 ) + floor( $c / 4.0 ) ) % 7
}

proc date:compress { mode args } {
	set simplify [list { x { y 0 } } { set x [string trimleft $x 0] ; if [string eq "" $x] { set x $y } ; return $x } ]
	set date [join $args]
	switch -exact -- [string tolower $mode] {
		comp - compress - enc - encode {
			if [string eq "" $date] { set date [clock seconds] }
			if ![isint $date] { set date [clock scan $date] }
			lassign [clock format $date -format "{%u} {%Y} {%m} {%d} {%H} {%M} {%S} {%z}"] dow year month day hour minute second zone
			regexp -nocase -- {^([+-])0*(\d{1,2})(\d\d)$} $zone - _polarity _hour _minute
			set offset ${_polarity}${_hour}
			set offset_minute $_minute
			set return ""
			set final @
			append final [string index [format %02X [ expr $dow | ( ( $day & 16 ) >> 1 ) ]] end]
			append final [string range [format %04X $year] end-1 end]
			#append final [string index [format %04X [ expr $year | ( [ isleapyear $year ] << 3 ) ]] 1]
			append final [format %01X [ expr 0x[string index [format %04X $year] 1] | ( [ isleapyear $year ] << 3 ) ]]
			append final [string index [format %01X $month] end]
			append final [string index [format %01X [ expr $day & 15 ]] end]
			foreach a [list hour minute second] b [list 48 72 96] { set $a [apply $simplify [set $a] $b] }
			set double [ expr ( ( $hour >= 12 ) && ( $hour <= 47 ) ) ? 2 : 1 ]
			append final [format %02X [ expr $hour * $double ]]
			append final [format %02X [ expr $minute * $double ]]
			append final [format %02X [ expr $second * $double ]]
			set zone_biased [ expr int( ( $offset + ( $offset_minute / 60.0 ) ) * 4 ) + 60 ]
			append final [format %02X $zone_biased]
			return $final
		}
		decomp - dec - decode {
			if ![regexp -nocase -- {^@([[:xdigit:]])([[:xdigit:]]{2})([[:xdigit:]])([[:xdigit:]])([[:xdigit:]])([[:xdigit:]]{2})([[:xdigit:]]{2})([[:xdigit:]]{2})([[:xdigit:]]{2})$} $date - 1 23 4 5 6 78 9a bc de] { error "Illegal pattern in code: $date" }
			set final ""
			append final [string totitle [lindex [list sun mon tues wednes thurs fri satur sun] [ expr 0x$1 & 7 ]]day]
			append final " "
			append final [format %04i [ expr ( ( 0x$4 & 7 ) << 8 ) | ( 0x$23 ) ]]
			append final -[format %02i [ expr ( 0x0$5 & 15 ) ]]
			append final -[format %02i [ expr 0x0$6 | ( ( 0x$1 & 8 ) << 1 ) ]]
			append final " "
			set double [ expr ( ( 0x$78 >= 12 ) && ( 0x$78 <= 46 ) ) ? 2 : 1 ]
			set _hour [ expr 0x$78 / $double ] ; if { $_hour == 48 } { set _hour 0 }
			append final [format %02i $_hour]
			append final :
			set _minute [ expr 0x$9a / $double ] ; if { $_minute == 72 } { set _minute 0 }
			append final [format %02i $_minute]
			append final :
			set _second [ expr 0x$bc / $double ] ; if { $_second == 96 } { set _second 0 }
			append final [format %02i $_second]
			append final " "
			append final [format %+05i [ expr int( ( int( 0x$de / 4.0 ) - 15 ) * 100 ) [lindex [list - +] [ expr ( ( 0x$de - 60 ) < 0 ) ? 0 : 1 ]] ( 0x$de % 4 ) * 15 ]]
			append final "[lindex [list " (leap year)"] [ expr ( 0x$4 & 8 ) ? 0 : 1 ]]"
			return $final
		}
		help { return "" }
		default { error "Syntax: date:compress <encode> <TS> or <decode> <@code>" }
	}
}

proc elapsed args {
	if [string eq -nocase HELP $args] { return "elapsed ?-milliseconds? <integer>" }
	set 0 [uniquematch -milliseconds [lindex [split $args] 0]]
	if [string eq -nocase $0 -MILLISECONDS] {
		set milliseconds [lindex [split $args] 1]
		if ![isnum -real $smilliseconds] return
		if { $milliseconds < 0 } return
		if { $milliseconds < 1000 } { return "$milliseconds ms" }
		set seconds [ expr $milliseconds / 1000.0 ]

	} {
		set seconds [lindex [split $args] 0]
	}
	if ![isnum -real $seconds] return
	if { $seconds < 1 } return
	if { $seconds < 60 } { return "$seconds s" }
	set minutes [ expr $seconds / 60.0 ]
	if { $minutes < 60 } { return "$minutes m" }
	set hours [ expr $minutes / 60.0 ]
	if { $hours < 60 } { return "$hours h" }
	set days [ expr $hours / 24.0 ]
	if { $days < 60 } { return "$days d" }
	set weeks [ expr $days / 7.0 ]
	if { $weeks < 52 } { return "$weeks w" }
	set years [ expr $weeks / 52.0 ]
	if { $years < 10 } { return "$years y" }
	set decades [ expr $years / 10.0 ]
	return "$decades e"
}

proc clock:scan { string { timezone "" } { format "" } } {
	# 2016-03-26 05:37:17 -0700
     # TCL 8.5 no longer supports "freeform clock scans" as well as it used to, due to the
	# number of ambiguities that exist. The "CLOCK SCAN <string> -format <format string>"
	# option was created to handle this (in clock.tcl). Thanks to thommey (freenode/#Eggdrop)
	# for the info on the changes in CLOCK.

	if ![regexp -nocase -- {^[A-Z0-9,/ \+\-\.\:]+$} $string] { return "" }
	if ![regexp -nocase -- {^(|[A-Z0-9,/ \+\-\.\:]+)$} $format] { return "" }
	set string "\{${string}\}"
	set options ""
	if { [info tclversion] >= 8.5 } {
		if ![string eq "" $format] { append options " -format \{${format}\}" }
		if ![string eq "" $timezone] { append options " -timezone $timezone" }
	} {
		if ![string eq "" $timezone] { append options " -gmt [falsetrue $timezone]" }
	}
#debug string options
	set error [ catch { set scan [eval clock scan $string $options] } errormsg ]
	if $error { error $errormsg } { return $scan }
}

proc SB:gmt_format { type gmt } {
	if [isempty gmt] return
	switch -exact -- [uniquematch [list decimal integer + - .] $type] {

		decimal - . {
			if [regexp -- {^[\+\-]?\d{1,2}(\.\d{1,2})?$} $gmt] {return $gmt}
			if ![regexp -- {^[\+\-]?\d{3,4}$} $gmt] {error "\[SB:GMT_FORMAT\] Illegal GMT format: \"${gmt}\" (should be in +\[0\]000 format)"}
			set sgn [sgn $gmt]
			set sign [lindex [list - + +] [incr sgn]]
			set gmt [string trimleft $gmt +-]
			set hour [nozero [left $gmt 2]]
			set minutes [mantissa [ fixmath [ right $gmt 2 ] / 60 ]]
			return [nozero ${sign}${hour}${minutes}]
		}

		integer - + - "-" {
			if [regexp -- {^[\+\-]?\d{3,4}$} $gmt] {return $gmt}
			if ![regexp -- {^[\+\-]?\d{1,2}(\.\d{1,2})?$} $gmt] {error "\[SB:GMT_FORMAT\] illegal GMT format: \"${gmt}\" (should be in +00, +00.0, or +00.00 format)"}
			set sgn [sgn $gmt]
			set sign [lindex [list - + +] [incr sgn]]
			regsub -- {^[\+\-]} $gmt "" gmt
			if [instr $gmt .] {set mantissa [ expr int( 60 * [ mantissa $gmt ] ) ]} {zero mantissa}
			int gmt
			return ${sign}[addzero $gmt 2][addzero $mantissa 2]
		}

		default {error "\[SB:GMT_FORMAT\] illegal option \"${type}\""}

	}
	return 0
}

proc SB:time_offset {time gmt_original gmt_target} { expr $time - ( ( [ SB:gmt_format decimal $gmt_original ] - [ SB:gmt_format decimal $gmt_target ] ) * 3600 ) }

interp alias {} printdate {} printtime
proc printtime {time handle} {formattime $time $handle}

interp alias {} formatdate {} formattime
proc formattime args {
	set arg [concat * $args]
	SB:setvariables -cliptimezone -gmt -timezone -format

	set list_arg $1e
	if [validflag -gmt] {set new_gmt $j1; set list_arg [lreplace $list_arg 0 0]}
	if [validflag -timezone] {set new_timezone $j1; set list_arg [lreplace $list_arg 0 0]}
	if [validflag -format] {set new_format $j1; set list_arg [lreplace $list_arg 0 0]}

	lassign $list_arg time handle additional; # $0 = "*"
	if [string eq * $time] {set time [clock seconds]}
	if [isempty time] {set time [clock seconds]}
	set time [nozero $time]
	int time
	if ![regexp -- {^\d+$} $time] {return ""}

	set format   [none [sbd:get CONFIG:GLOBAL-DATETIMEFORMAT] "%a, %d %b %Y, %H:%M:%S %Z"]
	set gmt      [SB:gmt_format decimal [none [sbd:get BOT:GMT] [clock format [clock seconds] -format %z]]]
	set timezone [none [sbd:get BOT:TIMEZONE] [clock format [clock seconds] -format %Z]]

	set clip_timezone [validflag -cliptimezone]

	if [validuser $handle] {
		set gmt_user [getdata $handle GMT]
		set format [none [getdata $handle DATETIME] $format]
		set timezone [none [getdata $handle TIMEZONE] $timezone]
	} {
		if [isempty handle] {set gmt_user $gmt} {
			if [regexp -- {^[\+\-]?(\d{1,2}|\d{4}|\d{1,2}\.\d{1,2})$} $handle] {
				set gmt_user $handle
			} {
				set gmt_user $gmt
				#error "\[SB:GMT_OFFSET\] illegal manual GMT offset: \"${handle}\""
			}
		}
	}

	if [notempty new_gmt] {set gmt [SB:gmt_format decimal $new_gmt]}
	if [notempty new_timezone] {set timezone $new_timezone}
	if [notempty new_format] {set format $new_format}

	set gmt_old $gmt
	set gmt [none $gmt_user $gmt]
	if [notempty gmt] {set gmt [SB:gmt_format decimal $gmt]}
	set offset [ fixmath ( [ nozero $gmt_user ] - $gmt_old ) * 3600 ]

	# Apply offset if known user & known user's GMT
	int offset
	incr time $offset

	# See if we needed user's data, but still require more offsetting? (TIME command)
	if [notempty additional] {
		if ![isint $additional] {error "\[FORMATTIME\] \"${additional}\" isn't an integer."}
		one clip_timezone
		incr time [ fixmath int( 3600 * ( $additional - $gmt_user ) ) ]
	}

	# Remove most likely combinations of text that surround GMT / timezone text
	# If you want to delete Stardate / Commonwealth year,
	# Change "%z" to "%[zq] and "%Z" to "%[ZQ]"
	if {$clip_timezone || [isempty gmt     ]} {regsub -all -- {[ ]?[<\(\[\{ ]?%z[ \}\]\)>]?[ ]?} $format " " format}
	if {$clip_timezone || [isempty timezone]} {regsub -all -- {[ ]?[<\(\[\{ ]?%Z[ \}\]\)>]?[ ]?} $format " " format}

	if ![validflag -format] { fixclock_z $format $gmt $timezone format }
	fixclock_q $time $format time format

	trim [clock format $time -format $format]
}

proc fixclock {time format gmt tz var_time var_format} {
	fixclock_q $time $format time format
	fixclock_z $format $gmt $tz format
	upvar 1 $var_time local_time $var_format local_format
	set local_time $time
	set local_format $format
	return [list $time $format]
}

proc fixclock_q {time format var_newtime var_newformat} {
	set stardate [clock format $time -format %Q]
	# Compensate for the "Leap Year December 31" bug
	if [string match -nocase "Stardate ??1000.?" $stardate] {
		regsub 1000 $stardate 999 stardate
		# Only replace if "??1000." because New Year's Day in the years
		# 2007, 2017, 2027, ... will pattern match "*1000.*" but SHOULD NOT
		# be corrected !!
	}

	# Compensate: day #1 -> 001, not 002 (by the internal math)
	if [string match *002.* $stardate] { regsub 002 $stardate 001 stardate }
	# set date 7/1; set year 1988 ; set day_n [clock format [clock scan "12/31/$year"] -format %j] ; decr day_n ; set i [ fixmath 1.0 / $day_n ] ; set s [decimal [noscinot [ expr $i * ( [ clock format [ clock scan "${date}/$year" ] -format %j ] -1 ) ]] 3 0]
	regsub -all %Q $format $stardate format

	# Commonwealth year [custom definition for %q option]
	set cy "CY[space][ expr [ clock format $time -format %Y ] +4920 ]" ; # 4926
	regsub -all %q $format $cy format
	upvar 1 $var_newtime local_newtime $var_newformat local_newformat
	set local_newtime $time
	set local_newformat $format
	return [list $time $format]
}

proc fixclock_z {format gmt tz var_newformat} {
	regsub -all -- %z $format [SB:gmt_format integer $gmt] format
	regsub -all -- %Z $format $tz format
	upvar 1 $var_newformat local_newformat
	set local_newformat $format
	return $format
}

proc convertgmt { { time "" } { from "" } { to "" } } { context ; error "\[CONVERTGMT\] I'm deprecated. Find me and remove me? (check CONTEXT log)" }

# --- Other ---

interp alias {} get:opts {} get:options
proc get:options command {
	switch -exact -- [stl $command] {
		regexp { catch { regexp -cow 1 1 } msg ; regexp -nocase -- {must be[:]? (.+)$} $msg - opts ; regsub -all -- {,|, or |\-\-} $opts "" opts ; return $opts }
		regsub { catch { regsub -cow 1 1 } msg ; regexp -nocase -- {must be[:]? (.+)$} $msg - opts ; regsub -all -- {,|, or |\-\-} $opts "" opts ; return $opts }
	}
}
	
proc SB:reload command { upvar 1 _____:command:_____ local ; set local $command ; uplevel 1 { SB:dispatch $nick $host $handle $chan "maintenance load [SB:commandlist src ${_____:command:_____}]" ; unset -nocomplain _____:command:_____ } }

proc header args {
	flags:simple $args [list -upper -lower -title -sentence -parenthesis -brackets -braces -anglebrackets] text flags
	# Defaults
	if ![validflag -parenthesis -brackets -braces -anglebrackets] { lappend flags -brackets }
	if ![validflag -upper -lower -title -sentence] { lappend flags -upper }
	# Determine parent command
	set level [info level]
	incr level -1
	if !$level { set header global } { set header [lindex [info level $level] 0] }
	# Additional text
	if [notempty text] { append header " [join $text]" } ; # Space!
	# Parse flags
	if [validflag -upper] { set header [stu $header] }
	if [validflag -lower] { set header [stl $header] }
	if [validflag -title] { set header [stt $header] }
	if [validflag -sentence] { set header [sts $header] }
	if [validflag -parenthesis] { lassign [list ( )] open close }
	if [validflag -brackets] { lassign [list \[ \]] open close }
	if [validflag -braces] { lassign [list \{ \}] open close }
	if [validflag -anglebrackets] { lassign [list < >] open close }
	return "${open}${header}${close}"
}

# Enter with: FINDCHAIN [list 1 2 3 4 5 6 7 8] <item to find>
# Each item pair should be: {item parent} {item parent} or else: the result list will be reversed
# Return "" if no links are found
proc findchain { list item } { # 1494144711
	if { [isempty list] || [isempty item] } return
	putlog "\[FINDCHAIN\] LIST($list):ITEM($item)"
	array set link $list
	set list [lstride [array get link] 2]
	empty chain
	# Backward:
	set _ $item
	while { [set m [lsearch -inline -exact -index 0 $list $_]] != "" } {
		if [isempty chain] { set chain $_ }
		set _ [lindex $m 1]
		lprepend chain $_
	}
	# Forward
	set _ $item
	while { [set m [lsearch -inline -exact -index 1 $list $_]] != "" } {
		if [isempty chain] { set chain $_ }
		set _ [lindex $m 0]
		lappend chain $_
	}
	lsort -unique -dictionary -increasing $chain
}


proc hexdump { text { size 16 } } {
	set max [string length [format %X [string length $text]]]
	set length [expr $size * 2] ; # Double it due to HEX
	set hexdump ""
	binary scan $text H* hex
	set count 0
	while { ![string eq "" $hex] } {
		set line [string range $hex 0 [expr $length - 1]]
		set hex [string range $hex $length end]
		empty chars
		set temp "\[[format %0${max}X $count]\] "
		for { set x 0 ; set y 1 } { $x < $length } { incr x 2 ; incr y 2 } {
			set char [string index $line $x][string index $line $y]
			if [isempty char] { append temp "-- " ; append chars " " ; continue }
			append temp "[string toupper $char] " ; # Space!
			regsub -- {^[\x00-\x1F\x80-\xFF]$} [binary format H* $char] . char
			append chars $char
		}
		append temp " | " $chars
		lappend hexdump $temp
		incr count $size
	}
	return [join $hexdump \n]
}

proc hexdump:file [list file [list chunk 16]] {
	if ![file exists $file] { error "\[HEXDUMP\] No such file: $file" }
	set size [file size $file]
	set size_hex [ expr ( $chunk * 3 ) - 1 ]
	set max [string length [format %X $size]]
	set r [open $file r]
	fconfigure $r -translation binary
	while { ![eof $r] } {
		set tell [tell $r]
		set data [read $r $chunk]
		binary scan $data H*@0a* hex asc
		regsub -all -- {[^\x20-\x7E\x80-\xFE]} $asc . asc ; # 127 isn't printable
		puts "[format "%0${max}X | %-#${size_hex}s | %-#${chunk}s" $tell [string toupper [regexp -inline -all -- .. $hex]] $asc]"
	}
	close $r
	return
}

interp alias {} SB:level_global {} globallevel
proc SB:level_global {handle {chan ""}} {
	if [SB:checkbot $handle] {return "bot"}
	foreach {level data} [sbdget USERLEVELS] {
		if [iswithin [lindex $data 2] [SB:accesslevel $handle $chan]] {
			set globallevel [lindex $data 4]
			regsub -all -- %GROUP% $globallevel [sbdget name:group] globallevel
			return $globallevel
		}
	}
	return ""
}

interp alias {} SB:level_channel {} chanlevel
proc SB:level_channel {handle chan} {
    set level ""
    if {[matchattr $handle b]} {return "bot"}
    if {[matchattr $handle |q $chan]} {set level "Auto-devoiced"}
    if {[matchattr $handle |d $chan]} {set level "Shitlisted: auto-deopped"}
    if {[matchattr $handle |k $chan]} {set level "Shitlisted: banned"}
    if {[matchattr $handle |f $chan]} {set level "[sbdget name:chanfriend]"}
    if {[matchattr $handle |v $chan]} {set level "[sbdget name:chanvoice]"}
    if {[matchattr $handle |H $chan]} {set level "[sbdget name:chanhalfop]"}
    if {[matchattr $handle |o $chan]} {set level "[sbdget name:chanop]"}
    if {[matchattr $handle |a $chan]} {set level "[sbdget name:chanop] (\037AUTOOPPED\037!!!!!)"}
    if {[matchattr $handle |m $chan]} {set level "[sbdget name:chanmaster]"}
    if {[matchattr $handle |n $chan]} {set level "[sbdget name:chanowner]"}
    if {[matchattr $handle b]} {set level "bot"}
    return "$level"
}

proc SB:checkbot bot {
	if [isbotnick $bot] {return 1}
	if [string eq -n $bot $::botnick] {return 1}
	if [matchattr $bot b] {return 1}
	return 0
}

proc isbetween {val lo hi} {
	# Reverse args if necessary because Domino is situationally dyslexic #
	if {$hi < $lo} {set l $lo; set lo $hi; set hi $l}

	if {($val <= $hi) && ($val >= $lo)} {return 1}
	return 0
}

proc iswithin {range value { bottom -1 } { top 9999 } } {
	set a [getrange $range -2147483648 2147483647]
	if ![string eq $a -1] {
		if { [l0 $a] < $bottom } { set bottom [l0 $a] }
		if { [l1 $a] > $top    } { set top    [l1 $a] }
	}
	set getrange [getrange $range $bottom $top]
	if [string eq $getrange -1] {return 0} { if [isbetween $value [l0 $getrange] [l1 $getrange]] { return 1 } { return 0 } }
}

proc longip [list a [list version 4]] { # Or "6"
	if [instr $a :] { # IPv6?
		# What the hell do we do with IPv6-encoded IPv4? (FE80::1.1.1.1)
		set s [split $a :]
		set l [llength $s]
		if { $l != 8 } {
			set m [lsearch $s ""]
			set s [lreplace $s $m $m {*}[lrepeat [ expr ( 8 + 1 ) - $l ] 0]] ; # Replacing the blank with "0" still keeps the count short-of-8; need to add 1!
		}
	}
	if [instr $a .] {
		if {[llength [split $a .]] != 4} {return -1}
		foreach {a b c d} [split $a .] break
		return [format %u [expr { ( $a << 24 ) + ( $b << 16 ) + ( $c << 8 ) + $d } ]]
	}
	# Credit (IPV4): LionPing at egglist@undernet.nu from EggHead list 14 Mar 03.
	switch -exact -- $version {
		4 { return [ expr { $a >> 24 & 255 } ].[ expr { $a >> 16 & 255 } ].[ expr { $a >> 8 & 255 } ].[ expr { $a & 255 } ] }
		6 {}
		default { error "\[LONGIP\] Unknown IP to convert: $a" }
	}
}

proc longip:convert ip {
	set version [ expr 4 + ( 2 * [string match *:* $ip] ) ]
	set shift [ expr 256 ** ( 1 + int( $version / 6 ) ) ]
	set rsequence [lreverse [set sequence [sequence 0 .. 7]]]
	lassign [split $ip .:] {*}$rsequence
	zero total
	negone c
	foreach _ $sequence {
		if [notempty $_] {
			incr c
			set total [ expr $total + ( 0x[set $_] * ( ${shift} ** $c ) ) ] 
		} 
	}
	return $total
}

proc SB:is_trigger_global args {
	set trigger $::lastbind
	if [notempty args] {set trigger [lindex $arg 0]}
	set m [lsearch -all -exact [string tolower [sbdget @TRIGGERS]] [string tolower $trigger]]
	if {$m > -1} {return 1} {return 0}
}

interp alias {} color {} SB:color
interp alias {} colour {} SB:color

# Enter with: [SB:COLOUR KILL OFF B(old) U(nderline) R(everse) 1,5] (any combination)
# Order matters !!
proc SB:color args {

	set 99 99,99; # \002\002
	set args [join $args]

	# Turning off colour via the legacy way?
	if [isempty args] {return "\003\002\002"}
	#%# Should that be \003$99 instead? (in case colour is turned on elsewhere)?
	#%# No! OUTPUT2 expects end-of-colour to match "\003[^\d]" so it knows
	#%# to restore the user's standard colour output preference!

	# Are colours disabled globally?
	set cb [sbd:nozero CONFIG:GLOBAL-COLORBLOCK]
	if ($cb) {return \003$99}; # Null out the colour code

	# How to pick "colour off" without affecting other options (bold, et al)?
	array set codes [list off \00399,99 kill \017 colour \003 bold \002 underline \037 reverse \026]

	empty code code_colour
	zero colour kill
	if [isempty args] {
		set a $99
	} else {
		foreach arg $args {
			if [left $arg 1 $codes(colour)] {set arg [mid $arg 2]}

			set um [uniquematch [list bold underline reverse off kill] $arg]
			if [notempty um] {set arg $um}

			switch -regexp -- [stl $arg] {

				{\d{1,2}(,\d{1,2})?} {
					set s [split $arg ,]
					empty t
					foreach a $s {lappend t [addzero $a 2]}
					set arg [join $t ,]
					append code $codes(colour)$arg
				}

				off {append code $codes(off)}

				kill {append code $codes(kill)}

				underline {append code $codes(underline)}

				bold {append code $codes(bold)}

				reverse {append code $codes(reverse)}

				default {error "\[SB:COLOUR\] Unknown control code: \"${arg}\""}

			}
		}
	}

	return $code
}

interp alias {} nocolor {} SB:remove_color
proc SB:remove_color text {regsub -all -- {\003([0-9]{1,2}(,[0-9]{1,2})?)?} $text ""}

interp alias {} nocodes {} SB:remove_codes
proc SB:remove_codes text {regsub -all -- \002|\037|\026|\035 [nocolor $text] ""}

interp alias {} home {} SB:home; # proc home (for finding only)
proc SB:home { { checkme "" } } {
	set home [sbdget CONFIG:GLOBAL-HOME]
	if [string eq -nocase NONE $checkme] { return [isempty home] }
	if [notempty checkme] { return [string eq -nocase $home $checkme] }
	if ![validchan $home] {
set frame [info frame] ; for { set _ 0 } { $_ <= $frame } { incr _ } { putlog "\[SB:HOME\] Frame trace #{$_}: [info frame $_]" }
		putlog "\00300,04\[SB6:HOME\] \002The HOME channel is an illegal channel: ${home}\002\003"
		return ""
	}
	return $home
}

proc seconds args {clock seconds}

proc jtext { reply { encoding "" } } {
	# These are encoded in UTF-8
	if [isempty reply] {return ""}

	array set japanese {
		shoosai_kibonnu       e8a9b3e7b4b02e2e2ee38080e382ade3839ce383b3e3838c212121
		bakabakashii_sentaku  e38182e3828ce381afe9a6ace9b9bfe9a6ace9b9bfe38197e79a84e381aae981b8e68a9ee381a7e69c89e3828be381abe38287efbc81
		panda_wa_kawaii_desu  e38391e383b3e38380e381afe58fafe6849be38184e381a7e38199e38082
		wasurechatta_no       e4bd95e3818be38292e5bf98e3828ce381a1e38283e381a3e3819fe381aeefbc81e38282e38186e4b880e5baa6e381a7e8a9a6e38197e381a6e4b88be38195e38184e38082
		nasakenai_yo          e3819de3828ce381afe68385e38191e381aae38184e38288efbc81
		anta_baka_?!          e38182e38293e3819fe9a6ace9b9bfefbc9fefbc81
		hen_no_kotae_ga_atta  e3819de3828ce381afe5a489e381aee7ad94e3818ce38182e381a3e3819fe381aee381a7e38199e381abe38287e38082
		kaisetsusho           e8a7a3e8aaace69bb8e38292e889afe3818fe381a7e8aaade381bfe4ba8be587bae69da5e381bee38199e3818befbc9fe38080
	}

	if [string eq -nocase LIST $reply] {return [array names japanese]}
	set um [uniquematch [array names japanese] $reply]
	if [notempty um] {set reply $um}
	if ![info exists japanese($reply)] {return ""}
	set jtext [binary format H* $japanese($reply)]

	# If we're already in UTF-8, don't translate any more: just return the text
	if [isempty encoding] { if [string eq -nocase UTF-8 [encoding system]] { return $jtext } }

	# Is this really necessary?
	if { !( [string eq -nocase ISO* [encoding system]] || [string match -nocase CP* [encoding system]] ) } { set jtext [encoding convertfrom [encoding system] $jtext] }

	if [isempty encoding] {
		if [string eq -nocase ISO8859-1 [encoding system]] { set jtext [encoding convertfrom utf-8 $jtext] }
	} {
		set jtext [encoding convertto $encoding $jtext]
	}
	return $jtext
}

interp alias {} fileread {} readfile
proc readfile { name { split "" } { binary "" } } {
	if [string eq -nocase BINARY $split] { set binary binary ; empty split }
	set r [open $name r]
	if [notempty binary] { fconfigure $r -translation binary }
	set data [read $r]
	close $r
	if [notempty split] { set data [split $data $split] }
	return $data
}

# Enter with DUMPFILE FILENAME DATA (optional: 1: data=LIST, send each alone.)
proc appendfile { a b { c false } { binary false } } {
	set w [open $a a+]
	set binary [string map [list 0 "" 1 binary] $binary]
	# This allows the user to specify "0" as usual, or specify the encoding [e.g.: utf-8].
	if {$binary != ""} {
		if ![string match -nocase *DOS* $binary] {
			fconfigure $w -translation $binary
			if { [lsearch -exact [list auto cr lf crlf] [stl $binary]] == -1 } { set c 0 }; # In binary write mode, do not LIST-SPLIT lines!
		}
	}
	if ($c) {
		foreach d $b {puts $w ${d}[lindex [list "" \r] [string match -nocase *DOS* $binary]]}
	} else {
		if [string eq "" $binary] {puts $w [regsub -all -- \n $b [lindex [list "" \r] [string match -nocase *DOS* $binary]]\n]} {puts -nonewline $w $b}
	}
	close $w
	return 0
}

interp alias {} writefile {} dumpfile
proc dumpfile { a b { c 0 } { binary 0 } } {
	if { [lsearch -exact -nocase [list BINARY CRLF CR LF UNICODE UTF-8 UTF-16 UTF-32] $c] != -1 } { set binary $c ; set c 0 }
	if { [lsearch -exact -nocase [list UNICODE UTF-8 UTF-16 UTF-32] $binary] != -1 } {
		set header [encoding convertto $binary \uFEFF]
		set w [open $filename w]
		fconfigure $w -translation binary
		# Unicode* header first
		switch -exact -- [string tolower $binary] {
			uni - unicode { set header \uFEFF }
			ucs2 - ucs-2 { set header \uFEFF }
			utf8 - utf-8 { set header \xEF\xBB\xBF }
			utf16le - utf16-le utf-16le utf-16-le - utf16 - utf-16 { set header \xFF\xFE }
			utf16be - utf16-be utf-16be utf-16-be - utf16 - utf-16 { set header \xFE\xFF }
			utf32le - utf32-le utf-32le utf-32-le - utf32 - utf-32 { set header \xFF\xFE\x00\x00 }
			utf32be - utf32-be utf-32be utf-32-be - utf32 - utf-32 { set header \x00\x00\xFE\xFF }
		}
		puts -nonewline $w $header
		flush $w
		if $list_mode {
			foreach line $data { puts $w ${line}\r }
		} {
			puts $w $data
		}
		flush $w
		close $w
		return 0
	} {
		set w [open $a w]
		set binary [string map -nocase [list 0 "" false "" 1 binary true binary] $binary]
		# This allows the user to specify "0" as usual, or specify the encoding [e.g.: utf-8].
		if {$binary != ""} {
			if ![string match -nocase *DOS* $binary] {
				fconfigure $w -translation $binary
				if { [lsearch -exact [list auto cr lf crlf] [stl $binary]] == -1 } { set c 0 }; # In binary write mode, do not LIST-SPLIT lines!
			}
		}
		if ($c) {
			foreach d $b {puts $w ${d}[lindex [list "" \r] [string match -nocase *DOS* $binary]]}
		} else {
			if [string eq "" $binary] {puts $w [regsub -all -- \n $b [lindex [list "" \r] [string match -nocase *DOS* $binary]]\n]} {puts -nonewline $w $b}
		}
		close $w
	}
	return 0
}

interp alias {} plural {} SB:plural
proc SB:plural args {# Backward-compatible upgrade
	SB:getflags $args [list -show -comma] text flags
	lassign $text word count ending replacement

	if [isempty word] return
	if [isempty count] return
	set plural $word
	if { $count != 1 } {
		if [right $plural 1 y] {set plural [left $plural -1]ie}
		if [right $plural 3 ium] {set plural [left $plural -2]a}
		if [notempty replacement] {set plural $replacement} {append plural [none $ending s]}
	}
	if [validflag -show] {
		set num $count
		if [validflag -comma] { set num [comma $num] }
		prepend plural "${num}[space]"
	}
	return $plural
}

# See replacement PROC below
proc `scinot args { # number { digits -1 }
	flags:simple $args -10 text flags
	lassign $text number digits
	if [isempty digits] { set digits -1 }
	if [isscinot $number] { return $number }
	set number [normalize $number]
	if { $number == 0 } { return 0 }
	if [string eq . [string index $number 0]] { set number "0$number" }
	if { [string first . $number] == -1 } { append number .0 }
	if [string match 0.* $number] {
		# Negative exponent
		set sign -
		regexp -- {^([0\.]+)} $number - zeroes
		set offset [ expr [ len $zeroes ] - 1 ]
		set number [mid $number [ expr $offset + 2 ]]
		set number [normalize [left $number 1].[none [mid $number 2] 0]]
		lassign [split $number .] integer mantissa
	} {
		# Positive exponent
		set sign +
		set offset [format %02d [ expr [instr $number .] - 2 ]]
		set integer [left $number 1]
		set mantissa [mid [regsub -all -- {\.} $number ""] 2]
	}

	if [validflag -10] { set exponent x10[exponent [string trimleft $sign$offset +]] } { set exponent E$sign$offset }
	# Common adjustment and output
	if { $digits == -1 } {
		return [normalize ${integer}.${mantissa}]$exponent
	} {
		if { $digits > [len $mantissa] } {
			return [format %.${digits}f ${integer}.[left ${mantissa}[string repeat 0 $digits] $digits]]$exponent
		} {
			return [format %.${digits}f ${integer}.[left $mantissa [incr digits]]1]$exponent
		}
	}
}

proc scinot [list number [list precision ""]] { #Tested with: 65504 -0.125 +.00001 25.78125 {+9.991 3} {-9.991 4} {-9.991 1} {+9991 1} {-9991 4}
	if [isscinot $number] { return $number }
	regexp -- {^([+-]?)0*(\d*)\.(\d*)\.?0*$} ${number}. - polarity integer mantissa
	if [allempty integer mantissa] { return 0.0E+00 }
	zero offset
	if [isempty integer] {
		if [isempty mantissa] { zero mantissa }
		regexp -- {^(0*)([0 1-9])(.*)0*$} $mantissa - mz mf mr
		set scinot ${polarity}${mf}.[format %01s $mr]E-[format %02i [ expr [string length $mz] + 1 + $offset ]]
	}  {
		regexp -- {^(\d)(.*)0*$} ${integer}${mantissa} - first remainder
		set scinot ${polarity}${first}.${remainder}E+[format %02s [ expr [string length $integer] - 1 ]]
	}
	if [notempty precision] {
		regexp -- {^([+-]?)(\d\.\d+)E([+-])0*(\d+)$} $scinot - p c s e
		set c [left $c [ expr $precision + 3 ]] ; # precision+(integer+period+rounding_digit)
putlog "\[SCINOT\] PRECISION($precision):C($c)"
		regexp -- {^(\d+)\.(\d+)$} [format %0.${precision}f $c] - i m
		if { $i > 9 } { 
			incr e
			set c [format %0.${precision}f [ expr ${i}.${m} / 10.0 ]]
			set scinot ${p}${c}E${s}[format %02s $e]
		} {
			set scinot ${p}[format %0.${precision}f $c]E${s}[format %02s ${e}]
		}
	}
	return $scinot
}

proc isscinot { number { loose false } } {
	set loose false
	if $loose {
		regexp -nocase -- {^[+-]?\d+(\.\d+)?E[+-]?\d+$} $number
	} {
		#regexp -- {^[+\-]?\d(\.\d+)?[Ee][+\-]?\d+$} $number
		regexp -- {^[+\-]?\d\.\d+[Ee][+\-]?\d+$} $number
	}
}

proc noscinot number { # Don't use DO ; causes an endless loop during the $_DO_MAX check
	if ![isscinot $number true] {return $number} ; # Parse as "loose"
	set sign ""
	set number [string trimleft $number +]
	if [left $number 1 -] {set sign -; set number [mid $number 2]}

	# Force sign upon exponent
	if ![regexp -- {[Ee][\+\-]} $number] {regsub -nocase -- E $number E+ number}

	# Force mantissa upon expression
	if ![instr $number .] {regsub -nocase -- E $number .0E number}

	set instr [instr [stl $number] e]

	set direction [mid $number [ expr $instr + 1 ] 1]
	set exponent  [nozero [mid $number [ expr $instr + 2 ]]]
	set number    [left $number [ expr $instr - 1 ]]
	prepend number [string repeat 0 $exponent]
	append  number [string repeat 0 $exponent]
	set dec [ expr [instr $number .] - 1 ]
	set number [left $number $dec][mid $number [ expr $dec + 2 ]]
	add dec ${direction}[nozero $exponent]; # Needed it keep from reverting into SCINOT
	int dec; # But, need to make it an INT for LEFT/MID.
	set number [left $number $dec].[mid $number [ expr $dec + 1 ]]
	return ${sign}[nozero $number]
}

proc nobr text {regsub -all -- " " $text \xA0}

proc SB:padindex {a b { fill "" } } {
	if ![isint $b] {error "\[SB:padindex\] Expected integer but got: $b"}
	if {[llength $a] >= $b} {return $a}
	for {set c 1} {$c <= $b} {incr c} {lappend a $fill}
	return [lrange $a 0 [incr b -1]]
}

proc not a { isfalse $a }

interp alias {} offon {} boolean:word -offon
interp alias {} onoff {} boolean:word -offon
interp alias {} noyes {} boolean:word -noyes
interp alias {} yesno {} boolean:word -noyes
interp alias {} falsetrue {} boolean:word -falsetrue
interp alias {} truefalse {} boolean:word -falsetrue
proc boolean:word args {
	flags:simple $args [list -offon -onoff -falsetrue -truefalse -noyes -yesno -upper -lower -title -sentence -case] value flags
	if ![validflag -offon -onoff -falsetrue -truefalse -noyes -yesno] { lappend flags -falsetrue] }
	set boolean [string is true -strict $value]
	if [validflag -falsetrue -truefalse] { set boolean [lindex [list false true] $boolean] }
	if [validflag -offon -onoff] { set boolean [lindex [list off on] $boolean] }
	if [validflag -noyes -yesno] { set boolean [lindex [list no yes] $boolean] }
	if [validflag -case] {
		set boolean [lindex [list [string tolower $boolean] [string toupper $boolean]] [string is true -strict $boolean]]
	} {
		if [validflag -upper] { set boolean [string toupper $boolean] }
		if [validflag -lower] { set boolean [string tolower $boolean] }
		if [validflag -title] { set boolean [stt $boolean] }
		if [validflag -sentence] { set boolean [string totitle $boolean] }
	}
	return $boolean
}

proc isleapyear year {
	if [ expr $year %   4 ] { return 0 }
	if [ expr $year % 100 ] { return 1 }
	if [ expr $year % 400 ] { return 0 }
	return 1
}

proc reverse text {
	if ![ catch { set reverse [string reverse $text] } ] { return $reverse }
	empty reverse
	foreach a [split $text ""] { prepend reverse $a }
	return $reverse
}

proc SB:deluserrecs {handle { newhandle "" } } {
	debug.print 2 "\[SB:DELUSERRECS\] \$handle -> \"$newhandle\""
	array set custom [sbdget @CUSTOM]
	foreach a [array names custom] {
		set b [lsearch -all -glob -inline $custom($a) *!*]
		if {$a == "user"} {set b $custom(user)}
		foreach c $b {
			set temp $c
			foreach chan [concat GLOBAL [channels]] {
				regsub -all ! $c $handle c
				regsub -all # $c $chan c
				foreach d [sbdfind $c] {
					if {$newhandle != ""} {
						regsub -all ! $temp $newhandle e
						regsub -all # $e $chan e
						sbdset $e [sbdget $d]
					}
					sbdset $d
				}
			}
		}
	}
	return 0
}

proc SB:addtimer  {a b { q "" } } {SB:killtimer $b $q;  timer $a $b}
proc SB:addutimer {a b { q "" } } {SB:killtimer $b $q; utimer $a $b}
proc SB:killtimer {procname { q "" } } {
	set count 0
	foreach utimer [utimers] { if [string match -nocase $procname [lindex $utimer 1]] { killutimer [lindex $utimer 2]; if [isempty q] {putlog "\[SB:KILLTIMER\] UTIMER [l2 $utimer]: [l1 $utimer]"}; incr count } }
	foreach  timer [ timers] { if [string match -nocase $procname [lindex $timer  1]] { killtimer  [lindex $timer  2]; if [isempty q] {putlog "\[SB:KILLTIMER\] TIMER [l2 $timer]: [l1 $timer]"}; incr count } }
	return $count
}

# See below for replacement
proc `urt {a { b "" } { c 256 } } {
	set a [duration $a]
	if {[l1 $a] == "years"} {
		set years [l0 $a]
		if {$years > 9} {
			set crap ""
			set decades [expr int($years / 10)]
			set years [expr $years - ($decades * 10)]
			if {$decades > 0} {set crap "$decades decade"}
			if {$decades > 1} {append crap "s"}
			if {$years > 0} {append crap " $years year"}
			if {$years > 1} {append crap "s"}
#?               set c "$crap [l2e $a]"
		}
	}
	set w1 "decade* year* week* day* hour* minute* second* tenths* hundreths* thousanths*"
	set w2 "e y w d h m s t n l"
	set count -1
	foreach junk $w1 {
		incr count
		set l [lsearch -glob $a $junk]
		if {$l > -1} {set a [lreplace $a $l $l [lindex $w2 $count]]}
	}
	# $C = number of elements to include
	set a [lrange $a 0 [ expr $c - 1 ]]
	regsub -all " " $a "" a
	if {$b != ""} {
		set temp ""
		foreach d [split $a ""] {
			append temp $d
			if [string match \[a-zA-Z\] $d] {append temp " "}
		}
		set a [join $temp $b]
	}
	return [join $a]
}

# Replacement PROC
proc urt { seconds { joiner "" } { number_of_terms 999 } } {
	# d[e]cades [y]ears [w]eeks [d]ays [h]ours [m]inutes [s]econds
	set duration [duration $seconds]
	array set terms [list]
	foreach regexp [regexp -inline -all -nocase -- {\d+ [a-zA-Z]} $duration] {
		lassign [split $regexp] value type
		if [string eq -nocase Y $type] { if { $value > 10 } { set terms(e) [ expr int( $value / 10 ) ] ; set value [ expr $value % 10 ] } }
		set terms([string tolower $type]) $value
	}
	empty urt
	foreach a [list e y w d h m s] { if [info exists terms($a)] { lappend urt $terms($a)$a } }
	if { [llength $urt] > $number_of_terms } {
		set urt [lrange $urt 0 [ expr $number_of_terms - 1 ]]+ ; # Position-vs(count)-index issue
	}
	join $urt $joiner
}

proc time:remaining { args } {
     flags:simple $args -raw temp flags
     lassign $temp items_complete items_total time_started time_current
     if [isempty time_current] { set time_current [clock seconds] }
	time_duration == $time_current - $time_started
     percentage == ( 100.0 * $items_complete ) / $items_total ; # Protects repeating fraction from rounding
     calc == ( ( 100.0 * $time_duration ) / $percentage ) - $time_duration
     int calc
     if [validflag -raw] { return $calc }
     lrange [urt $calc " "] 0 1
}

proc SB:os { { checkme "" } } {
	if [string eq -nocase VERSION $checkme] {
		set os $::tcl_platform(os)
		switch -glob -- [stl $os] {
			cygwin_nt-* {
				# https://en.wikipedia.org/wiki/List_of_Microsoft_Windows_versions
				array set ver [list 5.0 "Windows 2000" 5.1 "Windows XP" 6.0 "Windows Vista"]
				array set ver [list 6.1 "Windows 7" 6.2 "Windows 8" 6.3 "Windows 8.1"]
				array set ver [list 5.2 "Windows XP Professional x64"]
				regexp -nocase -- {cygwin_nt-([\d\.]+)} $os - me
				if ![info exists ver($me)] { set ver($me) "Windows $me" }
				return $ver($me)
			}
			cygwin* {
				array set ver [list 4.90 "Windows ME" 4.10 "Windows 98" 4.00 "Windows 95"]
				array set ver [list 3.11 "Windows for Workgroups 3.11"]
				regexp -nocase -- {(\d{1,2}\.\d{1,2}} $os - me
				if ![info exists ver($me)] { set ver($me) "Windows $me" }
				return $ver($me)
			}
			windows* { return $os }
			default  { return $os }
		}
	}

	if [string eq -nocase NIX $checkme] {set checkme *n?x}
	set platform $::tcl_platform(platform)
	if [string eq -nocase UNIX $platform] {
		switch -glob -- [string tolower $::tcl_platform(os)] {
			unix - linux - *bsd {set os unix}
			windows {set os windows}
			cygwin* {set os cygwin}
			mac* - darwin* {set os mac}
			default {error "\[SB:os\] what OS am I? platform ($::tcl_platform(platform)), os ($::tcl_platform(os))"}
		}
	}
	set os unix
	if [isempty checkme] { return $os }

	# Specific OS check
	if ![iswildcard $checkme] {append checkme *}
	string match -nocase $checkme $os
}

proc SB:os_ver { { checkme "" } } {
	putlog "\00301,00\[SB:OS_VER\]\003 \00300,01Invoked by:\003"
	set level [info level]
	putlog "[space 5]Levels: $level"
	for { set l $level } { $l > 0 } { incr l -1 } { putlog "[space 5]Level #${level}: [info level $l]" }
	return 1
}

proc testip {ip { v4only 0 } } {
	if !($v4only) {
		set ipv6 [testipv6 $ip]; # Returns "-1" if not a valid IPv6 address.
		if {$ipv6 > -1} {return $ipv6}
		if {$ipv6 == -2} {set ip [le [split $ip :]]}
	}
	if ![regexp -- {^[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}$} $ip] {return 0}
	set tmp [split $ip .]
	if {[llength $tmp] != 4} {return 0}
	foreach a $tmp { if { ($a < 0) || ($a > 255) } { return 0 } }
	return 1
}

# Returns: 0 (invalid IPv6), 1 (valid IPv6), -1 (not IPv6 format), or
#     -2 (IPv4 in IPv6 format)

proc testipv6 ip {
debug.print 2 "\[TESTIPV6\] 01: \$IP ($ip)"

	set error "-1"; # Using "-" in RETURN may be interpretted as a flag.

	# Valid formats:
	#    0:fe18:de1:31:C34A:Bd1f:0715:12
	#    BB18::1FE7:b18f:ecE
	#    ::0
	#    ::ffff:192.168.1.3
	#    (all of the above)/64

	# Check for, and ignore, the network node identifier ....
	set node [l1 [split $ip /]]
	if ![string eq "" $node] {
		if ![isint $node] {return -3}
		if ![isbetween $node 0 255] {return -3}
	}
	set ip [l0 [split $ip /]]
debug.print 2 "\[TESTIPV6\] 02: \$IP ($ip)"

	# Check for embedded IPv4 first.
	if {[string match ::*.*.*.* $ip] || [string match -nocase ::FFFF:*.*.*.* $ip]} {
debug.print 2 "\[TESTIPV6\] 03: \$IP ($ip) \[IPv4 check\]"
		# Looks like it could be IPv4. Let's check it ....
		set ipv4 [le [split $ip :]]
		set ipv4error 0
		if {[llength [split $ipv4 .]] != 4} {
			set ipv4error 1
		} else {
			foreach a [split $ipv4 .] {
				set a [nozero $a]
				if {($a < 0) || ($a > 255)} {set ipv4error 1}
			}
		}

debug.print 2 "\[TESTIPV6\] 04: \$IPV4ERROR ($ipv4error)"
		# End of IPv4 tests.
		if !($ipv4error) {return "-2"}

		# Is not an embedded IPv4, yet it has dots, so I can't presume it's
		# a valid IPv6 either. Return an error code.

		return $error
	}

debug.print 2 "\[TESTIPV6\] 05: \$IP ($ip)"

	# At this point, we're gonna presume it's a possible IPv6. Testing ....
	if ![regexp -nocase -- {^[\:0-9a-f]+$} $ip] {return $error}
	# Do not enclude "." since valid situations with it were handled above.

debug.print 2 "\[TESTIPV6\] 06: \$IP ($ip)"

	if [string match *::* $ip] {
		# Legally dropped "zero" (0) 16-bit words. Replace them completely.

debug.print 2 "\[TESTIPV6\] 07: \$IP ($ip)"
		set split [lsearch -all -exact [split $ip :] ""]
		if {[llength $split] > 1} {#return $error}
		# IPv6's with "::blah" or "blah:blah::" will cause an error here,
		# but the addresses are actually possibly still valid.

debug.print 2 "\[TESTIPV6\] 08: \$IP ($ip)"
		set doublecolon [string first :: $ip]
		set left [split [left $ip $doublecolon] :]
		set right [split [right $ip [ expr -1 * ( $doublecolon + 2 ) ]] :]
		# 2 = [len "::"]
debug.print 2 "\[TESTIPV6\] 09: \$IP ($ip)"

		array set new [list 1 0 2 0 3 0 4 0 5 0 6 0 7 0 8 0]
		set count 0; foreach a $left {incr count; set new($count) $a}
		set count [ expr 8 - [ llength $right ] ]; foreach a $right {incr count; set new($count) $a}

debug.print 2 "\[TESTIPV6\] 10: \$IP ($ip)"
		set ip ""
		for {set i 1} {$i < 9} {incr i} {lappend ip [stu [right 0000$new($i) 4]]}
debug.print 2 "\[TESTIPV6\] 11: \$IP ($ip)"
		set ip [join $ip ":"]
debug.print 2 "\[TESTIPV6\] 12: \$IP ($ip)"
		debug.print 2 "\[TESTIPV6\] final fix \$IP ($ip)"
	}
debug.print 2 "\[TESTIPV6\] 13: \$IP ($ip)"


	# By this time, we're in an 8-part IPv6 format [leading zeroes still gone].
	# Let's test this thing ....

	if {[llength [split $ip :]] != 8} {return $error}

	foreach a [split $ip :] {
		set dec [baseconv 16 10 0x$a]
		if {($dec < 0) || ($dec > 65535)} {return 0}
	}

	# If it survived that test, then it's an 8-part, 0-$FFFF address. Passed.
debug.print 2 "\[TESTIPV6\] final: \$IP ($ip)"

	return 1
}

interp alias {} SB:get_html {} www
interp alias {} SB:get_headers {} www:headers
proc SB:get_html_useragent args { error "\[SB:GET_HTML_USERAGENT\] Deprecated; WWW handles this internally now." }

proc www:var { url { query "" } { var_error "" } { var_html "" } { var_code "" } { var_headers "" } } {
	if [string eq * $query] { set query "" }
	foreach var [info vars var_*] {
		if [notempty $var] { upvar 1 $var [mid $var 5] }
	}
#	upvar 1 $var_error error $var_html html $var_code code $var_headers headers
	lassign [www $url $query] error html code headers
	list $error $html $code $headers
}

proc www { url { query "" } } {
	set http_package_error [ catch { package require http } ]
	if $http_package_error { error "\[WWW\] HTTP package is unavailable" }
	set code 0
	set error [catch {
		#::http::config -useragent "Mozilla/4.76 (Windows NT 8.1; U) Opera 30.0 \[en\]"; # Google hates the default of "Tcl http client package 2.4."
		::http::config -useragent "Mozilla/5.0 (Windows NT 6.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/49.0.2623.112 Safari/537.36 OPR/36.0.2130.80"
		# Allow it 10 seconds to connect
		if [isempty query] {
			set token [::http::geturl $url -timeout 10000 -headers {Accept-Language: en-us,en;q=0.5}]
		} {
			empty q
			foreach {a b} $query { if [notempty q] { append q & } ; append q [::http::formatQuery $a]=[::http::formatQuery $b] }
			if ![isnum -even [llength $query]] { error "\[WWW\] Query (POST) variables must be in even pairs: name-then-definition (failure in: $query)" }
			set token [::http::geturl $url -query $q -timeout 10000 -headers {Accept-Language: en-us,en;q=0.5} ]
		}
		set www [::http::data $token]
		array set temp [array get $token]
		array set temp $temp(meta); # Pull all META data into variables
		set code [lindex $temp(http) 1]
		::http::cleanup $token
	} ohshit]
	if $error { set www "I'm unable to get the information from $url at this time. Please try again later. (Error: $ohshit)" }
#debug error www code temp
	return [list $error $www $code [array get temp]]
}

proc www:headers url {
	set http_package_error [ catch { package require http } ]
	if $http_package_error { error "\[WWW:HEADERS\] HTTP package is unavailable" }
	set token [::http::geturl $url]
	array set temp [array get $token]
	array set temp $temp(meta); # Pull all META data into variables
	::http::cleanup $token
	return [array get temp]
}

proc userlist_ul { { range "1-1001" } { chan "" } } {
	array set userlist ""
	empty list
	foreach user [userlist -b] {lappend userlist([SB:accesslevel $user $chan]) $user}
	foreach level [lsort -real -dec [array names userlist]] {
		if [iswithin $range $level] {set list [concat $list [lsort -inc -uni -dict $userlist($level)]]}
	}
	return $list
}

proc userlist_list { { range "1-1001" } { chan "" } } {
	array set userlist ""
	empty list
	foreach user [userlist -b] {lappend userlist([SB:accesslevel $user $chan]) $user}
	foreach level [lsort -real -dec [array names userlist]] {
		if [iswithin $range $level] {set list [concat $list [list $level [lsort -inc -uni -dict $userlist($level)]]]}
	}
	return $list
}

proc SB:checkuseroutput {handle { mode "" } } {
	set default "notice"

	if [sbdexists CONFIG:GLOBAL-OUTPUT] {set default [sbdget CONFIG:GLOBAL-OUTPUT]}
	if ![validuser $handle] {return $default}
	set output [getdata $handle OUTPUT]
	if {$output == ""} {set output $default}
	if {$mode == ""} {return $output}
	# Next line depends on following values: "NOTice DOUble MSG CHAn (#chan)"
	return [string eq -nocase [left $output 3] [left $mode 3]]
}

proc nodup { list { nocase 0 } } {# $B = 0/1: NOCASE?
	empty nodup
	foreach check $list {
		if $nocase {
			set m [lsearch -glob [stl $nodup] [stl $check]]
		} {
			set m [lsearch -glob $nodup $check]
		}
		if {$m < 0} {lappend nodup $check }
	}
	return $nodup
}

proc nodup_last {list { nocase 0 } } {lreverse [nodup [lreverse $list] $nocase]}

proc `timeval {a {b "m"}} {
	   set valid "ltnsmhdwfoye"
	   if {$a == "check"} {return "$valid"}
	   if [string match *.*.* $a] {return -1}
	   if ![regexp -nocase -- {^\+(\d+(\.\d+)?[ltnsmhdwfoye])+$} $a] {return -1}
	   set times [regexp -inline -all -- {\d+(\.\d+)?[ltnsmhdwfoye]} [mid $a 2]]
	   zero time ; # Default time in seconds should the parsing fail.

	   foreach {v t} {
			 l .001
			 n .01
			 t .1
			 s 1
			 m 60
			 h 3600
			 d 86400
			 w 604800
			 f 1209600
			 o 2592000
			 y 31536000
			 e 315360000
	   } {set values($v) $t}

	   foreach {calculate null} $times {
		   set a [string trimleft $calculate +]
		   set junk ""
		   array set values ""

		   for {set i 0} {$i < [string length $a]} {incr i} {
				 set char [stl [string index $a $i]]
				 if [string match "*$char*" "0123456789 .e"] {
					append junk $char
				 } elseif [string match *$char* $valid] {
					    if [isempty junk] {return -1}
					    set mult 0
					    switch -exact -- $char {
							  l - n - t - s - m - h - d - w - f - o - y - e {set mult $values($char)}
							  default {return -1}
					    }

					    # Maximum value 32-bit TCL can handle: 2,147,483,647
					    # Equivelant value of: 68y5w3h14m7s
					    set time [expr $time + ($junk * $mult)]
					    set junk ""
					    set mult 0
				 } else {
					    return "-1"
				 }
		   }
	   }

	   if {$time > 0} {
			 switch -exact -- $b {
				    l - n - t - s - m - h - d - w - f - o - y - e {
						  set dec [lindex "." [string match *.* $time]]
						  set time [expr $time$dec / $values($b)]
				    }
			 }

	   }; # Default to minutes, not seconds (due to issues with BAN).
	   if [string match *.0 $time] {set time [expr int($time)]}
	   return $time
}

proc timeval { value { convert_to "m" } } {
	array set mult [list l .001 n .01 t .1 s 1 m 60 h 3600 d 86400 w 604800 y 31536000 e 315360000 c 3153600000]
	empty time marker
	if ![regexp -nocase -- {^([\+\-]*[\d\.]+[ceywdhmstnl]){1,}$} $value] return
	if [isempty marker] { set marker s }
	set timeval 0
	foreach regexp [regexp -inline -all -nocase -- {[\+\-]*[\d\.]+[lntsmhdwyec]} $value] {
		regexp -nocase -- {^([\+\-]?)[+-]*([\d\.]+)([lntsmhdwyec]?)$} $regexp - sign time marker
		if [isempty time] return
		if ![info exists mult($marker)] return
		set timeval [expr $timeval + ${sign}$time * $mult($marker)]
	}
	if [notempty convert_to] {
		if ![info exists mult($convert_to)] return
		set timeval [expr ( ${timeval} / 1.0 ) / $mult($convert_to)]
	}
	normalize $timeval
}

# URT is the opposite of TIMEVAL, returning "1d12o5d..." for any # of seconds.

# Replaced by ::NUMBER:EVAL
proc `numeval args {# Version 2
	# Information on SI prefixes taken from:
	# http://guymal.com/techCorner/powers.shtml

	set decimal [list p -51 0.<50>1 q -48 0.<47>1 r -45 0.<44>1 s -42 0.<41>1 t -39 0.<38>1 u -36 0.<35>1 v -33 0.<32>1 w -30 0.<29>1 x -27 0.<26>1 y -24 0.<23>1 z -21 0.<20>1 a -18 0.<17>1 f -15 0.<14>1 p -12 0.<11>1 n -9 0.<8>1 u -6 0.<5>1 m -3 0.<2>1 B 0 1 K 3 1<3> M 6 1<6> G 9 1<9> T 12 1<12> P 15 1<15> E 18 1<18> Z 21 1<21> Y 24 1<24> X 27 1<27> W 30 1<30> V 33 1<33> U 36 1<36> S 39 1<39> R 42 1<42> Q 45 1<45> P 48 1<48> ? 51 1<51>]

	set binary {
		p -51 0.<51>66819117752304891153513411678787046970379922002626217449048437304009966024678258966762456338983611203730106353759765625 q -48 0.<48>6842277657836020854119773355907793609766904013068924666782559979930620520927053718196475529111921787261962890625 r -45 0.<45>700649232162408535461864791644958065640130970938257885878534141944895541342930300743319094181060791015625 s -42 0.<42>71746481373430634031294954664443705921549411424077607513961896135157303433516062796115875244140625 t -39 0.<39>7346839692639296924804603357639035486366659729825547009429698164240107871592044830322265625 u -36 0.<33>752316384526264005099991383822237233803945956334136013765601092018187046051025390625 v -33 0.<33>77037197775489434122239117703397092741524065928615527809597551822662353515625 w -30 0.<30>7888609052210118054117285652827862296732064351090230047702789306640625
		x -27 0.<27>807793566946316088741610050849573099185363389551639556884765625 y -24 0.<24>82718061255302767487140869206996285356581211090087890625 z -21 0.<21>8470329472543003390683225006796419620513916015625 a -18 0.<18>867361737988403547205962240695953369140625 f -15 0.<15>88817841970012523233890533447265625 p -12 0.<12>9094947017729282379150390625 n -9 0.<9>931322574615478515625 u -6 0.<6>95367431640625 m -3 0.<3>9765625
		B 0 1 K 3 1024 M 6 1048576 G 9 1073741824 T 12 1099511627776 P 15 1125899906842624 E 18 1152921504606846976 Z 21 1180591620717411303424 Y 24 1208925819614629174706176 X 27 1237940039285380274899124224 W 30 1267650600228229401496703205376 V 33 1298074214633706907132624082305024 U 36 1329227995784915872903807060280344576 S 39 1361129467683753853853498429727072845824 R 42 1393796574908163946345982392040522594123776 Q 45 1427247692705959881058285969449495136382746624 P 48 1461501637330902918203684832716283019655932542976 ? 51 1496577676626844588240573268701473812127674924007424
	}

	SB:getflags $args [list -dec2dec -dec2bin -bin2dec -bin2bin] text flags
	set flags [le $flags] ; # Take last value specified
	if [isempty flags] { set flags -dec2dec }
	lassign $text number factor_to

	empty factor_from
	regexp -nocase -- {([A-Z])$} $number -> factor_from
	if [notempty factor_from] {
		set number [left $number -1]
		if [left $flags 4 -dec] { set var_from $decimal } { set var_from $binary }

		# Only calculate this if needed
		foreach {a b c} [regsub -all {\t|\n} $var_from ""] {
			if [string eq $a $factor_from] {
				zero zero
				regexp -- {<(\d{1,2})>} $c -> zero
				if $zero {regsub <${zero}> $c [string repeat 0 $zero] c}
				set number [::stringmath::mul $number $c]
				break
			}
		}
	}

	if [isempty factor_to] { set factor_to B }
	if [right $flags 3 dec] { set var_to $decimal } { set var_to $binary }

	# Only calculate this if needed
	foreach {a b c} [regsub -all {\t|\n} $var_to ""] {
		if [string eq $a $factor_to] {
			zero zero
			regexp -- {<(\d{1,2})>} $c -> zero
			if $zero {regsub <${zero}> $c [string repeat 0 $zero] c}
			set number [::stringmath::div $number $c]
			break
		}
	}
	return $number
}

proc SB:getflags { text validflags { var_text "" } { var_flags "" } } {
	empty flags

	while {[string eq [string index $text 0] -]} {
		set f [lindex $text 0]
		set text [lreplace $text 0 0]
		if [string eq $f --] break
		lappend flags $f
	}

	# ----- Version 2 -----
	if [string eq -nocase * $validflags] {
		set v $flags
	} {
		empty v
		foreach a $flags {
			set u [uniquematch $validflags $a]
			if [notempty u] { lappend v $u }
		}
	}

	if [notempty var_text] { upvar 1 $var_text lt ; set lt $text }
	if [notempty var_flags] { upvar 1 $var_flags lf ; set lf $v }
	return [list $text $v]
}

# Enter with: [LIST "+a target" "+n" "+q target"]
proc SB:combine_modes { modes { cutoff 4 } } {
#log "\00311,02\[SB:COMBINE_MODES\] MODES($modes)\003"
	if {$cutoff < 0} {zero cutoff}
	if !($cutoff) {
		set m [sbd:nozero @NETWORK:SERVER:MODES]
		if $m {
			set cutoff $m
		} else {
			if [info exists ::max-modes] {
				set cutoff ${::max-modes}
			} else {
				set cutoff 4
			}
		}
	}

	# Remove nulls and "+" created by SB:PARSE_CHANMODES
#!     set modes [ldestroy -exact -all -nonull $modes +]
	set l [lsearch -exact -all $modes +]
	if {$l != -1} { foreach a [lsort -int -decr $l] { set modes [lreplace $modes $a $a] } }

	# Which modes "require" extra parameters?
	set list_param [list b k l psmnti ov]
	# Note: +k can be removed by CHANSERV / U:line / channel-owner (on some IRCds) without a param

	if [sbd:notempty @NETWORK:SERVER:CHANMODES] {set list_param [concat [sbd:get @NETWORK:SERVER:CHANMODES] [sbd:get @NETWORK:SERVER:MODES:CHANOP]]}

	if 0 {
		# Who needs multiple cell values?
		empty multiple
		emptyarray cells
		foreach a [concat [alphabet lower split] [alphabet upper split]] {
			empty cell($a)
			if {[lsearch -exact [concat [l0 $list_param] f j [l4 $list_param]] $a] > -1} {append multiple $a}
		}
	}

	# Combine like signs and modes first (reduce how often we change polarity)
	#!set modes [lsort -inc -dict $modes]; # Do =NOT= use -UNIQUE !!
	#! DO NOT SORT! Will change the order of " -l+l 15 " - like sequences!

	empty combined cell sign extra
	zero counter
	do until {[isempty modes]} {
		set param [l0 $modes]
		lassign [l0 $modes] param0 param1
		if ![regexp -- {^[\+\-]} $param0] {prepend param0 +}
		if ![regexp -- {^[\+\-][a-zA-Z0-9\*][ ]?} $param0] {error "\[SB:COMBINE_MODES\] Unknown channel mode: $param0"}
		if ![string eq $sign [left $param0 1]] {
			set sign [left $param0 1]
			append cell $sign
		} {
			set param [right $param0 1]
		}
		set mode [right $param0 1]

		incr counter
		append cell $mode

		# Which needs an extra param?
		zero take
		if [instr [l0 $list_param] $mode] {one take}
		if [instr [l1 $list_param] $mode] {one take}
		if [instr [l2 $list_param] $mode] {set take [string eq $sign +]}
		if [instr [l3 $list_param] $mode] {zero take}
		if [instr [l4 $list_param] $mode] {one take}
#log "\00311,02\[SB:COMBINE_MODES\] LIST_PARAM($list_param):MODE($mode):TAKE($take):MODES($modes):PARAM0($param0):PARAM1($param1)\003"
		if $take {
			if [isempty param1] {error "\[SB:COMBINE_MODES\] Missing param for: $mode"}
			lappend extra $param1
		}; # If the param is present but not needed: quietly dump it

		if 0 {
			if [instr $multiple $mode] {
				if $take {lappend cells($mode) [list $sign $param1]} {lappend cells($mode) $sign}
			} {
				if $take {set cells($mode) [list $sign $param1]} {set cells($mode) $sign}
			}
		}

		if {$counter >= $cutoff} {
			lappend combined [trim "$cell [join $extra]"]
			set cell +
			empty extra
			zero counter
		}
		set modes [lreplace $modes 0 0]
	}
	if [notempty cell] {lappend combined [trim "$cell [join $extra]"]}
	return $combined
}

proc SB:shell_whereis cmd {
	set error [catch { set whereis [exec whereis $cmd] } null]
	if $error {return ""}
	if [string eq ${cmd}: [l0 $whereis]] {set whereis [lreplace $whereis 0 0]}
	return [l0 $whereis]
}

proc SB:shell_which cmd {
	set error [catch {set which [l0 [exec which [stl $cmd]]]} null]
	if $error {return ""}
	return $which
}

proc SB:shell_findcmd cmd {lindex [concat [SB:shell_which $cmd] [SB:shell_whereis $cmd]] 0}

proc SB:synonym {list item valid} {
	set m [lsearch -exact [stl $list] [stl $item]]
	if {$m < 0} {return $item} {return $valid}
}

proc synonym args {
	# Enter with: synonym [-nocase -glob -regexp -exact] <what to reply> <list of valid options> <user choice>
	if [isempty args] return
	flags:simple $args [list -nocase -glob -exact -regexp] temp flags
	if [isempty temp] return
	if ![validflag -glob -exact -regexp] { lappend flags -glob }
	flags:priority mode -glob -exact -regexp
	lassign $temp reply valid choice
	set case "st[lindex [list - l] [validflag -nocase]]"
	lappend list $reply ; # Shouldn't have to manually add it;.
	if { [lsearch -[mid $mode 2] [$case $valid] [$case $choice]] != -1 } { return $reply }
	return $choice
}

interp alias {} REM {} rem
proc rem args #:)

proc SB:isglobaltrigger trigger {
	empty list
	foreach level [list friend voice halfop op tech master owner] {set list [concat $list [sbdget SHORTCUT:$level]]}
	if {[lsearch -exact [string tolower $list] [string tolower $trigger]] < 0} {return 0} {return 1}
}

proc SB:namespaces_list args {
	set match [none $args *]
	empty namespaces
	set temp [namespace children :: *]
	while {[notempty temp]}  {
		set namespace [lindex $temp 0]
		set temp [lreplace $temp 0 0]
		if [string match -nocase $match $namespace] {lappend namespaces $namespace}
		set temp [concat $temp [namespace children $namespace *]]
	}
	return $namespaces
}

proc SB:match_components args {
	flags:simple $args [list -private -force] args flags
	if [isempty args] {return ""}
	empty return
	# Override!
	lappend flags -private
	foreach arg $args {
		if ![iswildcard $arg] {
			if [string eq -nocase CORE $arg] {set arg "./scripts/sb6.tcl"}
			if [string eq -nocase GPL $arg] {set arg "./scripts/sb6_gpl_v3_license.txt"}
			if [string eq -nocase license $arg] {set arg "./scripts/sb6_gpl_v3_license.txt"}
			if [string eq -nocase licence $arg] {set arg "./scripts/sb6_gpl_v3_license.txt"}
			if [string eq -nocase readme $arg] {set arg "./scripts/sb6_readme.txt"}
			if [string eq -nocase defs $arg] {set arg "./scripts/sb6_defs.tcl"}
			if [string eq -nocase rename $arg] {set arg "./scripts/sb6_defs_rename_me.tcl"}
			if [string eq -nocase rename_me $arg] {set arg "./scripts/sb6_defs_rename_me.tcl"}

			if { ( [file exists $arg] || [validflag -force] ) && ![file isdirectory $arg]} {lappend return $arg; continue}
			if [validflag -force] {
				set temp_dir_scripts ./scripts
				set temp_dir_beads ./scripts/beads
				if [validflag -private] { set temp_dir_beads ./scripts/beads/[string tolower $::nick] }
				if [string match sb6*.tcl $arg] { lappend return ${temp_dir_scripts}/$arg ; continue }
				if [string match sb6*.txt $arg] { lappend return ${temp_dir_scripts}/$arg ; continue }
				if [string match *.sb6 $arg] { lappend return ${temp_dir_beads}/$arg ; continue }
			}
			
			if ![string eq . [file dirname $arg]] {lappend return $arg; continue}
			if [validflag -private] {
				if [file exists ./scripts/beads/[stl $::nick]/${arg}] {set arg "./scripts/beads/[stl $::nick]/${arg}"}
				if [file exists ./scripts/beads/[stl $::nick]/${arg}.sb6] {set arg "./scripts/beads/[stl $::nick]/${arg}.sb6"}
			}
			if [file exists ./scripts/beads/${arg}] {set arg "./scripts/beads/${arg}"}
			if [file exists ./scripts/beads/${arg}.sb6] {set arg "./scripts/beads/${arg}.sb6"}
			if [file exists ./scripts/sb6_${arg}.tcl] {set arg "./scripts/sb6_${arg}.tcl"}
			if [file exists ./scripts/sb6_${arg}.txt] {set arg "./scripts/sb6_${arg}.txt"}
			if [file exists ./scripts/${arg}.tcl] {set arg "./scripts/${arg}.tcl"}
			if [file exists ./scripts/${arg}] {set arg "./scripts/${arg}"}
			if [string eq . [file dirname $arg]] {lappend return ./$arg; continue}
		}
		lappend return $arg
	}
	return $return
}

proc SB:short_component args { # NOt sure what this is even for ....
	flags:simple $args -private filename flags
	# What will this do?
	set is_path [string eq [list scripts beads] [lrange [file split [file dirname [file normalize $a]]] end-1 end]]
	set tail [file tail $filename]
	if [regexp -- {^(core|sb6(\.tcl)?)$} $tail] { return core }
	if [validflag -private] { if [file exists ./scripts/beads/[stl $::nick]/${tail}.sb6] { return $tail } }
	if [file exists ./scripts/beads/${tail}] { return $tail }
	if [file exists ./scripts/beads/${tail}.sb6] { return $tail }
	if [file exists ./scripts/${tail}] { return $tail }
	if [file exists ./scripts/${tail}.tcl] { return $tail }
	if [file exists ./scripts/sb6_${tail}.tcl] { return $tail }
	if [file exists ./scripts/${tail}.txt] { return $tail }
	if [file exists ./scripts/sb6_${tail}.txt] { return $tail }
	return $filename
}

proc SB:is_component [list filename [list expand_to_full_path false]] {
	zero ok
	if {
		[string match -nocase CORE $filename]
		|| [string match -nocase sb6*.tcl $filename]
		|| [string match -nocase sb6*.txt $filename]
		|| [string match -nocase *.sb6 $filename]
	} { 
		if $expand_to_full_path { return [SB:match_components $filename] } { return 1 }
	}
	return 0
}

proc SB:target_component [list file [list include_tail true]] {
	set tail [file tail $file]
	if [string match -nocase *.sb7 $file] { return ./scripts/beads[iff $include_tail /$tail] }
	if [string match -nocase *.sb6 $file] { return ./scripts/beads[iff $include_tail /$tail] }
	if [string match -nocase *.sbm $file] { return ./scripts/beads[iff $include_tail /$tail] }
	if [string match -nocase sb6*.tcl $file] { return ./scripts[iff $include_tail /$tail] }
	if [string match -nocase sb6.txt $file] { return ./scripts[iff $include_tail /$tail] }
	return .[iff $include_tail /$tail]
}

interp alias {} faf {} SB:findallfiles
proc SB:findallfiles { { findfile "*" } { dirs . } } { rglob $dirs $findfile }

proc `rglob {dirpath patterns { recursive 0 } { sizes 0 } } {
	# Custom-written by thommey [1347803875] :: http://paste.tclhelp.net/?id=bn7

	set rlist {}
	if [isempty patterns] { set patterns * }
	catch { foreach pattern $patterns { set rlist [concat $rlist [glob -nocomplain -types f -directory ${dirpath} -- $pattern]] } }
	catch { foreach pattern $patterns { set rlist [concat $rlist [glob -nocomplain -types [list f hidden] -directory ${dirpath} -- $pattern]] } }

	if !$recursive {
		if !$sizes { return [lsort -inc $rlist] }
		float total
		foreach a $rlist { add total [file size $a] }
		return [list [lsort -inc $rlist] $total]
	}

	# Note the recursive call to RGLOB
	# Recurse into directories
	if [notempty dirpath] {
		set dirs [concat [glob -nocomplain -types d -directory ${dirpath} -- *] [glob -nocomplain -types [list d hidden] -directory ${dirpath} -- *]]
		lremove -all -glob dirs . .. */. */..
		foreach dir $dirs { catch { set rlist [concat $rlist [rglob ${dir} ${patterns} $recursive]] } }
	}

	if $sizes {
		float total; # MUST BE FLOATING-POINT (stupid 32-bit issues)
		zero c
		set ll [llength $rlist]
		foreach file $rlist {
			incr c
			if ![file isdirectory $file] { catch { add total [file size $file] } }
		}
		return [list $rlist $total [array get size_list]]
	}
	return ${rlist}
}

proc `dglob { { dirpath . } { recursive true } } {
	# Custom-written by thommey [1347803875] :: http://paste.tclhelp.net/?id=bn7
	# Modified by ... me ....

	set dlist {}
	catch { set dlist [concat $dlist [glob -nocomplain -types d -directory ${dirpath} -- *]] }
	catch { set dlist [concat $dlist [glob -nocomplain -types {d hidden} -directory ${dirpath} -- *]] }
	lremove -all -glob dlist . .. */. */..
	if !($recursive) { return $dlist }

	# Note the recursive call to DGLOB
	# Recurse into directories
	if $recursive {
		set dirs [concat [glob -nocomplain -types d -directory ${dirpath} -- *] [glob -nocomplain -types [list d hidden] -directory ${dirpath} -- *]]
		lremove -all -glob dirs . .. */. */..
		foreach dir $dirs { catch { set dlist [concat $dlist [dglob ${dir} $recursive]] } }
	}
	return ${dlist}
}

proc fglob { { dir . } { match * } { recurse false } } { aglob file $dir $match $recurse }
proc rglob { { dir . } { match * } { recurse false } } { aglob file $dir $match $recurse }
proc dglob { { dir . } { match * } { recurse false } } { aglob dirs $dir $match $recurse }

proc aglob { { type files } { dir . } { match * } { recurse false } } {
	switch -exact -- [string tolower [string index $type 0]] {
		f { set types f }
		d { set types d }
		default { error "\[AGLOB\] Unknown option: $type" }
	}
	set aglob ""
	set aglob1 [glob -nocomplain -types $types -directory $dir -- $match]
	set aglob2 [glob -nocomplain -types [list $types hidden] -directory $dir -- $match]
	set aglob [aglob:clean [concat $aglob1 $aglob2]]
	if !$recurse { return $aglob }
	set dirs1 [glob -nocomplain -types d -directory $dir -- *] ; # "*" not $MATCH
	set dirs2 [glob -nocomplain -types [list d hidden] -directory $dir -- *] ; # "*" not $MATCH
	set dirs [aglob:clean [concat $dirs1 $dirs2]]
	foreach dir $dirs { set aglob [concat $aglob [aglob $type $dir $match $recurse]] }
	return $aglob
}

proc aglob:clean { list { clean "" } } {
	if [string eq "" $clean] { set clean [list . .. */. */..] }
	foreach item $clean {
		set m [lsearch -glob -all $list $item]
		if ![string eq "" $m] {
			foreach n [lsort -integer -decreasing $m] { set list [lreplace $list $n $n] }
		}
	}
	return $list
}

# Only used by BOTNET / SENDSC / STATUS
proc SB:list_triggers { { matchme "" } } {
	array set triggers ""
	empty list_triggers
	set binds [getthird [binds SB:dispatch_chan]]
	foreach a [lsort:list [regsub -all -nocase -- CONFIG:GLOBAL-COMMANDBINDS: [sbd:find CONFIG:GLOBAL-COMMANDBINDS:*] ""] [list owner master tech op halfop voice global all]] {
		set triggers($a) [sbd:get CONFIG:GLOBAL-COMMANDBINDS:$a]
		set list_triggers [concat $list_triggers $triggers($a)]
		set binds [ldestroy -all -nocase -exact -multiple -- $binds [sbd:get CONFIG:GLOBAL-COMMANDBINDS:$a]]
	}
	if [notempty binds] {
		set triggers(default) $binds
		set list_triggers [concat $list_triggers $binds]

	}
	if [notempty matchme] { if { [lsearch -exact $list_triggers $matchme] > -1 } { return 1 } { return 0 } }
	foreach a [array names triggers] {set triggers($a) [lsort -inc -uni -dict $triggers($a)]}
	return [array get triggers]
}

proc SB:matchhost { host { channels "" } } {
	empty users
	if [string eq "" $channels] {set channels [channels]}
	foreach channel $channels {
		set chanlist [chanlist $channel]
		foreach user $chanlist {
			if [string match -nocase $host ${user}![getchanhost $user $channel]] {lappend users $user}
		}
	}
	return $users
}

proc SB:reset_defaults {list defaults} {
	set list [SB:padindex $list [llength $defaults] ""]
	for {set x 0} {$x < [llength $defaults]} {incr x} {
		lset list $x [none [lindex $list $x] [lindex $defaults $x]]
	}
	return $list
}

proc SB:convert_html_codes {text { stripmode 0 } } {# Formerly: RSS.SBM :: SB:RSS_REMOVE_HTML_CODES
	set old $text

	foreach {before after} [trim {
		&times\\\; &#215\;
		&Oslash\\\; &#216\;
		&Ugrave\\\; &#217\;
		&Uacute\\\; &#218\;
		&Uuml\\\; &#220\;
		&Ucirc\\\; &#219\;
		&Yacute\\\; &#221\;
		&THORN\\\; &#222\;
		&szlig\\\; &#223\;
		&agrave\\\; &#224\;
		&aacute\\\; &#225\;
		&acirc\\\; &#226\;
		&atilde\\\; &#227\;
		&auml\\\; &#228\;
		&aelig\\\; &#230\;
		&aring\\\; &#229\;
		&ccedil\\\; &#231\;
		&egrave\\\; &#232\;
		&eacute\\\; &#233\;
		&ecirc\\\; &#234\;
		&euml\\\; &#235\;
		&igrave\\\; &#236\;
		&iacute\\\; &#237\;
		&icirc\\\; &#238\;
		&eth\\\; &#240\;
		&iuml\\\; &#239\;
		&lt\\\; &#60\;
		&ograve\\\; &#242\;
		&iexcl\\\; &#161\;
		&gt\\\; &#62\;
		&oacute\\\; &#243\;
		&cent\\\; &#162\;
		&ocirc\\\; &#244\;
		&pound\\\; &#163\;
		&otilde\\\; &#245\;
		&curren\\\; &#164\;
		&ouml\\\; &#246\;
		&yen\\\; &#165\;
		&divide\\\; &#247\;
		&brvbar\\\; &#166\;
		&brkbar\\\; &#166\;
		&oslash\\\; &#248\;
		&sect\\\; &#167\;
		&uacute\\\; &#250\;
		&ugrave\\\; &#249\;
		&uml\\\; &#168\;
		&die\\\; &#168\;
		&ucirc\\\; &#251\;
		&ordf\\\; &#170\;
		&copy\\\; &#169\;
		&uuml\\\; &#252\;
		&laquo\\\; &#171\;
		&yacute\\\; &#253\;
		&not\\\; &#172\;
		&thorn\\\; &#254\;
		&shy\\\; &#173\;
		&yuml\\\; &#255\;
		&reg\\\; &#174\;
		&macr\\\; &#175\;
		&hibar\\\; &#175\;
		&deg\\\; &#176\;
		&plusmn\\\; &#177\;
		&acute\\\; &#180\;
		&micro\\\; &#181\;
		&para\\\; &#182\;
		&middot\\\; &#183\;
		&cedil\\\; &#184\;
		&ordm\\\; &#186\;
		&raquo\\\; &#187\;
		&Egrave\\\; &#200\;
		&iquest\\\; &#191\;
		&Eacute\\\; &#201\;
		&Agrave\\\; &#192\;
		&Ecicr\\\; &#202\;
		&Aacute\\\; &#193\;
		&Euml\\\; &#203\;
		&Acirc\\\; &#194\;
		&Igrave\\\; &#204\;
		&Atilde\\\; &#195\;
		&Iacute\\\; &#205\;
		&Auml\\\; &#196\;
		&Icirc\\\; &#206\;
		&Aring\\\; &#197\;
		&Iuml\\\; &#207\;
		&Aelig\\\; &#198\;
		&ETH\\\; &#208\;
		&Dstrok\\\; &#208\;
		&Ccedil\\\; &#199\;
		&Ntilde\\\; &#209\;
		&Ograve\\\; &#210\;
		&Oacute\\\; &#211\;
		&Ocirc\\\; &#212\;
		&Otilde\\\; &#213\;
		&Ouml\\\; &#214\;
	}] {
		regsub -all $before $text \\$after text
		# \\ needed because of &.
	}

	for {set x 0} {$x < 256} {incr x} {
		set replace [binary format c $x]
		if {$x == 38} {prepend replace \\}
		regsub -all &#${x}\\\; $text $replace text
	}

	# Do this separately #
	regsub -all &amp\; $text \\\& text

	array set list ""
	set list(0) [list &amp\\\; \\\046 &#038\\\; \\\046]
	set list(1) [list \\\u2018 ' \\\u2019 ' &#8212\\\; - &#8216\\\; ']
	set list(2) [list &#8217\\\; ' &#8211\\\; - &#8216\\\; ' &#8220\\\; \"]
	set list(3) [list &#8221\\\; \" &#8243\\\; \" &#8230\\\; \.\.\. &hellip\\\; \.\.\.]
	set list(4) [list &#0133\\\; \.\.\. &#8217\\\; ' &nbsp\\\; " " &quot\\\; \"]
	set list(5) [list &ndash\\\; - &endash\\\; - &mdash\\\; - &emdash\\\; - &iquest\\\; Â¿]
	set list(6) [list &iexcl\\\; Â¡ &cent\\\; Â¢ &pound\\\; Â£  &curr\\\; Â¤ &yen\\\; Â¥]]
	set list(7) [list \\\u2014 -- \\\u2019 ' \\\u2018 ' &#8212\\\; -- \\\u2013 -]

	foreach x [lsort -int -inc [array names list]] {
		set item $list($x)
		foreach {before after} $item {regsub -all -- $before $text $after text}
	}
	regsub -all \xE2\x80\x99 $text ' text
	regsub -all \x81\x66 $text ' text

	while {$stripmode} {
		foreach regexp [regexp -inline -nocase -all -- {<.+?[\/]?>} $text] {
			regsub -- $regexp $text " " text
		}
		decr stripmode
	}

	foreach {a b} [list &lt\; < &gt\; >] {regsub -all $a $text $b text}

	foreach x [regexp -all -inline -nocase -- {&#x\d+?\;} $text] {
		set dec [dec 0[mid $x 3 -1]]
		regsub -all $x $text [binary format c* $dec] text
	}

	return $text
}

proc SB:parse_chanmodes {chan { include_beI 0 } } {
	set list ""
	set chanmodes [string trimleft [getchanmode $chan] +]; # Truncate the leading "+"
	set marker 1
	if ![string eq 0 $include_beI] {one include_beI}
	set exclude_beI [not $include_beI]; # EXCLUDE needed for LRANGE, below

	if !($exclude_beI) {
		# Add +beI to chunk #0, then LAPPEND, as appropriate, to entire LIST
		foreach {mode type} [list b ban e exempt I invite] {
			set list [chan${type}s $chan]
			if [notempty list] {
				foreach item $list {
					lset chanmodes 0 [l0 $chanmodes]$mode
					lassign $item mask who elapsed
					lappend chanmodes $mode $mask
				}
			}
		}
	}
	set need_params [join [sbd:lrange @NETWORK:SERVER:CHANMODES $exclude_beI 2] ""]

	foreach a [split [l0 $chanmodes] ""] {
		set extra ""
		if {[string first $a $need_params] > -1} {
			set extra [lindex $chanmodes $marker]
			incr marker
		}
debug.print 2 "\[SB:PARSE_CHANMODES\] \$A ($a), \$EXTRA ($extra), \$LIST ($list)"
		lappend list [list $a $extra]; # See the space? :)
	}

	return [join $list]
}

proc SB:url_encode { text { skipme "" } } {
	empty encoded
	for {set x 0} {$x < [len $text]} {incr x} {
		set char [string index $text $x]
		if [instr $skipme $char] {
			append encoded $char
		} {
			if [regexp -nocase -- {[A-Z0-9]} $char] {append encoded $char} {binary scan $char H* char2; append encoded %$char2}
		}
	}
	return $encoded
}

proc SB:url_decode encoded {
	foreach a [regexp -inline -all -nocase -- {%..} $encoded] {
		set replace [binary format H* [mid $a 2]]
		if [string eq %26 $a] {prepend replace \\}
		regsub -all -- $a $encoded $replace encoded
	}
	return $encoded
}

proc do args {
	# Code by Mai "Domino" Mizuno, inspired by: http://wiki.tcl.tk/3603
	# Coding assistance by Locke Shinseiko

	upvar 1 upvar_control control upvar_condition condition upvar_code code upvar_check check upvar_error_condition error_condition upvar_test_loop test_loop
	set errormsg "Syntax: do ?\[\"count\"\] <integer: count>|time <milliseconds>|while <condition>|until <condition>|require <condition>|unless <condition>|forever? code"
	empty control condition code
	one perform

	# Don't use LASSIGN (new version uses DO, which causes an inf* loop)
	# Don't use ISNUM (it uses SB:SETMISCVARIABLES which, itself, uses DO, which causes an inf* loop)
	set control   [l0 $args]
	set condition [l1 $args]
	set code      [l2 $args]

	set switch [none [uniquematch [list while until if unless count timer forever] $control] $control]

	zero _do_loop _do_counter _do_max _do_timer; # Need for DO <INTEGER>|TIME <milliseconds>

	switch -exact -- $switch {
		while - until #
		if - require {set control while; zero perform}
		unless {set control until; zero perform}
		forever {set control while; set code $condition; one condition}
		count {set control while; one _do_loop; set _do_max $condition; one condition}
		timer {
			set _do_timer [ expr [ clock clicks -milli ].0 + $condition ]
			if { $_do_timer >= [ expr 2147483648.0 * ( [ is64bit ] + 1 ) ]} { error "\[DO\] Requested time will go beyond the [ expr 32 * ( [ is64bit ] + 1 ) ]-bit signed integer limit." }
			set control while; one condition
		}
		default {
			if [regexp -- {^[\+\-]?(\d*\.)?\d+?$} $control] {
				if ![regexp -- {^\d+$} $control] { error "\[DO\] Loop count requires an integer, received: \"${control}\"" }
				if {$control < 1} { return 0 }
				if {$control < 1} { error "\[DO\] Illegal loop counter: $control" }
				# Override with "do nothing" response

				one _do_loop
				set _do_max $control
				set code $condition
				set control while
				set condition 1; # "$_do_counter <= $_do_max"
			} elseif {[isempty code] && [isempty condition]} {
				set code $control
				set control while
				one condition
			} else {
				puts "\[DO\] Syntax:"
				puts "\tDO \[FOREVER\] <code to run>"
				puts "\tDO <WHILE|UNTIL|IF|UNLESS> <loop condition> <code to run>"
				puts "\tDO <loop integer> <code to run>"
				return ""
			}
		}
	}
	if [info exists _do_max] { # Can't use ISNUM here! (see above re: inf*loop)
		if ![isint $_do_max] { error "\[DO\] Loop count requires an integer, received: \"${_do_max}\"" }
		if [expr ( int( $_do_max ) != $_do_max ) ? 1 : 0 ] { error "\[DO\] Loop count requres an integer, received: \"${_do_max}\"" }
	}

	if [isempty code] { error "\[DO\] Missing loop code" }

	while 1 {
		zero error
		if $_do_loop { incr _do_counter }
		if $perform {
			set error [catch {uplevel 1 $code} shit]
			switch -exact -- $error {0 - 4 # 3 return default { return -code $error $shit } }
		}
		one perform; # Allows "UNLESS/REQUIRE" to skip the first iteration

		uplevel 1 { set upvar_check [ catch { set upvar_test_loop [ expr ( $upvar_condition ) ? 1 : 0 ] } upvar_error_condition ] }

		# Check for errors in test condition
		switch -exact -- $check { 0 - 4 # 3 return default { return -code $check $error_condition } }

		# Counter- / timer-controlled loop?
		if $_do_loop { if { $_do_counter >= $_do_max } break }
		if $_do_timer { if { [clock click -milli] >= $_do_timer } break }

		# Loop?
		switch -exact -- ${test_loop}:[string tolower $control] { 0:while break 0:until continue 1:while continue 1:until break default { return -code error $errormsg } }
	}

	return ""; # Will we ever get here? (Shouldn't!)
}

# Syntax: LOOP <variable> ["FROM"] [start] ["TO"] <end/count> ["STEP"] [step count] <CODE>
# Returns: # of loops actually executed ( 1 to 7 step 3 = 3 loops [1, 4, & 7] )
# Loop variable, if not needed, can be assigned to "-" (like REGEXP)

# Replacement PROC below
proc `loop args {
	# Somehow, try to assure: <variable> <start> <end> <step> <code>
	# Positions will change as any terms are removed ....
	if [string eq -nocase FROM [l1 $args]] { set args [lreplace $args 1 1] }
	if [string eq -nocase TO   [l2 $args]] { set args [lreplace $args 2 2] }
	if [string eq -nocase STEP [l2 $args]] {# <end/count> STEP <step> <CODE>
		set args [linsert $args 1 1]
		# Fall-through to next check ....
	}
	if [string eq -nocase STEP [l3 $args]] { set args [lreplace $args 3 3] }

	lassign $args variable 0 1 2 3 4 ; # <variable> <start> <end> <step> <CODE>
	array set loop [list start 0 end 0 step 0 count 0 loops 0 code # incr incr]

	# The LLENGTH also includes $VARIABLE!
	switch -exact -- [llength $args] {

		0 - 1 { error "\[LOOP\] Missing arguments! Format: LOOP \[start \[TO\]\] \[end\] \[\[STEP\] <step value>\] <code> \[loop variable\]" }

		2 { set loop(start) 0 ; set loop(end) 0 ; set loop(step) 0 ; set loop(code) $0 }

		3 { set loop(start) 1 ; set loop(end) $0 ; set loop(step) 1 ; set loop(code) $1 }

		4 { set loop(start) $0 ; set loop(end) $1 ; set loop(step) 1 ; set loop(code) $2 }

		5 { set loop(start) $0 ; set loop(end) $1 ; set loop(step) $2 ; set loop(code) $3 }

		default { error "\[LOOP\] Too many arguments! Format: LOOP <variable> \[start \[TO\]\] \[end\] \[\[STEP\] <step value>\] <code>" }

	}
#debug.vars # loop(start) loop(end) loop(step) loop(count) loop(code) loop(loops)

	upvar 1 $variable counter
	if { ![isint $loop(start)] || ![isint $loop(step)] } { set loop(incr) add }
	for { set loop(count) $loop(start) } { $loop(count) <= $loop(end) } { $loop(incr) loop(count) $loop(step) } {
		incr loop(loops)
		set body $loop(code)
		set counter $loop(count)
		set error [ catch { uplevel 1 $body } complaint ]
		switch -exact -- $error { 0 - 4 continue 1 - 3 break default { return -code $error $complaint } }
	}
	return [nozero $loop(loops)]
}

# Replacement PROC
proc loop args {
	if [string eq -nocase HELP [lindex $args 0 0]] { return "loop ?-forever? ?-count count? ?-range start end? ?-step step? ?-variable variable? ?condition-code? loop-code" }
	flags -params $args [list -forever 0 -count 1 -range 2 -step 1 -variable 1] args flags
	set flags_used [validflag -forever -count -range]
	switch -exact $flags_used { 1 # 0 { one flags(-forever) } default { error "Loop: only one of these flags can be used at a time: -forever -count -range" } }

	lassign $args condition code
	if [isempty code] { swap code condition }
	set start -1
	set step 1
	if [notempty flags(-forever)] { lassign [list 1 2 0] start end step }
	if [notempty flags(-count)] {
		if ![isnum -integer $flags(-count)] { error "Loop -count requires an integer value" }
		lassign [list 1 $flags(-count) $step] start end step
	}
	if [notempty flags(-range)] {
		if { [llength $flags(-range)] == 1 } { error "Loop -range requires two parameters: starting value and ending value" }
		foreach a $flags(-range) { if ![isnum $a] { error "Loop -range requires numeric parameters" } }
		lassign $flags(-range) start end
	}
	if [notempty flags(-step)] { set step $flags(-step) } ; # Allow for all options, not just -RANGE
	if { $start == -1 } { error "Loop: Missing one of: -forever -range -count" }

	if [notempty flags(-variable)] { upvar 1 $flags(-variable) _LOOP }

	if [isempty code] return

	for { set _LOOP $start } { $_LOOP <= $end } { set _LOOP [normalize [ expr $_LOOP + $step ]] } {
		# Test the condition FIRST, so if it's patently false, don't execute the loop at all
		if [notempty condition] {
			set error [ catch { set result [uplevel 1 expr ( $condition ) ? 1 : 0] } info ]
			if $error { error "Loop: error in test condition ( $condition ): $info" }
#debug condition result info error
			if !$result break
		}

		# Do loop code
		set test [ catch { uplevel 1 $code } error ]
		switch -exact -- $test { 0 - 4 # 3 return default { return -code $test $error } }

		# Loop back to test condition
	}
}

proc SB:wildcard_to_regexp args {
	set wildcard $args
	set open ^; set close \$;
	empty header footer
	set split [split $wildcard]
	if [string eq -nocase [l0 $wildcard] -SOFT] {empty open close; set split [lreplace $split 0 0]}
	set wildcard [join $split]

	# For starting / terminal "*" we can't use the ".*?" restriction. ):
	if [left  $wildcard 1 *] {set header .*; set wildcard [mid  $wildcard  2]}
	if [right $wildcard 1 *] {set footer .*; set wildcard [left $wildcard -1]}

	return ${open}${header}[string map [list ? . * .*?] $wildcard]${footer}${close}
}

proc SB:regexp_markbad {text valid { errormarker \002 } } {
	# TEXT: text to check
	# VALID: valid chars (REGEXP format)
	# ERRORMARKER: how to indicate an invalid character (e.g.: \002)

	# If text fully matches, return a blank phrase
	set regexp [regexp -- ^${valid}\$ $text]
	if $regexp {return ""}

	# Return a marked string showing what worked and what failed
	empty new
	foreach a [split $text ""] {
		empty code
		if ![regexp -- ^${valid}\$ $a] {set code $errormarker}
		append new ${code}${a}${code}
	}
	return $new
}

proc SB:regexp_markok {text match { errormarker \002 } } {
	# TEXT: text to check
	# VALID: valid chars (REGEXP format)
	# ERRORMARKER: how to indicate an invalid character (e.g.: \002)

	# If text fully matches, return a blank phrase
	set regexp [regexp -- ^${match}\$ $text]
	if $regexp {return ""}

	# Return a marked string showing what worked and what failed
	empty new
	foreach a [split $text ""] {
		empty code
		if [regexp -- ^${match}\$ $a] {set code $errormarker}
		append new ${code}${a}${code}
	}
	return $new
}

interp alias {} SB:sendemail {} SB:send_email
proc SB:send_email message {
	# First off, test if we can even do this:
	set sendmail [SB:shell_findcmd sendmail]
	if [isempty sendmail] {return [list 0 "Unable to find SENDMAIL"]}

	if ![instr $message \n] {set message [join $message \n]}
	if [instr $message |] {regsub -all \\\| $message \n message}

	if ![file isdirectory ./sentmail] {
		file mkdir ./sentmail
		exec chmod 700 ./sentmail
	}

	set filename ./sentmail/email-[now]-[rand 99999].txt
	dumpfile $filename $message 0
	set error [catch {exec cat $filename | $sendmail -t -v} ohshit]
	if $error {
		putcmdlog "\[SB:SENDEMAIL\] Error: $ohshit"
		return [list 0 $ohshit]
	}

	return 1; # RETURN code for "email send successfully" to the calling PROC.
}

proc SB:timezone args {return "[sbd:get BOT:TIMEZONE] (GMT [sbd:get BOT:GMT])"}

proc SB:timenext {next timenext { hops 1 } } {
	# Will allow DECADE since the *nix 32-bit epoch counter goes into 2038 AD
	# No! Do "SB:TIMENEXT YEAR <time> 10" instead (February issue)
	if ![isint $hops] {error "\[SB:TIMENEXT\] ?hops? must be an integer value"}
	if [left $next 1 -] {
		# Let's truncate to allow UNIQUEMATCH to match properly
		set next [mid $next 2 2]
	}
	switch -exact -- [uniquematch [list decades years months weeks days hours minutes seconds] $next] {
		seconds {incr timenext [ expr $hops * 1]}
		minutes {incr timenext [ expr $hops * 60]}
		hours   {incr timenext [ expr $hops * 3600]}
		days    {incr timenext [ expr $hops * 86400]}
		weeks   {incr timenext [ expr $hops * 604800]}
		years   {incr timenext [ expr $hops * 31536000]}
		months  {# Special parsing necessary
			lassign [clock format $timenext -format "%Y %m %d %H %M %S"] year month day hour minute second
			set month [nozero $month]
			incr month $hops

			if {$month < 1} {decr year [ expr abs( int( $month / 12 ) ) ]; set month [ expr abs( int( $month % 12 ) ) ]; if !($month) {decr year; set month 12}}
			if {$month > 12} {incr year [ expr int( $month / 12 ) ]; set month [ expr int( $month % 12 ) ]}

			# illegal day?
			if [ catch { clock scan ${month}/${day}/${year} } ] {
				set null $month
				incr null
				if {$null > 12} {incr year; one null}
				set temp [clock scan "${null}/1/${year} 00:00:00"]
				decr temp; # Back up 1 second
				set day [clock format $temp -format %d]
			}
			set timenext [clock scan "${month}/${day}/${year} ${hour}:${minute}:${second}"]
		}
		default {error "\[SB:TIMENEXT\] illegal value \"${next}\""}
	}
	return $timenext
}

proc SB:regexp_remove {string deleteme { nocase 0 } { replacewith "" } } {
	if [nozero $nocase] {
		set regexp [regexp -all -inline -nocase -- $deleteme $string]
	} {
		set regexp [regexp -all -inline         -- $deleteme $string]
	}

	foreach regsub $regexp {regsub -all -- $regsub $string $replacewith string}
	return $string
}

proc SB:array_search {array value { target "" } } {
	if [isempty array] {return ""}
	upvar 1 $array temp
	if [notempty target] {upvar 1 $target return}
	empty return
	foreach element [lsort -inc -dict [array names temp]] {
		if [string match -nocase $value $temp($element)] {lappend return $element}
	}
	return $return
}

proc is64bit args { expr ( [lsearch -nocase -glob [list x86_64 amd64] [SB:uname m]] != -1 ) ? 1 : 0 }

proc SB:uname [list mode [list force false]] {
	set filename ./uname.txt
	if [string eq -nocase WINDOWS [SB:os]] {
		if [file exists $filename] { file delete $filename }
		return ; # Can't do UNAME on Windrop bots! TRAC #41
	}
	switch -exact -- [string tolower $mode] {
		update {
			if [sbd:boolean @UNAME:ERROR] return ; # Let's not even bother ....
			if [sbd:isempty @UNAME] {
				emptyarray unames
				# If an update is requested, forget the 'old' file .... Force an re-read!
				if { [file exists $filename] && !$force } {
					array set unames [readfile $filename]
					foreach uname [array names unames] { sbd:set @UNAME:$uname $unames($uname) }
				} {
					set error [sbd:set @UNAME:ERROR [ catch { set uname(a) [exec uname -a] } ]]
					if $error return
					foreach u [sequence:asc a .. z] { # alphabet -split -lower also works
						catch {
							sbd:set @uname:$u [set unames($u) [exec uname -$u]]
							debug unames($u)
						}
					}
					sbd:set @uname [array get unames]
					dumpfile $filename [array get unames] 0
				}
			}
			return
		}
		default { return [sbd:get @UNAME:$mode] }
	}
}

proc validhost {user matchme} {
	if ![validuser $user] {return 0}
	set match_nick [instr $matchme !]
	foreach host [getuser $user HOSTS] {
		if !($match_nick) {set host [l1 [split $host !]]}
		if [string match -nocase $host $matchme] { return 1 }
	}
	return 0
}

proc SB:parse_time args {
	set args [join $args]
	if [isint $args] {return $args}
	if [regexp -- {^\d\d[:]?\d\d$} $args] {return [clock scan $args]}
	if [regexp -- {^\d(\d)?[/\-]\d(\d)?([/\-]\d\d(\d\d)?)?$} $args] {return [clock scan $args]}
#     if [regexp -- {^\d\d\d\d[/\-]\d(\d)?[/\-]\d(\d)?$} $args] {set split [split $args /-]; return [clock scan "[l2 $split]/[l0 $split]/[l1 $split]"]}
	set error [ catch { set result [clock scan $args] } ]
	if !($error) {return $result}
	set error [ catch { get:time $args _ - } ]
	if !$error { return $_ }
	return ""
}

proc SB:detect_wan_ip args {
	set ip [SB:get_my_ip]
	if [isempty ip] return ; # Some kind of error occurred

	foreach a [list nat-ip my-ip vhost4 vhost6 listen-addr] {
		if [info exists ::$a] {
			if [string match -nocase *DETECT* [set ::$a]] { set ::$a $ip }
		}
	}
	# What about HOSTNAME? rDNS on the IP? Might be multiple responses ....
	return $ip
}

proc SB:get_my_ip args {
	lassign [SB:get_html http://ip.changeip.com] error temp
	set ip [l0 [split $temp \n]]
	if $error {return ""} {return $ip}
}

proc call args {
	empty return
	set command [jl0 [split $args]]
	if ![string eq "" [info commands $command]] {set return [eval $args]}
	return $return
}

proc SB:chon_process {handle idx} {
	setdccaway $idx "===== Note catcher: $handle ====="
	putdcc $idx "===== I'm marking you as /AWAY in DCC CHAT so you don't miss any notes. If you want to talk on the party line (or any of my internal BOTNET channels), just speak\; the /AWAY will cancel itself. ====="
	if [matchattr $handle vlo] {
		set checkme [list f p]
		empty added
		foreach check $checkme {
			if ![matchattr $handle $check] {append added $check; chattr $handle +$check}
		}
		if [notempty added] {
			putlog "\[CHON\] Automatically adding global +$added to: $handle"
			putidx $idx "\[CHON\] I've given you the global +$added [plural flag [len $added]] because, for some reason, some commands (both Eggdrop core & StormBot.TCL) fail without [plural it [len $added] them]."
			save
		}
	}
	return 0
}

proc SB:casematch args {# All-encompassing attempt to case-match a user
	empty list

	# Setup
	set userlist [userlist]
	set userlist_stl [stl $userlist]
	set bots [bots]
	set bots_stl [stl $bots]
	array set chanlist [list]
	array set chanlist_stl [list]
	foreach chan [channels] {
		set chanlist($chan) [chanlist $chan]
		set chanlist_stl($chan) [stu $chanlist($chan)]
	}

	foreach who $args {
		set stl [stl $who]

		# Start with the USERLIST first ....
		set m [lsearch -exact $userlist_stl $stl]
		if { $m > -1 } { lappend list [lindex $userlist $m] ; continue }

		# Hmmmm. Okay, let's try nicks on the channels ....
		foreach chan [channels] {
			set m [lsearch -exact $chanlist_stl($chan) $stl]
			if { $m > -1 } { lappend list [lindex $chanlist($chan) $m] ; continue }
		}

		# Dammit! Fine, let's try the BOTNET bots next ....
		set m [lsearch -exact $bots_stl $stl]
		if { $m > -1 } { lappend list [lindex $bots $m] ; continue }

		# Shit, I have no idea ....
		lappend list $who
		continue
	}
	return $list
}

proc SB:process_sigterm args {
	set die 1 ; # For SIGTERM, presume death is desired
	if [info exists ::die-on-sigterm] { if !${::die-on-sigterm} { set die 0 } }
	if $die {
		SB:botdatafile save
		save
		foreach a [SB:registerdeath -list] { catch { $a } }
		msghomer "\[SIGTERM\] Shutting down"
		putlog "\[SIGTERM\] Shutting down"
		after 1000
		die "SIGTERM received: shutting down"
		return 1
	}
	set message "\[SIGTERM\] SIGTERM received (config variable not set to die): RESTARTing instead ...."
	msghomer $message
	putlog $message
	after 1000
	restart
	return 1
}
proc SB:process_sigquit args { msghomer "\[SIGQUIT\] RESTARTing"; restart }
proc SB:process_sighup args {
	set die 0
	if [info exists ::die-on-sighup] { if ${::die-on-sighup} { set die 1 } }
	SB:botdatafile save
	save
	if $die {
		foreach a [SB:registerdeath -list] { catch { $a } }
		set message "\[SIGHUP\] SIGHUP received: config variable requires me to DIE ...."
		msghomer $message
		putlog $message
		after 1000
		die "SIGHUP received: shutting down"
		return 1
	}
	msghomer "\[SIGHUP\] REHASHing ...."
	rehash
	return 1
}

proc SB:process_sigill args {
	# Aggressive repair and restoration of the bot's functions ....
	flags:simple $args -debug - flags
	set debug [validflag -debug]
	catch { set home [SB:home] }
putquick "PRIVMSG $home :\[SIGILL\] Aggressive reset in progess ...." -next
	catch { after 1000; # Let msg to HOME be seen }

	# Remove BINDs (so nothing can be triggered)
if $debug { putquick "PRIVMSG $home :\[SIGILL\] Removing BINDs ...." -next }
	foreach bind [binds] { catch { unbind [l0 $bind] [l1 $bind] [l2 $bind] [l4 $bind] } }

	# Clear all queues
if $debug { putquick "PRIVMSG $home :\[SIGILL\] Clearing all output queues ...." -next }
	catch { clearqueue all }

	# Kill all LISTEN ports
if $debug { putquick "PRIVMSG $home :\[SIGILL\] Killing all LISTEN ports ...." -next }
catch { foreach dcc [dcclist telnet] { catch { listen [l1 [l4 $dcc]] off } } }

	# Unlink all bots
if $debug { putquick "PRIVMSG $home :\[SIGILL\] Unlinking all bots ...." -next }
	catch { foreach dcc [dcclist bot] { catch { killdcc [lindex $dcc 0] } } }
	catch { unlink * }

	# Kill all other DCC connections
if $debug { putquick "PRIVMSG $home :\[SIGILL\] Killing all DCC CHAT users ...." -next }
	foreach dcclist [dcclist] {
		set dcc [lindex $dcclist 0]
		set hand ""
		catch { set hand [idx2hand $dcc] }
		if [SB:isbotowner $hand] {
if $debug { putquick "PRIVMSG $home :\[SIGILL\][space 5]Protecting DCC CHAT #$dcc with: [idx2hand $dcc]" -next }
		} {
			catch { killdcc $dcc }
		}
	}

	# Remove timers
if $debug { putquick "PRIVMSG $home :\[SIGILL\] Killing all timers" -next }
	catch { SB:killtimer * quiet }

	# Remove all PROC commands
if $debug { putquick "PRIVMSG $home :\[SIGILL\] Removing all PROCs from memory ...." -next }
	set myproc [lindex [info level [info level]] 0]
	set myproc SB:process_sigill
	catch {
		foreach proc [info procs] {
			if { [lsearch -exact [list rehash return $myproc] $proc] < 0 } { rename $proc "" }
		}
	}

	##### From this point forward, use PUTQUICK only! #####

	# Remove all variables!
if $debug { putquick "PRIVMSG $home :\[SIGILL\] Clearing all NAMESPACEs & variables ...." -next }
	set namespaces [namespace children :: *]

	#lappend namespaces ""
	# [2010-07-18 14:12:12 -0700] <Dumbledore> [14:12] Tcl error [*dcc:tcl]: invalid command name "*dcc:tcl"
	# [2010-07-18 14:12:26 -0700] <Domino> mental note: never delete the global namespace!

	set var ""; set proc ""; set ::stormbot(nocomplain) ""
	foreach namespace $namespaces {
#rawprint "PRIVMSG $home :NAMESPACE($namespace):VARS"
		catch { foreach var [info vars ${namespace}::*] { unset $var } }
#rawprint "PRIVMSG $home :NAMESPACE($namespace):PROCS"
		catch { foreach proc [info procs ${namespace}::*] { rename $proc "" } }
		catch { namespace delete ::$namespace }
	}

#rawprint "PRIVMSG $home :\[SIGILL\] Clearing all global variables!"
	#catch { foreach var [info vars ::*] {unset $var} }

	# --- We should just now have EGGDROP core commands left ---
	putquick "PRIVMSG $home :\[SIGILL\] RESTARTing!" -next

	catch { unset -nocomplain ::stormbot } ; # Not needed (local variables): home var proc

	##### Nothing left but C-coded TCL and EGGDROP core commands #####
	after [ expr ( 5 - ( 4 * $debug ) ) * 1000 ]
	set error [ catch { after 500 restart } ohshit ]
	if $error restart
	return 1
}

proc SB:needsave type {
	set type [SB:synonym [list user userfile save] $type user]
	set type [SB:synonym [list chan chanfile channel channels] $type chan]
	set type [SB:synonym [list data datafile info] $type data]
	set type [SB:synonym [list * all] $type all]
	set type [SB:synonym [list back backup] $type back]
	sbd:lappend @NEEDSAVE $type
	SB:killtimer SB:needsave_timer quiet
	utimer 0 SB:needsave_timer
}

proc SB:needsave_timer args {
	SB:killtimer SB:needsave_timer quiet
	# Don't restore it; only needed for bootstrap!
	#utimer 1 SB:needsave_timer

	if [sbd:isempty @NEEDSAVE] { return 0 }
	if ![boolean [sbd:get @NEEDSAVE:NOSAVE]] {
	# Only need to do one of each type!
		foreach type [stl [lsort -unique [sbd:get @NEEDSAVE]]] {
			switch -exact -- [string tolower [left $type 4]] {
				* - all { SB:botdatafile write ; SB:save ; SB:savechannels }
				data { SB:botdatafile write }
				user { SB:save }
				chan { SB:savechannels }
				back { SB:botdatafile backup }
				default { error "\[SB:NEEDSAVE_TIMER\] Unknown type: $type" }
			}
		}
	}
	sbd:set @NEEDSAVE:NOSAVE
	sbd:set @NEEDSAVE
	return 0
}

proc SB:closeall args { foreach socket [ldestroy -gl -a [file channels] std*] { close $socket } }

proc ago time {duration [ expr [ now ] - $time ]}

proc alphabet args {
	flags:simple $args [list -lower -upper -split -reverse] mode flags
	# Legacy
	foreach a [list l u r s] b [list lower upper reverse split] {
		if { [lsearch -glob $mode ${a}*] != -1 } { lappend flags -$b }
	}
	set alphabet "abcdefghijklmnopqrstuvwxyz"
	if [validflag -lower] { set alphabet [stl $alphabet] }
	if [validflag -upper] { set alphabet [stu $alphabet] }
	if [validflag -reverse] { set alphabet [reverse $alphabet] }
	if [validflag -split] { set alphabet [split $alphabet ""] }
	return $alphabet
}

proc SB:replace_irc_codes text {
	# Catch all the colours:
	array set colorCs [list * </span>]
	array set temp [list * </span> 0 #FFFFFF 1 #000000 2 #000080 3 #008000 4 #FF0000 5 #800000 6 #800080 7 #FF8000 8 #FFFF00 9 #00FF00 10 #008080 11 #00FFFF 12 #0000FF 13 #FF00FF 14 #808080 15 #D0D0D0]
	for {set x 0} {$x < 16} {incr x} {
		for {set y -1} {$y < 16} {incr y} {
			if {$y < 0} {
				set colors($x) "<SPAN STYLE=\" color: $temp($x)\; \">"
			} {
				set colors(${x},${y}) "<SPAN STYLE=\" color: $temp($x)\; background-color: $temp($y)\; \">"
			}
		}
	}

	empty new
	foreach line [split $text \n] {
		# Replace colors first
		set regexp [regexp -inline -all -nocase -- {\003(\d(\d)?(,\d(\d)?)?)?} $line]
		foreach color $regexp {
			if ![left $color 1 \003] continue
			regsub -all -- $color $line $colors([none [mid $color 2] *]) line
		}

		# Now replace bold / underline / reverse codes
		foreach {find replace_on replace_off} [list \037 <U> </u> \002 <B> </b> \035 <I> </i> \017 </u></b></i> </u></b></i>] { ; # 026
			zero bit
			do if {[instr $text $find]} {
				# Don't use -ALL here!
				regsub -- $find $line [lindex [list $replace_on $replace_off] $bit] line
				set bit [not $bit]
			}
		}

		lappend new $line
	}
	set text [join $new \n]
	return $text
}

# Taken from private TCLs

# STRINGMATH will process with a high level of accuracy (good for scientific
# calculations up to 1,000 digits, except when pre-set values within TCL are
# used such as EXP / TAN / COS / SQR). Higher values of accuracy are
# available but at the sacrifice of CPU% / time. STRINGMATH was written and
# used to calculate PI to 1,000 digits in under 3 minutes.

namespace eval ::stringmath {

	proc add:all args { set args [lassign $args acc] ; foreach a $args { set acc [add $acc $a] } ; return $acc }
	proc sub:all args { set args [lassign $args acc] ; foreach a $args { set acc [sub $acc $a] } ; return $acc }
	proc mul:all args { set args [lassign $args acc] ; foreach a $args { set acc [mul $acc $a] } ; return $acc }
	proc div:all args { set args [lassign $args acc] ; foreach a $args { set acc [div $acc $a] } ; return $acc }
	proc pow:all args { set args [lassign $args acc] ; foreach a $args { set acc [pow $acc $a] } ; return $acc }

	proc add {a b} {
		if [string match *.*.* $a] {error "Error in ADD: only 1 decimal point allowed."}
		if [string match *.*.* $b] {error "Error in ADD: only 1 decimal point allowed."}

		if ![regexp -- {^([+\-])?[0-9\\.]+$} $a] {error "Error in ADD (\"${a}\"): only real numbers can be used with ADD."}
		if ![regexp -- {^([+\-])?[0-9\\.]+$} $b] {error "Error in ADD (\"${b}\"): only real numbers can be used with ADD."}
		if [string eq 0 $a] {return $b}
		if [string eq 0 $b] {return $a}

		set a [nozero $a]
		set b [nozero $b]

		set a [string trimleft $a +]
		set a_sign +
		if [left $a 1 -] {set a_sign -; set a [mid $a 2]}
		set a_integer $a
		set a_mantissa 0
		if [instr $a .] {set a_integer [l0 [split $a .]]; set a_mantissa [l1 [split $a .]]}
		set a_integer_len [len $a_integer]
		set a_mantissa_len [len $a_mantissa]

		set b [string trimleft $b +]
		set b_sign +
		if [left $b 1 -] {set b_sign -; set b [mid $b 2]}
		set b_integer $b
		set b_mantissa 0
		if [instr $b .] {set b_integer [l0 [split $b .]]; set b_mantissa [l1 [split $b .]]}
		set b_integer_len [len $b_integer]
		set b_mantissa_len [len $b_mantissa]

		# Determine how to perform addition
		set answer_sign_override +

		if [string eq ${a_sign}${b_sign} ++] {# Fall-through ....}
		if [string eq ${a_sign}${b_sign} +-] {return [sub $a $b]}
		if [string eq ${a_sign}${b_sign} -+] {return [sub $b $a]}
		if [string eq ${a_sign}${b_sign} --] {return -[add $a $b]}

		# Let's continue with "++" addition
		set max_integer [ expr ( ( $a_integer_len > $b_integer_len ) ? $a_integer_len : $b_integer_len ) + 1 ]
		set max_mantissa [ expr ( ( $a_mantissa_len > $b_mantissa_len ) ? $a_mantissa_len : $b_mantissa_len ) + 1 ]

		set number_1 [addzero $a_integer $max_integer].[left ${a_mantissa}[string repeat 0 $max_mantissa] $max_mantissa]
		set number_2 [addzero $b_integer $max_integer].[left ${b_mantissa}[string repeat 0 $max_mantissa] $max_mantissa]
		set number_1_reverse [reverse $number_1]
		set number_2_reverse [reverse $number_2]

		empty answer
		zero carry
		foreach a_digit [split $number_1_reverse ""] b_digit [split $number_2_reverse ""] {
			if [string eq .. ${a_digit}${b_digit}] {
				prepend answer .
				continue
			}
			set sum [ expr $a_digit + $b_digit + $carry ]
			set carry [ expr int( $sum / 10 ) ]
			set sum [ expr $sum % 10 ]
			prepend answer $sum
		}
		return [nozero $answer]
	}

	proc sub { a b } {
		if [string match *.*.* $a] {error "Error in SUB: only 1 decimal point allowed."}
		if [string match *.*.* $b] {error "Error in SUB: only 1 decimal point allowed."}

		if [string eq 0 $a] {
			if [left $b 1 -] {
				return [string trimleft $b -]
			} {
				return -$b
			}
		}
		if [string eq 0 $b] {return $a}

		set a [string trimleft $a +]
		set a_sign +
		if [left $a 1 -] {set a_sign -; set a [mid $a 2]}
		set a_integer $a
		set a_mantissa 0
		if [instr $a .] {set a_integer [l0 [split $a .]]; set a_mantissa [l1 [split $a .]]}
		set a_integer_len [len $a_integer]
		set a_mantissa_len [len $a_mantissa]

		set b [string trimleft $b +]
		set b_sign +
		if [left $b 1 -] {set b_sign -; set b [mid $b 2]}
		set b_integer $b
		set b_mantissa 0
		if [instr $b .] {set b_integer [l0 [split $b .]]; set b_mantissa [l1 [split $b .]]}
		set b_integer_len [len $b_integer]
		set b_mantissa_len [len $b_mantissa]

		# Determine how to perform substraction
		set answer_sign_override +

		if [string eq ${a_sign}${b_sign} ++] {
			if {$b > $a} {return -[sub $b $a]}
			# Otherwise, fall-through ....
		}
		if [string eq ${a_sign}${b_sign} +-] {return [add $a $b]}
		if [string eq ${a_sign}${b_sign} -+] {return -[add $a $b]}
		if [string eq ${a_sign}${b_sign} --] {
			if {$a < $b} {return [sub $b $a]} {return -[sub $a $b]}
		}

		# Let's continue with "++" subtraction

		set max_integer [ expr ( ( $a_integer_len > $b_integer_len ) ? $a_integer_len : $b_integer_len ) + 1 ]
		set max_mantissa [ expr ( ( $a_mantissa_len > $b_mantissa_len ) ? $a_mantissa_len : $b_mantissa_len ) + 1 ]

		set number_1 [addzero $a_integer $max_integer].[left ${a_mantissa}[string repeat 0 $max_mantissa] $max_mantissa]
		set number_2 [addzero $b_integer $max_integer].[left ${b_mantissa}[string repeat 0 $max_mantissa] $max_mantissa]

		set number_1_reverse [reverse $number_1]
		set number_2_reverse [reverse $number_2]

		empty answer
		zero carry
		foreach a_digit [split $number_1_reverse ""] b_digit [split $number_2_reverse ""] {
			if [string eq .. ${a_digit}${b_digit}] {
				prepend answer .
				continue
			}
			set sum [ expr $a_digit - $b_digit - $carry ]
			set carry [ expr ( $sum < 0 ) ? 1 : 0 ]
			if $carry { incr sum 10 }
			prepend answer $sum
		}

		return [nozero $answer]
	}

	proc mul {a b} {

		if [string match *.*.* $a] {error "Error in MUL: only 1 decimal point allowed per term."}
		if [string match *.*.* $b] {error "Error in MUL: only 1 decimal point allowed per term."}

		if [string eq 0 $a] {return 0}
		if [string eq 0 $b] {return 0}

		zero a_mantissa_count b_mantissa_count
		if [instr $a .] {set a_mantissa_count [len [l1 [split $a .]]]}
		if [instr $b .] {set b_mantissa_count [len [l1 [split $b .]]]}
		set mantissa_count [ expr $a_mantissa_count + $b_mantissa_count ]
		regsub -all -- \\\. $a "" a
		regsub -all -- \\\. $b "" b
		zero negative

		set a [string trimleft $a +]
		set b [string trimleft $b +]
		if [left $a 1 -] {set negative [ expr 1 - $negative ]; set a [mid $a 2]}
		if [left $b 1 -] {set negative [ expr 1 - $negative ]; set b [mid $b 2]}

		empty collection
		negone a_counter
		foreach a_digit [split [reverse $a] ""] {
			incr a_counter
			negone b_counter
			foreach b_digit [split [reverse $b] ""] {
				incr b_counter
# SCINOT gets introduced here ....
#				lappend collection [ expr $a_digit * $b_digit * ( pow( 10 , $a_counter ) * pow( 10 , $b_counter ) ) ]
				set item [ expr $a_digit * $b_digit ][string repeat 0 [ expr $a_counter + $b_counter ]]
				lappend collection $item
			}
		}

		zero tally
		if {[llength $collection] == 1} {
			set tally $collection
		} else {
			foreach a $collection {set tally [add $tally $a]}
		}

		if {$mantissa_count > [len $tally]} {prepend tally [string repeat 0 [ expr $mantissa_count - [ len $tally ] ]]}
		if $mantissa_count {set tally [left $tally -$mantissa_count].[right $tally $mantissa_count]}
		set tally [nozero $tally]
		if $negative {prepend tally -}
		return $tally
	}

	proc pow {a b} {
		if [string match *.*.* $a] {error "Error in POW: only 1 decimal point allowed."}
		if [string match *.*.* $b] {error "Error in POW: only 1 decimal point allowed."}

		if {$b == 0} {return 1}; # 0^1 is still 1!! This must come first.
		if {$a == 0} {return 0}
		if {$b == 1} {return $a}

		one pow
		for {set x 0} {$x < $b} {incr x} {set pow [mul $pow $a]}
		return $pow
	}


	proc div {dividend divisor} {
		set original_dividend $dividend
		set original_divisor  $divisor

		if [string match *.*.* $dividend] {error "Illegal number: $dividend"}
		if [string match *.*.* $divisor ] {error "Illegal number: $divisor"}

		set dividend [string trimleft $dividend +]
		set divisor  [string trimleft $divisor  +]
		set negative 0
		if [left $dividend 1 -] {set negative [ expr 1 - $negative ]; set dividend [string trimleft $dividend -]}
		if [left $divisor  1 -] {set negative [ expr 1 - $negative ]; set divisor  [string trimleft $divisor  -]}
		if $negative {set negative -} {set negative ""}

		# Shift decimal points
		set dividend [nozero $dividend]
		set divisor  [nozero $divisor]

		if [string eq 0 $divisor] {error "Division by zero"}
		if [string eq 0 $dividend] {return 0}
		if [string eq 1 $divisor] {return ${negative}[nozero $dividend]}

		# Pad a sufficient number of zeroes
		if ![instr $dividend .] {append dividend .}
		if ![instr $divisor  .] {append divisor  .}
		append dividend [string repeat 0 [len $divisor]]

		# .5 & 10.00
debug.print 2 DIVIDEND($dividend):DIVISOR($divisor)

		set split_divisor  [split $divisor  .]
		set split_dividend [split $dividend .]

		set instr_divisor  [instr $divisor  .]
		set instr_dividend [instr $dividend .]
debug.print 2 INSTR_DIVISOR($instr_divisor):INSTR_DIVIDEND($instr_dividend)
		set shift_divisor  [len [l1 $split_divisor ]]
debug.print 2 SHIFT_DIVISOR($shift_divisor)
		set shift_dividend [len [l1 $split_dividend]]
		if ![string eq 0 $shift_divisor] {
			regsub \\\. $divisor "" divisor
			regsub \\\. $dividend "" dividend
			incr instr_dividend $shift_divisor
debug.print 2 INSTR_DIVIDEND($instr_dividend)
			set dividend [left $dividend [ expr $instr_dividend - 1 ]].[mid $dividend $instr_dividend]
		}

		set divisor [nozero $divisor]; # Yes, it's needed again!
		set dividend [nozero $dividend]
		if ![instr $dividend .] {append dividend .}
		set decimal_point [instr $dividend .]
		regsub \\\. $dividend "" dividend
debug.print 2 DIVIDEND($dividend):DIVISOR($divisor)
debug.print 2 \n
debug.print 2 "Starting DIV loop...."
debug.print 2 \n

		# Start loop

		set position 1
		set size 1
		set remainder ""
		set quotient ""
		set base ""
		set getnext 0
		set decimalused 0
		set loop 0

		# Digits of accuracy
		set loop_max 100

		while 1 {
debug.print 2 WHILE
debug.print 2 REMAINDER($remainder)
			incr loop
			debug.print 2 "LOOP($loop)[string repeat \b 100]"
			flush stdout
			if {$loop > $loop_max} {
				#error "\[DIV\] $loop_max [plural loop $loop_max]! DIVIDEND($dividend):DIVISOR($divisor):QUOTIENT($quotient)"
				return ${negative}[nozero $quotient]
			}

			if {$position > [len $dividend]} {
				if ![string eq 0 $remainder] {append dividend 0} break
			}

			set base ${remainder}[mid $dividend $position $size]
debug.print 2 "-- BASE($base)"
			if {[len $base] < $size} {append base 0}

			if { ( $position + $size - 1 ) >= $decimal_point } {
				if !($decimalused) {one decimalused; append quotient .}
			}
			if [instr $base .] { # set decimalused 1; append quotient .; regsub \\\. $dividend "" dividend; incr size; continue }
debug.print 2 SIZE($size)

			if {[len $base] < $size} {
				set differential [ expr $size - [ len $base ] ]
				append base [string repeat 0 $differential]
			}

			# Need to leave BASE alone from this point through, as a leading "0" may appear (fake octal "08" error)
			set base [nozero $base]
debug.print 2 BASE($base)

			for {set x 9} {$x >= 0} {incr x -1} {
set expr [ mul $divisor $x ]
debug.print 2 X($x):EXPR($expr):BASE($base)
				if {$expr <= $base} {
					set mult [mul $divisor $x]
					set remainder [sub $base $mult]
					append quotient $x
debug.print 2 REMAINDER($remainder):QUOTIENT($quotient)
					incr position $size
					set size 1
					set getnext 1
					break
				}
			}

			if !($getnext) {incr size}
			set getnext 0
		}
		set quotient ${negative}[nozero $quotient]
		debug.print 2 "$original_dividend / $original_divisor = $quotient"
		return $quotient
	}

	proc test args {
		empty datafile
		lappend datafile "[using <A*$::tcl_precision "A / B"]\t[using <*$::tcl_precision DIV]\tD=E\t[using <*$::tcl_precision EXPR]\tD=F\t[using <*$::tcl_precision FIXMATH]\t"

		for {set x 1} {$x <= 100} {incr x 1} {
			for {set y 1} {$y <= 100} {incr y 1} {
			lappend datafile ""
				for {set x_sign 0} {$x_sign <= 1} {incr x_sign} {
					for {set y_sign 0} {$y_sign <= 1} {incr y_sign} {
						set a [lindex [list + -] $x_sign]$x
						set b [lindex [list + -] $y_sign]$y

						set div [div $a $b]
						set expr [ expr $a / $b ]
						set fixmath [ fixmath $a / $b ]

						if {$div == $expr} {set div_expr 1} {
							#set len [len $expr]
							set len $::tcl_precision
							decr len
							set div_expr [ fixmath [ left $div $len ] == [ left $expr $len ] ]
						}

						if {$div == $fixmath} {set div_fixmath 1} {
							set len [len $fixmath]
							decr len
							set div_fixmath [ fixmath [ left $div $len ] == [ left $fixmath $len ] ]
						}

						lappend datafile "[using A<*$::tcl_precision "$a / ${b}:"]\t[using A<*$::tcl_precision [left $div $::tcl_precision]]\t[noyes [ expr $div == $expr ]]\t[using A<*$::tcl_precision [left $expr $::tcl_precision]]\t[noyes [ expr $div == $fixmath ]]\t[using A<*$::tcl_precision [left $fixmath $::tcl_precision]]\t"
					}
				}
			}
		}
		dumpfile datafile.txt $datafile 1
	}

	# Calcualte PI using the BPP formula
	proc pi:calc k {
		set sum 0
debug.print 2 LOOP:
		for {set x 0} {$x <= $k} {incr x} {
puts "Loop #${x}: SUM($sum)"
debug.print 2 X($x)
			set part_0 [div 1 [pow 16 $k]]
debug.print 2 0($part_0)
			set part_1 [div 4 [add [mul 8 $k] 1]]
debug.print 2 1($part_1)
			set part_2 [div 2 [add [mul 8 $k] 4]]
debug.print 2 2($part_2)
			set part_3 [div 1 [add [mul 8 $k] 5]]
debug.print 2 3($part_3)
			set part_4 [div 1 [add [mul 8 $k] 6]]
debug.print 2 4($part_4)
			set part_5 [sub [sub [sub $part_1 $part_2] $part_3] $part_4]
debug.print 2 5($part_5)
			set part_6 [mul $part_0 $part_5]
debug.print 2 6($part_6)
			set sum [add $sum $part_6]
debug.print 2 SUM($sum)
debug.print 2 "Iteration #${x}: SUM($sum)"
		}
		if [file exists Z:/pi/pi.txt] {
			set pi [readfile Z:/pi/pi.txt]; # Read a file w/PI value for comparison
			set pi_piece [left $pi [len $sum]]
			return "[noyes [string eq $pi_piece $sum]]: $sum"; # Do they match so far?
		}
		return $sum
	}

	namespace export add sub mul div pow test pi:calc
}

# Call with: HIGHLIGHT [-NOCASE] [-REGEXP] <phrase> <what to highlight>
proc highlight args {
	zero nocase regexp

	# Check for flags (can't use SB:SETVARIABLES because the list elements can't be broken up!)
	do if {[left [l0 $args] 1 -]} {
		switch -glob -- [sl0 $args] {

			-nocase {one nocase; set args [lreplace $args 0 0]}

			-regexp {one regexp; set args [lreplace $args 0 0]}

			default {
				if [string eq [l0 $args] --] {set args [lreplace $args 0 0]}
				break
			}

		}
	}

	empty sentence match open close
	lassign $args sentence match open close
	if !($regexp) {
		if [iswildcard $match] {set match [SB:wildcard_to_regexp -soft $match]}
	}
debug.print 2 "\[HIGHLIGHT\] ARGS($args):SENTENCE($sentence):MATCH($match):NOCASE($nocase):REGEXP($regexp):OPEN($open):CLOSE($close)"

	if $nocase {
		regexp -all -nocase -- (${match}) $sentence - found
	} {
		regexp -all         -- (${match}) $sentence - found
	}

	if [isempty found] {return $sentence}
	set found [list $found]

	# Default to BOLD ON/OFF
	if [isempty open] {set open \002}
	if [isempty close] {set close $open}

	foreach item $found {
		set replace $item
		regsub -all -- $item $sentence ${open}${replace}${close} sentence
	}

	return $sentence
}

proc SB:idx_list handle {
	empty list
	set gf [dcclist chat]
	foreach entry $gf {
		lassign $entry idx user
		if [string eq -nocase $user $handle] {lappend list $idx}
	}
	return $list
}

proc SB:sendsmtp {smtp login pass email to text} {
	empty message
	foreach a [list "EHLO $::nick" "AUTH LOGIN" [::base64::encode $login] [::base64::encode $pass] "MAIL FROM: $email" "RCPT TO: $to" DATA] {
		lappend message $a
	}
	foreach a $text {
		if [string eq . $a] { set a .. }
		lappend message $a
	}
	lappend message "."
	lappend message "QUIT"

	empty smtp_reply
	set w [socket $smtp 25]
	fconfigure $w -blocking 0
	foreach a $message {
#%# debug.print 1 "> SB:SENDSMTP - MESSAGE: $a"
		for { set x 0 } { $x < 32 } { incr x } { regsub -all [binary format c* $x] $a "" a }
		puts $w ${a}\r
	}
debug.print "\[SB:SENDSMTP\] Server message: [read $w]"
	close $w
	return 0
}

proc validcmd  cmd {not [string eq "" [info commands $cmd]]}
proc validproc cmd {not [string eq "" [info proc     $cmd]]}

proc herhis handle {
	if [right [stl $handle] 1 s] {set herhis ${handle}'} {set herhis ${handle}'s}
	if ![validuser $handle] {return $herhis}
	set sex [stl [getdata $handle gender]]
	if [left $sex 1 m] {set herhis his}
	if [left $sex 1 f] {set herhis her}
	return $herhis
}

proc rawprint text {
	# Needed because SIGILL will wipe this out
	proc rawprint_validcmd cmd { expr [string eq "" [info commands $cmd]] ? 0 : 1 }
	if [rawprint_validcmd putraw] {
		putraw $text
	} elseif [rawprint_validcmd putnow] {
		if ![right $text 1 \n] {append text \n}
		putnow $text -oneline
	} elseif [rawprint_validcmd putdccraw] {
		if ![right $text 1 \n] {append text \n}
		set error [ catch { set result [ putdccraw 0 [len $text] $text ] } shit ]

		# This is version-dependent (1.6.20+)
		if $error { putnow $text -oneline }
	} else {
		putquick $text -next
	}
	return $text
}

proc SB:DCC:test {handle idx arg} {putidx $idx "\002\00303Yes, I'm still listening, ${handle}.\003\002"; return 0}

#############################################################################

namespace eval ::tdb {

	variable encryption_key [reverse stormbot.tdb]
	variable database ; array set database [list]

	proc data args {
		variable encryption_key
		variable database
		set encryption $encryption_key
		switch -exact -- [sl0 $args] {

			exists {
				set names [data names]
				if {[lsearch -exact $names [sl1 $args]] < 0} {return 0}
				return 1
			}

			names {
				empty names
				set matchme [none [sl1 $args] *]
				foreach a [array names database] {
					set b [decrypt $encryption $a]
					if [string match -nocase $matchme $b] {
						lappend names [decrypt $encryption $a]
					}
				}
				return [lsort -inc -uni -dict $names]
			}

			destroy {
				set variable [encrypt $encryption [sl1 $args]]
				if [info exists database($variable)] {unset database($variable)}
				return [sl1 $args]
			}

			get {
				set variable [encrypt $encryption [sl1 $args]]
				if [info exists database($variable)] {return [decrypt $encryption $database($variable)]}
				return ""
			}

			set {
				set variable [encrypt $encryption [sl1 $args]]
				set value [l2 $args]
				if [isempty value] {unset -nocomplain database($variable)} {set database($variable) [encrypt $encryption $value]}
				return $value
			}

			lset {
				set variable [sl1 $args]
				set index [l2 $args]
				set value [l3 $args]
				set data [data get $variable]
				if {$index < -1} { error "\[DATA\] Illegal index: $index" }
				if {$index > [llength $data]} { error "\[DATA\] Illegal index: $index" }
				lset data $index $value
				data set $variable $data
				return $value
			}

			lindex {
				set variable [sl1 $args]
				set index [l2 $args]
				set data [data get $variable]
				return [lindex $data $index]
			}

			lrange {
				set variable [sl1 $args]
				set index_start [l2 $args]
				set index_end   [l3 $args]
				set data [data get $variable]
				return [lrange $data $index_start $lindex_end]
			}

			rename {
				set variable_old [sl1 $args]
				set variable_new [sl2 $args]
				data set $variable_new [data get $variable_old]
				data set $variable_old
				return [sl2 $args]
			}

			match { set variable [l1 $args]; zero nocase; if [string eq -nocase [l2 $args] -NOCASE] {one nocase; set args [lreplace $args 2 2]}; set matchme [l2 $args]; if [isempty matchme] { error "\[DATA\] wrong # args: should be \"data match ?-nocase? data_element pattern\"" }; set data [data get $variable]; if $nocase { return [string match -nocase $matchme $data] } { return [string match $matchme $data] } }

			find {return [data names [l1 $args]]}

			clear { foreach a [data names [l1 $args]] { data set $a } }

			default { error "\[DATA\] Illegal command option: [su0 $args]" }

		}
	}

	proc tdb args {
		# This is an experimental text-based database. There WILL be bugs!

		# Need to syntax-check table & field names against ^[A-Za-z][A-Za-z0-9]*$

		lassign $args 1 2 3 4 5 6 7 8 9 10
		set splitter |
		variable encryption_key
		set field_chars [list % ^ & * ! @ \$]; # Can't use "#" (if at beginning of line, will be skipped as a comment)

		# Is HEADER format current?
		set header [data get INFO:$3]
		if [notempty header] {
			if {[llength $header] < 6} {# Old / preliminary format
				empty cby mby
				lassign $header ctime mtime autonumber fields meta
				set header [list $ctime $cby $mtime $mby $autonumber $fields $meta]
				data set INFO:$3 $header
				unset header ctime cby mtime mby autonumber fields meta
			}
		}

		# Return all functions with success value: 0 (fail) / 1 (OK)
		switch -exact -- [stl $1] {

			comp - compact - clean - cleanup {
				empty tables
				foreach table [lrange $args 1 end] {
					set tables [concat $tables [lsearch -glob -all -inline [tdb table names] [none [stl $table] *]]]
				}

				foreach table $tables {
					set fields [tdb field names $table]
					set records [tdb record names $table]
					empty destroy
					zero new

					foreach record $records {
						empty temp
						tdb record array $table $record crap
						foreach item [array names crap] {append temp $crap($item)}
						if [isempty temp] {lappend destroy $record}
					}

					set info [data get INFO:$table]
					lset info 1 [now]
					data set INFO:$table $info
				}
				return 1
			}

			renum - renumber {
				if ![tdb table exists $2] { error "\[TDB RENUMBER\] No such table: $2" }
				foreach table $2 {
					set fields [tdb field names $table]
					set records [tdb record names $table]
					empty purge
					emptyarray holding_pen
					zero count

					# I guess I'm blasting the %AUTONUMBER% field?

					foreach record $records {
						empty junk temp_record
						foreach field $fields {
							set name ${table}${splitter}${record}${splitter}$field
							set data [data get $name]
							lappend temp_record $data; # Is in $FIELDS order
							append junk $data
						}

						# Is there any data in the record?
						if [notempty junk] {
							incr count
							lappend purge $record
							foreach field $fields data $temp_record {
								# Auto-update information with new record number
								if {[lsearch -exact [stl $3] [stl $field]] > -1} {set data $count}
								set holding_pen(${table}${splitter}${count}${splitter}$field) $data
							}

							# Purge old records
							foreach record [data names ${table}${splitter}*] {data set $record}

							# Restore renumbered records
							foreach restore [array names holding_pen] {data set $restore $holding_pen($restore)}

							# Update the current record counts
							tdb header set $table autonumber $count
							tdb header set $table mtime [now]

							# Should be done.
						}
					}
				}
				return [llength $purge]

				empty tables
				foreach table [lrange $args 1 end] {set tables [concat $tables [lsearch -glob -all -inline [tdb table names] [none [stl $table] *]]]}

				foreach table $tables {
					set records [tdb record names $table]
					empty destroy
					zero new

					foreach record $records {
						empty temp
						tdb record array $table $record crap
						foreach item [array names crap] {append temp $crap($item)}
						if [isempty temp] {lappend destroy $record; continue}
						incr new
						if {$record == $new} continue
						tdb record rename $table $record $new

						# Blank the higher (old) entry
						foreach match [data find ${table}${splitter}${record}${splitter}*] {data destroy $match}
					}

					set highest [le [lsort -int -inc $destroy]]
					for {
						set x [ expr $new + 1 ]
					} {
						$x <= $highest
					} {
						incr x
					} {
						tdb record destroy $table $x
					}
					set info [data get INFO:$table]
					lset info 2 $new
					lset info 1 [now]
					data set INFO:$table $info
				}
				return $new
			}

			head - header {
				if [isempty 3] {error "\[TDB HEADER\] HEADER with no table name"}
				if ![tdb table exists $3] {error "\[TDB HEADER\] HEADER for non-existent table: $3"}

				# $2 = GET / SET
				set header [data get INFO:$3]

				switch -exact -- [stl $4] {

					ctime - created {set index 0}

					cby - cwho - createdby {set index 1}

					mtime - modified {set index 2}

					mby - mwho - modifiedby {set index 3}

					auto - count - number - counter - autonumber {set index 4}

					field - fields {set index 5}

					meta {set index 6}

					default { error "\[TDB HEADER\] Unknown header type: [stu $4]" }

				}

				switch -exact -- [stl $2] {

					get {return [lindex $header $index]}

					set {
						set new [lrange $args 4 end]
						lset header $index $new
						data set INFO:$3 $header
						return $new
					}

					default { error "\[TDB HEADER\] Unknown command: [stu $2]" }

				}
			}

			table {# INFO:<table name> = [LIST CTIME MTIME CURRENT_AUTONUMBER LIST_FIELDS]
				switch -exact -- [stl $2] {

					create {
						if [isempty 3] {error "\[TDB\] TABLE CREATE with no table name"}
						if [tdb table exists $3] {error "\[TDB\] TABLE CREATE with already existing table: $3"}
						data set INFO:$3 [list [now] [now] 0 "" ""]
						# LIST: created modified autonumber fields META
						return 1
					}

					destroy {
						if [isempty 3] {error "\[TDB\] TABLE DESTROY with no table name"}
						if ![tdb table exists $3] {error "\[TDB\] TABLE DESTROY on non-existent table: $3"}
						data set INFO:$3
						data clear ${3}${splitter}*
						return 1
					}

					names {return [stl [lsort -inc -uni -dict [regsub -all -nocase -- INFO: [data find INFO:[none $3 *]] ""]]]}

					exists {return [ expr ( [ lsearch -exact [ tdb table names ] [stl $3] ] > -1 ) ? 1 : 0 ]}

					rename {
						if [isempty 3] {error "\[TDB\] TABLE RENAME with no original table name"}
						if [isempty 4] {error "\[TDB\] TABLE RENAME with no target table name"}
						if ![data exists INFO:$3] {error "\[TDB\] TABLE RENAME with non-existent table: $3"}
						if [data exists INFO:$4] {error "\[TDB\] TABLE RENAME with already-existing target table: $4"}
						data rename INFO:$3 INFO:$4
						foreach a [data find ${3}${splitter}*] {
							regsub -nocase -- $3 $a $4 b
							data rename $a $b
						}
						data lset INFO:$4 1 [now]
						return 1
					}

					setup {# TABLE [LIST: field1 field2 field3 ...] TABLE [fields] TABLE [fields] ....
						foreach {table fields} [lrange $args 2 end] {
							if [isempty table] break
							if ![tdb table exists $table] {tdb table create $table}
							foreach field $fields {
								if [isempty field] break
								if ![tdb field exists $table $field] {tdb field create $table $field}
							}
						}
						return 1
					}

					default {error "\[TDB\] Unknown TABLE option: $2"}

				}
			}

			meta {
				# Syntax: TDB META SET <TABLE> <name> <data>
				if [isempty 3] {error "\[TDB META\] META with no table name"}
				if ![tdb table exists $3] {error "\[TDB META\] META with non-existent table: $3"}

				set info [data get INFO:$3]
				array set meta [lindex $info 6]

				switch -exact -- [stl $2] {

					set {
						if [isempty 4] { error "\[TDB META SET\] Missing META data name" }
						set name [stl $4]
						set data [lindex $args 4]
						set meta($name) $data
						if [isempty data] {unset meta($name)}
						lset info 6 [array get meta]
						lset info 2 [now]
						data set INFO:$3 $info

						return $data
					}

					get {
						if [isempty 4] { error "\[TDB META SET\] Missing META data name" }
						set name [stl $4]
						if [isempty meta($name)] {return ""} {return $meta($name)}
					}

					names - list {
						set matchme [none [stl $4] *]
						return [lsearch -glob -all -inline [stl [array names meta]] $matchme]
					}

					clear {
						lset info 6 ""
						lset info 2 [now]
						data set INFO:$3 $info
						return ""
					}

					destroy {
						if [isempty 4] { error "\[TDB META SET\] Missing META data name" }
						set name [none [stl $4] *]
						foreach entry [lsearch -glob -inline -all [stl [array names meta]] $name] {
							empty meta($entry)
							unset meta($entry)
						}
						lset info 6 [array get meta]
						lset info 2 [now]
						data set INFO:$3 $info
						return ""
					}

					default {error "\[TDB META\] Unknown META control option: $3"}

				}
			}

			record {
				switch -exact -- [stl $2] {

					create {
						if ![data exists INFO:$3] {error "\[TDB\] RECORD CREATE with non-existent table: $3"}
						if [isempty 3] {error "\[TDB\] RECORD CREATE with no table name"}
						# Allow blank entries (for later filling): if [isempty 4] {error "\[TDB\] RECORD CREATE with no field values"}

						if [notempty 4] {
							# Error check this before adding the new record
							set fields [tdb field names $3]
							foreach entry $4 {
								lassign $entry field value
								if {[lsearch -exact [stl $fields] [stl $field]] < 0} {error "\[TDB\] Invalid field in TABLE ${3}: $field"}
							}
						}
						set old [tdb header get $3 autonumber]
						set autonumber [ expr $old + 1 ]
						tdb header set $3 autonumber $autonumber

						# AUTONUMBER must be last, in case the user tries to
						# assign it; this will override whatever is forced
						foreach entry [concat $4 [list [list %AUTONUMBER% $autonumber]]] {
							lassign $entry field value
							data set ${3}${splitter}${autonumber}${splitter}${field} $value
						}
						tdb header set $3 mtime [now]
						return $autonumber
					}

					destroy {
						if ![data exists INFO:$3] {error "\[TDB\] RECORD DESTROY with non-existent table: $3"}
						if [isempty 3] {error "\[TDB\] RECORD DESTROY with no table name"}
						if [isempty 4] {error "\[TDB\] RECORD DESTROY with no ID number"}

						zero count
						foreach deleteme [data find ${3}${splitter}${4}${splitter}*] {
							incr count
							data set $deleteme
						}
						tdb header set $3 mtime [now]
						return $count
					}

					set {
						if ![data exists INFO:$3] {error "\[TDB\] RECORD SET with non-existent table: $3"}
						if [isempty 3] {error "\[TDB\] RECORD SET with no table name"}
						if [isempty 4] {error "\[TDB\] RECORD SET with no ID number"}
						if [isempty 5] {error "\[TDB\] RECORD SET with no field name"}
						if [string eq -nocase LAST $4] {set 4 [tdb record last $3]}
						if ![isint $4] { error "\[TDB RECORD SET\] I expect an integer for the record number, or the literal: \"LAST\"" }
						set fields [tdb field names $3]
						lassign $5 field value
						if {[lsearch -exact [stl $fields] [stl $field]] < 0} {error "\[TDB\] Invalid field in TABLE ${3}: $field"}

						# Allow blank entries (for later filling): if [isempty 6] {error "\[TDB\] RECORD SET with no new values"}
						data set ${3}${splitter}${4}${splitter}${5} $6
						tdb header set $3 mtime [now]
						return 1
					}

					get {
						if ![data exists INFO:$3] {error "\[TDB\] RECORD GET with non-existent table: $3"}
						if [isempty 3] {error "\[TDB\] RECORD GET with no table name"}
						if [isempty 4] {error "\[TDB\] RECORD GET with no ID number"}
						if [string eq -nocase LAST $4] {set 4 [tdb record last $3]}
						if ![isint $4] { error "\[TDB RECORD GET\] I expect an integer for the record number, or the literal: \"LAST\"" }
						if ![tdb record exists $3 $4] {return ""}

						empty record
						set fields [tdb field names $3]
						if {[llength $5] == 1} {
							set m [lsearch -exact [stl $fields] [stl $5]]
							if {$m > -1} {return [data get ${3}${splitter}${4}${splitter}$5]}
							error "\[TDB RECORD GET\] GET with unknown field name: $5"
						}

						empty new
						if [isempty 5] {
							set matchme $fields
						} {
							set names [stl [tdb field names $3]]
							foreach a [stl $5] {
								set new [concat $new [lsearch -glob -inline -all $names $a]]
							}
							set matchme $new
						}

						foreach field [lsort -inc -uni -dict $matchme] {
							set string ${3}${splitter}${4}${splitter}$field
							set data [data get $string]
							lappend record [list $field $data]
						}
						return $record
					}

					names {
						if [isempty 3] {error "\[TDB\] RECORD NAMES with no table name"}
						set data [data get INFO:$3]
						if [isempty data] {error "\[TDB\] RECORD NAMES with non-existent table: $3"}

						empty records
						foreach item [data names $3${splitter}*${splitter}*] {
							lappend records [l1 [split $item |]]
						}
						set records [stl [lsort -inc -uni -dict $records]]
						return $records
					}

					count {
						if [isempty 3] {error "\[TDB\] RECORD COUNT with no table name"}
						return [llength [tdb record names $3]]
					}

					exists {return [ expr ( [ lsearch -exact [ tdb record names $3 ] [stl $4] ] > -1 ) ? 1 : 0 ]}

					array {
						if ![data exists INFO:$3] {error "\[TDB\] RECORD ARRAY with non-existent table: $3"}
						if [isempty 3] {error "\[TDB\] ARRAY with no table name"}
						if [isempty 4] {error "\[TDB\] ARRAY with no record name"}
						if [isempty 5] {error "\[TDB\] ARRAY with no target variable name"}

						upvar 1 $5 local
						array set local [join [tdb record get $3 $4]]
						return [array get local]
					}

					current - last {
						if ![data exists INFO:$3] {error "\[TDB\] RECORD LAST with non-existent table: $3"}
						if [isempty 3] {error "\[TDB\] RECORD LAST with no table name"}
						set last [tdb header get $3 autonumber]
						#if !($last) { error "\[TDB RECORD LAST\] No records in table: $3" }
						return $last
					}

					rename {
						if ![data exists INFO:$3] {error "\[TDB\] RECORD RENAME with non-existent table: $3"}
						if [isempty 3] {error "\[TDB\] RECORD RENAME with no table name"}
						if [isempty 4] {error "\[TDB RECORD RENAME\] Missing record ID to rename"}
						if [isempty 5] {error "\[TDB RECORD RENAME\] Missing target record ID"}
						set fields [tdb field names $3]
						foreach field $fields {
							tdb record set $3 $5 $field [tdb record get $3 $4 $field]
						}
						tdb header set $3 mtime [now]
						return $5
					}

					isempty {
						if ![data exists INFO:$3] {error "\[TDB\] RECORD ISEMPTY with non-existent table: $3"}
						if [isempty 3] {error "\[TDB\] RECORD ISEMPTY with no table name"}
						if [isempty 4] {error "\[TDB\] RECORD ISEMPTY with no ID number"}
						set ll [tdb record last $3]
						if {$4 <   1} { error "\[TDB RECORD ISEMPTY\] Illegal record number: $2" }
						if {$4 > $ll} { error "\[TDB RECORD ISEMPTY\] Illegal record number: $2" }
						one isempty
						set record [tdb record get $3 $4]
						foreach entry $record {
							lassign $entry a b
							if [notempty b] { zero isempty }
						}
						return $isempty
					}

					default {error "\[TDB RECORD ISEMPTY\] Unknown RECORD option: $2"}

				}
			}

			field {
				switch -exact -- [stl $2] {

					add - create {
						if ![data exists INFO:$3] {error "\[TDB\] FIELD CREATE with non-existent table: $3"}
						if [isempty 3] {error "\[TDB\] FIELD CREATE with no table name"}
						if [isempty 4] {error "\[TDB\] FIELD CREATE with no field name"}
						set fields [data lindex INFO:$3 5]
						foreach entry [sl4e $args] {
							set m [lsearch -exact [stl $fields] [stl $entry]]
							if {$m > -1} continue
							if ![regexp -nocase -- {^[A-Z][A-Z0-9_]*$} $entry] {error "\[TDB\] Illegal field name in FIELD CREATE: \"${entry}\" - must start with A-Z, and otherwise consist of: A-Z _ 0-9"}
							lappend fields $entry
						}

						data lset INFO:$3 5 $fields
						data lset INFO:$3 2 [now]

						return $fields
					}

					drop - del - delete - destroy {
						if ![data exists INFO:$3] {error "\[TDB\] RECORD SET with non-existent table: $3"}
						if [isempty 3] {error "\[TDB\] FIELD DESTROY with no table name"}
						if [isempty 4] {error "\[TDB\] FIELD DESTROY with no field name"}
						lassign [ldestroy -both -all -multiple -glob -nocase -- [data lindex INFO:$3 3] [lrange $args 3 end]] fields destroyed

						if [notempty destroyed] {
							foreach record [tdb record names $3] {
								foreach field $destroyed {
									data destroy ${3}${splitter}${record}${splitter}$field
								}
							}
						}

						data lset INFO:$3 5 $fields
						data lset INFO:$3 2 [now]
						return [notempty destroyed]
					}

					names {
						if ![data exists INFO:$3] {error "\[TDB\] FIELD NAMES with non-existent table: $3"}
						set names [stl [lsort -inc -uni -dict [data lindex INFO:$3 5]]]
						return [lsearch -inline -all -glob $names [none $4 *]]
					}

					count {
						if ![data exists INFO:$3] {error "\[TDB\] FIELD COUNT with non-existent table: $3"}
						return [llength [tdb field names $3]]
					}

					exists {return [ expr ( [ lsearch -exact [stl [ tdb field names $3 ]] [stl $4] ] > -1 ) ? 1 : 0 ]}

					rename {
						if ![data exists INFO:$3] {error "\[TDB\] RECORD SET with non-existent table: $3"}
						if [isempty 3] {error "\[TDB\] FIELD RENAME with no table name"}
						if [isempty 4] {error "\[TDB\] FIELD RENAME with no field name"}
						if [isempty 5] {error "\[TDB\] FIELD RENAME with no target field name"}
						set fields [data lindex INFO:$3 5]
						set m [lsearch -exact [stl $fields] [stl $4]]
						if {$m < 0} {return 0}
						if ![regexp -nocase -- {^[A-Z][A-Z0-9_]*$} $5] {error "\[TDB\] Illegal field name: \"${5}\" - must start with A-Z, and otherwise consist of: A-Z _ 0-9"}
						set fields [lreplace $fields $m $m $5]
						foreach entry [data find ${3}${splitter}*${splitter}$4] {
							data rename $entry ${3}${splitter}*${splitter}$5
						}
						data lset INFO:$3 2 [now]
						return 1
					}

					default {error "\[TDB\] Unknown FIELD option: $2"}

				}
			}

			OLD_select {
				if [isempty 3] {error "\[TDB\] SELECT without target field(s)"}
				if ![string eq -nocase FROM $3] {error "\[TDB\] Syntax error in SELECT statement (requires FROM <table>)"}
				if ![data exists INFO:$4] {error "\[TDB\] SELECT FROM with non-existent table: $4"}

				empty select
				foreach record [tdb record names $4] {
					set data [tdb record get $4 $record]; # Nested LIST! ($RECORD could be a LIST of records to get)
					if [string eq * $2] {# No matching: return everything
						lappend select $data
					} elseif [isempty 5] {# Specific fields / no match text: return values (in field order)
						empty temp
						emptyarray array
						array set array [join $data]
						lappend temp $record
						foreach field [split [stl $2] ,] {
							if [info exists array([trim $field])] {lappend temp $array([trim $field])}
						}
						lappend select $temp
					} else {# Specific fields / match text: match everything, return record#
						if {[llength [split $2 ,]] != [llength [split $5 ,]]} {error "\[TDB\] Mismatch in lengths: number of fields ([llength [split $2 ,]]: ${2}) vs number of data values ([llength [split $5 ,]]: ${5})"}
						emptyarray array
						array set array [join $data]
						one matched
						foreach field [split [stl $2] ,] value [split $5 ,] {
							if ![info exists array($field)] {
								zero matched
							} {
								if ![string match -nocase $value $array($field)] {zero matched}
							}
						}
						if $matched {lappend select $record}
					}
				}
				return $select
			}

			select {
				if [isempty 2] {error "\[TDB\] SELECT without target field(s)"}
				if ![string eq -nocase FROM $3] {error "\[TDB\] Syntax error in SELECT statement (requires FROM <table>)"}
				set data [data get INFO:$4]
				if [isempty data] {error "\[TDB\] SELECT FROM with non-existent table: $4"}

				empty select
				if [string eq * $2] {
					set fields [lindex [data get INFO:$4] 5]
					set records [tdb record names $4]
					foreach record $records {
						set temp $record
						foreach field $fields {
							lappend temp [list $field [data get ${4}${splitter}${record}${splitter}$field]]
						}
						lappend select $temp
					}
					return $select
				}

				if [isempty 5] {
					set records [tdb record names $4]
					if [string eq * $2] {
						set fields [tdb header get $4 field]
						foreach record $records {
							set temp $record
							foreach field $fields {
								lappend temp [list $field [data get ${4}${splitter}${record}${splitter}$field]]
							}
							lappend select $temp
						}
						return $select
					} {
						set fields [split $2 ,]
						foreach record $records {
							set temp $record
							foreach field $fields {
								lappend temp [list $field [data get ${4}${splitter}${record}${splitter}$field]]
							}
							lappend select $temp
						}
						return $select
					}
					foreach record $records {
						empty temp
						foreach field $fields {
							lappend temp [list $field [data get ${4}${splitter}${record}${splitter}$field]]
						}
						lappend select $temp
					}
					return $select
				}

				# Default to: specific data matching

				if {[llength [split $2 ,]] != [llength [split $5 ,]]} {error "\[TDB\] Mismatch in lengths: number of fields ([llength [split $2 ,]]: ${2}) vs number of data values ([llength [split $5 ,]]: ${5})"}
				set split_field [split $2 ,]
				set split_match [split $5 ,]
				foreach record [tdb record names $4] {
					one match
					foreach check_field $split_field check_match $split_match {
						set name ${4}${splitter}${record}${splitter}$check_field
						set datum [data get $name]
						if ![string match -nocase $check_match $datum] {zero match}
					}
					if $match {lappend select $record}
				}
				return $select

				#####

				emptyarray array
				array set array [join $data]
				one matched
				foreach field [split [stl $2] ,] value [split $5 ,] {
					if ![info exists array($field)] {
						zero matched
					} {
						if ![string match -nocase $value $array($field)] {zero matched}
					}
				}
				if $matched {lappend select $record}
				return $select
			}


			find {
				switch -exact -- [stl $2] {

					datum - data - value - values {
						if [string eq "" [data get INFO:$3]] {error "\[TDB\] FIND with non-existent table: $3"}
						if [isempty 3] {error "\[TDB\] FIND VALUE with no table name"}
						if [isempty 4] {error "\[TDB\] FIND VALUE with no search value"}
						empty matches
						foreach checkme [data find ${3}${splitter}*] {
							if [data match -nocase $checkme $4] {
								lassign [split $checkme $splitter] table ID field
								if [left $table 1 %] {set table [mid $table 2]}
								if [left $field 1 %] continue
								lappend matches [list $table $ID $field [data get $checkme]]
							}
						}
						return $matches
					}

					default {error "\[TDB\] Unknown FIND option: $2"}

				}
			}

			load {# [tdb name]
				set filename ${::nick}.tdb
				if [notempty 2] {set filename $2}
				if ![file exists $filename] {return 0}
				set encryption stormbot.tdb
				zero count
				set r [open $filename r]
				do unless {[eof $r]} {
					set line [gets $r]
					if [left $line 1 #] continue
					set entries [split $line $field_chars]
					foreach entry $entries {
						if [isempty entry] continue
						lassign [decrypt $encryption $entry] item value
						#if ![left $item 1 %] continue
						data set $item $value
						incr count
					}
				}
				close $r
				return $count
			}

			save {# [tdb name] [tables]
				set filename ${::nick}.tdb
				if [notempty 2] {set filename $2}
				set encryption stormbot.tdb
				zero count
				set w [open $filename w]

				empty output matches
				set matchme *
				if [notempty 3] {
					regsub -all -- {,|\;} [lrange $args 2 end] " " matchme
					# in case that sucked ....
					if [isempty matchme] {set matchme *}

					empty matches
					foreach a [lrange $args 2 end] {
						lappend matches INFO:$a
						lappend matches ${a}${splitter}*${splitter}*
					}
				} {
					set matches *
				}

				foreach match $matches {
					foreach a [lsort -inc -dict [data find $match]] {
						incr count
						append output [lindex $field_chars [rand [llength $field_chars]]][encrypt $encryption [list $a [data get $a]]]
					}
				}

				puts $w [join [split "# StormBot.TCL V[sbd:get @VERSION] database #" ""]]
				puts $w [join [split "#! VERSION 1.0" ""]]
				puts $w $output
				close $w

				return $count
			}

			close - clear { data clear ; return 0 ; # Easy enough! }

			backup {
				set filename ${::nick}.tdb
				file copy -force $filename ${filename}~bak
				return 0
			}

			schema {
				empty schema
				foreach table [tdb table names [none $2 *]] {
					lappend schema [list $table [tdb record names $table] [tdb field names $table]]
				}
				return $schema
			}

			default {error "\[TDB\] Unknown command: $1"}

		}

		error "\[TDB\] Unknown directive: $args"
	}

	namespace export tdb data

}; # End of NAMESPACE EVAL

namespace import -force ::tdb::tdb ::tdb::data

proc tdb:death args { if ![string eq "" [sbd:find %*]] { tdb save } }
proc tdb:setup args { if [string eq "" [::tdb::data names]] { tdb load } }

#############################################################################

proc SB:shortdate { { date "" } { gmt false } { ymd 0 } } {
	if [istrue $ymd] {
		clock format [none $date [clock seconds]] -format "%Y-%m-%d %H:%M" -gmt [falsetrue $gmt]
	} {
		clock format [none $date [clock seconds]] -format "%m/%d/%Y %H:%M" -gmt [falsetrue $gmt]
	}
}

proc SB:lapsed {do { extra "" } } {
	switch -exact -- [stl $do] {

		set - mark - marker {
			array set lapsed [list 0 0]
			array set lapsed [sbd:get @LAPSED]
			incr lapsed(0)
			set lapsed($lapsed(0)) [clock clicks -milli]
			sbd:set @LAPSED [array get lapsed]
			return ""
		}

		print {
			zero count; one next
			array set lapsed [sbd:get @LAPSED]
			do forever {
				incr count; incr next
				if [info exists lapsed($next)] {
					debug.print 0 "\[SB:LAPSED: [none $extra default]\] Marker $count -> ${next}: [ expr $lapsed($next) - $lapsed($count) ]ms"
				} break
			}
			return 0
		}

		clear { sbd:empty @LAPSED; return ""}

		default { error "\[SB:LAPSED\] Unknown directive: [stu $do]"; return 0}

	}

	if 0 {
		set vals [lsort -int -inc [regsub -all -- time [info local time*] ""]]
		foreach val $vals {
			if [info exists time[ expr $val + 1 ]] {
				#log "\[COMMAND ADD [stu $command]\] Time lapsed ($val - [ expr $val + 1 ]): [ expr [ set time[ expr $val + 1 ] ] - [ set time$val ] ]ms"
			}
		}
	} ; # End IF 0

	error ?
}

proc SB:sbd_transfer {bot name data} {
	binary scan $data H* binary
	set data [encrypt SBD:TRANSFER [reverse $binary]]

	emptyarray sbd
	zero count
	do until {[isempty data]} {
		incr count
		# Increase BEFORE assigning: need first element to be "1" for
		# the receiving PROC (index #0 = stream meta data)
		set sbd($count) [left $data 256]
		set data [mid $data 257]
	}

	set id ${::botnet-nick}:[clock seconds]:[format %u [clock clicks -milli]]

	set list [list $id $name [llength [array names sbd]]]
	sbd:set @SBD:TRANSFER_COMPLETE:$id
	sbd:lappend @SBD:TRANSFER_COMPLETE:$id $list
	putbot $bot [list SBD:TRANSFER BEGIN $list]
	foreach count [array names sbd] {
		set stream [reverse $sbd($count)]
		if {$count == 5} continue
		putbot $bot [list SBD:TRANSFER DATA $id $count $stream]
		sbd:lappend @SBD:TRANSFER_COMPLETE:$id $stream
	}
	putbot $bot [list SBD:TRANSFER END $id $name]
	return 0
}

proc SB:sbd_receive {bot cmd data} {
	lassign $data cmd id name data
	switch -exact -- [stl $cmd] {

		begin {# ID data_name total_count
			lassign $id id name data
			sbd:set @SBD:TRANSFER_${bot}:$id
			empty list
			for {set x 0} {$x <= $data} {incr x} {lappend list ""}
			sbd:set @SBD:TRANSFER_${bot}:$id $list
			sbd:lset @SBD:TRANSFER_${bot}:$id 0 [list $id $name $data]
			return 0
		}

		data {
			sbd:lset @SBD:TRANSFER_${bot}:$id $name $data
			return 0
		}

		end {
			set meta [sbd:lindex @SBD:TRANSFER_${bot}:$id 0]
			lassign $meta null null count

			empty empty ; # (:
			for {set x 1} {$x <= $count} {incr x} {
				if [string eq "" [sbd:lindex @SBD:TRANSFER_${bot}:$id $x]] {lappend empty $x}
			}

			if [notempty empty] {
				putbot $bot [list SBD:CONFIRM RESEND START $id]
				foreach item $empty {
					putbot $bot [list SBD:CONFIRM RESEND DATA $id $item]
				}
				putbot $bot [list SBD:CONFIRM RESEND END $id]
				return 0
			}
			empty data
			for {set x 1} {$x <= $count} {incr x} {
				append data [reverse [sbd:lindex @SBD:TRANSFER_${bot}:$id $x]]
			}

			set data [binary format H* [reverse [decrypt SBD:TRANSFER $data]]]
			putlog "\[SB:SBD_RECEIVE\] Received data from ${bot}: $name"
			putbot $bot [list SBD:CONFIRM CLEAR $id]

			switch -exact -- [stl $name] {

				test { putlog "\[SB:SBD:RECEIVE\] TEST data received successfully: $data" ; return 0 }

				default { error "\[SB:SBD:RECEIVE\] Unknown target for SBD:TRANSFER data: [stu $name]" }

			}

			return 0
		}

		default { error "\[SB:SBD:RECEIVE\] Unknown command: [stu $cmd]" }

	}
	return 0
}

proc SB:sbd_confirm {bot cmd data} {
	lassign $data cmd name item extra
	switch -exact -- [stl $cmd] {

		resend {
			switch -exact -- [stl $name] {

				start {
					sbd:set @SBD:TRANSFER_RESEND:$name
					return 0
				}

				data {
					sbd:lappend @SBD:TRANSFER_RESEND:$item $extra
					return 0
				}

				end {
					lassign [sbd:lindex @SBD:TRANSFER_COMPLETE:$item 0] id name
						if [sbd:notempty @SBD:TRANSFER_RESEND:$item] {
						foreach count [sbd:get @SBD:TRANSFER_RESEND:$item] {
							putbot $bot [list SBD:TRANSFER DATA $item $count [sbd:lindex @SBD:TRANSFER_COMPLETE:$item [ expr $count - 1 ]]]
						}
						putbot $bot [list SBD:TRANSFER END $item $name]
					}
					sbd:set @SBD:TRANSFER_RESEND:$name
				}

				default { error "\[SB:SBD_CONFIRM\] Unknown RESEND instruction: [stl $name]" }
			}
			if [sbd:isempty @SBD:TRANSFER_COMPLETE:$id] { error "\[SB:SBD_CONFIRM\] $bot requested a resend of item #$item but there's no such data stream for ID #${id}!" }
		}

		clear {
			lassign [sbd:lindex @SBD:TRANSFER_COMPLETE:$name 0] id name count
			putlog "\[SB:SBD_CONFIRM\] Succesfully transferred (and verified) send to ${bot}: $name"
			sbd:set @SBD:TRANSFER_COMPLETE:$id
			return 0
		}

		default { error "\[SB:SBD_CONFIRM\] Unknown instruction: [stu $cmd]" }

	}
	return 0
}

proc SB:set_vhost {nick host handle { vhost "" } } {
	# Allow a specific VHOST to be manually set
	if [isempty vhost] {
		if ![validuser $handle] {return [list -1 "Unknown user"]}
		set vhost [getdata $handle vhost]
	}

	if [notempty vhost] {
		if [botismode Na] {
			set vident *
			if [instr $vhost @] {
				set vident [l0 [split $vhost @]]
				set vhost [l1 [split $vhost @]]
			}

			if [string match -nocase ${vident}@${vhost} $host] {
				return [list 0 "VHOST already in place: ${vident}@${vhost}"]
			} {
				lassign [split $host @] user_ident user_host
				if ![string match -nocase $vident $user_ident] {rawoutput "CHGIDENT $nick $vident"}
				if ![string match -nocase $vhost  $user_host ] {rawoutput "CHGHOST  $nick $vhost" }
				return [return 1 "Successfully set VHOST: ${vident}@${vhost}"]
			}
		} {
			return [list -2 "Unable to set VHOST: not Services Administrator (or higher)"]
		}
	}
	return [list -3 "Unknown error"]
}

# See replacement PROC below
proc `SB:get_count { list { limit 1000 } { combine 0 } } {
	if [isempty list] {return ""}
	if [isempty limit] {return ""}
	set get_count [splitnumberlist -range 0-$limit $list]

	if [string eq -nocase ALL [l0 $list]] {set get_count [splitnumberlist 1-$limit]}

	if [string eq -nocase FIRST [l0 $list]] {
		set first [l1 $list]
		if [string eq "" [l1 $list]] {set first 1}
		if {$first > $limit} {set first $limit}
		set get_count [splitnumberlist 1-$first]
	}

	if [string eq -nocase LAST [l0 $list]] {
		set last [l1 $list]
		if [string eq "" [l1 $list]] {set last 1}
		if ![isint $last] {printh $nick "\[SB:GET_COUNT\] Illegal value ( $last ): was expecting an integer"; return 0}
		if {$last > $limit} {set last $limit}
		set get_count [splitnumberlist [ expr ( $limit - $last ) + 1 ]-$limit]
	}

	if $combine { set get_count [makenumberlist $get_count] }
	return $get_count
}

# Replacement PROC
proc SB:get_count { list { lo 0 } { hi 1000 } { combine_list_into_ranges 0 } } {
	set new ""
	set final ""
	set firstlast 0
	regsub -all -- {\,|\;| [ ]+} $list " " list
	foreach a $list {
		# Due to this parsing format, possibilities like "last 3 or first 5" can't be done ):
		if [string eq -nocase ALL $a] {
			for { set x $lo } { $x <= $hi } { incr x } { lappend new $x }
			break
		} elseif [regexp -nocase -- {^FIRST[\:\/](\d+)$} $a - val] {
			# "LAST" is handled below
				for { set x [ expr $lo + ( $val - 1 ) ] } { $x >= $lo } { decr x } { if { $x <= $hi } { lappend new $x } }
		} elseif [regexp -nocase -- {^LAST[\:\/](\d+)$} $a - val] {
			# "LAST" is handled below
				for { set x [ expr $hi - ( $val - 1 ) ] } { $x <= $hi } { incr x } { if { $x >= $lo } { lappend new $x } }
		} elseif [regexp -nocase -- {^LAST(\-\d+)$} $a - val] {
			# "LAST" is handled below
				set end_hi $hi
				set end_lo [ expr $end_hi + $val ]
				if { $end_lo < $lo } { set end_lo $lo } { if { $end_hi > $hi } { set end_hi $hi } }
				lappend new $end_lo
		} elseif [regexp -nocase -- {^END(\-\d+)?$} $a - val] {
			set firstlast 0
			if [isempty val] {
				lappend new $hi
			} {
				set end_hi $hi
				set end_lo [ expr $end_hi + $val ]
				if { $end_lo < $lo } { set end_lo $lo } { if { $end_hi > $hi } { set end_hi $hi } }
				lappend new $end_lo
			}
		} elseif [regexp -nocase -- {^START(\+\d+)?$} $a] {
			set firstlast 0
			if [string eq -nocase START $a] {
				lappend new $lo
			} {
				set end_lo $lo
				set end_hi [ expr $end_lo + [ lindex [ split $a + ] 1 ] ]
				if { $end_hi > $hi } { set end_hi $hi }
				lappend new $end_hi
			}
		} elseif [string eq -nocase FIRST $a] { # First
#debug =first a new
			set firstlast 1
		} elseif [string eq -nocase LAST $a] { # Last
#debug =last a new
			set firstlast 2
		} elseif [regexp -- {^\-[\+\-]?\d+$} $a] { # -<integer>
#debug =range:-# a new
			regexp -- {^\-([\+\-]?\d+)$} $a - a1
			if { $a1 >= $lo } { for { set x $lo } { $x <= $a1 } { incr x } { lappend new $x } }
		} elseif [regexp -- {^[\+\-]?\d+\-$} $a] { # <integer>-
#debug =range:# a new
			regexp -- {^([\+\-]?\d+)\-$} $a - a1
			if { $a1 <= $hi } { for { set x $a1 } { $x <= $hi } { incr x } { lappend new $x } }
		} elseif [regexp -- {^\=?[\+\-]?\d+$} $a] { # Raw integer
#debug =integer a new
			regsub -- {^\=} $a "" a
			if $firstlast {
				switch -exact -- $firstlast {
					1 { for { set x $lo } { $x <= [ expr ( $lo + $a ) - 1 ] } { incr x } { lappend new $x } }
					2 { for { set x [ expr ( $hi - $a ) + 1 ] } { $x <= $hi } { incr x } { lappend new $x } }
					default { return ; error "\[SB:GET_COUNT\] Illegal value for \$FIRSTLAST: $firstlast" ; return }
				}
				set firstlast 0
				continue
			}
			lappend new $a
		} else { # Ranges
#debug =range:#-# a new
			if ![regexp -- {^([\+\-]?\d+)\-([\+\-]?\d+)$} $a] { return ; error "\[SB:GET_COUNT\] Illegal data atom: $a" ; return }
			regexp -- {^([\+\-]?\d+)\-([\+\-]?\d+)$} $a - a1 a2
			if { $a1 <= $a2 } { for { set x $a1 } { $x <= $a2 } { incr x } { lappend new $x } }
		}
	}
	# If "FIRST" or "LAST" is used without a number ....
	if $firstlast { switch -exact -- $firstlast { 1 { lappend new $lo } 2 { lappend new $hi } default { error "\[SB:GET_COUNT\] Illegal value for \$FIRSTLAST: $firstlast" ; return } } }

	# Sanity check
	foreach a $new { if { ( $a >= $lo ) && ( $a <= $hi ) } { lappend final $a } }
	set final [lsort -increasing -unique -real $final] ; # -INTEGER has 32-bit limit :(
	if $combine_list_into_ranges { set final [makenumberlist $final] }
	return $final
}

proc SB:areyousure { { word "clear" } } {return "Are you \002\037SURE\037\002?! Use \"[stu $word] [stu $word]\" to convince me."}

proc respect type {
	# For unknowns, allow them for now.
	switch -exact -- [stl $type] {

		hourly { return [sbd:nozero CONFIG:GLOBAL-HOURLY] }

		default { return 1 }
	}
}

#####
# Enter with: [LIST "-f1 param -f2 -f3 param -- text ...."] [LIST -f1 1 -f2 0 -f3 1 -flag4 0 -flag5 2]
# Returns: {ARRAY: flag (LIST:params)} {text}

proc SB:flags_process {text array} {
	array set params ""
	foreach {a b} $array {
		if ![left $a 1 -] {prepend a -}
		set params($a) $b
	}

	empty flags
	zero finished
	emptyarray processed

	do until {[isempty text]} {
		set flag [l0 $text]
		set text [lreplace $text 0 0]
		#if ![left $flag 1 -] {prepend flag -}
		if [string eq $flag --] {
		#set text [lreplace $text 0 0];
#msghome FLAG($flag):UM($um):TEXT($text)
		break
	}

		set um [uniquematch [array names params] $flag]
#msghome FLAG($flag):UM($um):TEXT($text)
		if [isempty um] {
			lprepend text $flag; # Put it back!
			break
		}
		if [notempty um] {
			if [isnull processed($um)] {empty processed($um)}
			set skip $params($um)
			if $skip {
				for {zero pass} {$pass < $skip} {incr pass} {
					set worf [l0 $text]
					if [left $worf 1 -] break; # Don't swallow the next flag; process it!
					lappend processed($um) $worf
					set text [lreplace $text 0 0]
				}
			}
		}
	}

	return [list [array get processed] $text]
}

proc SB:debug_marker arg {
	switch -exact -- [stl $arg] {

		clear - clean - begin - start {
			sbd:clear @DEBUG:*
			sbd:set @DEBUG:START [format %u [clock clicks -milli]]
			sbd:set @DEBUG:COUNT 0
		}

		set - mark {
			set count [sbd:get @DEBUG:COUNT]
			incr count
			sbd:set @DEBUG:COUNT $count
			sbd:set @DEBUG:$count [format %u [clock clicks -milli]]
		}

		end - report {
			sbd:set @DEBUG:END [format %u [clock clicks -milli]]
			set count [sbd:get @DEBUG:COUNT]
			msghomer "\[DEBUG TIMER\]:"
			for {one x} {$x <= $count} {incr x} {
				set message "\[DEBUG TIMER\]: #[comma $x]: [sbd:get @DEBUG:$x]"
				if [sbd:nozero @DEBUG:[ expr $x - 1 ]] {append message " ([ expr [ sbd:get @DEBUG:$x ] - [ sbd:get @DEBUG:[ expr $x - 1 ]] ] ms)"}
				msghomer $message
			}
			msghomer "\[DEBUG TIMER\]: Done"
			return ""
		}

		default { error "\[SB:DEBUG_PARKER\] Unknown command: [stu $arg]" }
	}
}

# See replacement PROC below
proc `SB:pingpong { { time "" } } {
	set pingpong_time 25
	switch -exact -- [string tolower $time] {

		start {
			set ::_sb_pingpong [clock seconds]
			rawoutput "PING :[clock seconds]"
			SB:pingpong:botnet
			return 0
		}

		end {
			unset -nocomplain ::_sb_pingpong
			rawoutput "PING :[clock seconds]"
			SB:pingpong:botnet
			return 0
		}

		default {
			if [info exists ::_sb_pingpong] {
				set p $::_sb_pingpong
				if { ( [ clock seconds ] - $p ) >= $pingpong_time } {
					rawoutput "PING :[clock seconds]"
					SB:pingpong:botnet
					set ::_sb_pingpong [clock seconds]
				}
				return 0
			}
			if [isempty time] { set time [now] }
			rawoutput "PING :$time"
			SB:pingpong:botnet
			return 0
		}

	}
	return 0
}

# Replacement PROC
proc SB:pingpong { { id "" } { reset false } } {
	set level [info level]
	decr level
	set last [info level $level]
	if [isempty id] {
		# Standard (immediate) mode
		#putcmdlog "\[SB:PINGPONG\] Manual mode: $last"
		rawoutput "PING :[clock seconds]"
		SB:pingpong:botnet
		return
	}
	# Let OLD definitions hold-over; first-invocations will react to
	# the really-old timer and trigger the first time within a loop
	# (which suits the purpose of this command in the first place)
	set id [stl $id]
	if $reset {
		sbd:zero @PINGPONG:$id
	} {
		if [sbd:isempty @PINGPONG:$id] { sbd:zero @PINGPONG:$id }
	}
	if [sbd:isempty PINGPONG:TIME] { sbd:set PINGPONG:TIME 30 }
	set time [sbd:get PINGPONG:TIME]
	set since [sbd:get @PINGPONG:$id]
	if { ( [clock seconds] - $since ) >= $time } {
		#putcmdlog "\[SB:PINGPONG\] Timered ([urt $time]): $last"
		rawoutput "PING :[clock seconds]"
		SB:pingpong:botnet
		sbd:set @PINGPONG:$id [clock seconds]
	}
	return
}

proc SB:pingpong:botnet args {
	foreach dcc [dcclist bot] { putidx [lindex $dcc 0] pi }
	if 0 { # Is thie actually needed / necessary?!
		foreach dcc [dcclist chat] {
			lassign $dcc idx user
			set console [console $idx]
			log "\[${user}\] pi" ; # +t will error if $::RAW-LOG = 0
		}
	} ; # End of IF 0
	return 0
}

proc SB:truncate_filename { file { length "72" } } {
	if {[string length $file] <= $length} {return $file}
	set file [file split $file]
	set file [file join [l0 $file] [l1 $file] [l2 $file] ... [lindex $file end-1] [lindex $file end]]
	return $file
}

proc SB:channel_color chan {
	if ![validchan $chan] {return ""}
	set color [sbd:get CONFIG:${chan}-COLOR]
	if [notempty color] {set color [sbd:get CONFIG:GLOBAL-COLOR]}
	return $color
}

interp alias {} asc {} ascii
interp alias {} ascw {} asciiw
proc ascii char {binary scan $char c* a; return $a}
proc asciiw char {
	binary scan $char S binary
	return [ expr $binary & 0xFFFF ]
}
proc ascr string {binary scan $string c* a; return $a}
proc aschr string {binary scan $string c* a; regexp -inline -all -- .. $a}

proc chr ascii {
	if {$ascii > 255} {chrw $ascii}
	binary format c* $ascii
}
proc chrw asciiw {binary format S $asciiw}
proc chrr list {
	empty chrr
	foreach a $list {
		append chrr [binary format c* $a]
	}
	return $chrr
}
proc chrhr string {
	empty chrhr
	foreach a [regexp -inline -all -- .. $string] {
		append chrhr [binary format H* $a]
	}
	return $chrhr
}

#proc next {number interval} {set factor [fixmath $number / $interval] ; fixmath $interval * ( floor( $factor ) + 1 ) }
# Replacement PROC (works on both sides of zero on number line)
# Will force integer results even if original number is float.
proc next [list number interval] { expr ( int( $number / $interval ) + 1 ) * $interval }
proc prev [list number interval] { expr int( ( $number - 1 ) / $interval ) * $interval }
proc nextequal [list value interval] { expr ( int ( ( $value - 1 ) / $interval ) + 1 ) * $interval }
proc prevequal [list value interval] { expr ( int ( ( $val - $interval ) / $interval ) + 1 ) * $interval }

proc interval args {
	flags:simple $args [list -next -previous -equal -integer] temp flags
	if [isempty flags] { lappend flags -next -equal } ; # Defaults
	lassign $temp value interval
	if [isempty interval] { one interval }
	if [validflag -integer] {
		if ![isint [normalize $value]] { # 1.2 --> 2, 1.0 --> 1
			int value
			lremove flags -equal
		}
	}
	if [validflag -next] {
		switch -exact -- [validflag -equal] {
			0 { return [ expr ( int( $value / $interval ) + 1 ) * $interval ] }
			1 { return [ expr ( int ( ( $value - 1 ) / $interval ) + 1 ) * $interval ] }
	
		}
	}
	if [validflag -previous] {
		switch -exact -- [validflag -equal] {
			0 { return [ expr int( ( $number - 1 ) / $interval ) * $interval ] }
			1 { return [ expr ( int ( ( $val - $interval ) / $interval ) + 1 ) * $interval ] }
	
		}
	}
	?
}


interp alias {} iif {} iff
# Enter with: SET VARIABLE [iff {condition} {true value} {false value}]
# Note: do not "{}" anything; let the TCL parser have both rounds of
# substitutions.
proc `iff {condition true { false "" } } {
	if [isempty condition] { return $false }

	# UPLEVEL because someone might do {$m > 0} instead of just "$m > 0"
	sbd:set @IFF:CONDITION $condition

	# MUST USE "!= 0" check: otherwise, values of "2" won't trigger TRUE
	# (due to -STRICT)
	# 1391962166: However, we ALSO have to allow "TRUE" "YES" & "ON" to pass
#		uplevel 1 { sbd:set @IFF:ANSWER [ expr ( [ sbd:get @IFF:CONDITION ] != 0 ) ] }
	uplevel 1 { sbd:set @IFF:ANSWER [ expr ( [ sbd:get @IFF:CONDITION ] ) ? 1 : 0 ] }
	set iff [sbd:get @IFF:ANSWER]; sbd:clear @IFF:*

	# DON'T use {$a} here: need the TCL interpreter to do both rounds
	# of variable substitutions! Otherwise, $A's value will be taken as
	# a full literal. If it's a command, it'll barph chunks:
	# if $A = "[isempty t]" then {$a} ("[isempty t]") will result in:
	# TCL error: invalid command name "[isempty t]"

	## 
	if [string is true -strict $iff] { return $true } { return $false }
}

# Replacement PROC
proc iff [list expression true [list false ""]] { expr [ uplevel expr $expression ] ? {$true} : {$false} }

#############################################################################

# [TREE code copyright notice]

# [2012-09-16 15:47:15-0700] <thommey> sure, I hereby grant you the irrevokable permission to use, modify, redistribute, relicense, re-assign copyright (not really legal in the US) and/or delete and/or do whatever you want with the code
# [2012-09-16 15:50:06-0700] <thommey> I dedicate any and all copyright interest in this code to the public domain. I make this dedication for the benefit of the public at large and to the detriment of our heirs and successors. I intend this dedication to be an overt act of relinquishment in perpetuity of all present and future rights to this code under copyright law.
# See below for additional code by Pixelz

# http://paste.tclhelp.net/?id=bnc

# Get an array: keys are bots and the values are lists of the direct children of that bot
proc SB:tree_getchildren {} {
	array set result {}
	foreach node [bots] {
		set result($node) [list]
	}
	# now append children to their parent
	foreach info [botlist] {
		foreach {bot uplink} $info break
		lappend result($uplink) $bot
		lappend result($bot) $uplink
	}
	return [array get result]
}

# Returns the bottree as a list of lines, the optional "root" arguments is the botnick of the bot you want on top. The other arguments are internal use only.
# defaults to $::botnet-nick
# Internal arguments:
# childrenarraystr is a one-time generated array representation of the direct children of bots, generated by [getallbotschildren] on the first call
# indentionlvl keep track of the current depth in the tree, it determines the "  " indention of the tree elements (and is used to detect the first call)
# endlvls keeps track of the indention levels that have already ended to determine whether to continue drawing "| | |" or not
# callsource prevents recursing into the direction we came from (botA -> botB -/-> botA)

# To "dummy" a tree with pre-set data (such as: a server map), set-up the
# variable $CHILDRENARRAYSTR (knowing it will be used by ARRAY SET) as a LIST:
# BOT {hub leaf leaf leaf ...} BOT {hub leaf leaf leaf ...} ...
# Use "" for root hub's hub: BOT {"" leaf1 leaf2 leaf2}

proc SB:tree { { root "" } { childrenarraystr "" } { indentionlvl "" } { endlvls "" } { callsource "" } { usebotlist 0 } } {
	if [isempty indentionlvl] {negone indentionlvl}
	set result ""
	# Initialize variables and children array from $childrenarraystr

	# How to check this against flat TCL?
	if {$root eq {}} {
		if [info exists ::botnet-nick] { set root ${::botnet-nick} }
	}

	# Let's avoid TCL8.5's -NOCASE for now (and check for flat TCL)
	if ![string eq "" [info commands bots]] {
		set m [lsearch -glob [stl [bots]] [stl $root]]
		if {$m > -1} {set root [matchcase $root [bots]]}
	}

	if {$childrenarraystr eq {}} { set childrenarraystr [SB:tree_getchildren] }
	if [isempty childrenarraystr] {
		if [isempty root] {return $::botnick} {return $root}
	}
	array set childrenarray $childrenarraystr
	set children $childrenarray($root)

	# Exclude direction we came from to avoid infinite recursion
	if {[set pos [lsearch -exact $children $callsource]] >= 0} { set children [lreplace $children $pos $pos] }
	# First line of the whole thing (root node)
	if {$indentionlvl < 0} {
		append result $root
	}

	# Indent and proceed with prefix-drawing ("| | | ..")
	incr indentionlvl
	set prefix "  "
	for {set i 0} {$i < $indentionlvl} {incr i} {
		if {[lsearch -exact $endlvls $i] >= 0} { append prefix " " } else { append prefix "|" }
		append prefix [space 4]
	}

	# In case this is being run as flat TCL (without Eggdrop)....
	if ![string eq "" [info commands botlist]] {set botlist [botlist]} {empty botlist}
	for {set i 0} {$i < [llength $children]} {incr i} {
		# Add a child to the list and recurse into its children
		set child [lindex $children $i]
		# not last child? "`--", else "|--"

		# If this is using the botnet, use BOTLIST to check sharing
		# (otherwise, ignore it)
		if $usebotlist {
			# If $BOTNET = "" (or NG) this will always fail.
			set m [lsearch -glob [stl $botlist] [stl [list $root $child * *]]]
			if {$m > -1} {set shareflag [lindex [lindex $botlist $m] end]}
		} {
			empty shareflag
		}

		if {$i < [llength $children]-1} {
			lappend result "$prefix|${shareflag}-$child"
		} else {
			lappend result "$prefix`${shareflag}-$child"
			lappend endlvls $indentionlvl
		}
		# Recursive call
		set result [concat $result [SB:tree $child $childrenarraystr $indentionlvl $endlvls $root]]
	}
	# Each iteration of this proc returns a list of lines to be added in the collector/caller

	return $result
}

proc SB:tree_hops {} {
	foreach bot [bots] {
		set trace [SB:hops_scan trace $bot list]
		set old $trace
		set trace [lreplace $trace 0 0]; # Delete main hub
		set trace [lreplace $trace end end]; # Delete bot itself
		if [isempty trace] {set trace $me}
		set tracePath($bot) $trace
	}
	return [array get tracePath]
}

# Earlier command (also created by thommey)
proc SB:tree_avghops {me} {
	upvar 1 tracePath tracePath
	zero totalbots; # ourselves, there's no tracepath for us
	zero hopstotal
	foreach {x path} [array get tracePath] {
		incr totalbots
		if [string eq -nocase $me [l0 $path]] { set path [l1e $path] }; # skip us as hop
		incr hopstotal [llength $path]
		incr hopstotal; # but we are a hop, so count us
	}
	return [format %.1f [expr {$hopstotal/double($totalbots)}]]
}

proc SB:findhub { { bot "" } { check "" } } {
	# Reflexive version
	if [isempty bot] {
		set hub [lindex [botlist] 0 0]
		# Flaw: a "main hub" will just list the 1st leaf (Chronologically) linked
		if [isempty check] { return $hub } { return [string eq -nocase $check $hub ] }
	}

	set botlist [botlist]
	set m [lsearch -glob [string tolower $botlist] [list [string tolower $bot] *]]
	if [string eq "" $check] {
		if { $m != -1 } { return [lindex $botlist $m 1] } return
	} {# RETURN boolean: is $CHECK actually $BOT's hub?
		if { $m != -1 } {
			if [string eq -nocase $check [lindex $botlist $m 1]] {
				return 1
			}
		}
		return 0
	}
	return
}

proc SB:tree_path {bot1 bot2} {
	set bots [bots]

	array set c [list 0 "" 1 ""]

	# What if "me" is one of the bots?
	if [string eq -nocase ${::botnet-nick} $bot1] {return [SB:hops_scan trace $bot2 list]}
	if [string eq -nocase ${::botnet-nick} $bot2] {return [lreverse [SB:hops_scan trace $bot1 list]]}

	# SB:HOPS_SCAN is part of SB6 [1287615239]
	set t_1 [SB:hops_scan trace $bot1 list]
	set t_2 [SB:hops_scan trace $bot2 list]

	# Remove self if necessary
	if {![string eq $bot1 ${::botnet-nick}] && ![string eq $bot2 ${::botnet-nick}]} {
		set t_1 [lreplace $t_1 0 0]
		set t_2 [lreplace $t_2 0 0]
	}
	set t_1 [lreverse $t_1]
	set t_2 [lreverse $t_2]

	set final [list]
	foreach a $t_1 {
		set m [lsearch -exact $t_2 $a]
		if {$m < 0} {
			lappend final $a
		} {
			set final [concat $final [lreverse [lrange $t_2 0 $m]]]
			break
		}
	}
	return $final
}

# Code written by Pixelz (2016-03-06 03:12:13 -0800): http://paste.tclhelp.net/?id=hq5
# Used with explicit permission of code author.
# [2016-03-06 03:35:22 -0800] <Pixelz> feel free to use it however you want

# Use tree:get_leaves to get a nested list of botnet hubs & leaves.
# Use the target bot's handle to "re-arrange" the botnet to the proper POV first,
# or "" to use the bot itself as the main hub (default).

# Requires TCL 8.4

proc tree:get_children {root} {
        set children [list]
        foreach subList [botlist] {
                lassign $subList bot uplink ver share
                if {[string equal -nocase $root $bot]} {
                        lappend children $uplink
                } elseif {[string equal -nocase $root $uplink]} {
                        lappend children $bot
                }
        }
        return $children
}

proc tree:get_leaves { { origin "" } } {
        if [string eq "" $origin] { set origin [lindex [bottree] 0 1] }

        # Get proper case origin bot
        set bots [bots]
        set m [lsearch -exact [string tolower [bots]] [string tolower $origin]]
        if { $m != -1 } { set origin [lindex $bots $m] }

        set result [list]
        set checked [list]
        lappend toCheck $origin
        while {[llength $toCheck] > 0} {
                set root [lindex $toCheck 0]
                if [string eq "" $root] break
                set toCheck [lrange $toCheck 1 end]
                lappend checked $root
                set children [list]
                foreach child [tree:get_children $root] {
                        if {[lsearch -exact $checked $child] == -1} {
                                if ![string eq -nocase $origin $child] {
                                        lappend children $child
                                }
                        }
                }
                if {[llength $children] > 0} {
                        lappend result [list $root $children]
                        #lappend toCheck {*}$children
                        set toCheck [concat $toCheck $children] ; # for TCL 8.4
                }

        }
        return $result
}

interp alias {} SB:bntr {} SB:botnet_traceroute
proc SB:botnet_traceroute {from to} {
	if ![isbotnick $from] { if ![islinked $from] { error "\[SB:BOTNET_TRACEROUTE\] ${from} is not linked to the botnet." } }
	if ![isbotnick $to] { if ![islinked $to] {error "\[SB:BOTNET_TRACEROUTE\] ${to} is not linked to the botnet." } }
	
	# Remove OBVIOUS situation ....
	if [string eq -nocase $to   [SB:findhub $from]] {return $to  }
	if [string eq -nocase $from [SB:findhub $to  ]] {return $from}

	# Okay, let's math this out ....
	emptyarray hubs
	foreach bot [bots] {set hubs($bot) [SB:findhub $bot]}

	# Watch out for endless loop nesting!
	zero count
	empty tr
	set current $to
	do {incr count; if {$count > [llength [bots]]} {error "\[SB:BOTNET_TRACEROUTE\] Endless loop detected!"}
		# if next hub = my hub, then: we're done:
		# ...  Attach "me" to the list and exit!
	}
	return $tr
}

proc SB:hops_scan { { traceme "" } { traceto "" } { listonly "" } } {
	upvar #0 botnet-nick bnn

	set botlist [botlist]
	set bots [bots]
	emptyarray hub hubs leaves
	empty header
	zero text_trace
	if [notempty traceto] {
		if [string eq -nocase TRACE $traceme] {
			one text_trace
			set header "Trace ($bnn -> %traceto%): "
			set traceme $traceto
		}
	}

	foreach link $botlist {
		set l [l0 $link]
		set h [l1 $link]

		if ![info exists hub($l)] {empty hub($l)}
		lappend hub($l) $h

		if ![info exists leaves($h)] {empty leaves($h)}
		if ![string eq -nocase $bnn $h] {lappend leaves($h) $l}
	}
	foreach leaf [array names leaves] {
		set data $leaves($leaf)
		set ll [llength $data]
		if $ll {set hubs($ll) $leaf}; # Remove empty reflexive
	}

	if [isempty traceme] {return [array get hubs]}

	if [string eq * $traceme] {set list_trace $bots} {set list_trace $traceme}

	empty result
	foreach traceme $list_trace {
		if [string eq -nocase $bnn $traceme] {set traceme $bnn} {set traceme [uniquematch $bots $traceme]}
		regsub -all -- %traceto% $header $traceme header_new
		if [string eq -nocase $traceme $bnn] {lappend result ${header_new}0; continue}

		if ![info exists hub($traceme)] {lappend result ${header_new}-1; continue}
		empty return stack
		zero loop
		set max [llength [bots]]
		set current $traceme
		set traceto $bnn

		do {
debug.print "> SB:HOPS_SCAN: LOOP: CURRENT($current):BNN($bnn)"
			if [string eq $current $traceto] {
				lappend stack $traceme
				if $text_trace {set header_new "[join $stack " -> "]: "}
				break
			}
			set current $hub($current)
			lprepend stack $current
			incr loop
			if {$loop > $max} {return -1}
		}
		lappend result ${header_new}$loop
	}
	if [isempty stack] {set stack $bnn}; # Presume a bot trace on itself, list mode
	if $text_trace { if [notempty listonly] { return $stack } }
	return $result
}

proc SB:tree_scan tree {# Presume that the entries here can NOT contain spaces ....
	# Let's force it into a drawn tree ....
	if ![instr $tree \n] { set tree [join $tree \n] }

	regsub -all [space]|\t $tree "" tree
	empty root
	emptyarray hubs levels
	set split [split $tree \n]

	# Read the tree, top-down
	# Version 3
	foreach line $split {
		if [isempty line] continue
		set s [split $line -]
		set bot [join [l1e $s] -]
		if [isempty bot] { set bot $line }
		if [isempty root] {set root $bot}
		set level [llength [regexp -inline -all -- {[|`]} [l0 $s]]]
		lappend levels($level) $bot
#msghome LINE($line):S($s):BOT($bot):ROOT($root):LEVEL($level):LEVELS($levels($level))
		if $level {
			set prev [ expr $level - 1 ]
			set le [lindex $levels($prev) end]
			lappend hubs($le) $bot
			lappend hubs($bot) $le
#msghome PREV($prev):LE($le):HUBS-${le}($hubs($le)):HUBS-${bot}($hubs($bot))
		}
	}
	# This RETURNed LIST can be used to rebuild the tree via SB:TREE.
	# (It's designed that way on purpose.)
	return [array get hubs]
}

proc SB:tree_trace args {
	flags:simple $args -hops [list from to] flags
	# If ($FROM != "") && ($TO == "") then $FROM->$TO; $FROM = self
	if { [isempty to] || [isbotnick $to] } {
		set to $from
		set from [none [var ::botnet-nick] $::nick]
	} {
		if ![islinked $to] { return "${to} is not linked." }
	}
	if ![isbotnick $from] {
		if ![islinked $from] { return "${from} is not linked." }
	}

	# Don't compensate for the bot's own hub: from its own POV, it's a hub to its own uplink.
	emptyarray hubs
	foreach botlist [botlist] {
		lassign $botlist bot hub version linked
		lappend hubs($hub) $bot
	}
	
	emptyarray trace
	foreach mode [list from to] {
		set bot [set $mode]
		lappend trace($mode) $bot
		do {
			foreach hub [array names hubs] {
				if { [set m [lsearch -exact -nocase $hubs($hub) $bot]] != -1 } { lappend trace($mode) $hub ; set bot $hub ; break }
			}
			if { $m == -1 } break
		}
	}
	# Arbitrary choice: find first common bot; that's the common hub
	empty common
	foreach _ $trace(from) {
		if { [set m [lsearch -exact -nocase $trace(to) $_]] != -1 } { set common $_ ; break }
	}
	if [isempty common] { error "\[SB:TREE_TRACE\] This should never happen: no common hub was found! TRACE(FROM): \{$trace(from)\} vs TRACE(TO): \{$trace(to)\}" }
	set n [lsearch -exact -nocase $trace(from) $common]
	set final [list {*}[lrange $trace(from) 0 $n] {*}[lreverse [lrange $trace(to) 0 [incr m -1]]]]
	if [validflag -hops] { return [ expr [llength $final] - 1 ] }
	return $final 
}

proc SB:tree_trace:hops args {
	if [notempty args] {
		lassign $args from to
		set hops [ expr [llength [SB:tree_trace $from $to]] - 1 ] 
	} {
		zero hops
		set bots [bots]
		foreach bot $bots {
			incr hops [ expr [llength [SB:tree_trace $bot]] - 1 ]
		}
		set hops [ expr ( $hops / double( [llength $bots] ) ) ]
	}
	format %.1f $hops
}

proc countchars args {
	if [left $args 1 -] {lassign $args flags string char pivot} {lassign $args string char pivot}
	if [isempty string] { error "\[COUNTCHARS\] syntax: COUNTCHARS <string to search through> <character(s) to find> \[characters to stop at, if found\]" }
	if [isempty char  ] { error "\[COUNTCHARS\] syntax: COUNTCHARS <string to search through> <character(s) to find> \[characters to stop at, if found\]" }
	set arg [concat * $args]
	SB:setmiscvariables -nocase

	# If the user was dumb enough to make $CHAR & $PIVOT the same ....
	if [string eq $char $pivot] {return 0}; # DON'T USE -NOCASE !! Allow for mixed case delimiters

	# Test if we even NEED to obey $PIVOT
	set m [instr $string $pivot]
	if !($m) {empty pivot}; # The limiter character doesn't exist: ignore it!

	if [isempty pivot] {
		if [validflag -nocase] {
			set regexp [regexp -nocase -inline -all -- $char $string]
		} {
			set regexp [regexp         -inline -all -- $char $string]
		}
		return [llength $regexp]
	} {# Here, we have to use INSTR....
		# Do we HAVE to do that? Can't we SPLIT against $PIVOT, then
		# run REGEXP against LINDEX #0? We can if $PIVOT is 1st character.
		# Otherwise, we'd have to run through SPLITM.
		zero total
		one eyed one horned flying purple people eater ; # (:
		do {
			set eyed [instr $string $char $eyed]
			if !($eyed) break
			if {$eyed > $m} break
			incr total
			incr eyed; # Otherwise, we'll keep matching the same character!
		}
		return $total
	}
}

proc rollchars { text args } {
	set rollchars ""
	set index 0
	if [string eq "" $args] { set args 13 }
	foreach char [split [stu $text] ""] {
		scan $char %c ascii
		set roll [lindex $args $index]
		incr index
		if [string eq "" $roll] { set roll [lindex $args 0] ; set index 0 }
		incr ascii [ expr ( -1 * $roll ) ]
		if { $ascii < 65 } { incr ascii 26 }
		if { $ascii > 90 } { incr ascii -26 }
		append rollchars [format %c $ascii]
	}
	return $rollchars
}

proc permutation string {
	int w x y z holder
	empty temp

	one maxcount
	for {set x [len $string]} {$x >= 1} {decr x} {
		set maxcount [ expr $maxcount * $x ]
	}

	emptyarray myarray chararray

	for {one x} {$x <= [len $string]} {incr x} { set chararray($x) [mid $string $x 1] }

	zero holder

	for {set w 2} {$w <= ([len $string] - 1)} {incr w} {
		for {one x} {$x <= [len $string]} {incr x} {
			for {one z} {$z <= ([len $string] - 1)} {incr z} {
				incr holder
				for {one y} {$y <= [len $string]} {incr y} {
					append myarray($holder) $chararray($y)
				}
				swap chararray($z) chararray([ expr $z + 1 ])
			}
		}
		if {$w <= [len $string]} {
			swap chararray($w) chararray([ expr $w + 1 ])
		}
	}

	empty perm
	for {one x} {$x <= $holder} {incr x} {lappend perm $myarray($x)}
	return $perm
	return [lsort -inc -uni -dict $perm]
}

# Replacement PROC below
proc `combination { set { n -1 } { presorted false } } {
	if [string eq "" [info commands dict]] { package require dict }
        if { [llength $set ] == 1 } { set set [split $set ""] }
        if !$presorted { set set [lsort -increasing -dictionary $set] }
        if { $n < 0 } { set n [llength $set] }
        if { $n < 1 } { # Why INCR?
                return
        } elseif { $n < 2 } {
                return $set
        }

        # Recursive call
        set res [combination $set [incr n -1] true]
        set result {}
        foreach item $set {
                foreach inner $res {
                        dict set result [ lsort [concat $item $inner] ] {}
                }
        }
        return [dict keys $result]
}

proc combination { set { n -1 } { presorted false } } {
     set set [join $set]
     if { [llength $set] == 1 } { set set [split $set ""] }
     if !$presorted { set set [lsort -increasing -dictionary $set] }
     if { $n < 0 } { set n [llength $set] }
     if { $n < 1 } { # Why INCR?
          return
     } elseif { $n < 2 } {
          return $set
     }

# Recursive call
     set res [combination $set [incr n -1] true]
     set result ""
     foreach item $set {
          foreach inner $res {
               array set combo [list [lsort [concat $item $inner]] ""]
               #dict set result [ lsort [concat $item $inner] ] {}
          }
     }
     return [array names combo]
     #return [dict keys $result]
}

proc getimginfo:value {hex { littleendian 0 } } {
	if [instr $hex [space]] { set hex [join $hex ""] }
	if $littleendian { set hex [join [lreverse [regexp -inline -all -- {.{1,2}} $hex]] ""] }
	dec 0x$hex
}

proc getimginfo { file var } {
	set timeformat {%Y-%m-%d %H:%M:%S}

	set ext [file extension $file]
	switch -exact -- [stl $ext] {

		.jpe - .jpeg - .jpg {
			# Look for JFIF tag
			set data [fileread $file "" 1] ; # Binary!

			if ![left $data 2 \xFF\xD8] { error "\[IMG.TCL\] Can't find JPEG signature FFD8: $file" }
			if {[instr $data JFIF] != 7} { error "\[IMG.TCL\] Can't find JFIF tag: $file" }
			binary scan $data H40 scale

			set scale_y [getimginfo:value [mid $scale 29 4]]
			set scale_x [getimginfo:value [mid $scale 33 4]]
			if {$scale_x != $scale_y} { error "\[IMG.TCL\] SCALE:X doesn't match SCALE:Y: $file" }

			# Look for 0xFFC0 or 0xFFC2 headers (big-endian)
			set in_se [instr $data \xff\xc0]
			if !($in_se) {
				set in_pe [instr $data \xff\xc2]
				if !($in_pe) { error "\[IMG.TCL\] Unable to find JPG data header FFC0 or FFC2: $file" }
				set in $in_pe
				set enc progressive
			} {
				set in $in_se
				set enc standard
			}
			incr in 2
			set header [mid $data $in]

			set in_ff [instr $header \xff]
			if !($in_ff) { error "\[IMG.TCL\] Unable to find JPG next data header or EOF: $file" }
			set header [left $header $in_ff]

			binary scan $header H* header
			lassign [split $header ""] - - - - - - y1 y2 y3 y4 x1 x2 x3 x4
#puts HEADER($header)
#puts X1($x1):X2($x2):X3($x3):X4($x4):Y1($y1):Y2($y2):Y3($y3):Y4($y4)
			set width [getimginfo:value ${x1}${x2}${x3}${x4}]
			set height [getimginfo:value ${y1}${y2}${y3}${y4}]

#puts "\[IMG.TCL\] ${file}: [regexp -inline -all -- .. $header]"
			set ctime [ctime $file]
			set mtime [file mtime $file]

#puts "\[IMG.TCL\] ${file}: \[JFIF\] ${width}x${height}x24 ([decimal [ fixmath $width / $height ] 3 1] ratio @ $scale_x DPI) ($enc encoding), CTIME:[clock format $ctime -format $timeformat] MTIME:[clock format $mtime -format $timeformat]"
			array set return [list width $width height $height depth 24 ratio [decimal [ fixmath $width / $height ] 3 1] dpi $scale_x ctime $ctime mtime $mtime encoding $enc]
		}

		.gif {
			set data [fileread $file "" 1] ; # Binary!
			binary scan $data H22 header

			# Unforunately, DPI can't be calculated (DPI isn't stored)
			# PSP5 keeps the value constant at 100 DPI despite resizing.
			# Other programs assume 72 DPI (a common default value)

			set version [binary format H* [mid $header 1 12]]
			set width   [getimginfo:value [mid $header 15 2][mid $header 13 2]]
			set height  [getimginfo:value [mid $header 19 2][mid $header 17 2]]
			set depth   [ expr int( pow( 2 , 1 + ( [ getimginfo:value [ right $header 1 ] ] & 7 ) ) ) ]
			zero interlace
			set ctime [     ctime $file]
			set mtime [file mtime $file]

			# Image descriptor?
			set in_id [instr $data ,]
			if $in_id {
				incr in_id 5; # The top-left & top-right can be skipped
				binary scan [mid $data $in_id 12] H10 header
				set width     [getimginfo:value [mid $header 3 2][mid $header 1 2]]
				set height    [getimginfo:value [mid $header 7 2][mid $header 5 2]]
				set value     [getimginfo:value [mid $header 9 2]]
				set depth     [ expr int( pow( 2 , 1 + ( $value & 7 ) ) ) ]
				set interlace [ expr ( $value & 64 ) > 0 ? 1 : 0]
			}

#puts "\[IMG.TCL\] ${file}: \[${version}\] ${width}x${height}x${depth} ([decimal [ fixmath $width / $height ] 3 1] ratio) (interlace:[noyes $interlace]), CTIME:[clock format $ctime -format $timeformat] MTIME:[clock format $mtime -format $timeformat]"
			array set return [list width $width height $height depth $depth ratio [decimal [ fixmath $width / $height ] 3 1] interlace $interlace ctime $ctime mtime $mtime]
		}

		.png {
			# Look for header "IHDR" (big-endian)
			# IHDR(32bit width)(32-bit height)(8-bit bit depth)
			# (8-bit colour type)(8-bit compression method)
			# (8-bit filter method)(8-bit interlace method)

			# Look for header "Creation time\000"
			# Look for header "tIME"

			set data [fileread $file "" 1] ; # Binary!
			binary scan [left $data 12] H* signature
			if ![string eq -nocase 89504E470D0A1A0A0000000D $signature] { error "\[IMG.TCL\] PNG header not valid or is corrupted: $file" }

			set ihdr [instr $data IHDR]
			if !($ihdr) { error "\[IMG.TCL\] Unable to find IDHR header in PNG file: $file" }
			incr ihdr 4
			binary scan [mid $data $ihdr] H26 header

			set width  [getimginfo:value [mid $header 1 8]]; # Little endian
			set height [getimginfo:value [mid $header 9 8]]
			set bit_depth [getimginfo:value [mid $header 17 2]]; # Remember: bits PER CHANNEL (RGBa)!
			set colour_type [getimginfo:value [mid $header 19 2]]
			array set colour_type_array [list 0 1 2 3 3 1 4 2 6 4]
			set bit_depth [ expr $bit_depth * $colour_type_array($colour_type) ]
			set colour_type_text [lindex [list grayscale * RGB INDEX grayscale+alpha * RGB+alpha] $colour_type]

			set method_compression [lindex [list LZ77] [getimginfo:value [mid $header 21 2]]]
			set method_filter [lindex [list filter:adaptive+5] [getimginfo:value [mid $header 23 2]]]
			set method_interlace [lindex [list non-interlaced Adam7-interlace] [getimginfo:value [mid $header 25 2]]]

			# Look for header "Creation Time\000"
			set in_ctime [instr $data "Creation Time\000"]
			if $in_ctime {
				incr in_ctime 14
				binary scan [mid $data $in_ctime] H14 header_ctime
				set ctime_year   [getimginfo:value [mid $header_ctime  1 4]]
				set ctime_month  [getimginfo:value [mid $header_ctime  5 2]]
				set ctime_day    [getimginfo:value [mid $header_ctime  7 2]]
				set ctime_hour   [getimginfo:value [mid $header_ctime  9 2]]
				set ctime_minute [getimginfo:value [mid $header_ctime 11 2]]
				set ctime_second [getimginfo:value [mid $header_ctime 13 2]]
				set ctime [clock scan [addzero $ctime_year 2]-[addzero $ctime_month 2]-[addzero $ctime_day 2] [addzero $ctime_hour 2]:[addzero $ctime_minute 2]:[addzero $ctime_second 2]]
			} {
				#error "\[IMG.TCL\] Unable to find CTIME for: $file"
				empty ctime
			}

			# Look for header "tIME"
			set in_mtime [instr $data "tIME"]
			if $in_mtime {
				incr in_mtime 4
				binary scan [mid $data $in_mtime] H14 header_mtime
				set mtime_year   [getimginfo:value [mid $header_mtime  1 4]]
				set mtime_month  [getimginfo:value [mid $header_mtime  5 2]]
				set mtime_day    [getimginfo:value [mid $header_mtime  7 2]]
				set mtime_hour   [getimginfo:value [mid $header_mtime  9 2]]
				set mtime_minute [getimginfo:value [mid $header_mtime 11 2]]
				set mtime_second [getimginfo:value [mid $header_mtime 13 2]]
				set mtime [clock scan [addzero $mtime_year 2]-[addzero $mtime_month 2]-[addzero $mtime_day 2] [addzero $mtime_hour 2]:[addzero $mtime_minute 2]:[addzero $mtime_second 2]]
			} {
				#error "\[IMG.TCL\] Unable to find CTIME for: $file"
				empty mtime
			}
			# Check for missing times (either CTIME or MTIME)
			if [isempty ctime] {set ctime $mtime} { if [isempty mtime] { set mtime $ctime } }

			set in_phys [instr $data pHYs]
			if $in_phys {
				binary scan [mid $data $in_phys] H26 phys
				set dpi_x [getimginfo:value [mid $phys  9 8]]
				set dpi_y [getimginfo:value [mid $phys 17 8]]
				set dpi_t [getimginfo:value [mid $phys 25 2]]
#puts X($dpi_x):Y($dpi_y)
				if {$dpi_x == $dpi_y} {
					# Square pixels
					if $dpi_t {
						# Unit = metre ( inch = 0.0254 m ( 2.54 cm ) )
						set dpi "[decimal [ fixmath $dpi_x * 0.0254 ] 0 1] DPI"
					} {
						set dpi "$dpi_x units (unknown)"
					}
				} {
					# Weird-shaped pixels
					if $dpi_t {
						# Unit = metre ( inch = 0.0254 m ( 2.54 cm ) )
						set dpi_xa [decimal [ fixmath $dpi_x * 0.0254 ] 0 1]
						set dpi_ya [decimal [ fixmath $dpi_y * 0.0254 ] 0 1]
						set dpi "${dpi_xa}:${dpi_ya} DPI"
					} {
						set dpi "${dpi_x}:${dpi_y} units (unknown)"
					}
				}
			} {
				set dpi unknown
			}

#puts "\[IMG.TCL\] ${file}: \[PNG\] ${width}x${height}x${bit_depth} $colour_type_text ([decimal [ fixmath $width / $height ] 3 1] ratio @ $dpi) ($method_compression $method_interlace $method_filter), CTIME:$ctime MTIME:$mtime"
			array set return [list width $width height $height depth $bit_depth dpi $dpi ctime $ctime mtime $mtime ratio [decimal [ fixmath $width / $height ] 3 1] colors $colour_type_text extra [list $method_compression $method_interlace $method_filter]]

		}

		.bmp {
			set data [fileread $file "" 1] ; # Binary!
			array set types [list BM Windows BA {OS/2 struct bitmap array} CI {OS/2 strict color icon} CP {OS/2 const color pointer} IC {OS/2 strict icon} PT {OS/2 pointer}]
			set e [catch { set type $types([left $data 2]) } shit]
			if $e { error "\[IMG.TCL\] Unknown BMP file type: $shit" }
			set type [l0 $type]
			binary scan [mid $data 15 40] H* header
			set size [getimginfo:value [left $header 8] 1]
			switch -exact -- $size {
				12 {
					# BITMAPCOREHEADER2
					# OS21XBITMAPHEADER
					# WIN2XBITMAPHEADER
					set width    [getimginfo:value [mid $header  9 8] 1]
					set height   [getimginfo:value [mid $header 17 8] 1]
					set width    [ expr $width  & 0x7FFFFFFF ]; # Signed 32-bit integer
					set height   [ expr $height & 0x7FFFFFFF ]; # High bit is reserved for interlace direction (top-down or bottom-up)
					set depth    [getimginfo:value [mid $header 29 4] 1]
					set dpi      unknown
					set compress unknown
				}
				40 {
					# BITMAPINFOHEADER
					# WIN3XBITMAPHEADER
					# WINNTBITMAPHEADER
					set width    [getimginfo:value [mid $header  9 8] 1]
					set height   [getimginfo:value [mid $header 17 8] 1]
					set width    [ expr $width  & 0x7FFFFFFF ]; # Signed 32-bit integer
					set height   [ expr $height & 0x7FFFFFFF ]; # High bit is reserved for interlace direction (top-down or bottom-up)
					set depth    [getimginfo:value [mid $header 29 4] 1]
					set comp     [getimginfo:value [mid $header 33 8] 1]
					set compress [lindex [list RGB RLE8-bit RLE4-bit BitFields JPEG PNG AlphaBitFields] $comp]
					set scale_x  [getimginfo:value [mid $header 49 8] 1]
					set scale_y  [getimginfo:value [mid $header 57 8] 1]
					set dpi [nozero [decimal [ fixmath $scale_x * 0.0254 ] 3 1]]
					if { $scale_x != $scale_y } {
						append dpi ":[nozero [decimal [ fixmath $scale_y * 0.0254 ] 3 1]]"
					}
				}

				52 {
					# BITMAPV2INFOHEADER
				}
				56 {
					# BITMAPV3INFOHEADER
				}
				64 {
					# BITMAPCOREHEADER2
					# OS22XBITMAPHEADER
					set width    [getimginfo:value [mid $header  9 8] 1]
					set height   [getimginfo:value [mid $header 17 8] 1]
					set width    [ expr $width  & 0x7FFFFFFF ]; # Signed 32-bit integer
					set height   [ expr $height & 0x7FFFFFFF ]; # High bit is reserved for interlace direction (top-down or bottom-up)
					set depth    [getimginfo:value [mid $header 29 4] 1]
					set comp     [getimginfo:value [mid $header 33 8] 1]
					set compress [lindex [list RGB RLE8-bit RLE4-bit BitFields JPEG PNG AlphaBitFields] $comp]
					set scale_x  [getimginfo:value [mid $header 49 8] 1]
					set scale_y  [getimginfo:value [mid $header 57 8] 1]
					set dpi [nozero [decimal [ fixmath $scale_x * 0.0254 ] 3 1]]
					if { $scale_x != $scale_y } {
						append dpi ":[nozero [decimal [ fixmath $scale_y * 0.0254 ] 3 1]]"
					}
				}
				108 {
					# BITMAPV4INFOHEADER
					# WIN4XBITMAPHEADER
					set width    [getimginfo:value [mid $header  9 8] 1]
					set height   [getimginfo:value [mid $header 17 8] 1]
					set width    [ expr $width  & 0x7FFFFFFF ]; # Signed 32-bit integer
					set height   [ expr $height & 0x7FFFFFFF ]; # High bit is reserved for interlace direction (top-down or bottom-up)
					set depth    [getimginfo:value [mid $header 29 4] 1]
					set comp     [getimginfo:value [mid $header 33 8] 1]
					set compress [lindex [list RGB RLE8-bit RLE4-bit BitFields JPEG PNG AlphaBitFields] $comp]
					set scale_x  [getimginfo:value [mid $header 49 8] 1]
					set scale_y  [getimginfo:value [mid $header 57 8] 1]
					set dpi [nozero [decimal [ fixmath $scale_x * 0.0254 ] 3 1]]
					if { $scale_x != $scale_y } {
						append dpi ":[nozero [decimal [ fixmath $scale_y * 0.0254 ] 3 1]]"
					}
				}
				124 {
					# BITMAPV5INFOHEADER
				}
				240 {
					# Unknown OS/2 2.X version header
				}
			}
			set ctime [     ctime $file]
			set mtime [file mtime $file]

#puts "\[IMG.TCL\] ${file}: ${width}x${height}x${depth} ([decimal [ fixmath $width / $height ] 3 1] ratio @ $dpi DPI) (compression:$compress), CTIME:[clock format $ctime -format $timeformat] MTIME:[clock format $mtime -format $timeformat]"
			array set return [list width $width height $height depth $depth dpi $dpi ctime $ctime mtime $mtime ratio [decimal [ fixmath $width / $height ] 3 1] colors {} extra $compress]
		}

		default { error "\[IMG.TCL\] Unknown image format: $ext ($file)" }

	}

	# Make sure all values are present: which are most likely to be missing?
	# Default: <depends>
	foreach a [list encoding] { if ![info exists return($a)] { set return($a) default } }
	# Default: blank
	foreach a [list colors depth extra] { if ![info exists return($a)] { empty return($a) } }
	# Default: ZERO
	foreach a [list dpi interlace] { if ![info exists return($a)] { zero return($a) } }

	upvar 1 $var local
	unset -nocomplain local
	array set local [array get return]
	return [array get return]
}

proc arraysplit {text splitter var { start 0 } } {
	upvar 1 $var local
	array set local [list]
	set c [ expr $start - 1 ]
	foreach a [split $text $splitter] { set local([incr c]) $a }
	return [array get local]
}

proc arrayjoin {array joiner { var "" } { start 0 } } { # $START won't be used anymore
	# This may be less logical as "var / start" than "start / var" order.
	upvar 1 $array local_array
	if [notempty var] { upvar 1 $var local } ; # Like REGSUB: set variable or return value

	# Sanity check: subscripts MUST BE INTEGERS (pairs with ARRAYSPLIT)
	foreach a [array names local_array] {
		if ![isnum -real $a] { error "\[ARRAYJOIN\] Subscripts of array \"${array}\" (\"${array}(${a})\") must be \037consecutive\037 integers" }
	}
	# Must be CONSECUTIVE sequence: skipped entries will not be ignored!
	empty arrayjoin
	foreach x [lsort -increasing -real [array names local_array]] { lappend local $joiner }
	set local [join $local $joiner] ; # Needed because of UPVAR above
}

proc time:tohex { { ts "" } { gmt true } } {
	if [isempty ts] { set ts [now] }
	if ![isint $ts] { set ts [clock scan [regsub -all ,  $ts ""]] }
	lassign [clock format $ts -format [list %u %Y %m %d %H %M %S] -gmt $gmt] u y m d h n s

	# Eliminate leading zeroes ( %k could replace %H )
#debug.vars [SB:home] =0 u y m d h n s
	foreach a [list u y m d h n s] { set $a [nozero [set $a]] }
	set double [ expr ( $h >= 12 ) ? 2 : 1 ]
#debug.vars [SB:home] =1 u y m d h n s double
	foreach a [list h n s] { set $a [ expr $double * [ set $a ] ] }
	foreach {a b} [list h 24 n 60 s 60] { if ![set $a] { set $a $b } }
#debug.vars [SB:home] =2 u y m d h n s double
	incr m [ expr $u * 16 ]
	incr d [ expr [ isleapyear $y ] * 128 ]
	incr h [ expr ( ( $gmt == true ) ? 1 : 0 ) * 128 ]
#debug.vars [SB:home] =3 u y m d h n s double
	return [right [hex $y] 2][addzero [hex $m] 2][addzero [hex $d] 2][addzero [hex $h] 2][addzero [hex $n] 2][addzero [hex $s] 2]
}

proc time:fromhex { { ts "" } { gmt false } } {
	if [isempty ts] { set ts [time:tohex [now]] }
	if ![regexp -nocase -- {^[0-9A-F]{2}[0-7][0-9A-F][0-9A-F]{2}[0-9A-F]{2}[0-9A-F]{2}[0-9A-F]{2}$} $ts] { error "\[TIME:TOSTD\] Illegally-formated timestamp: $ts" }

	regexp -nocase -- {^([0-9A-F]{2})([0-7])([0-9A-F])([0-9A-F]{2})([0-9A-F]{2})([0-9A-F]{2})([0-9A-F]{2})$} $ts -> y u m d h n s
	prepend y 07

	#foreach a [list y m d] { set $a [addzero [ dec 0x[set $a] ] 2]}
	foreach a [list y m d h n s] { set $a [ dec 0x[set $a] ] }
	if { $d & 0x80 } { set d [ expr $d & 0x7F ] }
	if { $h & 0x80 } { set h [ expr $h & 0x7F ] ; set gmt true }

	foreach {a b} [list h 24 n 60 s 60] { if { [set $a] == $b } { set $a 0 } }
	set test [dec 0x$h]
	foreach a [list h n s] {
		 if { $test >= 24 } { set $a [ expr [ set $a ] / 2 ]}
		 set $a [addzero [set $a] 2]
	}
#debug.vars u y m d h n s
	#return "[lindex [list Sun Mon Tues Wednes Thurs Fri Satur Sun] $u]day, ${y}-${m}-${d}, ${h}:${n}:${s}"
	return [clock format [clock scan "[lindex [list Sun Mon Tues Wednes Thurs Fri Satur Sun] $u]day ${y}-${m}-${d} ${h}:${n}:${s}" -gmt $gmt] -format "%Y-%m-%d %H:%M:%S"]
}

proc ispackage package { not [ catch { package present $package } ] }

proc getmd5 text {
	if ![ispackage md5] {
		set e [ catch { package require md5 } shit]
		if $e { error "\[GETMD5\] I can't load the MD5 package. My owner needs to install it!" }
	}
	set md5 [md5 $text]
	if ![regexp -nocase -- {^[0-9A-F]+$} $md5] {
		# This is an older version of MD5 that returned a 256-bit binary
		# version of the MD5 hash. Let's convert it to hexadecimal format
		# Somewhere between 2.0.2 & 2.0.7, the package was updated to
		# return the hex version.
		binary scan $md5 H* md5
	}
	return $md5
}

proc getmd5file file { getmd5 [readfile $file "" 0 ] }

proc nsfetch {ns command} {
	set ::nsfetch_command $command
	set error [ catch {
		namespace eval $ns { namespace export $::nsfetch_command }
		namespace import -force ${ns}::$command
	} shit ]
	unset ::nsfetch_command
	if $error { error "\[NSFETCH\] Unable to fetch \"${command}\" ($shit)" }
	if ![iscommand $command] { error "\[NSFETCH\] Unable to fetch \"${command}\" (unknown error)" }
	if ![string eq [info body $command] [info body ::${ns}::$command]] { error "\[NSFETCH\] Unable to fetch \"${command}\" (unknown error)" }
	return $command
}

proc explode args {
	#set args [lflatten $args]
	set args [join $args]
	SB:getflags $args [list -reverse -uppercase -lowercase -nospaces] text flags
	# Do operations on pre-exploded text: less characters (spaces) to process
	if [validflag -uppercase] { set text [stu $text] }; # Whichever is last wins
	if [validflag -lowercase] { set text [stl $text] }; # Whichever is last wins
	if [validflag -nospaces ] { regsub -all -- " " $text "" text }
	if [validflag -reverse  ] { set text [reverse $text] }
	split $text [null]
}

proc implode args {
	set args [join $args]
	SB:getflags $args [list -reverse -uppercase -lowercase -nonulls] text flags
	# Do operations on pre-exploded text: less characters (spaces) to process
	if [validflag -uppercase] { set text [stu $text] }; # Whichever is last wins
	if [validflag -lowercase] { set text [stl $text] }; # Whichever is last wins
	if [validflag -nonulls  ] { set text [lsearch -all -inline -not $text ""] }
	if [validflag -reverse  ] { set text [reverse $text] }
	join [join $text] [null]
}

proc showallenc {nick filename { size 10 } } {
	#set file [readfile $filename "" 1]
	foreach encoding [lsort -ascii [encoding names]] {
		set str [open $filename r]
		fconfigure $str -encoding $encoding
		print $nick "\[SHOWALLENC\] ${encoding}: [read $str $size]\[....\]"
		close $str
	}
}

proc checkrange { list range { bottom 0 } { top 2147483647 } } {
	empty checkrange
	foreach arg $list {
		if [iswithin $range $arg $bottom $top] { lappend checkrange $arg }
	}
	return $checkrange
}

proc findinvars text {
	empty o
	foreach _ [lsort -inc -uni -dict [info vars ::*]] {
		if [array exists $_] {
			foreach __ [lsort -inc -uni -dict [array names $_]] {
				set value [set ${_}($__)]
				if [string match -nocase *${text}* $value] { lappend o "${_}\[$__\]: $value" }
			}
		} {
			set value [set $_]
			if [string match -nocase *${text}* $value] { lappend o "${_}: [set $_]" }
		}
	}
	return $o
}

proc findinprocs text {
	empty list
	foreach proc [lsort -inc -uni -dict [info procs *]] {
		set body [info body $proc]
		zero count
		foreach line [split $body \n] {
			incr count
			if { [string match -nocase $text $line] || [instr $line $text] } { lappend list "${proc}\\${count}: $line" }
		}
	}
	set ll [llength $list]
	for { set _ 0 ; set c 1 } { $_ < $ll } { incr _ ; incr c } { lset list $_ "${c}/${ll}: [lindex $list $_]" }
	return $list
}

proc procdef { proc args } {
	# { tab2space 5 }
	if [isempty args] { set lines all } {
		if [string eq * $args] { set lines all } { set lines [splitnumberlist $args] }
	}
	set alias [interp alias {} $proc]
	if ![string eq "" $alias] { return "interp alias \{\} $proc \{\} $alias" }
	if [string eq "" [info procs $proc]] {
		if ![string eq "" [info commands $proc]] { error "\[PROCDEF\] Command \"${proc}\" is intrinsic (binary) code and can not be displayed." }
		error "\[PROCDEF\] No such PROC: $proc"
	}
	tab2space = 5
	if { [lsearch -exact [list tab \t] [string tolower $tab2space]] != -1 } {
		set tab \t
		set tab2space 0
	} {
		set tab [string repeat " " $tab2space]
	}
	empty args code
	foreach arg [info args $proc] {
		set def [info default $proc $arg a]
		if $def { lappend args "\{$arg \"${a}\"\}" } { lappend args $arg }
	}
	set args [join $args]
	set body [info body $proc]
	if $tab2space { regsub -all -- \t $body $tab body }
	empty source
	catch { # TCL 8.6 ....
		set disassemble [tcl::unsupported::disassemble proc $proc]
		set source [lsearch -inline -nocase -regexp [split $disassemble \n] {^[ ]*FILE }]
		if [notempty source] {
			lassign $source - filename - line
			set source "# Source: ${filename}:${line}"
		}
	}
	if [string eq -nocase ALL $lines] { return "${source}[iff [notempty source] \n]proc $proc \{ ${args} \} \{[regsub -all -- {[ \t]+\n} $body \n]\}" }
	set split [split $body \n]
	zero offset
	while { [string eq "" [l0 $split]] } { set split [lreplace $split 0 0] ; incr offset }
	empty o
	if [notempty source] { lappend o $source }
	foreach a [lsort -increasing -integer -unique $lines] {
		lappend o "\[${proc}\\${a}\] [string trimleft [lindex $split [ expr $a - 1 - $offset ]] " \t"]"
	}
	join $o \n
}

proc arrayisempty array {
	upvar 1 $array local
	if ![array exists local] { error "\[ARRAYISEMPTY\] Variable is scalar: $array" } ; # Is scalar
	if [string eq "" [array get local]] { return 1 }
	return 0
}

proc arrayhasempty array {
	upvar 1 $array local
	if ![array exists local] { error "\[ARRAYHASEMPTY\] Variable is scalar: $array" } ; # Is scalar
	if [string eq "" [array get local]] { return 1 }
	foreach a [array names local] { if [string eq "" $local($a)] { return 1 } }
	return 0
}

proc arrayfindempty array {
	upvar 1 $array local
	if ![array exists local] { error "\[ARRAYFINDEMPTY\] Variable is scalar: $array" } ; # Is scalar
	if [string eq "" [array get local]] { return 1 }
	empty arrayfindempty
	foreach a [array names local] { if [string eq "" $local($a)] { lappend arrayfindempty $a } }
	return $arrayfindempty ; # Null = all elements have data
}

proc sb:error { nick header { text "" } { print printh } } {
	# Optional RETURN-level? Hmmmm....
	if [isempty text] { swap header text }
	set level [ expr [ info level ] - 1 ]
	if [isempty header] { set header [l0 [info level $level]] }
	$print $nick "[iff [notempty header] "\[[stu $header]\] "]$text"
	return -code return 0 ; # The old method threw a "RETURN 0"
}

proc escape text { regsub -all -- {\\|\{|\}|\[|\]|\$|\;|\"} $text {\\&} }
proc escape args {
	flags:simple $args -double text flags
	set text [lindex $text 0] ; # Due to ARGS
	regsub -all -- {[\$\"\{\}\[\]\;\\]} $text {\\&} text
	if [validflag -double] { set text [escape $text] }
	return $text
}

# NotePad++ requires a "\" in front of the ";" to not comment-out 1/2 the line (in case it fails)
proc escape:unicode string { subst [string map {\n {\\u000a}} [regsub -all {[][{}\;#\\\$ \r\t\u0000-\u001F\u0080-\uFFFF]} $string {[format \\\\u%04x [scan "\\&" %c]]}]] }
proc unescape:unicode string { subst $string }

interp alias {} utf8uni {} utf8unicode
proc utf8unicode code {
	lassign [split $code ""] 1 2 3 4 5 6
	return ${2}[format %X [ expr ( ( 0x$3 & 3 ) << 2 ) | ( 0x$4 >> 2 ) ]][format %X [ expr ( ( 0x$4 & 3 ) << 2 ) | ( 0x$5 & 3 ) ]]$6
}

foreach _LOOP [list + - * / %] { proc $_LOOP { 1 2 } "expr \$1 $_LOOP \$2" } ; unset _LOOP
proc ^ { 1 2 } { expr pow( $1 , $2 ) }
proc ?: { 1 2 3 } { expr $1 ? $2 : $3 }

proc SB:filt { idx text } {
	if [regexp -nocase -- {^[\(\[][\d\-\/\:\. ]+[\]\)] [<\-][^ ]+ (.+)} $text - int] {
#		putlog "\[SB6:FILT\] Intercepted text: $int"
	}
	if [notempty int] { return $int }
	return $text
	# return "" to make the bot not process the line, or return the text to process ($text = original)
}

### Trace commands ###
### Stack commands ###

proc stack:trace { args } {
	if [isnum -integer $args] { set deep $args } { set deep [info level] }
	empty o
	lappend o "\[STACK:TRACE\] Tracing to: [info level [ expr [ info level ] - 1 ]]"
	for { set x 1 } { $x < $deep } { incr x } { lappend o "\[STACK:TRACE\] [string repeat " " [ expr ( $x - 1 ) * 3 ]]Level #${x}: [info level $x]" }
	return $o
}

proc trace:me args {
	# Enter with: TRACE <code> (TCL will perform a round of substitution before arriving)
	# Enter with: TRACE {<code>} (TCL will not perform substitution; text will arrive exactly as coded)

	# See also: PRINT:IDENT to indent for heirarchy
	set deep [info level]
	empty o
	lappend o "\[TRACE - level #${deep}\] $args"
	for { set x [ expr $deep - 1 ] } { $x > 0 } { incr x -1 } { lappend o "\[TRACE - level #${x}\] [info level $x]" }
	uplevel 1 eval $args
	return $o
}

proc trace:log args {
	upvar 1 __args moo
	set moo $args ; # UPVAR won't allow an already-defined variable be used directly
	uplevel 1 { set __tracelog [trace $__args] }
	upvar 1 __tracelog tracelog
	set c 0
	foreach a [lreverse $tracelog] { incr c 3 ; putlog "[string repeat " " $c]$a" }
	return 0
}

proc get:levels args {
	if [string eq -nocase HELP $args] { return "get:levels ?-show -last?" }
	flags:simple $args [list -show -last] args flags
	l == [info level] - 2
	empty level
	for { set _ 1 } { $_ <= $l } { incr l } {
		lappend level [iff [validflag -show] [info level $l] [l0 [info level $_]]]
	}
	incr l
	lappend level [iff [validflag -show -last] [info level $l] [l0 [info level $l]]]
	join $level " -> "
}

#############################################################################
#############################################################################
#############################################################################
#############################################################################
#############################################################################

proc @encrypt:bitch { command nick } {
	if [isfalse [sbd:get @ENCRYPT:compensated]] {
		noticer $nick "\[[stu ${command}]\] [stu $command] failure: command no longer available."
		noticer $nick "Due to Eggdrop core making the blowfish->ENCRYPT/DECRYPT mode change (ecb->cbc) in Eggdrop 1.8.1+sslserverlist, ENCRYPTION IS BROKEN IN A NON-BACKWARD-COMPATIBLE WAY!"
		noticer $nick "Because of this change, ENCRYPT works like MD5 where different hashes are created (but all verify to the original message). As such, I can not safely transmit encrypted data to other bots. The [stu $command] command is broken until this situation is fixed properly."
		noticer $nick "For more information, see:"
		noticer $nick "[space 5]https://github.com/eggheads/eggdrop/issues/384"
		noticer $nick "[space 5]https://github.com/eggheads/eggdrop/commit/33b12cf5297aeb3895facbb2f6a5a484f576fdc9"
		return 1
	}
	return 0
}

proc SB:get_version args {
	set ::stormbot(version) 6.0
	set ::stormbot(distro) http://www.stormbot.org/
	if [file exists scripts/sb6_version.txt] {
		set data [readfile scripts/sb6_version.txt]
		set ts [l0 $data]
		set ver [l1 $data]
		set www [l2 $data]
		regsub -all -- {stormbot|.tar.gz} $ver "" ver
		set ::stormbot(version) "$ver :: [formatedate $ts] ($www)"
		set ::stormbot(distro) $www
	}
}

proc notify:coder [list nick msg] {
	set level [info level]
	decr level
	printh $nick "An error has occurred: $msg"
	printh $nick "File a bug at: http://bug.stormbot.org/"
	if [SB:checkaccess [SB:commandlist level debug] [nick2hand $nick]] {
		printh $nick "Include the output of the DEBUG command (you have access to the command: \"$::botnick debug all\")"
	}
	printh $nick "Include the following in the bug report:"
	printh $nick "\"LEVEL ${level}: [info level $level]\""
	printh $nick ""
	printh $nick "This was a \"trapped\" error\; the command will continue executing now ...."
	return 0
}

proc array2scalar array {
	set ::_array2scalar_array $array
	uplevel 1 {
		if [array exists $::_array2scalar_array] {
			foreach [list ::_scalar2array_name ::_scalar2array_value] [array get $::_array2scalar_array] {
				set $::_scalar2array_name $::_scalar2array_value
			}
		}
	}
	foreach _ [info global ::_array2scalar*] { unset $_ }
	return
}

proc scalar2array { array args } {
	set ::_scalar2array_array $array
	set ::_scalar2array_args $args
	uplevel 1 {
		foreach ::_scalar2array_arg $::_scalar2array_args {
			if [info exists $::_scalar2array_arg] {
				array set $::_scalar2array_array [list $::_scalar2array_arg [set $::_scalar2array_arg]]
			}
		}
	}
	foreach _ [info vars ::_scalar2array*] { unset $_ }
	return
}

proc dot args { return \xE2\x80\xA2 }

#############################################################################

proc log { level { text "" } } {
	if [string eq -nocase HELP $level] { return "log ?\"file\" | log level ( 6 | d | c | o | <custom> )? <text to log>" }
	if [regexp -nocase -- {^(LOG)?FILE$} $level] {
		set a [open ${::nick}.log a+]
		puts $a "\[[clock format [clock seconds] -format {%Y-%m-%d %H:%M:%S %z}]\] $text"
		flush $a
		close $a
		# Fall-through ....
	}
	if [isempty text] {
		set text $level
		set level 6
	}
	if ![sbd:nozero @SCRIPT_IS_LOADED] { set level o }
	if ![catch { putloglev $level * "\[[clock format [clock seconds] -format {%Y-%m-%d %H:%M:%S %z}]\] $text" } ] return
	if ![catch { putloglev 6      * "\[[clock format [clock seconds] -format {%Y-%m-%d %H:%M:%S %z}]\] $text" } ] return
	if ![catch { putloglev d      * "\[[clock format [clock seconds] -format {%Y-%m-%d %H:%M:%S %z}]\] $text" } ] return
	if ![catch { putloglev c      * "\[[clock format [clock seconds] -format {%Y-%m-%d %H:%M:%S %z}]\] $text" } ] return
	if ![catch { putloglev o      * "\[[clock format [clock seconds] -format {%Y-%m-%d %H:%M:%S %z}]\] $text" } ] return
	putlog "\[[clock format [clock seconds] -format {%Y-%m-%d %H:%M:%S %z}]\] $text"
	return
}


proc log:debug { level { text "" } } {
#%#
}


#############################################################################

# Must be last: SB:BOOTSTRAP
proc SB:bootstrap args {#\u53\102\u3atest\u5f\x6eu\x6dbe\x72
	# 2018-08-10 21:02:00 -0700
	# Need to immediately load the channel file data. Reason: SERVMON (and some day, others)
	# invoke SB:HOME. If the bot's config doesn't have the CHANNEL ADD command, but it's in
	# the CHANNEL file, the call to SB:HOME in between will cause an "illegal home channel"
	# warning (triggered by SB:HOME, during the bootstrap call for SERVMON). The only
	# solutions are: (1) tell bot owners to update their config files with CHANNEL ADDs,
	# (2) forcefully l oad the chanfile here before we arrive here, or, (3) put this on a
	# delay (AFTER 1 SB:BOOTSTRAP, for example). I don't know what option #3 would do, 
	# damage-wise. Found by Dat.
	if [info exists ::chanfile] { if [file exists $::chanfile] { uplevel #0 source $::chanfile } }

	# Version defaults
	catch { if [sbd:isempty @VERSION] { sbdset @VERSION 6.0.[file mtime [info script]] } }
	catch { if [sbd:isempty @UPDATED] { sbdset @UPDATED [file mtime [info script]] } }

	SB:register variable user debuglevel
	SB:register death tdb:death
	SB:uname update

	# https://github.com/eggheads/eggdrop/issues/384 -- I win!
	# https://github.com/eggheads/eggdrop/issues/397 -- I win!
	if ![info exists blowfish-use-mode] { set blowfish-use-mode ecb }

	utimer 0 tdb:setup
	utimer 0 SB:detect_wan_ip ; # Must be on a timer (will I/O-block bot start up)

	# Generically needed binds
	bind BOT - SBD:TRANSFER SB:sbd_receive
	bind BOT - SBD:CONFIRM  SB:sbd_confirm
	bind BOT - SB:PIPETOIDX SB:pipetoidx
	bind dcc tm debuglevel  debug.level
	bind dcc tm debug.level debug.level
	bind RAW  - MODE SB:mode_opstatus_raw_mode
	bind RAW  - 353  SB:mode_opstatus_raw_353
	bind SIGN - *    SB:mode_opstatus_sign
	bind SPLT - *    SB:mode_opstatus_splt
	bind KICK - *    SB:mode_opstatus_kick
	bind PART - *    SB:mode_opstatus_part
	bind NICK - *    SB:mode_opstatus_nick
	bind JOIN - *    SB:mode_opstatus_join
	bind RAW  - 367  SB:mode_opstatus_renew ; # [+b] nick chan *!*@* who when
	bind RAW  - 348  SB:mode_opstatus_renew ; # [+e] nick chan *!*@* who when
	bind RAW  - 346  SB:mode_opstatus_renew ; # [+I] nick chan *!*@* who when
	bind RAW  - 324  SB:mode_opstatus_renew ; # [+-] nick chan +ntrlL 30 #CowHead
	bind RAW  - 221  SB:mode_botmodes
	bind RAW  - 008  SB:mode_botmodes_snomask
	bind chon - * SB:chon_process
	bind dcc - "" SB:DCC:test
	bind dcc - q *dcc:quit ; # QUOTE bead disallows ".q" to be unique alias to ".quit"
	bind filt n * SB:filt

	# Bind /NOTICE commands?
	set notice false
	if [info exists ::stormbot_setup(@bind:notice)] { set notice $::stormbot_setup(@bind:notice) }
	if [info exists ::stormbot(config:global-bind:notice)] { set notice $::stormbot(config:global-bind:notice) }
	if $notice { bind notc - * SB:dispatch_notice } {  catch { unbind notc - * SB:dispatch_notice } }

	# Bind events
	catch { bind EVNT - SIGHUP  SB:process_sighup  }
	catch { bind EVNT - SIGTERM SB:process_sigterm }
	catch { bind EVNT - SIGILL  SB:process_sigill  }
	catch { bind EVNT - SIGQUIT SB:process_sigquit }

	# Purge !LAST data from purged last.sb6 bead
	set last [sbd:find LAST:*]
	if [notempty last] {
		putlog "\[SB6:BOOTSTRAP\] Purging \"!LAST\" data from memory ...."
		foreach a [sbd:find LAST:*] { sbd:set $a } ; SB:needsave data
	}

	foreach a [info procs `*] { rename $a "" } ; # Remove "replaced" PROCs

	# ENCRYPT/DECRYPT ECB (old) vs CBC (new) issue
	# Do not combine with other version checks below
	if [left [encrypt testing 0] 1 *] {
		foreach _ [list en de] {
			if [string eq "" [info commands ${_}crypt:original]] {
				rename ${_}crypt ${_}crypt:original
				proc ${_}crypt { key string } [subst { ${_}crypt:original ecb:\$key \$string }]
			}
		}
	}
	sbd:set @ENCRYPT:compensated true

	if { $::numversion > 1080103} {
		# GETUSER patch should be in place
		empty o
		if { [llength [userlist -b]] > 0 } {
			set user [string trim [lindex [split [string trim $::owner] {, ;}] 0]]
			foreach { a b } [lsort -stride 2 -index 0 [getuser $user]] { lappend o $a }
			sbd:set @getuser:valid $o
		}
	} {
		# 2017-03-30 16:53:00 -0700: wedges for GETUSER/SETUSER param capture
		foreach _ [list g s] {
			if [string eq "" [info commands ${_}etuser:original]] {
				rename ${_}etuser ${_}etuser:original
				interp alias {} ${_}etuser {} ${_}etuser:wedge
			}
		}
	}

	# Packages
	empty p_ok p_bad
	set packages [list http tls pop3 cmdline]
	set p__core [list "This is part of Eggdrop core. If this is missing, that means Eggdrop mis-identified where your libraries are. You might need to re-compile or re-install Eggdrop" http]
	set p__tcllib [list "Download and install TCLLIB to get this package" pop3 cmdline]
	set p__other [list "Download TLS (sourceforge) to get this package. It requires a compile (either locally, or, as root for the system in general)" tls]
	foreach package $packages {
		if [ catch { package require $package } msg ] {
			lappend p_bad [list $package $msg]
		} {
			lappend p_ok $package
		}
	}
	# TLS registration
	if { [lsearch -exact $p_ok http] != -1 } {
		if { [lsearch -exact $p_ok tls] != -1 } {
			 set error [ catch { ::http::register https 443 [list ::tls::socket -require 0 -request 1] } msg ]
			 if $error { putlog "\[SB6:BOOTSTRAP\] Failure while registering TLS for HTTPS: $msg" }
		}
	}
	if [isempty p_bad] {
		putlog "\[SB6:BOOTSTRAP\] All packages successfully loaded: [ajl $packages]"
	} {
		putlog "\[SB6:BOOTSTRAP\] Successful package loads: [ajl $p_ok]"
		foreach _ [lsort -inc -dict -index 0 $p_bad] {
			lassign $_ package reason
			empty solution  
			foreach var [info vars p__*] {
				set m [lsearch -exact -nocase -start 1 [set $var] $package]
				if { $m != -1 } { set solution [lindex [set $var] 0] }
			}
			putlog "[space 5]Bad package: $package ($reason).[iff [notempty solution] "Solution: $solution"]"
		}
	}

	# https://github.com/eggheads/eggdrop/issues/342 -- I win!
	# New configuration variables
	SB:get_version ; # Isn't reading true version; only getting 6.0.TS version
	set ver_egg "[lreplace $::version 1 1] ($::numversion)"
	set ver_tcl "[info tclversion] ([info patchlevel])"
	set ver_lib [info library]
	set ver_sb  v[regsub -nocase -- {^StormBot(\.TCL)[ ]?} [sbd:get @version] ""]
	set confargs <unknown>
	foreach - [list "" -] { if [info exists ::configure${-}args] { set confargs [string trim [regsub -all -- {' '} [set ::configure${-}args] " "] '] } }
	putlog "\[SB[left [sbd:get version] 1]\] Eggdrop $ver_egg -- Library: $ver_lib -- TCL $ver_tcl -- StormBot.TCL $ver_sb -- configured with: $confargs"

	stormbot set www http://www.StormBot.Org/
	stormbot set www:bug http://Bug.StormBot.Org/
	stormbot set www:paste http://Paste.StormBot.Org/
	stormbot set irc [list IRC.StormBot.Org:6667 IRC.StormBot.Org:+6697]
	stormbot set irc:channel #StormBot
	stormbot set version [sbd:get version]
	return 0
}

 #############################################################################
 #############################################################################
 #############################################################################
 #############################################################################
 #############################################################################

#####
# Bootstrap:

SB:bootstrap ; # SERVMON is calling SB:HOME, which (if config doesn't have CHANNEL ADD, but chan file does, the HOME chan might be illegal in between the two)
} ; #</tail>

# Load other script fragments (tail first: loader needs some stuff)
sb6:bootstrap:sb6_tail.tcl
sb6:bootstrap:sb6_load.tcl
sb6:bootstrap:sb6_disp.tcl

#<CORE>
# Removal information
if ![info exists ::stormbot(#post:vars)    ] {set ::stormbot(#post:vars)     [info vars    ]}
if ![info exists ::stormbot(#post:commands)] {set ::stormbot(#post:commands) [info commands]}
if ![info exists ::stormbot(#post:timers)  ] {set ::stormbot(#post:timers)   [timers       ]}
if ![info exists ::stormbot(#post:utimers) ] {set ::stormbot(#post:utimers)  [utimers      ]}
if ![info exists ::stormbot(#post:binds)   ] {set ::stormbot(#post:binds)    [binds        ]}

# --- End: transfer control back to EGGDROP ---
return 0
#</core>>>>
